/*****************************************************************************
  $Id: $
  File:		lib/COMBINAT/dyckWords.mu
  Authors:	Nicolas M. Thiry <nthiery@users.sourceforge.net>
  Created:	16/02/2001
  License:	LGPL
*****************************************************************************/

domain combinat::dyckWords

    category combinat::structure;
    // category inlinedoc::cat;

    Name := "combinat::dyckWords";

    info := "Library 'combinat::dyckWords': utilities for dyck Words";

    introduction := "
 A Dyck word of size n is a word with n zeroes and n ones such that in
 any prefix there is more zeroes than ones. If a zeroes is interpreted
 as an opening parenthesis and a one as a closing parenthesis, we get
 a word of well balanced parenthesis. Usually, a Dyck word is
 interpreted as a discrete path in the plane NxN, going from (0,0) to
 (2n,0), with right-up steps (zeroes) and right-down steps (ones),
 without going below the x-axis. Dyck words are fundamental, because
 they are in bijection with so many other discrete structures like
 ordered trees, binary trees, complete binary trees, n-gone 
 triangulations, 2-line tableaux, non crossing set partitions, ...

 Examples:
  - [1,1,1,0,0,0]		((()))
  - [1,0,1,0,1,0]		()()()
  - [1,0,0,1,1,0]		())(() // Invalid!
  - [1,1,1,1,0,0]		(((()) // Invalid!
                                             
 The real generator still needs to be implemented.

 See also combinat::structures.
";

    synopsis := "
 - count(n)		number of dyckWords of size n
 - list(n)		list of the dyckWords of size n
 - list(n1,n2)		list of the dyckWords with n1 '(' and n2 ')'
 - generator(n)		generator for the dyckWords of size n
 - generator(n1,n2)	generator for the dyckWords with n1 '(' and n2 ')'

 - toNonCrossingPartition(w)	Bijection of Biane with non crossing partitions
 - fromNonCrossingPartition(p)	(see combinat::nonCrossingPartitions)
 - toString(w)		Conversion with strings of well balanced parenthesis
 - fromString(s)	

 Not yet implemented:
 - first(n)		first dyckWord of size n
 - Next(c)		next dyckWord
 - previous(c)		previous dyckWord
 - last(n,k)		last dyckWord of size n
";

    interface := dom::interface_Structure union 
        {
         hold(toNonCrossingPartition), hold(fromNonCrossingPartition),
         hold(toString), hold(fromString)
         };

    type := Type::ListOf(Type::Integer);

    /// Symbol used for representing an opening parenthesis "("
    openSymbol   := 1;	// sym
    /// Symbol used for representing a closing parenthesis "("
    closeSymbol := 0;

    //////////////////////////////////////////////////////////////////////////
    // count
    //////////////////////////////////////////////////////////////////////////

    count := combinat::catalan;

    list := 
    proc(k1: Type::NonNegInt, k2=0: Type::NonNegInt) : Type::ListOf(dom::type)
        option remember;
        local dycks, d, d1, d2, i;
    /*
       Returns the list of all dyck words with k1 ")" and k2 "(". 
       They are listed in the order: ["()()()", ... "((()))"]
       corresponding to the inductive generation.
    */
    begin
        if args(0)=1 then
            k2:=k1;
        end_if;
        if k1 < k2 then error("k1 must be < than k2"); end_if;
        if k2 = 0 then return([[dom::openSymbol $ k1]]); end_if;
        if k1 = 1 then return([[dom::openSymbol,dom::closeSymbol]]); end_if;
        if k1 = 0 then return([[]]); end_if;
        dycks:=[];
        if k1 > k2 then
            dycks:=dycks.[[dom::openSymbol]. d $ d=dom::list(k1-1,k2)];
        end_if;
        for i from 0 to k2-1 do
            for d1 in dom::list(i,i) do
                for d2 in dom::list(k1-i-1, k2-i-1) do
                    dycks:=dycks.[[dom::openSymbol].d1.[dom::closeSymbol].d2];
                end_for;
            end_for;
        end_for;
        revert(sort(dycks));
    end_proc;

    //////////////////////////////////////////////////////////////////////////
    // Methods
    //////////////////////////////////////////////////////////////////////////

    /// Bijection of Biane from dyck words to non crossing partitions
    //
    // Thanks to Mathieu Dutour for describing the bijection
    toNonCrossingPartition :=
    proc(l: dom::type)
        : combinat::nonCrossingPartitions::type
        local partition, stack,
              i, j, p, nz;
    begin
        partition := {};
        stack := [];
        i := 1;
        p := 1;
        // Invariants:
        //  - l[i]=0
        //  - p is the number of the opening parenthesis at position i
        while i<=nops(l) do
            stack:=append(stack,p);
            j := i+1;
            while j<=nops(l) and l[j]=dom::closeSymbol do
                j := j+1;
            end_while;
            // Now j points to the next 1, or past the end of l
            nz := j-(i+1); // number of )'s between i and j
            if nz > 0 then
                // Remove the nz last elements of stack, and make a
                // new part in partition.
                if nz>nops(stack) then
                    error("Incorrect dyck word");
                end_if;
                partition := partition 
                            union {[op(stack, nops(stack)-nz+1..nops(stack))]};
                stack := [op(stack, 1..nops(stack)-nz)];
            end_if;
            i := j;
            p := p+1;
        end_while;
        if nops(stack)>0 then
            error("Incorrect dyck word");
        end_if;
        partition;
    end_proc;

    /// Bijection of Biane from non crossing partitions to dyck words
    fromNonCrossingPartition :=
    proc(partition: combinat::nonCrossingPartitions::type)
        : dom::type
        local l, v;
    begin
        l:=[0 $ _plus(nops(v) $ v in partition)];
        for v in partition do
            l[v[nops(v)]] := nops(v);
        end_for;
        [ (dom::openSymbol, dom::closeSymbol $ i) $ i in l ]:
    end_proc;

    /// conversion from dyck words to well-balanced strings of parenthesis
    toString :=
    proc(word: dom::type) : DOM_STRING
    begin
        _concat(if c=dom::openSymbol then
                    "("
                else
                    ")"
                end_if 
                $ c in word);
    end_proc;

    /// conversion from well-balanced strings of parenthesis to dyck words
    fromString :=
    proc(string: DOM_STRING) : dom::type
    begin
        [ if string[i]="(" then 
              dom::openSymbol;
          else
              dom::closeSymbol; 
          end_if
         $ i = 0..length(string)-1 ];
    end_proc;

begin
end_domain:

stdlib::LoadlibState["combinat::dyckWords"] := TRUE:
