/*****************************************************************************
  $Id: generators.mu,v 1.7 2001/11/29 17:52:50 nthiery Exp $
  File:		lib/COMBINAT/generators.mu
  Authors:	Nicolas M. Thiry <nthiery@users.sourceforge.net>
  Created:	25/02/2001
  License:	LGPL
*****************************************************************************/

domain combinat::generators

    // category inlinedoc::cat;

    Name := "combinat::generators";

    info := "Library 'combinat::generators': utilities for generators";

    introduction := "
 A generator is a device that allow for iterating over a large
 quantity of objects without having to produce and store all of them
 at once. This typically is very handy in combinatorics.

 This package provides tools to produce and manipulate generators. The
 convention is that a generator is any MuPAD object g such that the
 function call g() is defined. Each call g() returns returns either a
 new object, or FAIL if no more objects are available.

 Example:
 >> g:=combinat::generators::fromList([a,a,c]):
 >> g();
					a
 >> g();
					a
 >> g();
					c
 >> g();
                                        FAIL

 The effect of calling again a generator g after the FAIL value has
 been returned is undefined (i.ee.g. possibly disastrous).

 Currently, the easiest way to define a generator in MuPAD is to use
 the so called closure-effect. Here is for example how to construct a
 generator for the sequence 1^2, 2^2, 2^2, ...

 >> power_generator:=
 >> proc()
	local i;
        option escape;
    begin
        i:=0;
	return(proc()
	       begin
		   i:=i+1;
		   return(i^2)
	       end_proc)
    end:
 >> g:=power_generator(2):
 >> g();
					1
 >> g();
					4
 >> g();
					9

 What happens here is that the variable i in the procedure g returned
 by power_generator is lexically bound to the context of
 power_generator. Hence, it is not reseted inbetween calls of g.

 The same generator can be also constructed alternatively as follow:
 >> g:=combinat::generators::fromNext(1, n->n+1);
 >> h:=combinat::generators::compose(n->n^2, g);

 Generators cannot be constructed with most MuPAD elements, since
 there is no reference effect. The only exceptions to this rule are
 closures, domains (not domains elements!), and objects defined in
 dynamic modules.

 Generators have a reference effect! Assume g is defined as above:
 >> g();
					1
 >> h:=g: h();
					4
 >> g();
					9
";

    synopsis := "
 - type				type for generators

 - toList(generator)		list of the elements generated by generator
 - count(generator)		number of elements generated by generator

 - empty()			the trivial generator
 - repeater(value, n)		generator repeating n times value
 - repeater(value)		generator repeating value forever
 - fromRange(i..j)		generator for the elements of the range i..j
 - fromList(list)		generator for elements of list
 - fromNext(first, next)	generator from first value and next operation
 - compose(f, g)		composition of the generator g with the function f
 - pipe(g1, g2)			pipe of the generators g1, g2, ...
 - concat(g1, g2, ...)		concatenation of the generators g1, g2, ...
 - cartesian(g1, g2, ...)	cartesian product of the generators g1, g2, ...
				(inelegant implementation in MuPAD 2.0.0)
 - select(g, predicate <,opts>) generator for the elements e generated by g
                                satisfying predicate(e, opts)
";

    interface := {hold(type), hold(toList), hold(count), 
                  hold(empty), hold(repeater),
                  hold(fromList), hold(fromNext), hold(fromRange),
                  hold(concat), hold(compose), hold(select), 
                  hold(pipe), hold(cartesian)
		};

    type := DOM_PROC;
    typeConstructor := DOM_PROC;

    empty :=
    proc() : dom::type
    begin
        return(()->FAIL)
    end_proc;

    repeater :=
    proc(value, n) : dom::type
        option escape;
    begin
        if args(0) = 1 then
            return(()->value);
        else
            return(proc()
                   begin
                       if n>0 then
                           n:=n-1;
                           return(value);
                       else
                           return(FAIL);
                       end_if;
                   end_proc
                  );
        end_if;
    end_proc;

    /**
       fromRange(range)		Generator iterating over range
    **/
    fromRange :=
    proc(range: DOM_EXPR) : dom::type
        local i;
        option escape;
    begin
        i:=op(range,1)-1;	// Avoids using a first_time
        return(proc(): Type::Union(Type::Integer, DOM_FAIL)
               begin
                   i:=i+1;
                   if i<=op(range,2) then
                       return(i);
                   else
                       return(FAIL);
                   end_if;
               end_proc
              );
    end_proc;

    /** fromList(values)	Generator iterating over the elements of values
    **/
    fromList :=
    proc(list: Type::Union(DOM_LIST, DOM_SET)) : dom::type
        local index;
        option escape;
    begin
        index:=0;
        return(proc()
               begin
                   index:=index+1;
                   if index<=nops(list) then
                       return(op(list,index));
                   else
                       return(FAIL);
                   end_if;
               end_proc
              );
    end_proc;

    /** toList(generator)	List of the elements generated by generator
    **/
    toList :=
    proc(generator: dom::type) : DOM_LIST
        local result, value;
    begin
        result:=[];
        while((value:=generator())<>FAIL) do
            result:=append(result,value);
        end_while;
        result;
    end_proc;

    /** count(generator)	Count the number of elements generated by generator
    **/
    count :=
    proc(generator: dom::type) : Type::NonNegInt
        local n;
    begin
        n:=0;
        while(generator()<>FAIL) do
            n:=n+1;
        end_while;
        return(n);
    end_proc;

    /** concat(generator1, generator2, ...)

        The concatenation of the generators generator1, generator2, ...
        That is the generator returning successively the elements
        generated by generator1, generator2 and so on.
    **/
    concat :=
    proc()
        local gens, i;
        option escape;
    begin
        gens := [args()];
        if nops(gens)=0 then
            return(combinat::generators::empty());
        end_if;
        i:=1; // invariant: gens[i] is a valid generator
        return(proc()
                   local x;
               begin
                   while (x:=gens[i]()) = FAIL do
                       i:=i+1;
                       if i>nops(gens) then
                           return(FAIL);
                       end_if;
                   end_while;
                   x;
               end_proc)
    end_proc;

    /** compose(f, g)	Composition of the generator g by the function f

	A generator for f(x_1), f(x_2), ..., where x_1, x_2, ...
        are the elements generated by generator.
    **/
    compose :=
    proc(f: Type::Union(DOM_PROC, DOM_FUNC_ENV), generator: dom::type)
        local options;
        option escape;
    begin
        options := args(3..args(0));
        return(proc()
                   local x;
               begin
                   if (x:=generator()) = FAIL then
                       FAIL
                   else
                       f(x, options)
                   end_if;
               end_proc)
    end_proc;

    /**	   pipe(generator, generator_constructor <, options>)

	Concatenation of the elements generated successively by
	the generators generator_constructor(generator(), args);
    **/
    pipe :=
    proc(generator: dom::type, generator_constructor: dom::type)
        local options, gen;
        option escape;
    begin
        options := args(3..args(0));
        gen := generator();
        if gen=FAIL then
            return(dom::empty());
        end_if;
        gen := generator_constructor(gen, options);
        
	proc()
            local x;
        begin
            while TRUE do
                if (x:=gen()) <> FAIL then
                    return(x)
                end_if;
                if (gen:=generator()) = FAIL then
                    return(FAIL);
                end_if;
                gen := generator_constructor(gen, options)
            end_while;
        end_proc
    end_proc;

    fromNext :=
    proc(first, Next: Type::Union(DOM_PROC, DOM_FUNC_ENV))
        local first_time, arguments;
        option escape;
    begin
        first_time := TRUE;
        arguments := args(3..args(0));
        proc()
        begin
            if first_time then
                first_time:=FALSE;
            else
                first:=Next(first, arguments);
            end_if;
            return(first);
        end_proc;
    end_proc;

    cartesian :=
    if domtype(copyClosure) = DOM_FUNC_ENV then
        proc()
            option escape;
            local generators,
                  gen_copies,
                  current,
                  first_time,
                  n;
        begin
            n := args(0);
            generators := [args()];
            gen_copies := map(generators, copyClosure);
            first_time := TRUE;
            proc()
                local i,j;
            begin
                if first_time then
                    current := [ FAIL $ n ];
                    for i from 1 to n do
                        if (current[i]:=generators[i]()) = FAIL then
                            return(FAIL);
                        end_if;
                    end_for;
                    first_time := FALSE;
                    return(current);
                else
		    // Invariants:
                    //  - n is the number of generators
	            //  - gen_copies contains a copy of the generators in
	            //	  their original state
	            //  - generators contains the generators in their 
	            //	  current state
	            //  - The generators are non empty
	            //  - current contains the last generated element
                    for i from n downto 1 do
                        if (current[i]:=generators[i]()) <> FAIL then
                            for j from i+1 to n do
                                generators[j]:=copyClosure(gen_copies[j]);
                                current[j] := generators[j]();
                            end_for;
                            return(current);
                        end_if;
                    end_for;
                    return(FAIL);
                end_if;
            end_proc;
        end_proc;
//         proc()
//             option escape;
//             local gen_copies, 
//                   generators, 
//                   current,
//                   i;
//         begin
//             generators := [args()];
//             gen_copies := map(generators, copyClosure);
//             current := [ FAIL $ args(0) ];
//             for i from 1 to nops(generators)-1 do
//                 if (current[i]:=generators[i]()) = FAIL then
//                     return(dom::empty());
//                 end_if;
//             end_for;
// 	    // Invariants (n is the number of generators):
//             //  - The n-1 first terms of current are "valid"
//             //  - gen_copies contains a copy of the generators in their
//             //	  original state
//             //  - generators contains the generators in their 
//             //	  current state
//             //  - The first n-1 generators are non empty
//             proc()
//             begin
//                 for i from nops(generators) downto 1 do
//                     if current[i]:=generators[i]() <> FAIL then
//                         for j from i+1 to nops(generators) do
//                             generators[j]:=copyClosure(gen_copies[j]);
//                             current[j] := generators[j]();
//                         end_for;
//                         if current[nops(current)] = FAIL then
//                             return(FAIL);
//                         end_if;
//                         return(current);
//                     end_if;
//                 end_for;
//                 return(FAIL);
//             end_proc;
//         end_proc;
    else
	proc(/* sequence of dom::type */ ) 
            // Temporary inelegant implementation when copyClosure
            // is not available (mupad 2.0.0)
        begin
            combinat::cartesianProduct::generator
            (dom::toList(l) $ l in [args()]);
        end_proc;
    end_if;

    /** select(g, predicate <,opts>)	selects the elements generated by g satisfying predicate

	A generator for the elements e generated by g satisfying
	predicate(e, opts).
    **/
    select :=
    proc(generator: dom::type, predicate: Type::Union(DOM_PROC, DOM_FUNC_ENV))
        option escape;
        local options;
    begin
        options := args(3..args(0));
        proc()
            local x;
        begin
            while (x:=generator()) <> FAIL do
                if predicate(x, options) then
                    return(x)
                end_if;
            end_while;
            return(FAIL);
        end_proc;
    end_proc;

begin
end_domain:

stdlib::LoadlibState["combinat::generators"] := TRUE:
