/*****************************************************************************
  $Id: getOptions.mu,v 1.2 2001/12/03 17:28:27 nthiery Exp $
  File:		lib/PROG/getOptions.mu
  Authors:	Nicolas M. Thiry <nthiery@users.sourceforge.net>
  Purpose:	combinat::getOptions -- Generic option parsing procedure.
  Created:	16/02/2001
  License:	LGPL
*****************************************************************************/

/**
   combinat::getOptions(allOptions
                    <optionsType>,
                    offset: Type::NonNegInt,
		    < Opt >, < Opt = Value >, ...) : Type::OptionsTable

   allOptions:	  list or set of all options, or table with an entry
                  'option = default value' for each option
		  (default value can be NIL)
   optionTypes:   table with an entry
                  'option = type of value' for each option
                  (if option can have values)
   offset         integer position to skip the 'offset'-th arguments of 'ARGS'

   Opt:		  the name of an option: an object of type DOM_IDENT.
   Value:	  the value of an option: an object of any type.

   Return value: a table of options constructed as follow:
    - If Opt does not appear in allOptions, raise an error "unknown option".
    - If Value=NIL, don't change the default value for Opt.
    - If Value is not specified, set the value for Opt to TRUE.
    - If Value does not match the type described in optionsType, raises
      an error "incorrect type"
    - Else set the value for Opt to Value.
    - Leave the other options defined in defaultValues unchanged.

   When a default value for a given option is not provided, then NIL
   is used.

   Examples:

   >> combinat::getOptions(table(Opt1=1), 0, Opt2);

      error: unknown option: Opt2

   >> combinat::getOptions(table(Opt1=FALSE, Opt2=2, Opt3=3), 0,
			Opt1, Opt2=NewValue, Opt3=NIL);

			table(Opt1=TRUE, Opt2=NewValue, Opt3=3);

   Caveats:
    - If the type of a non-mandatory option is specified, say
      Type::NonNegInt, then a default value should also be specified. 
      Otherwise, the value of the option will be NIL, which will
      raise an error. This might be slightly non-intuitive, but allows
      for the programmer to always make the assumption that the arguments
      are of the correct type.

   Bugs:
    - Error messages should specify the calling function, not the
      function itself.
    - When testargs() is interactive, no argument checking will be
      done for interactive functions, since combinat::getOptions will be
      called non-interactively by those.
**/
combinat::getOptions :=
proc(options: Type::Union(Type::OptionsTable, 
                          Type::SetOf(Type::Option),
                          Type::ListOf(Type::Option)),
     optionsType: Type::Union(Type::OptionsTable, Type::NonNegInt)
    ) : Type::OptionsTable
    local opt, key, value, offset, ARGS;
begin
    // Initialization
    if domtype(options) <> DOM_TABLE then
        options := table(op(map(options, o->o=NIL)))
    end_if;
    if domtype(optionsType) = DOM_INT then
        offset := optionsType;
        optionsType := table();
        ARGS := [args(3+offset..args(0))];
    else
        if testargs() then
            if args(0) < 3 then
                error("missing offset argument");
            end_if;
            if not testtype(args(3), Type::NonNegInt) then
                error("incorrect type for 3rd argument");
            end_if;
        end_if;
        offset := args(3);
        ARGS := [args(4+offset..args(0))];
    end_if;

    // Option analysis
    for opt in ARGS do
        if testtype(opt, Type::Equation(DOM_IDENT, Type::AnyType)) then
	    key:=op(opt,1);
            value:=op(opt,2);
        elif testtype(opt, DOM_IDENT) then
            key:=opt;
            value:=TRUE;
        else
            error("Bad option: ".expr2text(opt));
        end_if;
	if contains(options, key) then
            if value <> NIL then
                options[key]:=value;
            end_if;
	else
            error("Unknown option: ".expr2text(key));
        end_if;
    end_for;
    
    // Option checking: mandatory / types
    if testargs() then
        for opt in options do
            key := op(opt,1);
            value := op(opt,2);
            if value = hold(Mandatory) then
                error("Option ".expr2text(key)." is mandatory");
            end_if;
            if contains(optionsType, key) 
               and not testtype(value, optionsType[key]) then
                error("Bad type for option ".expr2text(key));
            end_if;
        end_for;
    end_if;
    return(options);
end_proc:
