/*****************************************************************************
  $Id: $
  File:		lib/COMBINAT/integerListsLexBuilder.mu
  Authors:	Florent Hivert <Florent.Hivert@univ-mlv.fr
		Nicolas M. Thiry <nthiery@users.sourceforge.net>
  Created:	16/06/2001
  License:	LGPL
*****************************************************************************/

domain combinat::integerListsLexBuilder

    category combinat::structure;
    // category inlinedoc::cat;

    Name := "combinat::integerListsLexBuilder";

    info := "Library 'combinat::integerListsLexBuilder': Lexicographic generation of lists of integers under constraints";

    introduction := "
   Notations:
   n: sum of the composition

   Definitions:

   Informally, we call here composition any finite list of
   nonnegative integers. By abuse, in some cases we even allow for
   +infinity parts.
   The sum of a composition is the sum of its parts.
   The length of a composition is the number of its parts

   Several orders between compositions:

    - Point by point: given two compositions Comp1 and Comp2 of same
      length, Comp1 is *below* Comp2 if Comp1[i]<Comp2[i].
    - Lexicographic:
    - Refinement:

   The goal is to build all the compositions satisfying a given
   set of constraints, through a generator. The constraints can be of
   the following form:

    - Regularity conditions:
      Let minSlope < maxSlope be integers or +-infinity.
      A composition Comp is regular if:
		minSlope <= Compo[i+1]-Compo[i] <= maxSlope, forall i.
      Regular functions are defined similarily.

    - Length conditions:
      Let minl <= maxl be nonnegative integers (or +infinity for maxl)
      A composition Comp satisfies the condition if:
		minl <= nops(Compo) <= maxl

    - Upper/lower conditions:
      Let floor and ceiling be regular functions from [ 1 .. maxl ] to N.
      A composition Comp satisfies the condition if:
		Compo[i] >= floor(i),      forall i in 1..nops(Compo)
		Compo[i] <= ceiling(i),    forall i in 1..nops(Compo)

    - Refinement conditions:

   A composition is valid if it satisfies all the imposed conditions.
   Furthermore, two compositions are considered identical if they only
   differ by trailing zeroes at positions > minl. To get a unique
   representative, we add the condition that if nops(Compo)>minl, then
   Compo[nops(Compo)]>0.

   Currently the algorithm implemented here deals with regularity
   conditions, length conditions and upper/lower conditions.
   Its complexity is amortized linear over the number of generated
   compositions

   Example:

    - minl	= 9
    - maxl	= 9
    - floor	= [1,0,3,1,0,1,0,0]
    - ceiling	= [7,7,5,5,7,7,2,7]
    - minSlope	= -2
    - maxSlope	= 0

	   v  v  v  v  v  v  v  v  v  
	7|       v  v           v   | 
	6|       v  v           v   | 
	5|                      v   | 
	4|                      v   | 
	3|       ^              v   | 
	2|       ^                  | 
	1| ^     ^  ^     ^         | 
	   ^  ^  ^  ^  ^  ^  ^  ^  ^  
	   1  2  3  4  5  6  7  8  9  

   [6,4,3,3,3,3,1,1,1] is a valid composition of 25; 
   graphically, it contains all the `^' and none of the `v':
	   v  v  v  v  v  v  v  v  v  
	7|       v  v           v   | 
	6| O     v  v           v   | 
	5| O                    v   | 
	4| O  O                 v   | 
	3| O  O  ^  O  O  O     v   | 
	2| O  O  ^  O  O  O         | 
	1| ^  O  ^  ^  O  ^  O  O  O| 
	   ^  ^  ^  ^  ^  ^  ^  ^  ^  
	   1  2  3  4  5  6  7  8  9  

   [3,3,3,1,1,1,0,0,0] is the valid composition of minimal sum (minCompo);
   it can be constructed by `smoothing' the lower bound in order to
   satisfy the slope constraints:

	   v  v  v  v  v  v  v  v  v  
	7|       v  v           v   | 
	6|       v  v           v   | 
	5|                      v   | 
	4|                      v   | 
	3| O  O  ^              v   | 
	2| O  O  ^                  | 
	1| ^  O  ^  ^  O  ^         | 
	   ^  ^  ^  ^  ^  ^  ^  ^  ^  
	   1  2  3  4  5  6  7  8  9  

   [7,7,5,5,5,5,4,2,2] is the valid composition of maximal sum (maxCompo);
   It can be constructed by `smoothing' the upper bound in order to
   satisfy the slope constraints:

	   v  v  v  v  v  v  v  v  v  
	7| O  O  v  v           v    |
	6| O  O  v  v           v    |
	5| O  O  O  O  O  O     v    |
	4| O  O  O  O  O  O  O  v    |
	3| O  O  ^  O  O  O  O  v    |
	2| O  O  ^  O  O  O  O  O  O |
	1| ^  O  ^  ^  O  ^  O  O  O |
	   ^  ^  ^  ^  ^  ^  ^  ^  ^ 
	   1  2  3  4  5  6  7  8  9 

   Actually, we can replace the original upper and lower bound by
   respectively the valid composition of maximal sum, and the valid
   composition of minimal sum:

	   v  v  v  v  v  v  v  v  v  
	7|       v  v  v  v  v  v  v |
	6|       v  v  v  v  v  v  v |
	5|                   v  v  v |
	4|                      v  v |
	3| ^  ^  ^              v  v |
	2| ^  ^  ^                   |
	1| ^  ^  ^  ^  ^  ^          |
	   ^  ^  ^  ^  ^  ^  ^  ^  ^ 
	   1  2  3  4  5  6  7  8  9 

   To obtain the lexicographically maximal compositions, we can simply
   fill the diagram in the order below:

	   v  v  v  v  v  v  v  v  v 
	7| 18 24 v  v  v  v  v  v  v |
	6| 16 21 v  v  v  v  v  v  v |
	5| 14 17 23 28 33 38 v  v  v |
	4| 13 15 20 26 30 35 40 v  v |
	3| ^  ^  ^  22 27 32 37 v  v |
	2| ^  ^  ^  19 25 29 34 39 42|
	1| ^  ^  ^  ^  ^  ^  31 36 41|
	   ^  ^  ^  ^  ^  ^  ^  ^  ^ 
	   1  2  3  4  5  6  7  8  9 

   For example, [7,7,5,5,4,3,1,0,0] is the lexicographically maximal
   composition of 32 (first):

	   v  v  v  v  v  v  v  v  v 
	7| 18 24 v  v  v  v  v  v  v |
	6| 16 21 v  v  v  v  v  v  v |
	5| 14 17 23 28       v  v  v |
	4| 13 15 20 26 30       v  v |
	3| ^  ^  ^  22 27 32    v  v |
	2| ^  ^  ^  19 25 29         |
	1| ^  ^  ^  ^  ^  ^  31      |
	   ^  ^  ^  ^  ^  ^  ^  ^  ^ 
	   1  2  3  4  5  6  7  8  9 

   Problems with degenerate cases:

    - If maxl=infinity and maxSlope>0, testing whether there exists a
      valid composition of n is only semi-decidable! For example, for
      floor(i)=0, maxl=maxSlope=infinity, minSlope=-infinity, there
      exists a valid composition of 1 if and only if ceiling(i) > 0
      for some i. This can only be checked by running through all
      positive integers i, which may take forever.

   See also combinat::structure.
";

    synopsis := "
 - type:			list of nonnegative integers

 - constraints:	minLength, maxLength, floor, ceiling, minSlope, maxSlope

   minLength, maxLength:	NonNegInt or + - infinity
   floor, ceiling:		regular functions
   minSlope, maxSlope		NonNegInt or + - infinity

 - count(n, constraints)	number of generic compositions of n
 - list(n, constraints)	list of all generic compositions of n
 - generator(n, constraints)	generator for all generic compositions of n

 - first(n, constraints)	first generic composition of n
 - Next(c, constraints)		next generic composition of n
";

    examples := "
 - List all partitions of 7:
    combinat::integerListsLexBuilder::list
	(7, 0, infinity, ()->0, ()->infinity, -infinity, 0);

 - List all partitions of n with parts at least 2:
    combinat::integerListsLexBuilder::list
	(n, 0, infinity, ()->2, ()->infinity, -infinity, 0);

 - List all partitions of n which are contained in a given partition p:
    combinat::integerListsLexBuilder::list
	(n, 0, infinity, ()->0, i->p[i], -infinity, 0);

 - List all partitions of n which contain a given partition p:
    combinat::integerListsLexBuilder::list
	(n, 0, infinity, ()->0, i->p[i], -infinity, 0);

 - List all compositions of n
    combinat::integerListsLexBuilder::list
	(n, 0, infinity, ()->1, ()->infinity, -infinity, infinity):

 - List all vectors of integers of sum n and length l:
    combinat::integerListsLexBuilder::list
	(n, l, l, ()->0, ()->infinity,	-infinity, infinity):

 - List all monomials of degree d which divide a given monomial m:
   (the monomial being represented by its degree vector)
    combinat::integerListsLexBuilder::list
	(d, nops(m), nops(m), ()->0, i->m[i], -infinity, infinity):

 - List all monomials of degree d which are divisible by a given monomial m:
   (the monomial being represented by its degree vector)
    combinat::integerListsLexBuilder::list
	(d, nops(m), nops(m), i->m[i], i->infinity, -infinity, infinity):
";

    interface := dom::interface_Structure union 
        {hold(first),
         hold(Next), hold(rightmostPivot),
         hold(upperRegular), hold(lowerRegular),
         hold(comp2ceiling), hold(comp2floor)
         };

    type		:= Type::ListOf(Type::NonNegInt);
    typeIntBound	:= Type::Union(Type::NonNegInt, stdlib::Infinity);
    typeCompositionBound:= Type::ListOf(dom::typeIntBound);
    typeSlopeBound	:= Type::Union(Type::Integer, stdlib::Infinity);
    typeRegularFunction	:= Type::Union(DOM_PROC, DOM_FUNC_ENV);

    //////////////////////////////////////////////////////////////////////////
    // first
    //////////////////////////////////////////////////////////////////////////

    first :=
    // Returns the lexicographically smallest valid composition of n
    // Preconditions: 
    //  - floor and ceiling need to satisfy the slope constraints
    //    e.g. be obtained from comp2floor or comp2ceil
    //  - floor must be below ceiling to ensure the existence a
    //    valid composition
    proc(n: Type::Integer,
         minl: dom::typeIntBound, maxl: dom::typeIntBound,
         floor: dom::typeRegularFunction, ceiling: dom::typeRegularFunction,
         minSlope: dom::typeSlopeBound, maxSlope: dom::typeSlopeBound)
        : Type::Union(dom::type, DOM_FAIL)
        local lowX,  lowY,
              highX, highY,
              N, result, i;
    begin
        // Increases minl until n<=_plus(ceiling(i) $i=1..minl)
        // may run forever!
        N := _plus(ceiling(i)$i=1..minl);
        while N < n do
            minl:=minl+1;
            if minl > maxl then
                return(FAIL);
            end_if;
            if ceiling(minl) = 0 and maxSlope <= 0 then
            // N will never increase again
                return(FAIL);
            end_if;
            N := N + ceiling(minl);
        end_while;

        n := n - _plus(floor(i)$i=1..minl);
        if n < 0 then return(FAIL); end_if;

	// Now we know that we can build the composition inside the
	// `tube' [1..minl] x [floor, ceiling]


        if minSlope=-infinity then		// Trivial case
            result:=[];
            for i from 1 to minl do
                if n<= ceiling(i) - floor(i) then
                    result:=append(result,floor(i)+n);
                    break;
                else
                    result:=append(result,ceiling(i));
                    n:=n-(ceiling(i) - floor(i));
                end_if;
            end_for;
            result:=result.[floor(j) $ j=i+1..minl]
        else
            // remaining number of boxes to add
            lowX  := 1; lowY  := floor(1); // Not in the list
            highX := 1; highY := floor(1);
    
            while n > 0 do
                // Invariant after each iteration of the loop:
                // [lowX,lowY] is the coordinate of the rightmost point of the
                // current diagonal such that floor(lowX) < lowY
    
                lowY := lowY + 1;
                while lowX < minl and
                      lowY + minSlope > floor(lowX+1)  do
                    lowX := lowX + 1; lowY := lowY + minSlope;
                end_while;

                highY := highY + 1;
                while highY > ceiling(highX) do
                    highX := highX + 1; highY := highY + minSlope;
                end_while;
            
                n := n - (lowX - highX + 1);
            end_while;
    
            [ceiling(i) $ i = 1..highX-1,
             (highY + minSlope * i - 1) $ i =  0..-n-1,
             (highY + minSlope * i    ) $ i = -n..lowX-highX,
             floor(i) $ i = lowX+1..minl];
        end_if;
    end_proc;

    //////////////////////////////////////////////////////////////////////////
    // next
    //////////////////////////////////////////////////////////////////////////

    rightmostPivot :=
    proc(compo: dom::type,
         minl:  dom::typeIntBound, maxl: dom::typeIntBound,
         floor: dom::typeRegularFunction, ceiling: dom::typeRegularFunction,
         minSlope: dom::typeSlopeBound, maxSlope: dom::typeSlopeBound)
        : Type::Union(Type::ListProduct(Type::PosInt, Type::NonNegInt),
                      DOM_FAIL)
         // Return the rightmost pivot position, and a lower bound for
         // the pivot value

        local x, y, temp,
              floorx_x, ceilingx_x,
              oldfloorx_x, oldceilingx_x,
              F, G,
              lowX, highX,
              minSlope_lowX, maxSlope_highX;
    begin

	// Step 1: Look for available space after nops(compo).
	//
	// Let y be minimal such that y>nops(compo) and ceiling(y)>0.
	// Then, for a given x, there exists i>nops(compo) such that
	// ceiling_x(i)>0 iff and only if ceiling_x(y)>0. Moreover in this
	// case ceiling_x(i)=0 for all i, nops(compo) < i < y.
	//
	// This part may loop forever if maxl=infinity and ceiling(i)=0
	// for all i > nops(compo)
        y := nops(compo) + 1;
        while y <= maxl do
            if ceiling(y) > 0 then break; end_if;
            if maxSlope <= 0 then y:=maxl+1; break; end_if;
            y := y + 1;
        end_while;
	// Assertions:
	//  - ceiling(y)=0 forall i, nops(compo) < i <= y
	//  - if y < maxl+1 then ceiling(y) > 0

	// Step 2: Search for the rightmost pivot position x from 
	//         nops(compo) downto 1.
	//
	// Invariants at beginning of loop:
	//  - floorx_x = 	floor_x(x)
	//  - ceilingx_x = 	ceiling_x(x)
	//  - F =   _plus(compo[i]     $ i=x..nops(compo))
	//        - _plus(floor_x(i)   $ i=x..nops(compo))
	//  - G =   _plus(ceiling_x(i) $ i=x..nops(compo))
	//        - _plus(compo[i]     $ i=x..nops(compo))
	//
	// x is the rightmost pivot position iff
	//  - ceiling_x(x) >= floor_x(x) and (G >= 0 or F >= floor_x(y))
	//
	// Remarks:
	//  - Obviously, F only needs to be maintained when y < maxl+1
	//  - G may decrease, but at most by one at each step
	//  - F only increases

        x := nops(compo);
        if x = 0 then
            return(FAIL);
        end_if;
        ceilingx_x := compo[x]-1;
        floorx_x   := floor(x);
        if x > 1 then
            floorx_x := max(floorx_x, compo[x-1]+minSlope);
        end_if;
        F := compo[x] - floorx_x;
        G := ceilingx_x - compo[x]; // this is -1

	// Further invariants at beginning of loop:
	//  - highX is the largest integer such that x <= highX <= nops(compo)
	//    and (ceiling_x(highX) < ceiling(highX) or x=highX)
	//	  Note:^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^ is true for x=highX
	//  - lowX is the largest integer such that x <= lowX <= nops(compo)
	//    and (floor_x(lowX) > floor(lowX) or x=lowX)
	//  - minSlope_lowX =  _plus(minSlope(i) $ i=x..lowX-1)
	//  - maxSlope_highX = _plus(maxSlope(i) $ i=x..highX-1
	// Remarks:
	//  - both lowX and highX are decreasing

        highX := x;
        lowX  := x;

        while not (ceilingx_x >= floorx_x and
                   (G>=0 or
                    (y < maxl+1 and
                     F - max(floor  (y), floorx_x   + (y-x)*minSlope) >= 0 and
                     G + min(ceiling(y), ceilingx_x + (y-x)*maxSlope) >= 0
                    ))) do
            if x = 1 then
                return(FAIL);
            end_if;
            x:=x-1;

            oldceilingx_x:=ceilingx_x;
            oldfloorx_x:=floorx_x;
            ceilingx_x := compo[x]-1;
            floorx_x := floor(x);
            if x>1 then
                floorx_x := max(floorx_x, compo[x-1]+minSlope);
            end_if;

            minSlope_lowX  := minSlope * (lowX -x);
            maxSlope_highX := maxSlope * (highX-x);

    	// update G
            if maxSlope = infinity then
        	// In this case, we have:
        	//  - ceiling_x(i) = ceiling(i), for i>x
        	//  - G >= 0 or G = -1
                G := G + (ceiling(x+1)-compo[x+1]);
            else
                G := G + (highX-x) * ( (compo[x]+maxSlope) - compo[x+1] )
                    - 1;
                while highX > x and
                      (temp := (ceilingx_x + maxSlope_highX) - ceiling(highX))
                      >= 0 do
                    G := G - temp;
                    highX := highX - 1;
                    maxSlope_highX := maxSlope * (highX-x);
                end_while;
            end_if;
            if G >= 0 and compo[x] > floorx_x then
        	// By case 1, x is the rightmost pivot position
                break;
            end_if;
        
            if y < maxl+1 then
        	// update F
                F := F + (compo[x] - floorx_x);
                if minSlope <> -infinity then
                    F := F + (lowX-x) * ( oldfloorx_x - (floorx_x+minSlope) );
                    while lowX > x and
                          (temp := floor(lowX) - (floorx_x + minSlope_lowX))
                          >= 0 do
                        F := F - temp;
                        lowX := lowX - 1;
                        minSlope_lowX  := minSlope * (lowX-x);
                    end_while;
                end_if;
            end_if;
        end_while;

        return([x, floorx_x]);
    end_proc;

    Next :=
    proc(compo: dom::type,
         minl:  dom::typeIntBound, maxl: dom::typeIntBound,
         floor: dom::typeRegularFunction, ceiling: dom::typeRegularFunction,
         minSlope: dom::typeSlopeBound, maxSlope: dom::typeSlopeBound)
        : Type::Union(dom::type, DOM_FAIL)

        local x,			// pivot position
              low, high,		// bounds on the pivot value
              newFloor, newCeiling;	// wrappers
    begin

        x := dom::rightmostPivot(args());
        if x=FAIL then 
            return(FAIL);
        else
            [x,	low ] := x;
            high := compo[x]-1;
        end_if;

        // Build wrappers around floor and ceiling to take into
        // account the new constraints on the value of compo[x].
        //
        // Efficiency note: they are not wrapped more than once, since
        // the method Next calls first, but not the converse.
        
        newFloor   := 
        if minSlope=-infinity then
            i -> floor(x+(i-1));
        else
            i -> max(floor  (x+(i-1)), low  + (i-1)*minSlope);
        end_if;

        newCeiling := 
        if maxSlope=infinity then
            i -> if i=1 then compo[x]-1 else ceiling(x+(i-1)) end_if;
        else
            i -> min(ceiling(x+(i-1)), high + (i-1)*maxSlope);
        end_if;

        [op(compo,1..x-1),
         op(dom::first(_plus(op(compo,x..nops(compo))),
                       max(minl-x+1, 0), maxl-x+1,
                       newFloor, newCeiling,
                       minSlope, maxSlope))
        ];
    end_proc;

    //////////////////////////////////////////////////////////////////////////
    // previous
    //////////////////////////////////////////////////////////////////////////

    //////////////////////////////////////////////////////////////////////////
    // generator
    //////////////////////////////////////////////////////////////////////////

    generator :=
    proc(n: Type::Integer, 
         minl:  dom::typeIntBound, maxl: dom::typeIntBound,
         floor: dom::typeRegularFunction, ceiling: dom::typeRegularFunction,
         minSlope: dom::typeSlopeBound, maxSlope: dom::typeSlopeBound)
        : combinat::generators::type
    begin
        combinat::generators::fromNext(dom::first(args()),
                                       dom::Next, args(2..args(0)));
    end_proc;

    //////////////////////////////////////////////////////////////////////////
    // Utilities for regular compositions and functions
    //////////////////////////////////////////////////////////////////////////

    lowerRegular :=
    proc(comp: dom::typeCompositionBound,
         minSlope: dom::typeSlopeBound, maxSlope: dom::typeSlopeBound)
        : dom::type
        // Returns the uppest regular composition below comp

        local i;
    begin
        for i from 2 to nops(comp) do
            comp[i] := min(comp[i], comp[i-1] + maxSlope);
        end_for;

        for i from nops(comp) - 1 downto 1 do
            comp[i] := min(comp[i], comp[i+1] - minSlope);
        end_for;
        comp;
    end_proc;

    upperRegular :=
    proc(comp: dom::typeCompositionBound,
         minSlope: dom::typeSlopeBound, maxSlope: dom::typeSlopeBound)
        : dom::type
    // Returns the lowest regular composition above comp
        local i;
    begin
        for i from 2 to nops(comp) do
            comp[i] := max(comp[i], comp[i-1] + minSlope);
        end_for;

        for i from nops(comp) - 1 downto 1 do
            comp[i] := max(comp[i], comp[i+1] - maxSlope);
        end_for;
        comp;
    end_proc;

    // TODO: compo2ceil

    comp2floor :=
    proc(floor: dom::typeCompositionBound,
         minSlope: dom::typeSlopeBound, maxSlope: dom::typeSlopeBound)
        : dom::typeRegularFunction
    // Given a composition, returns the lowest regular function N->N above it
        option escape;
    begin
        if nops(floor)=0 then
            return(i->0)
        end_if;
        floor := dom::lowerRegular(floor, minSlope,maxSlope);
        return(proc(i: Type::PosInt) : Type::NonNegInt
               begin
                   if i<=nops(floor) then
                       return(floor[i]);
                   else
                       return(max(0,floor[nops(floor)]-(i-nops(floor))*minSlope));
                   end_if;
               end_proc);
    end_proc;

    positions :=
    // Returns the list of the coordinates of the rightmost point of the
    // diagonals that is above floor
    // Not really used anymore; just here for historical reference.
    proc(maxl: dom::typeIntBound,
         floor: dom::typeRegularFunction, ceiling: dom::typeRegularFunction,
         minSlope: dom::typeSlopeBound)
        : Type::ListOf(Type::ListOf(Type::NonNegInt))
        local lowX,  lowY,
              highX, highY,
          result;
    begin
        result := [];
        lowX  := 1; lowY  := floor(1); // Not in the list
        highX := 1; highY := floor(1);
	// end of loop invariant: [lowX,lowY] is the coordinate of the
	// rightmost point of the current diagonal such that
	// floor(lowX) < lowY

        while TRUE do

            lowY := lowY + 1;
            while lowX < maxl and
                  lowY + minSlope > floor(lowX+1)  do
                lowX := lowX + 1; lowY := lowY + minSlope;
            end_while;

            if lowY > ceiling(lowX) then
                return (result);
            end_if;

            highY := highY + 1;
            while highY > ceiling(highX) do
                highX := highX + 1; highY := highY + minSlope;
            end_while;

            result := append(result,[lowX,lowY,highX,highY]);
        end_while;
    end_proc;

begin
end_domain:

stdlib::LoadlibState["combinat::integerListsLexBuilder"] := TRUE:
