/*****************************************************************************
  $Id: integerListsLexBuilding.mu,v 1.3 2002/01/11 19:34:25 nthiery Exp $
  File:		lib/COMBINAT/integerListsLexBuilding.mu
  Authors:	Florent Hivert <Florent.Hivert@univ-mlv.fr
		Nicolas M. Thiry <nthiery@users.sourceforge.net>
  Created:	16/06/2001
  License:	LGPL
*****************************************************************************/

category combinat::integerListsLexBuilding

    category combinat::structure;
    // category inlinedoc::cat;

    info := "Category 'combinat::integerListsLexBuilding': wrapper around combinat::integerListsLexBuilder";

    introduction := "

   This is a wrapper category that make it easy to build special
   purpose libraries using the lexicographic generator of
   combinat::integerListsLexBuilder. This category inherits from
   combinat::structure.

   See also: combinat::structure, combinat::integerListsLexBuilder.
";

    synopsis := "

    Required entries:
     - parseOptions: a function returning a proper sequence of
                     constraints for combinat::integerListsLexBuilder.

    Provided entries (in addition to those of combinat::structure):
     - size
     - first, last, Next

    Additional utilities:
     - parseOptions_IntegerListLexBuilding:	Default option parser
";

    examples := "
    Creates a domain for fixed-length strict partitions:
    (length: k, lower bound on a part: 1, maximal slope: 0)

    domain kpartitions
	category combinat::integerListsLexBuilding;

        parseOptions := k -> (k, k, ()->1, ()->infinity, -infinity, 0);
    begin
    end_domain;

    The first partition of 4 of length 2:
    l := kpartitions::first(4, 2);

    The next one:
    l := kpartitions::Next(l, 2);

    That was the last one:
    l := kpartitions::Next(l, 2);

    A generator for all partitions of 6 of length 3:
    g := kpartitions::generator(6, 3);

    All partitions of 7 of length 2:
    l := kpartitions::list(7,2);
";

    interface_IntegerListLexBuilding := dom::interface_Structure union 
        {hold(size),
         hold(first), hold(Next)
         };
    interface := dom::interface_IntegerListLexBuilding; // to allow for reuse

    /// Returns constraints for integerListsLexBuilder from user-friendly options
    parseOptions := toBeDefined;

    /**	Helper for constructing the option parser dom::parseOptions

        This generic parser makes it easy for a domain of the category
        combinat::IntegerListLexBuilding to build his own option parser.

	Example:

        domain ivectors
	    category combinat::integerListsLexBuilding;

	    parseOptions :=
	    proc(k)
	    begin
		parseOptions_IntegerListLexBuilding
		   (table(
                          hold(Length)	= k,
                          hold(MinPart)	= 0,
               		  hold(MaxPart)	= infinity,
               		  hold(Inner)	= NIL,
               		  hold(Outer)	= NIL,
	       		  hold(MinSlope)= -infinity,
	       		  hold(MaxSlope)= infinity
			 ),
		    args());
	    end_proc;
	end_domain;
	ivectors::list(MinLength=4, Outer=[1,3,4], MaxSlope=4);

        It's up to the user (or the domain!) to ensure that Inner and
        Outer are compatible with MinPart, MaxPart, MinSlope and MaxSlope.
    **/
    parseOptions_IntegerListLexBuilding :=
    proc(defaultOptions: Type::OptionsTable)
        : Type::SequenceOf(Type::AnyType,6,6)
         // sequence of 6 constraints
        local options, inner, outer, minpart;
        option escape;
    begin
        options := combinat::getOptions(defaultOptions,
                                    dom::parseOptions_OptionTypes,
                                    1, args());
        if options[hold(Length)] <> NIL then
            options[hold(MinLength)] := options[hold(Length)];
            options[hold(MaxLength)] := options[hold(Length)];
        end_if;
        minpart := options[hold(MinPart)];
        if options[hold(Inner)] <> NIL then
            if domtype(options[hold(Inner)]) = DOM_LIST then
                inner := options[hold(Inner)];	// slight optimization ?
                options[hold(Inner)] :=
                    i->if i>nops(inner) then minpart else inner[i]; end_if;
                options[hold(MinLength)] := max(options[hold(MinLength)],
                                                nops(inner));
            end_if;
        else
            options[hold(Inner)] := fp::unapply(minpart,hold(i));
        end_if;
        if options[hold(Outer)] <> NIL then
            if domtype(options[hold(Outer)]) = DOM_LIST then
                outer := options[hold(Outer)];	// slight optimization ?
                options[hold(Outer)] := i-> outer[i];
                options[hold(MaxLength)] := min(options[hold(MaxLength)],
                                                nops(outer));
            end_if;
        else
            options[hold(Outer)] := fp::unapply(options[hold(MaxPart)],hold(i));
        end_if;
        
        options[hold(MinLength)], options[hold(MaxLength)],
        options[hold(Inner)], options[hold(Outer)],
        options[hold(MinSlope)], options[hold(MaxSlope)];
    end_proc;

    parseOptions_OptionTypes :=
    table(
          hold(MinLength) = Type::NonNegInt,
          hold(MaxLength) = combinat::integerListsLexBuilder::typeIntBound,
	  hold(Length)    = Type::Union(Type::NonNegInt, DOM_NIL),
          hold(MinPart)   = Type::NonNegInt,
          hold(MaxPart)   = combinat::integerListsLexBuilder::typeIntBound,
	  hold(Inner)     = Type::Union(combinat::integerListsLexBuilder::type,
                                        DOM_NIL),
//	  hold(Inner)     = combinat::integerListsLexBuilder::type,
          hold(Outer)     = Type::Union(combinat::integerListsLexBuilder::typeCompositionBound,
                                        DOM_NIL),
	  hold(MinSlope)  = combinat::integerListsLexBuilder::typeSlopeBound,
	  hold(MaxSlope)  = combinat::integerListsLexBuilder::typeSlopeBound
         );

    //////////////////////////////////////////////////////////////////////////

    type := Type::ListOf(Type::NonNegInt);

    size := l -> nops(l);

    // Not very useful yet, as domains of this category could as well
    // call dom::count_FromGenerator right away. But if some time in
    // the future we can improve integerListsLexBuilder::count, we will
    // just have to make a modification here to let all domains of
    // this category get the benefit.

    count_integerListsLexBuilding := dom::count_FromGenerator;
    count := count_integerListsLexBuilding;

    //////////////////////////////////////////////////////////////////////////
    // wrapper methods around combinat::integerListsLexBuilder
    // first, Next, ...
    //////////////////////////////////////////////////////////////////////////

    first :=
    proc(n: Type::Integer)
        : Type::Union(dom::type, DOM_FAIL)
    begin
	combinat::integerListsLexBuilder::first(n, dom::parseOptions(args(2..args(0))));
    end_proc;

    last :=
    proc(n: Type::Integer) 
        : Type::Union(dom::type, DOM_FAIL)
    begin
	combinat::integerListsLexBuilder::last(l, dom::parseOptions(args(2..args(0))));
    end_proc;

    Next :=
    proc(l: dom::type) : Type::Union(dom::type, DOM_FAIL)
    begin
	combinat::integerListsLexBuilder::Next(l, dom::parseOptions(args(2..args(0))));
    end_proc;

    previous :=
    proc(l: dom::type) 
        : Type::Union(dom::type, DOM_FAIL)
    begin
        combinat::integerListsLexBuilder::previous(l, dom::parseOptions(args(2..args(0))));
    end_proc;

    generator :=
    proc(n: Type::Integer)
        : combinat::generators::type
        local constraints;
    begin
        constraints := dom::parseOptions(args(2..args(0)));
        combinat::generators::fromNext
        (combinat::integerListsLexBuilder::first(n, constraints),
         combinat::integerListsLexBuilder::Next, constraints);
    end_proc;

begin
end_category:

stdlib::LoadlibState["combinat::integerListsLexBuilding"] := TRUE:
