/*****************************************************************************
  $Id: $
  File:		lib/COMBINAT/integerVectors.mu
  Authors:	Nicolas M. Thiry <nthiery@users.sourceforge.net>
  Created:	16/02/2001
  License:	LGPL
*****************************************************************************/

domain combinat::integerVectors

    category combinat::integerListsLexBuilding;
    // category inlinedoc::cat;

    Name := "combinat::integerVectors";

    info := "Library 'combinat::integerVectors': utilities for integer vectors";

    introduction := "
 An integer vector of sum n with k parts is a list of k nonnegative
 integers which sum up to n.

 Integer vectors are naturally ordered lexicographically.
 Several functions of this library are directly inherited from
 combinat::integerListsLexBuilding, and accept generic
 constraints:

  - size of parts: 		MinPart = , MaxPart =
  - slope:			MinSlope = , MaxSlope =
  - containment:		Inner = , Outer =

 See the documentation of combinat::integerListsLexBuilding and the
 examples for details on those options.

 See also combinat::integerListsLexBuilding, combinat::structure.
";

    synopsis := "
 - type:			fixed length list of nonnegative integers

 - count    (n, k, <Options>)	number of integer vectors
 - list     (n, k, <Options>)	list of all integer vectors
 - generator(n, k, <Options>)	generator for all integer vectors

 - first    (n, k, <Options>)	first integer vector
 - last     (n, k, <Options>)	last integer vector
 - Next     (c, k, <Options>)	next integer vector
 - previous (c, k, <Options>)	previous integer vector

 Available options: MinPart, MaxPart, Inner, Outer, MinSlope, MaxSlope.
 See combinat::integerListsLexBuilding for their meaning.
";

    example := "
 All integer vectors of sum 5 and 3 parts, that are bounded by [4,1,1]:

	>> combinat::integerVectors(5, 3, Outer=[4,1,1]);
		[[4,1,0], [4,0,1], [3,1,1]]
";

    interface := dom::interface_Structure union 
        {hold(first), hold(last), 
         hold(Next), hold(previous)
        };

    type := Type::ListOf(Type::NonNegInt);

    parseOptions :=
    proc(k: Type::NonNegInt)
    begin
        dom::parseOptions_IntegerListLexBuilding
        (table(hold(Length)	= k,
               hold(MinPart)	= 0,
               hold(MaxPart)	= infinity,
               hold(Inner)	= NIL,
               hold(Outer)	= NIL,
	       hold(MinSlope)	= -infinity,
	       hold(MaxSlope)	= infinity
              ),
         args(2..args(0))
         ):
    end_proc;

    //////////////////////////////////////////////////////////////////////////
    // count
    //////////////////////////////////////////////////////////////////////////

    count :=
    proc(n: Type::Integer, k: Type::NonNegInt) : Type::NonNegInt
    begin
        if args(0) > 2 then
            // There is no efficient counter yet when there are
            // further constraints.
            return(dom::count_IntegerListLexBuilding(args()));
        end_if;

        if n >= 0 then
            binomial(n+k-1,n)
        else
            0;
        end_if;
    end_proc;

    Next :=
    proc(l: dom::type) : Type::Union(dom::type, DOM_FAIL)
    begin
	combinat::integerListsLexBuilder::Next(l, dom::parseOptions(nops(l), args(2..args(0))));
    end_proc;

begin
end_domain:

stdlib::LoadlibState["combinat::integerVectors"] := TRUE:
