/*****************************************************************************
  $Id: $
  File:		lib/COMBINAT/integerVectorsWeighted.mu
  Authors:	Nicolas M. Thiry <nthiery@users.sourceforge.net>
  Created:	16/02/2001
  License:	LGPL
*****************************************************************************/

domain combinat::integerVectorsWeighted

    category combinat::structure;
    // category inlinedoc::cat;

    Name := "combinat::integerVectorsWeighted";

    info := "Library 'combinat::integerVectorsWeighted': utilities for weighted integer vectors";

    introduction := "
 Let w be a weight vector, i.e. a list of k positive integers, and let
 n be a nonnegative integer. A w-weighted integer vector of sum n is a
 list c of k nonnegative integers such that

			c[1]*w[1] + ... + c[k]*w[k] = n.

 Example: [1,2,0] is a [3,1,1]-weighted integer vector of sum 5.

 If w[1]=...=w[k]=1, then a w-weighted integer vector of sum n is a
 usual integer vector of sum n (we suggest to use the much faster
 integerVectors library in this case). 
                         
 If x[1], ..., x[k] are variables with degree given by w[1],...,w[k]
 then a w-weighted integer vector of sum n is the exponnent vector of
 a monomial of degree n in those variables.

 Anyway, this package needs a full rewrite using generators, and the
 preceeding restriction ought to be taken care of automatically by the
 algorithm.

 See also combinat::integerVectors and combinat::structure.
";

    synopsis := "
 - type:		List of nonnegative integers

 - count(w, n)		number of w-weighted integer vectors of sum n
 - list(w, n)		list of all w-weighted integer vectors of sum n
 - generatingSeries(n, <z>)	generating series by sum for the w-weighted integer vectors
";

    interface := dom::interface_Structure union
        {generatingSeries};

    type := Type::ListOf(Type::NonNegInt);
    typeWeightVector := Type::ListOf(Type::PosInt);

    //////////////////////////////////////////////////////////////////////////
    // count
    //////////////////////////////////////////////////////////////////////////

    // Algorithm: computation from the generating series.
    // Isn't that actually even worst than generating them ?

    count :=
    proc(n: Type::NonNegInt, w: dom::typeWeightVector) : Type::NonNegInt
    begin
        coeff(series(dom::generatingSeries(w,hold(z)), hold(z), n+1), n);
    end_proc;

    //////////////////////////////////////////////////////////////////////////
    // generatingSeries
    //////////////////////////////////////////////////////////////////////////

    generatingSeries :=
    proc(w: dom::typeWeightVector, z=hold(z): DOM_IDENT)
    begin
        1/_mult((1-z^w[i]) $ i=1..nops(w));
    end_proc;

    //////////////////////////////////////////////////////////////////////////
    // list
    //////////////////////////////////////////////////////////////////////////

    list :=
    proc(n: Type::NonNegInt, l: dom::typeWeightVector)
        : Type::ListOf(dom::type)
        local recfun;
    begin

        recfun:=
        proc(n: Type::NonNegInt, l: dom::typeWeightVector)
            : Type::ListOf(dom::type)
            local w, d, result;
        begin
            result:=[];
            w:=op(l,nops(l));
            delete(l[nops(l)]);
            if l=[] then
                d:=n div w;
                if n mod w =0 then
                    return([[d]])
                else
                    return([]);		// bad branch ...
                end_if;
            end_if;
            for d from floor(n/w) downto 0 do
                result:=result. map(recfun(n-d*w, l), x->append(x,d))
            end_for;
            result;
        end_proc:
        
        recfun(n,l);
    end_proc;

begin
end_domain:

stdlib::LoadlibState["combinat::integerVectorsWeighted"] := TRUE:
