/*****************************************************************************
  $Id: $
  File:		lib/COMBINAT/partitions.mu
  Authors:	Nicolas M. Thiry <nthiery@users.sourceforge.net>
		Authors of the combinat package of MuPAD 2.0.0
		Authors of mu-EC
  Created:	16/02/2001
  License:	LGPL
*****************************************************************************/

domain combinat::partitions

    category combinat::integerListsLexBuilding;
    // category inlinedoc::cat;

    Name := "combinat::partitions";
    // Bugware: Name is only specified here as a workaround for a bug
    // in DOM_DOMAIN::printInfo, which is fixed in MuPAD 2.0.0 dev

    info := "Library 'combinat::partitions': utilities for integer partitions";

    introduction := "
 A partition of n is a decreasing list of positive integers which sum
 up to n.

 Partitions are naturally ordered lexicographically.
 Several functions of this library are directly inherited from
 combinat::integerListsLexBuilding, and accept generic
 constraints:

  - length of the partition:	MinLength = , MaxLength = , Length =
  - size of parts: 		MinPart = , MaxPart =
  - slope:			MinSlope = , MaxSlope =
  - containment:		Inner = , Outer =

 Note that specifying MinPart=0 allows for 0 parts. This can be
 useful in conjunction with the option Length.

 See also combinat::integerListsLexBuilding, combinat::structure.
";

    synopsis := "
 - type				Decreasing list of nonnegative integers
 - Constraints:			MinLength =, MaxLength =, Length =
				MinPart =, MaxPart =

 - count(n, <Constraints>)	number of partitions of n
 - list(n, <Constraints>)	list of all partitions of n

 - first(n, <Constraints>)	first partition of n
 - Next(p, <Constraints>)	Next partition

 - conjugate(p)			the conjugate partition of p
 - hooklengths(p)		the hook lengths of p (ALPHA)
 - toExp(p)			multiplicities of the parts of a partition
 - toExp(p,k)			multiplicities of the parts of a partition
 - fromExp(p)			partition from the multiplicities of its parts

 - conjugacyClassSize(p)	Size of the conjugacy class of the
                                symmetric group indexed by p (ALPHA)
 - centralizerSize(p)		Size of the centralizer of the conjugacy class
				of the symmetric group indexed by p (ALPHA)

 Not yet implemented:
 - last(n, <Constraints>)	last partition of n
 - previous(p, <Constraints>)	previous partition

 Deprecated features:
 - new is an alias for count to emulate calls of the form
   combinat::partitions(n) for backward compatibility
";

    interface := dom::interface_Structure union 
        {hold(size),
         hold(first), hold(last), 
         hold(Next), hold(previous),
         hold(conjugate), hold(hooklengths),
	 hold(toExp), hold(fromExp)
         };

    type := Type::ListOf(Type::NonNegInt);

    parseOptions :=
    proc()
    begin
        dom::parseOptions_IntegerListLexBuilding
        (table(hold(MinLength)	= 0,
               hold(MaxLength)	= infinity,
               hold(Length)	= NIL,
               hold(MinPart)	= 1,
               hold(MaxPart)	= infinity,
               hold(Inner)	= NIL,
               hold(Outer)	= NIL,
	       hold(MinSlope)	= -infinity,
	       hold(MaxSlope)	= 0
              ),
         args()
         ):
    end_proc;

    //////////////////////////////////////////////////////////////////////////
    // count
    //////////////////////////////////////////////////////////////////////////

    /**
...
    **/

    count :=
    proc(n: Type::NonNegInt) : Type::NonNegInt
        local k;
    begin
        if args(0) > 1 then
            // There is no efficient counter yet when there are
            // further constraints.
            dom::count_integerListsLexBuilding(args());
        end_if;
        if n<100 then // use Euler's pentagonal formula
            // the following avoids stacks problems under MACOS
            for k from n mod 25 to n step 25 do
                dom::count_Euler(k)
            end_for
        else // use Hardy-Ramanujan-Rademacher formula
            dom::count_HRR(n)
        end_if
    end_proc;

    /** number of partitions of n, using Euler's pentagonal formula

      p(n) + sum((-1)^k*(p(n-w(k))+p(n-w(-k))),k=1..infinity) = 0

      where w(k)=(3*k^2+k)/2. The asymptotic expansion of p(n) is:

                             1/2    /    / 2 n \1/2 \
                            3    exp| PI | --- |    |
                                    \    \  3  /    /
                     p(n) ~ -------------------------
                                      12 n

      Reference: G. Andrews, The Theory of Partitions, Addison-Wesley, 1976.
    **/
    count_Euler :=
    proc(n: Type::NonNegInt) : Type::NonNegInt
        local s,t,k;
        option remember;
    begin
        if n=0 then 
            return(1);
        end_if;
        s := 0;
        for k from 1 to n do
            t := (3*k^2-k)/2;
            if n < t then return(abs(s)) end_if;
            s := dom::count_Euler(n-t)-s;
            t := t+k;
            if n < t then return(abs(s)) end_if;
            s := dom::count_Euler(n-t)+s
        end_for;
    end_proc;

    /** Number of partitions of n using Hardy-Ramanujan-Rademacher formula.
	It is more efficient than Euler's pentagonal formula for n>=100.
    **/
    count_HRR := 
    proc(n: Type::NonNegInt) : Type::NonNegInt
        local s,k,t,f,x,X,C,zeroes,Akn,r,rt,ok;
        save DIGITS;
    begin
        // first compute an approximation to set DIGITS
        s:=float(1/(4*n*sqrt(3))*exp(PI*sqrt(2*n/3)));
        s:=ln(s)/ln(10.0);
        DIGITS:=ceil(s+ln(s)+10);
        C:= float(1/PI/sqrt(2));
        X:= genident();
        x:= genident();
        f:= subs( diff(sinh(PI/X*(2/3*(x-1/24))^(1/2))/(x-1/24)^(1/2),x),x=n);

        s:=0; zeroes:=1; r:=0.0; ok:=0;
        k:=1; 
        repeat
            if igcd(zeroes,k)=1 then
                Akn:=dom::HRR_A(k,n);
                if Akn=0 then
                    zeroes:=zeroes*k
                else
                    t:=r+C*Re(float( Akn*k^(1/2)*eval(subs(f,X=k)) ));
                    rt:=round(t); s:=s+rt; r:=t-rt;
                    DIGITS:=ceil(0.4343*ln(abs(t))+10);
                    userinfo(1,"iteration ".k.
                             " last term=".expr2text(rt).
                             " rest=".expr2text(r));
                    if rt=0 and abs(r)<0.1 then 
                        ok:=ok+1 
                    else
                        ok:=0
                    end_if
                end_if
            end_if;
            k:=k+1;
        until ok=3 end_repeat;
        s
    end_proc;

    HRR_A := 
    proc(k,n)
        local s,h;
    begin
        s:=0;
        for h from 0 to k-1 do
            if igcd(h,k)=1 then
                dom::HRR_omega(h,k)/2-n*h/k;
                s:=s+(if type(%)=DOM_INT then 
                          1 
                      else
                          cos(2*PI*mods(op(%))/op(%,2)) 
                      end_if);
            end_if
        end_for;
        s
    end_proc;

    HRR_omega := 
    proc(h,k) 
        local mu;
    begin
        _plus((frac(mu/k)-1/2)*(frac(h*mu/k)-1/2)$mu=1..k-1)
    end_proc;

    //////////////////////////////////////////////////////////////////////////
    // Former implementation of first
    //////////////////////////////////////////////////////////////////////////

    first_old := 
    proc(n: Type::NonNegInt,
         k=FAIL: Type::Union(Type::NonNegInt, DOM_FAIL), 
         m=FAIL: Type::Union(Type::NonNegInt, DOM_FAIL))
        : Type::Union(dom::type, DOM_FAIL)
    begin
        if testargs() and args(0) > 3 then
            error("Wrong number of arguments")
        end_if;
        if n=0 then
            return([]);
        end_if;
        if k=0 then
            return(FAIL);
        end_if;
        if m=FAIL then
            return([n])
        end_if;
        if k <> FAIL and m*k < n then
            return(FAIL);
        end_if;
        return([m $ n div m, n mod m]);
    end_proc;

    //////////////////////////////////////////////////////////////////////////
    // Former implementation of Next
    //////////////////////////////////////////////////////////////////////////

    Next_old :=
    proc(p: dom::type,
         k=FAIL: Type::Union(Type::NonNegInt, DOM_FAIL), 
         m=FAIL: Type::Union(Type::NonNegInt, DOM_FAIL))
        : Type::Union(dom::type, DOM_FAIL)
        local n, r, index;
    begin
        if testargs() and args(0) > 3 then
            error("Wrong number of arguments")
        end_if;

        // Actually, m is never used, since the partition p is
        // supposed to already satisfy the conditions, and the
        // operation only decreases the sizes of the part. 
        // m is recycled as a local variable.
        
        // Algo: find the first index where it's possible to decrease
        // an entry, and redistribute n:=l[index]+l[index+1]+... as
        // the smallest partition of n with parts of size at most
        // l[index]-1
        n:=0;
        for index from nops(p) downto 1 do
            n:=n+p[index];
            if p[index]>1 then
                m := p[index]-1;	// maximal value of the parts
                if k <> FAIL and m*(k-index+1) < n then
                    next;
                end_if;
                r := n mod m;		// size of the last part
                return([p[i] $ i=1..index-1,
                        m $ n div m,
                        (if r>0 then r else null() end_if)]);
            end_if;
        end_for;
        return(FAIL);
    end_proc;

    //////////////////////////////////////////////////////////////////////////
    // generator
    //////////////////////////////////////////////////////////////////////////

    generator_old :=
    proc(n: Type::NonNegInt,
         k=FAIL: Type::Union(Type::NonNegInt, DOM_FAIL), 
         m=FAIL: Type::Union(Type::NonNegInt, DOM_FAIL))
        : combinat::generators::type
    begin
        combinat::generators::fromNext(dom::first_old(args()),
                                       dom::Next_old, args(2..args(0)));
    end_proc;

    //////////////////////////////////////////////////////////////////////////
    // methods
    //////////////////////////////////////////////////////////////////////////

    /// The conjugate partition of p
    // Taken from mu-EC
    conjugate :=
    proc(p: dom::type) : dom::type
        local l;
    begin
        if (p = []) then
            [];
        else
            l := nops(p);
            [l $ p[l], l-i $ p[l-i]-p[l-i+1] $ i=1..l-1];
        end_if;
    end_proc;

    /// The list of the hook lengths of p
    // Should this list be sorted ?
    hooklengths :=
    proc(p: dom::type) : Type::ListOf(Type::PosInt)
        local pconj;
    begin
        pconj := dom::conjugate(p);
        [ p[i]-i + pconj[j]-j+1 $ j=1..p[i] $ i=1..nops(p) ];
    end_proc;

    /// Multiplicities of the parts of a partition
    //	Cf. the documentation from PARTPart2Exp.help in mu-EC
    toExp :=
    proc(p: dom::type, k=0: Type::Union(Type::NonNegInt, DOM_FAIL))
        : Type::ListOf(Type::NonNegInt)
        local a;
    begin
	if nops(p)>0 then
	    k := max(k, p[1]);
        end_if;
        a := [0 $ k];
        ( a[i] := a[i]+1 ) $ i in p;
        a;
    end_proc;

    /// partition from the multiplicities of its parts
    //  Cf. the documentation from PARTExp2Part.help
    fromExp :=
    proc(l: Type::ListOf(Type::NonNegInt))
        : dom::type
    begin
        revert([ i $ l[i] $ i=1..nops(l) ]);
    end_proc;

    /**
    centralizerSize(p) -> Type::PosInt

    The size of the centralizer of the conjugacy class of the
    symmetric group indexed by the partition p.

    Ref: A.~Kerber, "Algebraic Combinatoric Via Finite Group Action", 1.3 p 24
    **/
    centralizerSize :=
    proc(p: dom::type) : Type::PosInt
        local a;
    begin
        a:=combinat::partitions::toExp(p);
        _mult(i^a[i]*fact(a[i]) $ i=1..nops(a))
    end_proc;

    /**
    combinat::conjugacyClassSize(p) -> Type::PosInt

    The size of the conjugacy class of the symetric group
    indexed by the partition p.

    Ref: A.~Kerber, "Algebraic Combinatoric Via Finite Group Action", 1.3 p 24
    **/
    conjugacyClassSize :=
    proc(p: dom::type) : Type::PosInt
    begin
        _plus(op(p))! / dom::centralizerSize(p);
    end_proc;

begin
end_domain:

stdlib::LoadlibState["combinat::partitions"] := TRUE:
