/*****************************************************************************
  $Id: $
  File:		lib/COMBINAT/permutations.mu
  Authors:	Nicolas M. Thiry <nthiery@users.sourceforge.net>
  		Florent Hivert <Florent.Hivert@univ-mlv.fr>
                Vincent Proper (some things are reused from mu-EC)
  Created:	16/02/2001
  License:	LGPL
*****************************************************************************/

domain combinat::permutations

    inherits combinat::words;
    category combinat::structure;
    // category inlinedoc::cat;

    Name := "combinat::permutations";

    info := "Library 'combinat::permutations': utilities for permutations";

    introduction := "
 A permutation of a list l is a list of same size obtained by some
 reordering of the elements of l.
 Examples:
  - [1,4,3,2] and [2,3,1,4] are distinct permutations of [1,2,3,4].
  - [a,a,b,c] and [b,c,a,a] are distinct permutations of [a,c,b,a].
 A Standard permutation is a permutation of [1,2,...,n], for some integer n.
                                                            
 Note: there may be problems with lists of elements that do not have a
 proper internal order in the system kernel. Is it always guaranted
 that equal elements are contiguous in sort(l) ?

 See also combinat::structure and combinat::words.
";

    synopsis := "
 - type				any list (DOM_LIST)

 - count(n)			number of permutations of [1..n]
 - count(l)			number of permutations of the list l
 - count(l, Duplicate)		number of permutations of the list l, 
                                with duplicates
 - list(n)			list of all permutations of [1..n]
 - list(l)			list of all permutations of the list l
 - list(l, Duplicate)		list of all permutations of the list l
				with duplicates
 - generator(n)			generator for the permutations of [1..n]
 - generator(l)			generator for the permutations of the list l
 - generator(l, Duplicate)	generator for the permutations of the list l
                                with duplicates

 - first(n)			first permutation of [1..n]
 - first(l)			first permutation of the list l
 - last(n)			last permutation of [1..n]
 - last(n)			last permutation of the list l
 - Next(l)			next permutation of the list l
 - previous(l)			previous permutation of the list l

 - random(n)			random permutation of [1..n]
 - random(l)			random permutation of the list l

 - inversions(p)		list of the inversions of the permutation p
 - inversionsNumber(p)		number of inversions of the permutation p
 - descents(p)			list of the descents of the permutation p
 - majorIndex(p)		major-index of the permutation p
 - descentPolynomial(p,z)	descent polynomial in z of the permutation p

The following methods are only defined for standard permutations:
                                                        
 - toMatrix(sp)			permutation matrix of the permutation sp
 - inverse(sp)			inverse of the permutation sp
";

    interface := dom::interface_Structure union 
                 combinat::words::interface union 
        {hold(typeStandard),
	 hold(first), hold(last), 
         hold(Next), hold(previous),
	 hold(random),
	 hold(toMatrix), hold(inverse),
         hold(inversions),hold(inversionsNumber),
         hold(descents), hold(majorIndex), hold(descentPolynmial)};

    type := DOM_LIST;
    typeStandard := Type::ListOf(Type::PosInt); // Does not check that
                                                // the elements are distinct

    //////////////////////////////////////////////////////////////////////////
    // count
    //////////////////////////////////////////////////////////////////////////

    count :=
    proc(n: Type::Union(Type::NonNegInt, DOM_LIST))
        local list, r, s, i;
    begin

        if testargs() then
            if args(0) > 2 then
                error("wrong number of arguments !");
            end_if;
            if args(0) = 2 and args(2) <> Duplicate then
                error("Unknown option '".expr2text(args(2))."'") ;
            end_if;
        end_if;
  
        if domtype(n)=DOM_LIST then
            list:=n;
        else
            return(n!);
        end_if;

        if args(0) = 2 then
            return(nops(list)!);
        else
            // returns n!/i1!i2!...ik!, where i1, i2, ..., ik are the
            // numbers of elements in the list of each given value
            r:=nops(list)!;
            list:=sort(list);
            s:=1;			// number of successive equal elements
            for i from 2 to nops(list) do
                if list[i]=list[i-1] then
                    s:=s+1;
                    r:=r/s;
                else
                    s:=1;
                end_if;
            end_for;
            return(r);
        end_if ;
    end_proc;

    //////////////////////////////////////////////////////////////////////////
    // first
    //////////////////////////////////////////////////////////////////////////

    first :=
    proc(l: Type::Union(Type::NonNegInt,DOM_LIST)) : dom::type
    begin
        if testargs() then
            if args(0) > 2 then
                error("wrong number of arguments !");
            end_if;
            if args(0) = 2 and args(2) <> Duplicate then
                error("Unknown option '".expr2text(args(2))."'") ;
            end_if;
        end_if;

        if domtype(l)=DOM_LIST then
            if args(0) = 2 then		// We assume option duplicate is on
                return(sort(l));
            else
                return(sort([op(l)]));
            end_if
        else
            return([$1..l]);
        end_if
    end_proc;

    //////////////////////////////////////////////////////////////////////////
    // last
    //////////////////////////////////////////////////////////////////////////

    last :=
    proc(l: Type::Union(Type::NonNegInt,DOM_LIST)) : dom::type
    begin
        revert(dom::first(args()));
    end_proc;

    //////////////////////////////////////////////////////////////////////////
    // next
    //////////////////////////////////////////////////////////////////////////

    Next :=
    proc(p: dom::type)
        : Type::Union(dom::type, DOM_FAIL)
        local q, i, j;
    begin
        if testargs() and args(0) > 3 then
            error("Wrong number of arguments")
        end_if;

        // Look for the first strict climb p[i-1]<p[i] in the
        // permutation. Obviously, on the left of i, the elements are
        // sorted decreasingly. We look for the first element there
        // which is strictly smaller than p[i], and exchange it with
        // p[i]. We also reverse the order of the other elements on the
        // right of p. In the final list q, there is no strict climb
        // between i-1 and i, and only weak climbs on the left of i.
        for i from 2 to nops(p) do
            if p[i-1]<p[i] then
                q:=p;
                for j from i-2 downto 1 do
                    if p[j]>=p[i] then // '<=' make it work with duplicates
                        break;
                    else
                        q[i-1-j]:=p[j+1];
                    end_if;
                end_for;
                // p[j+1] is the first element on the left of i which
                // is bigger or equal to p[i]
                q[i]:=p[j+1];
                q[i-1-j]:=p[i];
                for j from j downto 1 do
                    q[i-j]:=p[j];
                end_for;
                return(q);
            end_if;
        end_for;
        return(FAIL); // There is no descent, so p was the last permutation
    end_proc;

    //////////////////////////////////////////////////////////////////////////
    // previous
    //////////////////////////////////////////////////////////////////////////

    previous :=
    proc(p: dom::type)
        : Type::Union(dom::type, DOM_FAIL)
        local q, i, j;
    begin
        if testargs() and args(0) > 3 then
            error("Wrong number of arguments")
        end_if;

        // Look for the last strict descent p[i]>p[i+1] in the
        // permutation. Obviously, on the right of i, the elements are
        // sorted increasingly. We look for the last element there
        // which is bigger than p[i], and exchange it with p[i]. We
        // also reverse the order the other elements on the right of
        // p. In the final list q, there is no strict descent between
        // i and i+1, and only weak descents on the right of i.
        for i from nops(p)-1 downto 1 do
            if p[i]>p[i+1] then
                q:=p;
                for j from i+2 to nops(p) do
                    if p[j]>=p[i] then // make it work with duplicates
                        break;
                    else
                        q[nops(p)-(j-i-2)]:=p[j-1];
                    end_if;
                end_for;
                // p[j-1] is the last element on the right of i which
                // is bigger or equal to p[i]
                q[i]:=p[j-1];
                q[nops(p)-(j-i-2)]:=p[i];
                for j from j to nops(p) do
                    q[nops(p)-(j-i-1)]:=p[j];
                end_for;
                return(q);
            end_if;
        end_for;
        return(FAIL); // There is no descent, so p was the last permutation
    end_proc;

    //////////////////////////////////////////////////////////////////////////
    // generator
    //////////////////////////////////////////////////////////////////////////

    generator :=
    proc(list: Type::Union(Type::NonNegInt,DOM_LIST)) 
        : combinat::generators::type
        option escape;
        local lindex, i, j, n;
    begin
        if testargs() then
            if args(0) > 2 then
                error("wrong number of arguments !");
            end_if;
            if args(0) = 2 and args(2) <> Duplicate then
                error("Unknown option '".expr2text(args(2))."'") ;
            end_if;
        end_if;
        if domtype(list)=DOM_LIST and 
           ((args(0) = 2) or not testtype(list, Type::ListOf(Type::Real))) then
            // Next and previous cannot deal with duplications, nor
            // with lists of elements that cannot be ordered by <. So,
            // we create a generator for the permutations of a list
            // lindex of integers corresponding the indices of the
            // elements of l.
            list:=sort(list);
            n:=nops(list);
            if args(0) = 2 then
                // With the option duplicate, the integers are simply 1..n
                lindex:=[$1..n];
            else
                // Without the option duplicate, we mimic the
                // repetitions of an element in l by repeating the
                // index of the first element having this value. E.g.
                // [a,b,b,b,c,d,d] -> [1,2,2,2,5,6,6]
                lindex:=[1$n];
                j:=1;
                for i from 2 to n do
                    if list[i-1] <> list[i] then
                        j:=i;
                    end_if;
                    lindex[i]:=j;
                end_for;
            end_if;

            return(combinat::generators::compose
                   (p->[list[i] $ i in p],
                    combinat::generators::fromNext(lindex, dom::Next)));
        end_if;
        return(combinat::generators::fromNext(dom::first(args()),
                                              dom::Next));
    end_proc;

    /// random(n) - random permutation of [1..n]
    /// random(l) - random permutation of the list l

    random :=
    proc(l: Type::Union(Type::NonNegInt,DOM_LIST)) : dom::type
	local beenthere, rnd, i, n;
    begin
        if domtype(l)<>DOM_LIST then
	    l:=[$1..l];
        end_if;
        n := nops(l);
	if n=0 then // special case since n=0 is an invalid argument for random
            return([]);
        end_if;
	beenthere:= [FALSE $ n];
	rnd:=random(n);	   // random generator for 0..n-1
        [ (repeat i:=rnd()+1 until not beenthere[i] end_repeat;
           beenthere[i]:=TRUE;
           l[i];
          ) $ j=1..n
        ];
    end_proc;

    //////////////////////////////////////////////////////////////////////////
    // Methods
    //////////////////////////////////////////////////////////////////////////

    /// toMatrix(sp) - permutation matrix of the standard permutation sp
    toMatrix :=
    proc(sigma: dom::typeStandard)
        local M, m, i;
    begin
        m:=nops(sigma);
        M:=Dom::Matrix()(m,m);
        for i from 1 to m do
            M[sigma[i],i]:=1;
        end_for;
        M;
    end_proc;


    /// inverse of the permutation
    inverse := 
    proc(x: dom::typeStandard)
	local result;
    begin
	result:=x; // just to create a list of same size
	(result[x[i]]:=i) $ i=1..n;
        result;
    end_proc;


    /// list of the inversions of a permutation
    inversions := 
    proc(p: dom::type): Type::ListOf(DOM_SET)
        local
              i,j,result, 
              len; 
    begin
        len := nops(p);
        result := [];
        for i from 1 to len-1 do
            for j from i+1 to len do
                if (p[i]>p[j]) then
                    result := append(result,{i,j});
                end_if;
            end_for;
        end_for;
        result;
    end_proc;


    /// number of inversions of a permutation
    inversionsNumber := 
    proc(p: dom::type): Type::NonNegInt
        local
              i,j,result, 
              len; 
    begin
        len := nops(p);
        result := 0;
        for i from 1 to len-1 do
            for j from i+1 to len do
                if (p[i]>p[j]) then
                    result := result + 1;
                end_if;
            end_for;
        end_for;
        result;
    end_proc;


    /// sign(p) - sign of a permutation
    sign := 
    proc(p: dom::type): Type::Integer
    begin
        if dom::inversionsNumber(p) mod 2 = 0 then 1 else -1; end_if;
    end_proc;

/**
    descents(p)
    Returns the descent List of the permutation p
 
    That is the list of the i for which p[i]>p[i+1]
**/
 
    descents :=
    proc(p: dom::type):  Type::ListOf(Type::PosInt)
        local result, i;
    begin
        result:=[];
        for i from 1 to nops(p)-1 do
            if p[i]>p[i+1] then
                result:=append(result,i);
            end_if
        end_for;
        result;
    end_proc;

/**
    majorIndex(p)
    Returns the major index of the permutation p (sum of the descents)
    References: Mac-Mahon (1915) reprint 1960
**/

    majorIndex :=
    proc(p: dom::type):  Type::NonNegInt
    begin
        _plus(op(dom::descents(p)));
    end_proc;


/**
    descentPolynomial(p, z)
 
    Returns the descent polynomial of the permutation p. That is the
    product of all the (z[p[1]]..z[p[i]]), where i are descents of p.
 
    References: Garsia and Stanton 1984
**/
 
    descentPolynomial :=
    proc(p: dom::type, z=hold(z): DOM_IDENT)
        local result, Pol, i;
    begin
        result:=1;
        Pol:=1;
        for i from 1 to nops(p)-1 do
            Pol:=Pol*z[p[i]];
            if p[i]>p[i+1] then
                result:=result*Pol;
            end_if
        end_for;
        result;
    end_proc;

begin
end_domain:

stdlib::LoadlibState["combinat::permutations"] := TRUE:
