/*****************************************************************************
  $Id: $
  File:		lib/COMBINAT/structure.mu
  Authors:	Nicolas M. Thiry <nthiery@users.sourceforge.net>
  Created:	16/02/2001
  License:	LGPL
*****************************************************************************/

category combinat::structure

    // category inlinedoc::cat;

    info := "Category 'combinat::structure': generic utilities for combinatorial structures";

    introduction := "
Many operations, like counting or generating, are common to all
combinatorial structures. This category defines a standardized
interface, and generic tools.
";

    synopsis := "
 Standardized entries (not all of them need to be implemented):
 - type				MuPAD type for the structure
 - testtype(p)			test if p has the correct type
 - check ???			thorough test ?
 - size(s)			size of the structure s
 - count(<size>)		number of structures of size <size>
 - generatingSeries(<ident>)	generating series by size
 - generator(<size>)		generator for all the structures of size <size>
 - list(<size>)			list of all the structures of size <size>
 - first(<size>)		first structure of size <size>, or FAIL
 - last(<size>)			last structure of size <size>, or FAIL
 - Next(s)			next structure of size <size>, or FAIL
 - previous(s)			previous structure of size <size>, or FAIL
 - rank(s)			rank of the structure s
 - unrank(<size>, n)		n-th structure of size <size>, or FAIL
 - random(<size>)		random structure of size <size>
 - _less(s1,s2)			comparison of the structures s1 and s2

 Moreover, calling directly combinat::mystructure(<size>) is a
 shortcut for combinat::mystructure::list(<size>). See
 combinat::partitions for an exception to this rule.

 Required entries:
 - type
 - generator or list

   Warning: if none of them is provided, the default implementations
   that are provided will call each other, and cause an infinite loop!

 Default implementations are provided for:
 - count
 - list
 - new
 - generator
";

    interface_Structure :=
    { hold(type), //hold(testtype),
      hold(count), hold(list), hold(generator)
    };

    interface := dom::interface_Structure; // to allow for reuse

    type;
    testtype  := (x) -> testtype(x,type);

    count     := dom::count_FromGenerator;
    list      := dom::list_FromGenerator;
    new	      := dom::list;
    generator := (if dom::first <> FAIL and dom::Next <> FAIL then
                      dom::generator_FromNext;
                  elif dom::last <> FAIL and dom::previous <> FAIL then
                      dom::generator_FromPrevious;
                  else
                      dom::generator_FromList;
                  end_if);

    //////////////////////////////////////////////////////////////////////////
    // Default implementations for count
    //////////////////////////////////////////////////////////////////////////

    count_FromGenerator :=
    proc() : Type::NonNegInt
    begin
        combinat::generators::count(dom::generator(args()));
    end_proc;

    //////////////////////////////////////////////////////////////////////////
    // Default implementations for list
    //////////////////////////////////////////////////////////////////////////

    list_FromGenerator :=
    proc() : Type::ListOf(dom::type)
    begin
        combinat::generators::toList(dom::generator(args()));
    end_proc;

    //////////////////////////////////////////////////////////////////////////
    // Default implementations for generator
    //////////////////////////////////////////////////////////////////////////

    generator_FromList :=
    proc() : combinat::generators::type
    begin
        combinat::generators::fromList(dom::list(args()));
    end_proc;

    generator_FromNext :=
    proc() : combinat::generators::type
    begin
        combinat::generators::fromNext(dom::first(args()),
                                       dom::Next);
    end_proc;

    generator_FromPrevious :=
    proc() : combinat::generators::type
    begin
        combinat::generators::fromNext(dom::last(args()),
                                       dom::previous);
    end_proc;

begin
end_category:

stdlib::LoadlibState["combinat::structure"] := TRUE:
