/*****************************************************************************
  $Id: subsets.mu,v 1.5 2001/11/21 21:25:22 nthiery Exp $
  File:		lib/COMBINAT/subsets.mu
  Authors:	Nicolas M. Thiry <nthiery@users.sourceforge.net>
  Created:	16/02/2001
  License:	LGPL
*****************************************************************************/

domain combinat::subsets

    category combinat::structure;
    // category inlinedoc::cat;

    Name := "combinat::subsets";

    info := "Library 'combinat::subsets': utilities for subsets of a set";

    introduction := "
 A subset of a set s is, well, a subset. This package serves both for
 subsets of a fixed size, or subsets of any size (i.e. the powerset).

 Right now, sets are generated lexicographically w.r.t. the internal
 order of the set provided as argument. Should we specify the order in
 which sets are returned ? Should we use the natural order of the
 elements instead ?
";

    synopsis := "
 Powerset of a set:

 - count(n)		number of subsets of {1..n}
 - count(s)		number of subsets of the set s
 - list(n)		list of all subsets of {1..n}
 - list(s)		list of all subsets of the set s
 - generator(n)		generator for the subsets of {1..n}
 - generator(s)		generator for the subsets of the set s

 - first(n)		first subset of {1..n}
 - first(s)		first subset of the set s
 - last(n)		last subset of {1..n}
 - last(s)		last subset of the set s

 Subsets of size k of a set:

 - count(n,k)		number of subsets of size k of {1..n}
 - count(s,k)		number of subsets of size k of the set s
 - list(n,k)		list of all subsets of size k of {1..n}
 - list(s,k)		list of all subsets of size k of the set s
 - generator(n,k)	generator for the subsets of size k of {1..n}
 - generator(s,k)	generator for the subsets of size k of the set s

 - first(n,k)		first subset of size k of {1..n}
 - first(s,k)		first subset of size k of the set s
 - last(n,l)		last subset of size k of {1..n}
 - last(s,k)		last subset of size k of the set s

 - ??? Next(subset, s)	Next subset of the set s
 - ??? previous(subset, s)	previous subset of the set s
";

    interface := dom::interface_Structure union 
        {hold(first), hold(last), 
         hold(Next), hold(previous)
         };

    type := DOM_SET;
    typeSet  := Type::Union(Type::NonNegInt, DOM_SET);
    typeSize := Type::Union(Type::NonNegInt, DOM_FAIL);

    //////////////////////////////////////////////////////////////////////////
    // count
    //////////////////////////////////////////////////////////////////////////

    count :=
    proc(set: dom::typeSet, k=FAIL: dom::typeSize)
        : Type::NonNegInt
        local n;
    begin
        if testargs() then
            if args(0) > 2 then
                error("Wrong number of arguments") ;
            end_if ;
        end_if;
        if not testtype(set, DOM_SET) then
            n:=set;
        else
            n:=nops(set);
        end_if;
        if k=FAIL then
            2^n
        else
            binomial(n,k);
        end_if;
    end_proc;

    //////////////////////////////////////////////////////////////////////////
    // first
    //////////////////////////////////////////////////////////////////////////
    
    first :=
    proc(set: dom::typeSet, k=FAIL: dom::typeSize)
        : Type::Union(dom::type, DOM_FAIL)
        local w;
    begin
        if testargs() then
            if args(0) > 2 then
                error("Wrong number of arguments") ;
            end_if ;
        end_if ;

        if domtype(set)<>DOM_SET then
            set:={$1..set}
        end_if;

        w:=combinat::subwords::first([$1..nops(set)], args(2..args(0)));
        if w=FAIL then
            FAIL
        else
            dom::fromIndexList(set, w)
        end_if;
    end_proc;

    //////////////////////////////////////////////////////////////////////////
    // last
    //////////////////////////////////////////////////////////////////////////
    
    last :=
    proc(set: dom::typeSet, k=FAIL: dom::typeSize)
        : Type::Union(dom::type, DOM_FAIL)
        local w;
    begin
        if testargs() then
            if args(0) > 2 then
                error("Wrong number of arguments") ;
            end_if ;
        end_if ;

        if domtype(set)<>DOM_SET then
            set:={$1..set}
        end_if;

        w:=combinat::subwords::last([$1..nops(set)], args(2..args(0)));
        if w=FAIL then
            FAIL
        else
            dom::fromIndexList(set, w)
        end_if;
    end_proc;

    //////////////////////////////////////////////////////////////////////////
    // fromIndexList
    //////////////////////////////////////////////////////////////////////////

    fromIndexList := 
    proc(set, indexList)
    begin
        {op(set, i) $ i in indexList};
    end_proc;
    //////////////////////////////////////////////////////////////////////////
    // generator
    //////////////////////////////////////////////////////////////////////////
    
    generator :=
    proc(set: dom::typeSet, k=FAIL: dom::typeSize)
        : combinat::generators::type
        option escape;
    begin
        if testargs() then
            if args(0) > 2 then
                error("Wrong number of arguments") ;
            end_if ;
        end_if;

        if domtype(set)<>DOM_SET then
            set:={$1..set}
        end_if;
        // We construct a generator for the subwords of [1..nops(n)]
        // and map each of them into a subset of set
        return(combinat::generators::compose
               (indexList->{op(set, i) $ i in indexList},
                combinat::subwords::generator([$1..nops(set)],
                                              args(2..args(0)))));
    end_proc;

begin
end_domain:

stdlib::LoadlibState["combinat::partitions"] := TRUE:
