/*****************************************************************************
  $Id: subwords.mu,v 1.4 2001/11/21 22:47:10 nthiery Exp $
  File:		lib/COMBINAT/subwords.mu
  Authors:	Nicolas M. Thiry <nthiery@users.sourceforge.net>
  Created:	16/02/2001
  License:	LGPL
*****************************************************************************/

domain combinat::subwords
    category combinat::structure;
    // category inlinedoc::cat;

    Name := "combinat::subwords";

    info := "Library 'combinat::subwords': utilities for subwords of a word";

    introduction := "
 A subword of a word w is a word obtaining by deleting the letters at
 some of the positions in w.

 Example:
  - [b,n,j,r] is a subword of [b,o,n,j,o,u,r]
  - [n,r] is a subword of [b,o,n,j,o,u,r]
  - [r,n] is not a subword of [b,o,n,j,o,u,r]
  - [n,n,u] is not a subword of [b,o,n,j,o,u,r]
  - [n,n,u] is a subword with repetitions of [b,o,n,j,o,u,r]

 See also: combinat::words, combinat::structures.
";

    synopsis := "
 - type		any list (DOM_LIST)

 All subwords of a word w:

 - count(w)			number of subwords
 - list(w)			list of all subwords
 - generator(w)			generator for the subwords

 - first(w)			first subword
 - last(w)			last subword

 Subwords of length k of a word w:

 - count(w,k)			number of subwords
 - list(w,k)			list of all subwords
 - generator(w,k)		generator for the subwords

 - first(w,k)			first subword
 - last(w,k)			last subword

 Subwords of length k of a word w, with repetitions allowed:

 - list(w,k, Repeat)		list of all subwords
 - generator(w,k, Repeat)	generator for the subwords
";

    interface := dom::interface_Structure union 
        {hold(first), hold(last)
         };

    type := DOM_LIST;
    typeSize := Type::Union(Type::NonNegInt, DOM_FAIL);

    //////////////////////////////////////////////////////////////////////////
    // count
    //////////////////////////////////////////////////////////////////////////

    count :=
    proc(w: dom::type, k=FAIL: dom::typeSize)
        : Type::NonNegInt
    begin
        if testargs() then
            if args(0) > 2 then
                error("Wrong number of arguments") ;
            end_if ;
        end_if;
        if k=FAIL then
            2^(nops(w))
        else
            binomial(nops(w),k);
        end_if;
    end_proc;

    //////////////////////////////////////////////////////////////////////////
    // first
    //////////////////////////////////////////////////////////////////////////
    
    first :=
    proc(w: dom::type, k=FAIL: dom::typeSize)
        : Type::Union(dom::type, DOM_FAIL)
    begin
        if testargs() then
            if args(0) > 2 then
                error("Wrong number of arguments") ;
            end_if ;
        end_if ;

        if k=FAIL then
            return([]);
        elif k>nops(w) then
            return(FAIL);
        end_if;
        return([op(w, 1..k)]);
    end_proc;

    //////////////////////////////////////////////////////////////////////////
    // last
    //////////////////////////////////////////////////////////////////////////
    
    last :=
    proc(w: dom::type, k=FAIL: dom::typeSize)
        : Type::Union(dom::type, DOM_FAIL)
    begin
        if testargs() then
            if args(0) > 2 then
                error("Wrong number of arguments") ;
            end_if ;
        end_if;

        if k=FAIL then
            return(w);
        elif k>nops(w) then
            return(FAIL);
        end_if;
        return([op(w, nops(w)-k+1..nops(w))]);
    end_proc;
    
    //////////////////////////////////////////////////////////////////////////
    // generator
    //////////////////////////////////////////////////////////////////////////
    
    generator_nk :=
    proc(n: Type::NonNegInt, k: Type::NonNegInt)
        : combinat::generators::type
        option escape;
        local subword, first_time, dif;
        /* 
           generator_nk(n,k)
		generator for the subwords of length k of the word [1..n]

           The subwords are generated lexicographically. Algorithm:
           find in the current subword the biggest element i that can
           be increased without collision, increase it, and reset all
           the bigger elements to be the immediate successors i+1,
           i+2, ... of i. For example: with n=9 and l=[1,3,7,8,9], we
           find that i is 3, and so the next subword is [1,4,5,6,7].

           The algorithm is actually slightly more general. We
           generate all words of length k of [1..n] such that
           subwords[i]+dif <= subwords[i+1].

	   Setting dif=1, this generates all the subwords, and setting
	   dif=0, all the subwords with repetitions.
	*/

    begin
        if testargs() then
            if args(0) > 3 or (args(0)=3 and args(3) <> hold(Repeat)) then
                error("Wrong number of arguments");
            end_if ;
        end_if;

        if args(0)=3 then // repetitions are allowed
            dif := 0;
        else
            dif := 1;
        end_if;
        
        if k=0 then
            return(combinat::generators::repeater([],1));
        end_if;
        if n<1+(k-1)*dif then
            return(combinat::generators::empty());
        else
            subword:=[ 1+i*dif $ i=0..k-1];
        end_if;
        
        first_time := TRUE;
        return(proc()
                   local i,j;
               begin
                   if first_time then
                       first_time:=FALSE;
                       return(subword);
                   end_if;
                   
                   // Find the biggest element that can be increased
                   if subword[k]<n then
                       subword[k]:=subword[k]+1;
                       return(subword);
                   end_if;
                   for i from k-1 downto 1 do
                       if subword[i]+dif<subword[i+1] then
                           subword[i]:=subword[i]+1;
                           // reset the bigger elements
                           for j from 1 to k-i do
                               subword[i+j]:=subword[i]+j*dif;
                           end_for;
                           return(subword);
                       end_if;
                   end_for;
                   return(FAIL);
               end_proc)
    end_proc;

    generator :=
    proc(w: dom::type, k=FAIL: dom::typeSize)
        : combinat::generators::type
        option escape;
        local generator;
    begin
        if testargs() then
            if args(0) > 3 or (args(0)=3 and args(3) <> hold(Repeat)) then
                error("Wrong number of arguments") ;
            end_if ;
        end_if;

        // Case 1: build recursively a generator for all subwords
        if k=FAIL then
            return(combinat::generators::concat(dom::generator(w,k) $k=0..nops(w)));
        end_if;

        // Case 2: build a generator for the subwords of length k
        // We construct a generator for the subwords of [1..nops(n)]
        // and map each of them into a subword of w
        if args(0)<3 then
            generator:=dom::generator_nk(nops(w),k);
        else
            generator:=dom::generator_nk(nops(w),k, hold(Repeat));
        end_if;
        return(combinat::generators::compose((subword)->[op(w,i) $ i in subword],
                                   generator));
    end_proc;

begin
end_domain:

stdlib::LoadlibState["combinat::subwords"] := TRUE:
