/*****************************************************************************
  $Id: $
  File:		lib/COMBINAT/tableaux.mu
  Authors:	Nicolas M. Thiry <nthiery@users.sourceforge.net>
  		Florent Hivert <Florent.Hivert@univ-mlv.fr>
  Created:	31/10/2001
  License:	LGPL
*****************************************************************************/

domain combinat::tableaux

    // category inlinedoc::cat;
    category combinat::structure;

    Name := "combinat::tableaux";

    info := "Library 'combinat::tableaux': utilities for tableaux";

    introduction := "
 The tableau are stored a lists of lists of ordered objects. 

 Example of tableau:          [[c],[b,c],[a,b,b]]
 Example of filling:          [[a],[c,a],[b,a,b]]
 Example of tabloid:          [[a],[b,c],[a,b,b]]
 Example of standard tableau: [[6],[4,5],[1,2,3]]

 See also combinat::structure
";

    synopsis := "
 - type:		list of lists of positive integers

 - shape(T):		the shape of the tableau T
 - toWord(T):		the word corresponding to the tableau T
 - fromWord(w):		left inverse function of toWord
 - conjugate(f):	the conjugate of the filling f

 - count(lambda)	number of standard Tableaux of shape lambda
 - list(lambda)	list of the standard Tableaux of shape lambda
 - generator(lambda)	generator for the standard Tableaux of shape lambda
 In the future the last three methods will work also for general tableau. 
                                                               
 - canonical(lambda)	canonical standard tableau of shape lambda
                        (filled by row)
 - columnCanonical(lambda) column canonical standard tableau of shape lambda
                        (filled by columns)

 - indexFilling(sT)	index filling of the standard tableau T
 - cocharge(sT)		cocharge of the standard tableau T
 In the future the last two methods will work also for general tableau. 
";

    interface := dom::interface_Structure union 
        {
         hold(canonical), hold(columnCanonical), 
	 hold(indexFilling), hold(cocharge),
         hold(conjugate), hold(toWord), hold(fromWord), hold(shape)
         };

    typeTableau  := Type::ListOf(DOM_LIST); // does not check order constraints
    typeStandard := Type::ListOf(Type::ListOf(Type::PosInt));  // idem
    typeTabloid  := Type::ListOf(DOM_LIST); // idem
    typeFilling  := Type::ListOf(DOM_LIST);
    type         := dom::typeTableau;

    //////////////////////////////////////////////////////////////////////////
    // generator
    //////////////////////////////////////////////////////////////////////////

    generator :=
    proc(lambda: combinat::partitions::type)
        : combinat::generators::type
        option escape;
        local n, corow, row, slambda, gen, thedom;
    begin
        if nops(lambda)=0 then
            return(combinat::generators::fromList([[]]));
        end_if;

        thedom := dom; // Usual hack since dom is not lexically bound
        	       // inside subfunctions.

        n:=_plus(op(lambda));
        row:=1;
        corow:=nops(lambda);
        slambda := [op(lambda, 1..nops(lambda)-1)];
        if (lambda[nops(lambda)])>1 then
            slambda:=append(slambda,lambda[nops(lambda)]-1);
        end_if;
        gen:=dom::generator(slambda);
        // Invariant:
        //  - gen is a valid generator for the tableaux with one less
        //    cell on the row row; those are of shape slambda
        //  - corow = nops(lambda)-row+1
        proc() : Type::Union(dom::typeStandard, DOM_FAIL)
            local t;
        begin
            while (t:=gen())=FAIL do
                repeat
                    row:=row+1;
                    corow:=corow-1;
                    if corow=0 then
                        return(FAIL);
                    end_if;
                until lambda[corow]>lambda[corow+1] end;
                slambda := lambda;
                slambda[corow] := slambda[corow]-1;
                gen := thedom::generator(slambda);
            end_while;
            if nops(slambda) < nops(lambda) then // better test ?
                t:=_concat([[n]],t);
            else
                t[row]:=append(t[row], n);
            end_if;
            return(t);
        end_proc;
    end_proc;

    canonical :=
    proc(lambda: combinat::partitions::type): dom::typeStandard
        // Canonical tableau of shape lambda, filled with 1..n from
        // the bottom row to the upper row.
        //
        // Example: canonical([3,2]) = [[4,5], [1,2,3]]
    local result, i, j;
    begin
        result := [];
        i:=_plus(op(lambda));
        for j in revert(lambda) do
            result := append(result, [ $ i-j+1..i ] );
            i:=i-j;
        end_for;
        result;
    end_proc;

    columnCanonical :=
    proc(lambda: combinat::partitions::type): dom::typeStandard
        // column-canonical standard tableau of shape lambda, filled
        // with 1..n from the leftmost column to the rightmost column
        //
        // Example: columnCanonical([3,2]) = [[2,4], [1,3,5]]
    local result, i, row, col;
    begin
        if nops(lambda)=0 then
            return([]);
        end_if;
        // Allocates directly the space required for the resulting tableau
        // That's probably faster than appending the elements to the rows
        // one by one
        result := [[NIL$i] $ i in revert(lambda)];
        i:=1;
        for col from 1 to lambda[1] do
            for row from nops(result) downto 1 do
                if col>nops(result[row]) then 
                    break;
                end_if;
                result[row][col]:=i;
                i:=i+1;
            end_for;
        end_for;
        result;
    end_proc;


    indexFilling :=
    proc(S: dom::type) : dom::typeFilling
        // Returns the index filling I(S) of the standard tableau S.
        // Reference: Higher Specht Polynomials for the symmetric group
        // and the wreath product, S. Ariki, T. Terasoma, H. Yamada.
        //
        // Example:
        //   indexFilling([[3,5],[1,2,4]]) = [[1, 2], [0, 0, 1]]
        local n, result, row_indexes, previous_col, row, col, i, index;
    begin
        n := _plus(op(map(S, nops)));
        result := S; 		// Tableau of same shape at S
        
        row_indexes := [1 $ nops(S)-1,2];
        previous_col :=1;
        index:=0;
        result[nops(result)][1]:=index;
        for i from 2 to n do
            // Invariants:
            //  - previous_col is the column of the cell containing i-1 in S
            //  - the cells [r,c] of S with c<row_indexes[r] contains 1..i-1
	    //  - result is properly filled in the cells containing 1..i-1 in S

            // Look for the position of i in S
            for row from nops(S) downto 1 do
                col := row_indexes[row];
                if col <= nops(S[row]) and
                   S[row][col] = i then
                    row_indexes[row]:=col+1;
                    break;
                end_if;
            end_for;
            // Since S is standard, we know that S[row][col]=i;
            if S[row][col]<>i then
                error("the tableau is not standard");
            end_if;
            if col <= previous_col then
                index:=index+1;
            end_if;
            result[row][col] := index;
            previous_col := col;
        end_for;
        result;
    end_proc;

    cocharge :=
    proc(S: dom::typeStandard) : Type::NonNegInt
        // The cocharge of the tableau S (or the charge ???)
        // Reference:
    begin
        _plus(op(map(dom::indexFilling(S), op)));
    end_proc;


    fromWord :=
    proc(p: combinat::words::type) : dom::typeFilling 
        local T, i;
    begin
        if nops(p)=0 then
            return([]);
        end_if;
        T:=[[p[1]]];
        for i from 2 to nops(p) do
            if p[i-1]>p[i] then
                T:=append(T,[p[i]]);
            else
                T[nops(T)]:=append(T[nops(T)], p[i]);
            end_if;
        end_for;
        T;
    end_proc;

    toWord :=
    proc(T: dom::typeFilling) : combinat::words::type
    begin
        map(T, op);
    end_proc;

    shape :=
    proc(T: dom::typeFilling)
    begin
        revert(map(T, nops));
    end_proc;

    conjugate :=
    proc(T: dom::typeFilling) : dom::typeFilling
        local result, col, row, Row;
    begin
        if nops(T)=0 then
            return([]);
        end_if;
        result := [];
        for col from nops(T[nops(T)]) downto 1 do
            row := nops(T);
            Row := [];
            while row >=1 and col <= nops(T[row]) do
                Row := append(Row, T[row][col]);
                row := row - 1;
            end_while;
            result := append(result, Row);
        end_for;
        result;
    end_proc;
    

begin
end_domain:

stdlib::LoadlibState["combinat::standardTableaux"] := TRUE:
