/*****************************************************************************
  $Id: $
  File:		lib/COMBINAT/words.mu
  Authors:	Nicolas M. Thiry <nthiery@users.sourceforge.net>
  		Florent Hivert <Florent.Hivert@univ-mlv.fr>
  Created:	16/02/2001
  License:	LGPL
*****************************************************************************/

domain combinat::words

    category combinat::structure;
    // category inlinedoc::cat;

    Name := "combinat::words";

    info := "Library 'combinat::words': utilities for words";

    introduction := "
 Let A be an alphabet. A word of length k over A is a list of k
 elements of A.

 So far, an alphabet A can be represented by a set (DOM_SET) a list
 (DOM_LIST). The order between the letters is specified by the
 order of the operands in A. The result is undefined when A contains
 twice the same letter.

 See also combinat::structure.
";

    synopsis := "
 - type			any list (DOM_LIST)

 - count(A, k)			number of words of length k over A
 - count(n, k)			number of words of length k over {1..n}
 - list(A, k)			list of all words of length k over A
 - list(n, k)			list of all words of length k over {1..n}
 - generator(A, k)		generator for the words of length k over A
 - generator(n, k)		generator for the words of length k over {1..n}

 - evaluationTable(word)	evaluation of word, as a table (ALPHA)
 - evaluationPartition(word)	evaluation of word, as a partition (ALPHA)

 - alphabetToOrder(A)		ordering function (<) of A
 - standard(word)		standard permutation of word
 - standard(word, A)		standard permutation of word wrt. A 
 - standard(word, o)		standard permutation of word wrt. the order o 

";

    interface := dom::interface_Structure union
        {//hold(evaluationTable), hold(evaluationPartition),
         hold(alphabetToOrder), hold(standard)
         };


    type := DOM_LIST;
    alphabetType := Type::Union(DOM_LIST, DOM_SET, Type::NonNegInt);
    orderType := Type::Union(DOM_PROC, DOM_FUNC_ENV);

    //////////////////////////////////////////////////////////////////////////
    // count
    //////////////////////////////////////////////////////////////////////////

    count :=
    proc(n: dom::alphabetType, k: Type::NonNegInt) : Type::NonNegInt
    begin
        if testargs() and args(0) <> 2 then
            error("Wrong number of arguments") ;
        end_if ;

        if not testtype(n, Dom::Integer) then
            n:=nops(n);
        end_if;
        return(n^k);
    end_proc;

    //////////////////////////////////////////////////////////////////////////
    // generator (special case of the generator of combinat::cartesian)
    //////////////////////////////////////////////////////////////////////////

    generator :=
    proc(A: dom::alphabetType, k: Type::NonNegInt)
        : combinat::generators::type
    /* 
       generator(A, k) 		generator of all words of length k of A
       generator(n, k)	 	generator of all words of length k of [1..n]

       Each call of the generator yields the following word
       lexicographically or FAIL if there is no more. There is
       absolutely no waranty if the generator is called again after
       FAIL was returned.

       Is the result reasonable when a A contains twice an element ? */
        local word, word_int, n, first_time;
        option escape;
    begin
        if testargs() and args(0) <> 2 then
            error("Wrong number of arguments") ;
        end_if ;

        if testtype(A, Dom::Integer) then
            A:=[$1..A];
        end_if;
        n:=nops(A);
        if k=0 then
            return(combinat::generators::repeater([],1));
        end_if;
        if n=0 then
            return(combinat::generators::empty());
        end_if;
        word_int:=[1 $ k];
        word:=[op(A,1) $ k];
        first_time:=TRUE;
        // Invariants:
        //  - word_int is a word of k integers in [1..n]
        //  - word is the corresponding word of k elements in A
        //  - first_time is true if and only if it's the first run
        return(proc(): Type::Union(dom::type, DOM_FAIL)
                   local index;
               begin
                   if first_time then
                       first_time:=FALSE;
                       return(word);
                   end_if;
                   for index from k downto 1 do
                       if word_int[index]<n then
                           word_int[index]:=word_int[index]+1;
                           word[index]:=op(A, word_int[index]);
                           return(word);
                       else 
                           word_int[index]:=1;
                           word[index]:=op(A,1);
                       end_if;
                   end_for;
                   return(FAIL);
               end_proc)
    end_proc;

    //////////////////////////////////////////////////////////////////////////
    // Methods
    //////////////////////////////////////////////////////////////////////////

    /**
	evaluationTable(word) - evaluation of word, as a table

        This is a table such that for each letter l appearing in word,
	t[l] is the number of occurences of l in word.

	Example:
		combinat::words::evaluationTable([b,a,d,b,c,d,b]);
			table(a=1, b=3, c=1, d=2)

    **/
    evaluationTable :=
    proc(word: dom::type) : DOM_TABLE
        local l,t;
    begin
	t:=table();
        for l in word do
            if contains(t,l) then
                t[l] := t[l]+1;
            else
                t[l] := 1;
            end_if;
        end_for;
	t;
    end_proc;

    /**
	evaluationPartition(word) - evaluation of word, as a partition

        This is the list of number of occurences of each letters in
	word, sorted decreasingly.

	Example:
		combinat::words::evaluationPartition([b,a,d,b,c,d,b]);
			[3,2,1,1]
    **/
    evaluationPartition :=
    proc(word: dom::type) : combinat::partitions::type
    begin
        revert(sort(map([op(dom::evaluationTable(word))], op, 2)));
    end_proc;


    //////////////////////////////////////////////////////////////////////////
    // Standardization
    //////////////////////////////////////////////////////////////////////////

    alphabetToOrder := 
    proc(A: dom::alphabetType) : dom::orderType
        option escape;
        local rank;
    begin
        if testtype(A, Dom::Integer) then
            _less;
        else
            (rank(op(A,i)):=i)$i=1..nops(A);
            (x,y) -> (rank(x) < rank(y));
        end_if;
    end_proc;

    standard :=
    proc(word         : dom::type, 
         order = FAIL : Type::Union(dom::orderType, dom::alphabetType, DOM_FAIL)
        ) : Type::ListOf(Type::PosInt)
        local t, k, keys, tmp, offset;
    begin
        if order = FAIL
            then order := _less;
        elif testtype(order, dom::alphabetType)
            then order := dom::alphabetToOrder(order);
        end_if;

        t:=dom::evaluationTable(word);
        
        keys := sort(map([op(t)], op, 1), order);
        offset:=0;
        tmp:=0;
        for k in keys do
            tmp   := t[k];
            t[k]  := offset;
            offset := offset+tmp;
        end_for;
        
        map(word, l->(t[l]:=t[l]+1));
    end_proc;

begin
end_domain:

stdlib::LoadlibState["combinat::words"] := TRUE:
