alias(DAT(x)=detools::data[x],
      CDAT(x)=contains(detools::data,x)):

/*
    For CKsolve, currently, it is assumed that there is one dependent variable 
    to worry about (call it u), and several indepedent variables (call them
    x_1,x_2,...,x_n,t).
    The first argument will be a differential expression.  If the 
     expression is order n, it will be necessary to be able to solve for 
     the nth derivative of u with respect to x_k (for some k) and have
     the solution be of order < n.
    The second (optional) argument will be an integer; the degree of the 
     resulting series.  If this argument is given, the degree will be 5.
    The third (optional) argument is the center of the resulting power
     series (a list), by default it is [0,0].
    The fourth (optional) argument are initial conditions.
     This should be a list with two elements; 
      The first element should be the initial surface, it should be of
       the form x_k = a (Note that in this case it will be necessary to
       solve the differential expression for u_{x_k...x_k}, and that
       a should be the kth component of the center).
      The second element should be a list of n expressions, indicating
       u, u_{x_k}, ..., evaluated at x_k = a.
*/

detools::CKsolve :=
    proc()
     local pde, DE:DOM_DOMAIN, DV:DOM_DOMAIN, dvar, ivars, numvars:DOM_INT, 
	   deorder:DOM_INT, ctr:DOM_LIST, haveCtr:DOM_BOOL, deg:DOM_INT, 
	   haveDeg:DOM_BOOL, haveInitConds:DOM_BOOL, makeIndexList:DOM_PROC,
	   npde, surface, init, der:DOM_PROC, dtable:DOM_TABLE, 
	   milist:DOM_LIST, vtable:DOM_TABLE, value:DOM_PROC, idval:DOM_PROC, 
	   tmpvars:DOM_LIST, sol, j:DOM_INT, optarg, i:DOM_INT, l;
     begin
	 // Analysis of input
	 detools::getdata(testargs(), args());
	 if DAT("System") then error("cannot handle systems") end_if;
	 // First, a utility program.
	 // A method to create index lists.  Stolen from DVarCat.mu, with the 
	 // order of the elements of the lists reversed. 
	 // q stands for the order of the elements of the lists, and n for 
	 // the length of the lists.  All lists of length n whose order is 
	 // 1<= order<= q will be created in increasing order.
	 makeIndexList :=
	     proc(q:DOM_INT, n:DOM_INT):DOM_LIST
	     local ilist:DOM_LIST,nlist:DOM_LIST,j:DOM_INT,k:DOM_INT;
	     begin
		 nlist := [];
		 for j from 1 to q do
		     ilist := [[]];
		     for k from 1 to (n-1) do
			 ilist := map(ilist,
			      ((l,j) -> ([i].l $ i=0..(j-_plus(op(l))))),j);
		     end_for;
		     ilist := map(ilist, ((l,j) -> ([j-_plus(op(l))].l)), j);
		     nlist := nlist.ilist;
		 end_for;
		 return(nlist);
	     end_proc;
	 // Now, get some needed information.
	 // The equation to solve:
	 pde := DAT("Eqs")[1];
	 // The input domain:
	 DE := DAT("DFin");
	 // The differential variables domain
	 DV := DAT("DVin");
	 // The dependent variable
	 dvar := expr(op(DV::depVariables));
	 if nops(dvar) > 1 then
	    error("Cannot handle more than one function");
	 end_if;
	 // The independent variables:
	 ivars := map(DV::indVariables, expr);
	 // The number of independent variables:
	 numvars := DV::numIndVar;
	 // The order of the differential equation:
	 deorder := DE::order(pde);
	 // Check for optional arguments:
	 ctr := [0$numvars]; haveCtr := FALSE;
	 deg := ORDER; haveDeg := FALSE;
	 haveInitConds := FALSE;
	 if nops(DAT("Parms")) > 0 then
	    for j from 1 to nops(DAT("Parms")) do
		optarg := DAT("Parms")[j];
		// See if the optional parameter is the degree:
		if domtype(optarg) = DOM_INT then
		    if haveDeg then
			error("Illegal optional arguments");
		    else
			deg := optarg;
			haveDeg:= TRUE;
		    end_if;
		elif domtype(optarg) = DOM_LIST then
		    if nops(optarg) = numvars and
		       ({op(map(optarg, domtype))} union {DOM_INT, DOM_IDENT})
		                        = {DOM_INT, DOM_IDENT} then
		       if haveCtr then
			    error("Illegal optional argument");
		       else
			   ctr := optarg;
			   haveCtr := TRUE;
		       end_if;
		    elif (nops(optarg) = 2) 
		             and (domtype(optarg[2]) = DOM_LIST)
			     and (nops(optarg[2]) = deorder)  then
		       if haveInitConds then
			  error("Illegal optional argument");
		       else
		          init := optarg;
			  haveInitConds := TRUE;
		       end_if;
		     end_if;
		else
		    error("Illegal optional arguments");
		end_if;
	    end_for;
	 end_if;
	 // Now, put the pde in the form u_{tt...} = lower order stuff
	 npde := FAIL;
	 if haveInitConds then
	    surface := init[1];
	    i := contains(ivars, op(surface,1));
	    if not (op(surface,2) = ctr[i]) then
		error("Center of series not on initial surface");
	    end_if;
	    npde := DE::solve(pde, dvar([ivars[i]$deorder]));	    
	    init := init[2];
         else
	    i := 0;
	    while npde = FAIL and i < numvars do
		i := i+1;
		npde := DE::solve(pde, dvar([ivars[i]$deorder]));
	    end_while;
	 end_if;
	 if npde = FAIL then
	     error("Cannot solve for a highest order term of the right form");
	 end_if;
	 if DE::order(npde) = deorder then
	     error("PDE contains more than one term of highest order");
	 end_if;
	 // Now, a method to find necessary derivatives.
	 // l will represent a multi-index indicating the 
	 // derivative desired.
	 der :=
	     proc(l:DOM_LIST)
	     local ord:DOM_INT, k:DOM_INT, sol;
	     begin
		 ord := _plus(op(l));
		 // If the order of the derivative of the variable (t)
		 // is less than or equal to deorder, and no derivatives of
		 // other variables are taken, get the derivative from eqns,
		 // otherwise take the derivative of a lower order term.
		 if haveInitConds then
		     if l[i] <= deorder then
			 if l[i] = ord then
			     if l[i] < deorder then
				 return(init[ord+1]);
			     else
				 return(npde);
			     end_if;
			 end_if;
			 k:= 1;
			 while l[k] = 0 or k = i do
			     k := k + 1;
			 end_while;
			 l[k] := l[k] - 1;
			 return(DE::totalDiff(dtable[l],k));
		     else
			 l[i] := l[i] - 1;
			 return(DE::totalDiff(dtable[l],i));
		     end_if;
		 else
		     if l[i] = deorder then
			 if ord = deorder then
			     return(npde);
			 else
			     k := 1;
			     while l[k] = 0 or k = i do
				 k := k + 1;
			     end_while;
			     l[k] := l[k] - 1;
			     return(DE::totalDiff(dtable[l], k));
			 end_if;
		     elif l[i] > deorder then
			 l[i] := l[i] - 1;
			 return(DE::totalDiff(dtable[l], i));
		     else
			 return(genident("A"));
		     end_if;
		 end_if;
	     end_proc; /* der */
	// Now, find the derivatives.
	dtable := table();
	if haveInitConds then
	    dtable[[0$numvars]] := init[1];
	else
	    dtable[[0$numvars]] := genident("A");
	end_if;
	milist := makeIndexList(deg, numvars);
	for l in milist do
	    dtable[l] := der(l);
	end_for;
	// Next, we need to evaluate all the derivatives at x_k = 0,
	// t = 0.
	// The values will be kept in vtable
	vtable := table();
	// A method to evaluate the derivatives.
	value :=
	    proc(l:DOM_LIST)
	    local fn, vs:DOM_LIST;
	    begin
		fn := dtable[l];
		if domtype(fn) = DE then
		    vs := [op(DE::indets(fn, DiffVars))];
		    vs := map(vs, (v) -> (v = idval(v)));
		    vs := 
		    fn := DE::subs(fn, vs);
		end_if;
		fn := expr(fn);
		vs := zip(ivars, ctr, _equal);
		fn := subs(fn, vs);
		return(fn);
	    end_proc; /* value */
	// A method to find the values of the indets needed for val
	// if the variable v is independent, then replace it by 0,
	// otherwise look it up in vtable.
	idval :=
	    proc(v)
	    begin
		if DV::vartype(v) = "Indep" then
		    return(ctr[DV::index(v)]);
		else
		    return(vtable[DV::multiIndex(v)]);
		end_if;
	    end_proc; /* idval */
	// Next, put the values in vtable.
	vtable[[0$numvars]] := value([0$numvars]);
	for l in milist do
	    vtable[l] := value(l);
	end_for;
	// Now, create the polynomial
	milist := map(milist, l -> [expr(vtable[l])/_mult(op(map(l,fact))),l]);
	milist := [[expr(vtable[[0$numvars]]),[0$numvars]]] . milist;
	tmpvars := [genident()$k=1..numvars];
	sol := expr(poly(milist,tmpvars));
	sol := subs(sol,
	    [tmpvars[i]=(ivars[i]-ctr[i])$i=1..numvars]);
	return(Dom::MultivariateSeries()(sol, zip(ivars, ctr, _equal), deg));
    end_proc; /* CKsolve */
