/*++
Cat::DifferentialFunction -- category of functions of differential variables


Domains in Cat::DifferentialFunction(DV) are a Cat::AbelianSemiGroup and
contain functions that depend on differential variables from a domain DV, an
element of Cat::DifferentialVariable.  (We cannot use Cat::Rng etc., as we
must allow for linear functions where multiplication is not possible.
Nonlinear domains should also be in Cat::PartialDifferentialRing).

The following methods must be provided by domains in this category:
	"diff" :     differentiate with respect to an element of DV or
		     with respect to a parameter
	"indets" :   return indeterminates of function as a set;
	             with option "DiffVars" only those in DV
	"subs" :     substitute an expression for an element of DV
	"solve" :    solve for an element of DV;
	             if not possible return FAIL
	"has" :      checks whether function depends on differential variable
	"Variables" :  the domain DV
and the methods of a Cat::AbelianSemiGroup.

For these methods default implementations are provided:
	"order" :      order of a function
	"class" :      class of a function
	"leader" :     leader of a function (with respect to order in DV)
	"diffSubs" :   substitute for a differential variable and all its 
		       derivatives
	"isConst" :    checks whether a function is a constant (wrt to DV)
	"multiDiff" :  computes efficiently several derivatives of the same
		       functions
	"jacobian" :   Jacobian of a list of functions
	"totalDiff" :  total derivative with respect to an independent
	               variable (can be given by its index)
	"autosimplify" :   purely algebraic simplification
	"autoreduce" : simplification including differential operations
	"eval" :     eval the differential function for a given list
		     of expression ("check a solution")
	"linear":    TRUE, if the domain represents linear functions
	"polynomial": TRUE, if the domain represents polynomial functions
Some of these are not very efficient/powerful; some are valid only under
additional assumptions.

The entry "jacobianType" returns the domain for Jacobian matrices. By default 
this is the domain Dom::SparseMatrix(dom). However, some domains (e.g. for linear 
functions) may use other domains.


Authors: Eduard Niess, Werner M. Seiler, Uni Karlsruhe/Mannheim
         (wms@ira.uka.de, seiler@euler.math.uni-mannheim.de)
Created: ??? (1996)
Last update: 	Thu Nov 29 15:02:12 2001
++*/


alias(TYPE_ERR=error("wrong type of argument"),
      ARG_ERR=error("wrong number of arguments")):


category Cat::DifferentialFunction(DV)

    local DVcoerce:DOM_PROC, DVdiff:DOM_PROC, DVorder:DOM_PROC, 
	  DVclass:DOM_PROC, DVderiv:DOM_PROC, DVtype:DOM_PROC, 
	  DVindex:DOM_PROC, DVless:DOM_PROC, DVgreater:DOM_PROC, 
	  DVmulti:DOM_PROC, DVrepeated:DOM_PROC, DVr2m:DOM_PROC, 
	  DVnotation:DOM_PROC,
	  mm:DOM_INT, nn:DOM_INT, remDup:DOM_PROC, smerge:DOM_PROC, 
	  merge:DOM_PROC, sortinsert:DOM_PROC;

    category Cat::AbelianSemiGroup;


    /* ================= */
    /* Undefined Methods */
    /* ================= */

    diff;
    indets;
    solve;
    has;

    /* ============== */
    /* Simple Methods */
    /* ============== */

    Variables := DV;
    
    linear := FALSE;
    
    polynomial := FALSE;
        
    jacobianType := Dom::SparseMatrix(dom);


    isConst :=
        proc(df:dom):DOM_BOOL
	begin iszero(nops(dom::indets(df,hold(DiffVars)))) end_proc;


    order :=
	proc(df:dom):DOM_INT
	local tmp:DOM_SET;
	begin
	    tmp := dom::indets(df,hold(DiffVars));
	    if tmp={} then -2 else max(op(map(tmp,DVorder))) end_if
	end_proc;


    class :=
	proc(df:dom):DOM_INT
	local tmp:DOM_SET,q:DOM_INT;
	begin
	    q := dom::order(df);
	    tmp := dom::indets(df,hold(DiffVars));
	    if tmp={} then return(0);
	    else
	        tmp := select(tmp, ((jv,q) -> bool(DVorder(jv)=q)), q);
	        max(op(map(tmp,DVclass)));
	    end_if;
	end_proc;


    leader :=
        proc(df:dom)
        local tmp:DOM_LIST, res:DV, dv:DV;
        begin
            tmp := [op(dom::indets(df,hold(DiffVars)))];
            if tmp=[] then return(FAIL)
	    else
		res := tmp[1];
		delete tmp[1];
		for dv in tmp do
		    if DVless(res,dv) then res := dv end_if;
		end_for;
		return(res);
	    end_if;
        end_proc;

    
    diffVars := 
	proc(df:dom)
	begin
	    sort([op(dom::indets(df, hold(DiffVars)))], DVgreater);
	end_proc;


    /* ============= */
    /* Substitutions */
    /* ============= */
    
    // In order to facilitate the handling of multiple substitutions, here
    // in the category a method `multiSubs' is implemented which requires the
    // existence of a method `subsOne' in the domain. It is recommended that
    // in domains the following approach is taken:
    //		subs := dom::multiSubs;
    //		subsOne := ....
    // where `subsOne' must only handle single substitutions without any
    // argument checks. This somewhat roundabout construction is necessary,
    // as a implementation of `subs' is already provided in Dom::BaseDomain
    // and this overwrites any categorial definition of `subs'.
    
    // Takes care of argument checks for a single equation.
    // Returns FAIL, if something is not correct;
    // otherwise some coercions are performed.
    checkEq :=
	proc(eq)
	local y, ex;
	begin
	    if type(eq)<>"_equal" then return(FAIL) end_if;
	    y := DVcoerce(op(eq,1));
	    if y=FAIL then y := op(eq,1) end_if;
	    ex := op(eq,2);
	    if iszero(ex) then ex := dom::zero
	    elif ex=1 then ex := dom::one
	    else
		ex := DVcoerce(ex);
		if ex=FAIL then
		    ex := dom::coerce(op(eq,2));
		    if ex=FAIL then
			ex := op(eq,2);
		    end_if;
		end_if;
	    end_if;
	    return(y=ex);
	end_proc;
	
	
    // This method translates multiple substitutions into a sequence of
    // single substitutions.
    multiSubs :=
	proc(de:dom):dom
	local res, eq, eql:DOM_LIST;
	begin
	    case args(0)
	    of 0 do ARG_ERR;
	    of 1 do return(de);
	    of 2 do 
		case type(args(2))
		of "_equal" do 
		    eq := args(2);
		    eq := dom::checkEq(eq);
		    if eq=FAIL then TYPE_ERR end_if;
		    return(dom::subsOne(de,eq));
		of DOM_LIST do 
		of DOM_SET do 
		of DOM_TABLE do 
		    eql := [op(args(2))];
		    res := de;
		    eql := map(eql,dom::checkEq);
		    if not(iszero(contains(eql,FAIL))) then 
			TYPE_ERR;
		    end_if;
		    for eq in eql do
			res := dom::subsOne(res,eq);
		    end_for;
		    return(res);
		otherwise TYPE_ERR;
		end_case;
	    otherwise 
		eql := [args(2..args(0))];
		res := de;
		eql := map(eql,dom::checkEq);
		if not(iszero(contains(eql,FAIL))) then TYPE_ERR end_if;
		for eq in eql do
		    res := dom::subsOne(res,eq);
		end_for;
		return(res);
	    end_case;
	end_proc;


    /* ================================= */
    /* Methods Based on Derivative Trees */
    /* ================================= */

    // Computation of several derivatives of the same expression.
    // Input: domain element and list of multi (or repeated) indices
    // Output: list of derivatives
    multiDiff :=
	proc(ex:dom,der:DOM_LIST):DOM_LIST
	local derivTree:DOM_PROC, tree:DOM_LIST;
	begin
	    // compute derivative tree
	    if DVnotation="repeated" then
		tree := detools::derList2Tree(map(der,DVr2m));
	    else
		tree := detools::derList2Tree(der);
	    end_if;
	    // recursive procedure for working down the tree
	    derivTree :=
		proc(f,T:DOM_LIST,res:DOM_LIST,DOM:DOM_DOMAIN):DOM_LIST
		local df, l, i:DOM_INT, vars:DOM_LIST;
		begin
		    df := f;
		    for l in T do
			if testtype(l,Type::ListOf(Type::NonNegInt)) then
			    // node with derivative -> compute it
			    vars := [];
			    for i from 1 to nn do
				vars := vars . [ i $ l[i] ];
			    end_for;
			    if vars<>[] then
				df := DOM::totalDiff(df,vars); 
			    end_if;
			elif testtype(l,Type::NonNegInt) then
			    // leaf -> return result
			    res[l] := df;
			else
			    // branch -> recursive call
			    res := derivTree(df,l,res,DOM);
			end_if;
		    end_for;
		    return(res);
		end_proc;
	    // finally, call this procedure with the given data
	    derivTree(ex,tree,[ 0 $ nops(der) ],dom);
	end_proc;
	

    // `eval' enters a "solution" into a given differential expression;
    // the user must take care that a reasonable domain is used.
    eval :=
        proc(sys,sol)
	local DOM:DOM_DOMAIN, SDOM:DOM_DOMAIN, sysflag:DOM_BOOL,
	      j:DOM_INT, allvars:DOM_SET, vars:DOM_LIST, dersol:DOM_LIST, 
	      res, slist:DOM_LIST, mm:DOM_INT;
        begin
	    if domtype(sys)<>DOM_LIST then
		sysflag := FALSE;
		sys := [sys];
	    else sysflag := TRUE end_if;
	    mm := DV::numDepVar;
	    DOM := dom;
	    if mm=1 and domtype(sol)<>DOM_LIST then
		sol := [sol];
	    end_if;
	    SDOM := domtype(sol[1]);		// domain of "solution"
	    sys := map(sys,dom::coerce);
	    if testargs() then
		if not(iszero(contains(sys,FAIL))) then TYPE_ERR end_if;
		if nops(sol)<>mm then
		    error("wrong number of functions given");
		end_if;		
	    end_if;
	    if SDOM::hasProp=FAIL or 
	       not(SDOM::hasProp(Cat::DifferentialFunction(DV))) then
		SDOM := Dom::DifferentialExpression(
			    Dom::RestrictedDifferentialVariable(DV,
					Types="Indep"));
		sol := map(sol,SDOM::convert);
		if not(iszero(contains(sol,FAIL))) then TYPE_ERR end_if;
	    end_if;
	    
	    res := [ 0 $ nops(sys) ];
	    // get all contained derivatives
	    allvars := _union(op(map(sys,dom::indets,hold(DiffVars))));
	    allvars := select(allvars,
			      ((jv) -> contains({"Dep","Deriv"},DVtype(jv))));
	    // compute derivatives via derivative tree
	    vars := [ 0 $ mm ];
	    dersol := [ 0 $ mm ];
	    slist := [];
	    for j from 1 to mm do
		vars[j] := [op(select(allvars, 
				      ((jv,k) -> bool(DVindex(jv)=k)),
				      j))];
		if DVnotation="repeated" then
		    dersol[j] := SDOM::multiDiff(sol[j],
					     map(vars[j],DVrepeated));
		else
		    dersol[j] := SDOM::multiDiff(sol[j],
					     map(vars[j],DVmulti));
		end_if;
		slist := slist . zip(vars[j],dersol[j],_equal);
	    end_for;
	    // enter derivatives in expressions
	    res :=map(sys,dom::convert_to,DOM_EXPR);
	    res := map(res,subs,op(slist));
	    res := map(res,SDOM::convert);
	    if not(iszero(contains(res,FAIL))) then
		error("internal error");
	    end_if;
      	    if sysflag then res else res[1] end_if;
	end_proc;


    // "differential substitution" based on `subsOne'
    diffSubs :=
        proc(f:dom,eq):dom
	local nof:dom, nf:dom, JVar:DOM_LIST, DerList:DOM_LIST,
	      jv, ex, i:DOM_INT, n:DOM_INT, DOM:DOM_DOMAIN;
	begin
	    // extract parts of eq
	    if testargs() then
		if (args(0)<>2) then ARG_ERR end_if;
		if type(eq)<>"_equal" then TYPE_ERR end_if;
	    end_if;
	    jv := DV::coerce(op(eq,1));
	    if jv=FAIL then TYPE_ERR end_if;
	    ex := dom::coerce(op(eq,2));
	    if ex=FAIL then TYPE_ERR end_if;

	    // ordinary substitution
	    DOM := dom;
	    nof := dom::zero;
	    nf := dom::subsOne(f,jv=ex);
	    
	    // compute needed derivatives of ex
	    JVar := [op(select(dom::indets(nf,hold(DiffVars)),
			       ((dv,i) -> (DVtype(dv)="Deriv" and 
			                   DVindex(dv)=i)), 
			       DVindex(jv)))];
	    i := 0;
	    n := nops(JVar);
	    DerList := [ [] $ n ];
	    while i<n do
		i := i+1;
		DerList[i] := DV::derivativeOf(JVar[i],jv);
		if DerList[i]=[] then
		    delete JVar[i];
		    delete DerList[i];
		    i := i-1;
		    n := n-1;
		end_if;
	    end_while;
	    if iszero(n) then return(nf) end_if;  // no derivatives occur
	    DerList := zip(JVar,dom::multiDiff(ex,DerList),_equal);
	    
	    // perform substitutions
	    while nf<>nof do
		nof := nf;
		for eq in DerList do
		    nf := dom::subsOne(nf,eq);
		end_for;
	    end_while;
	    
	    nf;
	end_proc;


    /* ===================== */
    /* Total Differentiation */
    /* ===================== */

    // A default implementation is provided only for the case that `dom'
    // is a (commutative) ring, as multiplications are needed.
    
    // We use `D' as an abbreviation for `totalDiff'
    D := 
	proc(l,df:dom):dom
	begin dom::totalDiff(df,l) end_proc;
    
    
    // First argument is single expression or list of expression.
    // Second argument is either independent variable or its index
    //		or a list of either.
    // Optional third and fourth arguments are Jacobian of first argument and
    // the ordered lists of differential variables indexing the columns.
    totalDiff :=
        if dom::hasProp(Cat::CommutativeRing) then
	    proc(sys,mu)
            local sysflag:DOM_BOOL, DOM:DOM_DOMAIN, n:DOM_INT, jmi, vars:DOM_LIST, 
		i:DOM_INT, tmp;
            begin
		// ====== analyse input =====
		sysflag := bool(domtype(sys)=DOM_LIST);
                if not(sysflag) then sys := [sys] end_if;
		case domtype(mu)
		of DOM_INT do 
		    if testargs() and
		       ((mu<1) or (mu>DV::numIndVar)) then TYPE_ERR end_if;
		    mu := [mu];
		    break;
		of DV do
		    if testargs() and DVtype(mu)<>"Indep" then TYPE_ERR end_if;
		    mu := [DV::index(mu)];
		    break;
		of DOM_LIST do
		    if testargs() then
			mu := map(mu, (i -> case domtype(i)
					    of DOM_INT do
						if (i<1) or (i>DV::numIndVar)
						then TYPE_ERR end_if;
						return(i);
					    of DV do 
						if DVtype(i)<>"Indep"
						then TYPE_ERR end_if;
						return(DV::index(i));
					    otherwise TYPE_ERR;
					    end_case));
		    else
			mu := map(mu, (i -> case domtype(i)
					    of DOM_INT do return(i);
					    of DV do return(DV::index(i));
					    otherwise TYPE_ERR;
					    end_case));
		    end_if;
		    break;
		otherwise TYPE_ERR;
		end_case;
		// further argument checks
		if testargs() then
		    if args(0)>3 then ARG_ERR end_if;
		    if {op(map(sys,domtype))}<>{dom} then TYPE_ERR end_if;
		    if (args(0)=3) and 
		        (domtype(args(3)) <> DOM_LIST 
		        or domtype(args(3))[1] <> dom::jacobianType
		        or domtype(args(3))[2] <> Type::ListOf(Type::SetOf(DV))) then
			    TYPE_ERR;
		    end_if;
		end_if;
		
		// ===== initialisation =====
		DOM := dom;
		n := nops(sys);
		// compute Jacobian if not given
		if args(0)=3 then
		    jmi := (args(3), args(4));
		else
		    vars := map(sys, dom::indets, DiffVars);
		    if vars=[ {} $ n ] then     // check for constant system
			if sysflag then
			    return([ dom::zero $ n ]);
			else return(dom::zero) end_if;
		    end_if;
		    jmi := dom::jacobian(sys);
		end_if;
		
		tmp := (sys, map(sys, dom::indets, DiffVars));

		// ===== compute total differentiation =====
		for i in mu do
		    if _and(op(map(op(tmp, 2), l -> bool(l = {})))) then
			if sysflag then return([dom::zero $ nops(sys)])
			else return(dom::zero) end_if;
		    end_if;
		    if i = 1 then
			tmp := dom::totalDiffJM(i, jmi);
		    else
			tmp := dom::totalDiffJM(i, dom::jacobian(op(tmp, 1)));
		    end_if;
		end_for;
		if sysflag then 
		    if domtype(sys) = dom::jacobianType then
			(dom::jacobianType)::convert_to(op(tmp, 1), DOM_LIST);
		    else
			op(tmp, 1);
		    end_if
		else 
		    op(tmp, 1)[1]; 
		end_if;
	    end_proc;
	else
	    toBeDefined;
	end_if;
	    
    totalDiffJM :=
	proc(mu:DOM_INT, jac, vars)
	local DOM:DOM_DOMAIN, sys, n;
	begin
	    DOM := dom;
	    vars := (dom::jacobianType)(map(vars, l -> 
		(if DVtype(l) = "Indep" then
		    if DVindex(l) = mu then
			DOM::one;
		    else
			DOM::zero;
		    end_if;
		else
		    DOM::convert(DVdiff(l, mu));
		end_if)));
	    sys := jac * vars;
	    n := jac::dom::matdim(sys)[1];
	    ([(dom::jacobianType::convert_to(sys, DOM_ARRAY))[i, 1] $ i = 1..n], 
	     [dom::indets(sys[i], DiffVars) $ i = 1..n]);
	end_proc;
	
//	local n:DOM_INT, rows:DOM_LIST, dvars:DOM_LIST, tdvars:DOM_LIST, vars:DOM_LIST, 
//	    i:DOM_INT, rvars:DOM_LIST, j:DOM_INT, res:DOM_INT, k:DOM_INT, t:DOM_LIST,
//	    pos:DOM_INT, DOM:DOM_DOMAIN;
		
// 	    DOM := dom;
// 	    n := nops(sys);
// 	    jac := (dom::jacobianType)::convert_to(jac, DOM_ARRAY);
// 	    nvars := map([(dom::jacobianType)::convert_to(nvars, DOM_ARRAY)[i, 1] 
// 		$ i = 1..nops(nvars)], dom::convert_to, DV);
// 	    rows := [select(zip(nvars, [jac[i, j] $ j = 1..nops(nvars)], 
// 		(l1, l2) -> [l1, l2]), l -> not(DOM::iszero(l[2]))) 
// 		$ i = 1..n];
// 	    // collect variable lists
// 	    nvars := select(nvars,   // variables that need chain rule
//                      ((jv) -> contains({"Deriv","Dep"},DVtype(jv))));
// 	    dvars := map(nvars,DVdiff,mu); // their derivatives
// 	    tdvars := map(dvars,dom::convert);
// 	    vars := [ {} $ n ];     // differential variables of result
// 	    // iterate over all functions
// 	    for i from 1 to n do
// 		if rows[i]=[] then	// constant function
// 		    sys[i] := dom::zero;
// 		    next;
// 		end_if;
// 		// collect appearing variables
// 		rvars := map(rows[i],op,1);
// 		j := nops(rvars);
// 		vars[i] := [0 $ j];
// 		res := [ dom::zero $ j ];
// 		j := 0;
// 		k := 0;
// 		// compute derivative iterating over variables
// 		for t in rows[i] do
// 		    case DVtype(t[1])
// 		    of "Dep" do
// 		    of "Deriv" do	// apply chain rule
// 			pos := contains(nvars,t[1]);
// 			k := k+1;
// 			res[k] := dom::_mult(t[2],tdvars[pos]);
// 			j := j+1;
// 			vars[i][j] := dvars[pos];
// 			break;
// 		    of "Indep" do	// partial derivative
// 			if DV::index(t[1])=mu[ind] then
// 			    k := k+1;
// 			    res[k] := t[2];
// 			end_if;
// 			break;
// 		    otherwise error("this cannot happen");
// 		    end_case;
// 		end_for;
// 		// analyse results
// 		if iszero(k) then	// no terms present
// 		    sys[i] := dom::zero;
// 		    vars[i] := {};
// 		    next;
// 		end_if;
// 		// sum up and collect variables
// 		sys[i] := dom::_plus(op(res,1..k));
// 		vars[i] := {op(vars[i], 1..j)} union {op(rvars)};
// 	    end_for;
// 	    (sys, vars);
//	end_proc;


    /* =============== */
    /* Jacobian Matrix */
    /* =============== */

    // Again a default implementation is provided only for the case that
    // `dom' is a ring which is here just a simple test for non-linear
    // functions. If the elements of `dom' represent linear functions,
    // another domain must be used for the Jacobian. This can be handled
    // reasonably only directly in this domain.
    
    jacobian :=
        if dom::hasProp(Cat::CommutativeRing) then
            proc(sys)
            local v, vars, varlist, f, DOM;
            begin
		if args(0)>1 then vars := args(2) end_if;
		if domtype(sys)<>DOM_LIST then
		    sys := [sys];
		    if args(0)>1 then vars := [vars] end_if;
		end_if;
                if testargs() then
		    if args(0)<1 or args(0)>2 then ARG_ERR end_if;
		    if {op(map(sys,domtype))}<>{dom} then TYPE_ERR end_if;
		    if args(0)=2 then
			if {op(map(vars,domtype))}<>{DOM_SET} then
			    TYPE_ERR
			end_if;
			for v in vars do
			    if {op(map(v,domtype))}<>{DV} then TYPE_ERR end_if;
			end_for;
		    end_if;
                end_if;
		
                if args(0)=1 then 
		    vars := map(sys, dom::indets, DiffVars);
		end_if;
		varlist := [op({op(map(vars, op))})];		
                if varlist=[] then error("empty matrix") end_if;
                // should return zero matrix - but of what size?
		
		DOM := dom;
		f := proc(i:DOM_INT, j:DOM_INT)
		    begin
			if not contains(vars[i], varlist[j]) then
			    return(DOM::zero);
			else
			    return(DOM::diff(sys[i], varlist[j]));
			end_if;
		    end_proc;
		    
		((dom::jacobianType)(nops(sys), nops(varlist), f), 
		 varlist);		
	    end_proc
        else toBeDefined end_if;


    /* ============================ */
    /* Simplification and Reduction */
    /* ============================ */
    
    // Simplification is here understood as a purely algebraic operations,
    // i.e. no differentiated equations are used. In contrast, reductions
    // also use differentiated equations and are thus in general more
    // expensive.

    // The default implementation of `autoreduce' depends decisively on
    // the method `solve', as it solves all equations for their leaders
    // and enters them into the other equations. Equations which cannot
    // be solved are put into a special list in the hope that further
    // reductions render them solvable.
    autoreduce := 
	proc(sys)
        local new_l:DOM_LIST, notSolved_l:DOM_LIST, solved_l:DOM_LIST, eq:dom, 
	    ld:DV, k:DOM_INT, oldSolved_l:DOM_LIST, sub:DOM_LIST, neq:DOM_LIST,
	    ic:dom, tmp:DOM_LIST, d:DOM_LIST, old_l:DOM_LIST, eq2:dom,
	    DOM:DOM_DOMAIN; 
       begin
	    // analyse input
	    if domtype(sys)<>DOM_LIST then sys := [sys] end_if;
	    if testargs() then
		if args(0)<>1 then ARG_ERR end_if;
		if {op(map(sys,domtype))}<>{dom} then TYPE_ERR end_if;
	    end_if;
	    // perform first an algebraic simplification
	    // for a single equation this is all that can be done
	    sys := map(sys, dom::simplify); 
	    DOM := dom;
	    if nops(sys) = 1 then return(sys) end_if; 
	    // initialise the three basic lists
	    new_l := sys;		// not yet analysed equations
	    notSolved_l := [];		// equations not solvable
	    solved_l := [];   		// solve equations
	    while new_l <> [] do
		for eq in new_l do
		    // try to solve new equation for its leader
		    ld := dom::leader(eq);
       		    if (k := dom::solve(eq, ld)) = FAIL then 
			notSolved_l := append(notSolved_l, eq)
		    else 
			solved_l := sortinsert(solved_l, [ld, k],
				((l1, l2) -> DVless(l1[1], l2[1])));
		    end_if;
		end_for;
		// treat all solved equations
		new_l := [];
		oldSolved_l := [];
		while solved_l <> [] do
		    // take first equation
		    sub := solved_l[1];
		    oldSolved_l := append(oldSolved_l, sub);
		    delete solved_l[1];
		    // iterate over all other solved equations in order
		    // to achieve triangular form
		    while solved_l <> [] do
			neq := solved_l[1];
			// compare leader
			if sub[1]<>neq[1] then break end_if;
			delete solved_l[1];
			// compute difference of right hand sides
			// store as new equation
			ic := neq[2] - sub[2];
			if not(dom::iszero(ic)) then
			    new_l := append(new_l, ic);
			end_if;
		    end_while;
		    tmp := [];
		    // check for differential dependencies
		    for neq in solved_l do
			// compare leaders
			d := DV::derivativeOf(neq[1], sub[1]);
			if d = [] then 		// no derivative of sub[1]
			    // eliminate all derivatives of sub[1] in
			    // the equation neq
			    tmp := append(tmp, [neq[1], dom::diffSubs(neq[2], 
			                  sub[1] = sub[2])]);
			else			// derivative of sub[1]
			    // compute integrability condition
			    // store as new equation
			    ic := dom::diffSubs(
				    dom::totalDiff(sub[2], d) - neq[2], 
				    sub[1] = sub[2]);
			    if not(dom::iszero(ic)) then
				new_l := append(new_l, ic)
			    end_if;
			end_if;
		    end_for;
		    solved_l := tmp;
		end_while;
		solved_l := oldSolved_l;
		old_l := [];
		// try to simplify unsolved equations
		for eq in notSolved_l do
		    eq2 := eq;
		    for sub in solved_l do
			// eliminate all derivatives of sub[1] in eq2
			eq2 := dom::diffSubs(eq2, sub[1] = sub[2]);
		    end_for;
		    if eq2 = eq then	// no changes
			old_l := append(old_l, eq)
		    else		// put changed equation to new equation
			if not(dom::iszero(eq2)) then
			    new_l := append(new_l, eq2)
			end_if;
		    end_if;
		end_for;
		notSolved_l := old_l;
		old_l := [];
		// simplify new equations
		for eq in new_l do
	       	    for sub in solved_l do
			// eliminate all derivatives of sub[1] in eq
			eq := dom::diffSubs(eq, sub[1] = sub[2]);
		    end_for;
		    if not(dom::iszero(eq)) then
			old_l := append(old_l, eq);
		    end_if;
		end_for;
		new_l := old_l;
	    end_while;
	    // collect results
	    notSolved_l . map(solved_l, ((l) -> (DOM::convert(l[1]) - l[2])));
	end_proc;


    // The implementation of `autosimplify' is almost identical with the
    // one of `autoreduce'. Only the calls to `diffSubs' are replaced by
    // calls to `subs'.
    autosimplify := 
	proc(sys)
        local new_l:DOM_LIST, notSolved_l:DOM_LIST, solved_l:DOM_LIST, eq:dom, 
	    ld:DV, k:DOM_INT, oldSolved_l:DOM_LIST, sub:DOM_LIST, neq:DOM_LIST,
	    ic:dom, tmp:DOM_LIST, old_l:DOM_LIST, eq2:dom, DOM:DOM_DOMAIN;
        begin
	    if domtype(sys)<>DOM_LIST then sys := [sys] end_if;
	    if testargs() then
		if args(0)<>1 then ARG_ERR end_if;
		if {op(map(sys,domtype))}<>{dom} then TYPE_ERR end_if;
	    end_if;
	    sys := map(sys, dom::simplify); 
	    DOM := dom;
	    if nops(sys) = 1 then return(sys) end_if; 
	    new_l := sys;
	    notSolved_l := [];
	    solved_l := [];   
	    while new_l <> [] do
		for eq in new_l do
		    ld := dom::leader(eq);
		    if (k := dom::solve(eq, ld)) = FAIL then
			notSolved_l := append(notSolved_l, eq)
		    else
			solved_l := sortinsert([ld, k], solved_l,  
				((l1, l2) -> DVless(l1[1], l2[1])));
		    end_if;
		end_for;
		new_l := [];
		oldSolved_l := [];
		while solved_l <> [] do
		    sub := solved_l[1];
		    oldSolved_l := append(oldSolved_l, sub);
		    delete solved_l[1];
		    while solved_l <> [] do
			neq := solved_l[1];
			if sub[1] <> neq[1] then break end_if;
			delete solved_l[1];
			ic := neq[2] - sub[2];
			if not(dom::iszero(ic)) then
			    new_l := append(new_l, ic)
			end_if;
		    end_while;
		    tmp := [];
		    for neq in solved_l do
			tmp := append(tmp, [neq[1], dom::subs(neq[2], 
			                  sub[1] = sub[2])]);
		    end_for;
		    solved_l := tmp;
		end_while;
		solved_l := oldSolved_l;
		old_l := [];
		for eq in notSolved_l do
		    eq2 := eq;
		    for sub in solved_l do
			eq2 := dom::subs(eq2, sub[1] = sub[1]);
		    end_for;
		    if eq2 = eq then
			old_l := append(old_l, eq)
		    else
			if not(dom::iszero(eq2)) then
			    new_l := append(new_l, eq2)
			end_if;
		    end_if;
		end_for;
		notSolved_l := old_l;
		old_l := [];
		for eq in new_l do
		    for sub in solved_l do
			eq := dom::subs(eq, sub[1] = sub[2]);
		    end_for;
		    if not(dom::iszero(eq)) then
			old_l := append(old_l, eq2);
		    end_if;
		end_for;
		new_l := old_l;
	    end_while;
	    notSolved_l . map(solved_l, ((l) -> (DOM::coerce(l[1])) - l[2]));
	end_proc;

        
begin

    if args(0)<>1 then ARG_ERR end_if;
    if (domtype(DV)<>DOM_DOMAIN) or (DV::hasProp=FAIL) or
       not(DV::hasProp(Cat::DifferentialVariable)) then
	error("arg must be in Cat::DifferentialVariable");
    end_if;
  
    DVcoerce := DV::coerce;
    DVdiff := DV::totalDiff;
    DVorder := DV::order;
    DVclass := DV::class;
    DVindex := DV::index;
    DVderiv := DV::derivativeOf;
    DVtype := DV::vartype;
    DVless := DV::_less;
    DVgreater := (dv1,dv2) -> DVless(dv2,dv1);
    DVmulti := DV::multiIndex;
    DVrepeated := DV::repeatedIndex;
    DVr2m := DV::r2m;
    DVnotation := DV::notation;
    
    nn := DV::numIndVar;
    mm := DV::numDepVar;
    smerge := listlib::singleMerge;
    merge := listlib::merge;
    sortinsert := listlib::insert;
    remDup := listlib::removeDuplicates;
    
end_category:

unalias(TYPE_ERR,ARG_ERR,DV):

