/*****************************************************************************
  $Id: PermutationGroup.mu,v 1.2 2001/12/10 15:04:57 nthiery Exp $
  File:		lib/DOMAINS/CATEGORY/PermutationGroup.mu
  Authors:	Nicolas M. Thiry <nthiery@users.sourceforge.net>
		Original version by Stephan Wehmeier and the MuPAD group
  Taken from:	PerMuVAR, 02/12/2001
  Taken from:	MuPAD 1.4/2.0, 1998-2000
  Created:	08/08/1999
  License:	LGPL
*****************************************************************************/

/*++
Cat::PermutationGroup(n) -- The category of permutation groups

That is the group of groups of permutations of some finite set, or the
category of finite groups acting faithfully on finite sets.

Default implementations are provided for card, size, generators. Some
of these are not very efficient!
++*/

category Cat::PermutationGroup
    category Cat::Group;

/*--
    allElements

    List or set of all the elements of the group.
--*/

    allElements;

/*--
    size

    Size of the group. 
    Overload if possible!
--*/

    size := nops(dom::allElements);

/*--
    generators

    A list or set of generators of the group.
    Overload if possible!
--*/

    generators := dom::allElements;

/*--
    set

    Finite set on which this permutation group acts.

    If it's a list, this defines the enumeration e1, e2, ... en of the
    elements of the set. If it's a set, this enumeration will be defined
    by MuPAD's internal order.

--*/ 

    set;

/*--
    invertSet

    Table containing for each element of the set its index in the enumeration
    of the set. Therefore, if e is an element of the set, the expression
    op(dom::set, dom::invertSet[e]) will evaluate back to e.
--*/

    invertSet := table(op(dom::set,i)=i $ i = 1..nops(dom::set));

/*--
    card

    Size of the set on which this permutation group acts.
--*/

    card := nops(dom::set);

/*--
    compactName(e:dom::set) -> DOM_STRING

    Return some compact name for the element e of the set. This can
    be usefull for example when you want to create variables which are
    indexed by the elements of the set, and want to keep the name
    of those variables short.

    See Dom::PermutationGroupModule(...)::variables.
--*/

    compactName := 
    proc(e)
	option remember;
    begin
	expr2text(e);
    end_proc;

/*--
    act(g: dom, e: dom::set)

    Action of an element g of the group on an element e of the set

    Should func_call be equivalent ?
--*/

    act;

/*--
    Action on lists

    Recall that dom::set defines an enumeration e1..en of the
    elements of S. Suppose you want to implement some kind of object V
    whose content is indexed by S: for each e in S, you have some
    value V_e. An example of such object would be a formal linear
    combinations of the elements of S. One way is to use a table:
			V[e]=V_e.
    Another, more efficient, is to use a list:
			[V_e1, V_e2,..., V_en].

    The action of G on S extends to an action of G on such objects,
    defined by (g.V)[g.e] = V[e]

    actOnList(g: dom, l:DOM_LIST) -> DOM_LIST

    This method implements the action of g on lists representing
    objects whose content is indexed by S.

    actOnListFunc(g) -> DOM_PROC

    Returns an optimized procedure that takes a list and returns it
    after action of g.

    There was several way to construct such procedures, and I choosed
    the one that generated the fastest procedures (benchmarks with
    MuPAD-1.4.2, see "timeperm.mu"). This is likely to change with
    MuPAD-1.5.
--*/

    actOnList := 
    proc(g: dom, l: DOM_LIST)
	local l2, inv;
    begin
	l2 := [ 0 $ dom::card ];
	inv:=dom::invertSet;
	for e in dom::set do
	    l2[inv[g(e)]]:=l[inv[e]];
	end_for;
	new(dom, l2);
    end_proc;

    actOnListFunc := 
    proc(g:dom)
    begin
	subsop(()-> NIL,
	       4 = dom::actOnList(g,[hold(op)(DOM_VAR(0,2),i)
				     $ i=1..dom::card])):
    end_proc;

    fastGroup := 
    proc()
	option remember;
    begin
	userinfo(5, "Generation of a faster copy of the group");
	map(dom::allElements, dom::actOnListFunc)
    end_proc;

    listLess := dom::listLexLess;

    listLexLess :=
    proc(x1:DOM_LIST, x2:DOM_LIST)
	local i;
    begin
	i := 1;
	while (i<=dom::card) and (x1[i]=x2[i]) do
	    i := i+1;
	end_while;
	if (i>dom::card)
	   then FALSE
	else bool(x1[i]<x2[i])
	end_if;
    end_proc;

    listRevLexLess :=
    proc(x1:DOM_LIST, x2:DOM_LIST)
	local i;
    begin
	i := dom::card;
	while (i>0) and (x1[i]=x2[i]) do
	    i := i-1;
	end_while;
	if (i=0)
	   then FALSE
	else bool(x2[i]<x1[i])
	end_if;
    end_proc;

    isCanonic :=
    proc(x:DOM_LIST)
	local g;
    begin
	    // By default, use lexicographic order
	userinfo(10, "Checking if ". expr2text(x) . " is canonic");
	for g in dom::fastGroup() do
	    if dom::listLess(x,g(x)) then return(FALSE); end_if;
	end_for;
	return(TRUE);
    end_proc;
	    
    canonic := 
    proc(x: DOM_LIST)
	local g, x2, xmax;
    begin
	// By default, use lexicographic order
	userinfo(10, "Calculating orbit and canonic representative of ". expr2text(x));
	xmax:=x;
	for g in dom::fastGroup() do
	    x2:=g(x);
	    if dom::listLess(xmax,x2) then
		xmax:=x2;
	    end_if;
	end_for;
	xmax
    end_proc;

    orbit := 
    proc(x:DOM_LIST)
    begin
	[ f(x) $ f in dom::fastGroup() ];
    end_proc;

    orbitSet := 
    proc(x:DOM_LIST)
    begin
	{op(dom::orbit(x))}:
    end_proc;

    orbitSize := 
    proc(x:DOM_LIST)
    begin
	nops(dom::orbitSet(x));
    end_proc;

    canonicCompositions :=
    proc(n: Type::NonNegInt)
    begin
	select(combinat::integerVectors(n,dom::card),
	       dom::isCanonic);
    end_proc;

    // canonicListsByShape(shape: DOM_LIST) -> Type::ListOf(DOM_LIST)
    //
    // Returns all canonic lists of shape shape.
    
    canonicListsByShape :=
    proc(shape: DOM_LIST)
    begin
	select(combinat::permutations(shape),
	       dom::isCanonic);
    end_proc;

    canonicBinaryLists :=
    proc(n: Type::NonNegInt)
    begin
	select(combinat::integerVectors(n, dom::card, hold(MaxPart)=1),
	       dom::isCanonic);
    end_proc;

/*--
    cycles

    The permutation in cycle representation (list of its cycles)
--*/

    cycles;

/*--
    cycleType(g) -> DOM_LIST

    Returns the cycle type of an element of the group. Its a list
    looking like [3,0,0,1,0]. Here it indicates that there are 3
    cycles of length 1 and 1 cycle of length 4. The length of the list
    is of course the size of dom::set. (There cannot be bigger cycles).
--*/

    cycleType := 
    proc(g: dom)
	local a,c;
    begin
	a:=[0 $ dom::card];
	for c in dom::cycles(g) do
	    a[nops(c)]:=a[nops(c)]+1;
	end_for;
	a;
    end_proc;

/*--
    charPolynomial(s: permutation of [1..dom::card], z: free variable) -> polynomial in z
    Returns the characteristic polynomial of a permutation s.
    eg: det(s-z.Id), viewing s as a permutation matrix.
    Its value is the product over the cycles of s 
    of the (1-z^l) where l is the length of the cycle.
    Ref: Sturmfels "Algorithms in Invariant Theory, p 72".
    Bug: Maybe should return a polylib::Poly(z).
--*/

    charPolynomial := 
    proc(g:dom, z: DOM_IDENT)
	local a;
    begin
	a:=dom::cycleType(g);
	_mult((1-z^i)^a[i] $ i = 1..nops(a));
    end_proc;

/*--
    cycleTypes:
    
    Count the elements of the group by cycle type.
    
    It returns a table looking like: table([2,1,0]=34, ...).

    [2,1,0] is the cycle type of at least one of the elements of the
    group, and 34 is the number of elements of the group having this
    cycle type.

    The implementation below is an inefficient stub which requires to
    generate the group, and to compute the cycle type of each of its
    elements. Please overload!
--*/

    cycleTypes :=
    proc()
	local a,g,t;
	option remember;	       
    begin
	t:=table();
	for g in dom::allElements do
	    a:=dom::cycleType(g);
	    if contains(t,a) then
		t[a]:=t[a]+1;
	    else
		t[a]:=1;
	    end_if;
	end_for;
	t;
    end_proc;

/*--
    cycleIndicatorPoly(z:DOM_IDENT) -> polylib::Poly([z[i] $ i=1..dom::card])
    
	Ref: A. Kerber, Algebraic Combinatoric Via Finite Group Action
	2.2 p 70

    This method compute the Cycle Indicator Polynomial ot the module,
    using the cycle types of the elements of the group
    (dom::cycleTypes()).

    It uses the following formula:
    
	1/|G| \sum_{g\inG} \prod_{i=1}^{|X|} z_i^{a_i(g)}

    Its optimized by grouping the elements of the group by cycle type.

    See also dom::cycleTypes().
--*/	
		
    cycleIndicatorPoly := 
    proc(z:DOM_IDENT)
	option remember;
	local Pol, Polya, equa;
    begin
	userinfo(3, "cycleIndicatorPoly: starting computation");
	Pol:=polylib::Poly([z[i] $ i = 1..dom::card], Dom::Rational);
	
	Polya:=Pol(0);
	for equa in dom::cycleTypes() do
	    userinfo(5, "cycleIndicatorPoly: working on cycletype",
		     op(equa,1));
	    /* Cool, the internal representation of a cycle type is
	       exactly the same as the corresponding monomial.
	       This makes for a simple conversion.
	       op(equa,1): cycle type
	       op(equa,2): number of elements of the group with dom cycle type
	    */
	    Polya:=Polya + Pol([[op(equa,2), op(equa,1)]]);
	end_for;
	Polya * Pol(1/dom::size);
    end_proc;

/*-- 
    PolyaSubs(P:DOM_POLY, weights:DOM_LIST)

    Polya substitute the weights in the polynom P. That is if the
    weights are [y_1..y_k], then z_i is substituted by (\sum_j y_j^i).

    P must be a polynomial in the z[1],z[2],... where z is a free identifier.

    Ref: A. Kerber, "Algebraic Combinatoric Via Finite Group Action", 2.2 p 71


    This is just a routine, somewhat independant of this category.
    The best in the future would be to include this into some general
    package dealing with generating series.
--*/

    PolyaSubs :=
    proc(P:DOM_POLY, weights:DOM_LIST)
	local var, eqns;
    begin
	userinfo(3, "Polya Substitution");
	eqns:=[];
	// Hypothesis: The variables of P are of the form z[1], z[2], ...
	// Where z is an identifier. This uses op(z[i],2) to get i.
	for var in op(P,2) do
	    eqns:=eqns.[var = _plus(op(map(weights, id^op(var,2))))];
	end_for;
	subs(expr(P), eqns);
    end_proc;

/*--
    weightEnum(weights:DOM_LIST) -> polynomial expression in the weights
    weights: list of the weights of the element of Y.

    Returns the generating series by weight of the G-classes in
    Y^X. It proceeds by Polya substitution in the cycle indicator
    polynomial.

    Note: Y has to be finite !

    Ref: A. Kerber, Algebraic Combinatoric Via Finite Group Action 2.2 p 71

    Examples:

     - weightEnum([1,1]) returns the number of G-classes of "graphs"
     (that is Y has 2 elements 0 and 1 of same weight).

     - weightEnum([1,z]) returns the number of G-classes of "graphs",
     counted by number of edges (Y has two elements 0 et 1 of
     respectiv weight 1 and z).

--*/

    weightEnum := 
    proc(weights: DOM_LIST)
	option remember;
    begin
	dom::PolyaSubs(dom::cycleIndicatorPoly(hold(z)), weights);
    end_proc;

    // numberOfSetPartitionOrbits(partition: Type::ListOf(Type::NonNegInt))
    //	-> Type::NonNegInt
    //
    // Preconditions: p is a partition, i.e. a decreasing list of
    // nonnegative integers.
    //
    // Returns the number of orbits of unordered set partitions
    // such that p[i] is the size of the ith largest part.

    numberOfSetPartitionsOrbits :=
    proc(partition: Type::ListOf(Type::NonNegInt))
	option remember;
    begin
	_plus(op(map(poly2list(dom::cycleIndicatorPoly(hold(z))),
		     proc(term, colors)
			 local partition, coef;
		     begin
			 ([coef, partition]):=term;
			 coef*
			 combinat::numberOfSetPartitionColorings(partition, colors)
		     end_proc,
		     partition)));
    end_proc;

    // numberOfCanonicListsByShape(l: DOM_LIST) -> DOM_INT
    //
    // returns the number of orbits of lists having the same shape as l.

    numberOfCanonicListsByShape :=
    proc(l: DOM_LIST) : Type::PosInt
    begin
	dom::numberOfSetPartitionsOrbits
	    (combinat::words::evaluationPartition(l));
    end_proc;


/*-- Conversions to / from other software packages: --*/

/*--
    Invar()
    
    Returns a set of generators of the group for the maple package
    invar in the form:
	permgroup(card, { g1, g2, g3 })
    where g1, g2, g3 are in disjoint cycle notation, with cycles of
    length one removed.
--*/

    Invar := 
    proc()
	local l,g;
    begin
	l:={};
	for g in dom::generators do
	    l:=l union
	       { select(dom::cycles(g), c->bool(nops(c)>1)) };
	end_for;
	expr2text(hold(permgroup)(dom::card, l));
    end_proc;

/*--
    Invar2()
    
    Returns a set of generators of the group for the second version of
    the maple package Invar. Looks like:
	pgroup({ g1, g2, g3 }, 0, 0, card)
    where g1, g2, g3 are in disjoint cycle notation, with cycles of
    length one removed.
    
    This can be called with up to two optionnal arguments that will be 
    passed as argument 2 and 3 of pgroup. See ?pgroup in maple.
--*/

    Invar2 := 
    proc()
	local l,g;
    begin
	l:={};
	for g in dom::generators do
	    l:=l union
	       { select(dom::cycles(g), c->bool(nops(c)>1)) };
	end_for;
	if args(0)>2 then error("Too many arguments"); end_if;
	expr2text(hold(pgroup)(l, args(), 0 $ 2-args(0), dom::card));
    end_proc;

    magma :=
    proc()
    begin
        _concat("PermutationGroup<",
                expr2text({$1..dom::card})," | [",
                expr2text(dom::generators[1]),
                ", ". expr2text(dom::generators[i])
                $ i=2..nops(dom::generators), "]>"
               );
    end_proc;

/*--
    isCyclic()

    Returns true if the group is cyclic, that is generated by one of
    its elements.
--*/

    isCyclic :=
    proc()
        local t, o, i;
    begin
        for t in dom::cycleTypes() do
            t := op(t,1);
            o := 1;
            for i from 1 to nops(t) do
                if t[i] <> 0 then
                    o := lcm(o,i);
                end_if;
            end_for;
            if o = dom::size then
                return(TRUE);
            end_if;
        end_for;
        return(FALSE);
    end_proc;

begin

end_category:

stdlib::LoadlibState["Cat::PermutationGroup"] := TRUE:
