#++
Cat::SolvableAlgebra -- the category of solvable algebras

Author: Marcus Hausdorf, Uni Mannheim
        (hausdorf@math.uni-mannheim.de)
Created: 	Tue Oct  1 13:43:28 2001
Last update: 	Thu Jan 31 15:52:01 2002
++#

#++ 
Cat::SolvableAlgebra(R) is the category of solvable algebras of polynomial type, i.e. its 
elements are polynomial rings with a (not necessarily commutative) multiplication '.' 
satisfying a.f = a*f and le(f.g)=le(f)+le(g) for all a in R and the chosen term order. In such 
algebras the computation of Groebner and involutive bases is possible.
    
Methods:

lcoeff, lmonomial, lterm, nterms, nthterm, nthcoeff, nthmonomial, tcoeff, coeff, multcoeffs,
mapcoeffs: see the corresponding polynomial methods
coeffRing: returns R
variables: returns the polynomial indeterminates
numVars: returns the number of polynomial indeterminates
exponent2monomial: converts a multiindex into a monomial
monomial2exponent: converts a monomial into a multiindex
exponentOrder, monomialOrder, elemOrder: -1/0/1 order methods
exponentLess, monomialLess, elemLess: TRUE/FALSE order methods
reduce: reduce an element with respect to another element
autoreduce: autoreduce a list of elements
nForm: compute the normal form of an element with respect to a list of elements
involutiveNormalForm: involutive normal form computation
involutiveComplete: involutive completion
++#

alias(A_ERR = error("wrong type of arguments"),
      N_ERR = error("wrong number of arguments"));
  
category Cat::SolvableAlgebra(R)
    
    category Cat::Ring, Cat::LeftModule(R);
    
    //==================
    // undefined methods
    //==================
    
    lcoeff;		
    lmonomial;
    lterm;
    nterms;
    nthterm;
    nthcoeff;
    nthmonomial;
    tcoeff;
    coeff;
    multcoeffs;
    mapcoeffs;
    coeffRing;
    variables;
    numVars;
    exponent2monomial;
    monomial2exponent;
    exponentOrder;
    reduce;
    
    //================
    // generic methods
    //================
    
    exponentLess := ((l1, l2) -> (bool(dom::exponentOrder(l1, l2) = -1)));
    
    monomialOrder := ((l1, l2) -> (dom::exponentOrder(dom::monomial2exponent(l1)),
	                           dom::monomial2exponent(l2)));
						 
    monomialLess := ((l1, l2) -> (bool(dom::exponentOrder(dom::monomial2exponent(l1),
						          dom::monomial2exponent(l2)) = -1)));
						     
    lexponent := (l -> (dom::monomial2exponent(dom::lmonomial(l, (args(2..args(0)))))));
    
    elemOrder := proc(f:dom, g:dom):DOM_INT
	local tmp:DOM_INT;
	begin
	    if not iszero((tmp := dom::exponentOrder(dom::lexponent(f), 
						     dom::lexponent(g)))) 
	    then
		return(tmp);
	    elif dom::nterms(f) = 1 then
		if dom::nterms(g) = 1 then
		    return(0)
		else
		    return(-1)
		end_if;
	    elif dom::nterms(g) = 1 then
		return(1);
	    else
		return(dom::elemOrder((dom::lmonomial(f, Rem))[2],
		    (dom::lmonomial(g, Rem))[2]));
	    end_if; 
	end_proc; 
	
    elemLess := ((l1, l2) -> (bool(dom::elemOrder(l1, l2) = -1)));
    
    autoreduce := proc(F:Type::ListOf(dom))
	local n:DOM_INT, G:DOM_LIST, g:dom, f:dom, reduced:DOM_BOOL, tmp, hr_f:DOM_BOOL,
	    DOM:DOM_DOMAIN;
	begin
	    if args(0) < 1 or args(0) > 2 then
		N_ERR;
	    end_if;
	    F := sort(F, dom::elemLess);
	    DOM := dom;
	    tmp := map(F, l -> [DOM::lcoeff(l), DOM::lexponent(l)]);
	    if args(0) = 2 and args(2) = hold(Head) then
		hr_f := TRUE;
		F := zip(F, tmp, (l1, l2) -> table("elem" = l1, "mons" = [l2], "lmon" = l2));
	    else
		hr_f := FALSE;
		F := zip(F, tmp, (l1, l2) -> table("elem" = l1, "mons" = DOM::poly2list(l1),
		    "lmon" = l2));
	    end_if;
	    repeat
		reduced := FALSE;
		n := nops(F) + 1;
		repeat
		    n := n - 1;
		    G := F;
		    f := F[n];
		    delete G[n];
		    g := dom::intNForm(f, G, hr_f);
		    reduced := bool(f <> g);
		until (n = 1 or reduced) end_repeat;
		if reduced then
		    delete F[n];
		    F := listlib::insert(F, g, 
			(l1, l2) -> DOM::elemLess(l1["elem"], l2["elem"]));
		end_if
	    until not reduced end_repeat;
	    map(F, l -> l["elem"]);
	end_proc;
	
    nForm := proc(f:dom, F:Type::ListOf(dom))
	local hr_f:DOM_BOOL, a1, a2, DOM:DOM_DOMAIN;
	begin
	    if args(0) < 2 or args(0) > 3 then
		N_ERR;
	    end_if;
	    if args(0) = 3 and args(3) = hold(Head) then
		hr_f := TRUE;
		a1 := table("elem" = f, "mons" = [[dom::lcoeff(f), dom::lexponent(f)]]);
	    else
		hr_f := FALSE;
		a1 := table("elem" = f, "mons" = dom::poly2list(f));
	    end_if;
	    DOM := dom;
	    a2 := map(F, l -> table("elem" = l, "lmon" = [DOM::lcoeff(l), DOM::lexponent(l)]));
	    (dom::intNForm(a1, a2, hr_f))["elem"];
	end_proc;
	
    intNForm := proc(f, G, hr_f)
	local reduced:DOM_BOOL, d, g, t, isDivisor:DOM_PROC; 
	begin
	    isDivisor := proc(mu1, mu2)
	    local tmp, k;
	    begin
		tmp := zip(mu2, mu1, _subtract);
		for k from 1 to nops(mu1) do
		    if tmp[k] < 0 then
			return(0);
		    end_if;
		end_for;
		tmp;
	    end_proc;	    
	
	    repeat
		reduced := FALSE;
		for g in G do
		    for t in f["mons"] do
			if not iszero((d := isDivisor(g["lmon"][2], t[2]))) then
			    reduced := TRUE;
			    f := table(
				"elem" = dom::reduce(f["elem"], g["elem"], g["lmon"], t, d),
				"lmon" = [dom::lcoeff(f["elem"]), dom::lexponent(f["elem"])]); 
			    if hr_f then
				f["mons"] := [f["lmon"]];
			    else
				f["mons"] := dom::poly2list(f["elem"]);
			    end_if;
			end_if;
			if reduced then break end_if;
		    end_for;
		    if reduced then break end_if;
		end_for;
	    until (dom::iszero(f["elems"])) or not reduced end_repeat;
	    f;
	end_proc;			

    
    //=======================
    // involutive normal form
    //=======================
    
    // involutiveNormalForm(f, F, L) computes the involutive normal form of f modulo the set F 
    // w.r.t. the involutive division L. The switch HeadReduce (default: FALSE) determines 
    // whether full or only head reductions are used. 
     
    involutiveNormalForm := proc(f:dom, F:Type::ListOf(dom), L:DOM_DOMAIN)
	local hr_f:DOM_BOOL, a1, a2, DOM:DOM_DOMAIN;
	begin
	    if args(0) < 3 or args(0) > 4 then
		N_ERR;
	    end_if;
	    if args(0) = 4 and args(4) = hold(Head) then
		hr_f := TRUE;
		a1 := table("elem" = f, "mons" = [[dom::lcoeff(f), dom::lexponent(f)]]);
	    else
		hr_f := FALSE;
		a1 := table("elem" = f, "mons" = dom::poly2list(f));
	    end_if;
	    DOM := dom;
	    a2 := map(F, l -> table("elem" = l, "lmon" = [DOM::lcoeff(l), DOM::lexponent(l)]));
	    (dom::intInvNForm(a1, a2, L::separationTableMult([a2[i]["lmon"][2] $ i = 1..nops(a2)]),
		L, hr_f))["elem"];
       end_proc; 
	    	
    intInvNForm := proc(f, G, sep_t, L, hr_f)
	local reduced:DOM_BOOL, g, t, d;
	begin
	    repeat
		reduced := FALSE;
		for g in G do
		    for t in f["mons"] do
			if not iszero((d := L::isInvDivisor(g["lmon"][2], t[2], 
				sep_t[g["lmon"][2]]))) 
			then
			    reduced := TRUE;
			    f := dom::reduce(f["elem"], g["elem"], g["lmon"], t, d);
			    f := table("elem" = f, 
				"lmon" = [dom::lcoeff(f), dom::lexponent(f)]);
			    if hr_f then
				f["mons"] := [f["lmon"]];
			    else
				f["mons"] := dom::poly2list(f["elem"]);
			    end_if;
			end_if;
			if reduced then break end_if;
		    end_for;
		    if reduced then break end_if;
		end_for;
	    until (dom::iszero(f["elems"])) or not reduced end_repeat;
	    f;
	end_proc;			
    
    involutiveComplete := proc(F:Type::ListOf(dom), L)
	local DOM:DOM_DOMAIN, T:DOM_LIST, Q:DOM_LIST, sepm_t:DOM_TABLE, sepnm_t:DOM_TABLE,  
	    doneNMP_t:DOM_TABLE, criterion:DOM_PROC, g, h, hr_f:DOM_BOOL, tmp, k:DOM_INT, 
	    noMore:DOM_BOOL;
	begin
	    // argument checks 
	    if args(0) > 3 or args(0) < 2 then
		N_ERR;
	    end_if;
	    if args(0) = 3 and args(3) = hold(Head) then
		hr_f := TRUE;
	    else
		hr_f := FALSE;
	    end_if;
	    if (domtype(L) <> DOM_DOMAIN
		or L::hasProp = FAIL
		or not L::hasProp(Cat::InvolutiveDivision))
	    then
		A_ERR;
	    end_if;
	    DOM := dom;
	    // Buchberger's chain (resp. second) criterion
	    criterion := proc(g)
		local f;
		begin
		    for f in T do
			if (not iszero(L::isInvDivisor(f["lmon"][2], g["lmon"][2], 
			    sepm_t[f["lmon"][2]]))
			    and DOM::exponentLess(zip(g["crit"], f["crit"], 
				(l1, l2) -> max(l1, l2)), g["lmon"][2]))
			then
			    print(Unquoted, "The C R I T E R I O N has been invoked");
			    return(TRUE);
			end_if;
		    end_for;
		    FALSE;
		end_proc;
	    // autoreduce F. The result is an ordered list.
	    F := dom::autoreduce(F);
	    // set up T and Q; doneNMP_t lists all nonmultiplicative prolongations
	    // carried out so far.
	    // sepm_t contains the multiplicative variables for the elements in T.
	    F := map(F, l -> table("elem" = l, "lmon" = [DOM::lcoeff(l), DOM::lexponent(l)],
		"crit" = DOM::lexponent(l)));
	    T := [F[1]];
	    Q := [F[i] $ i = 2..nops(F)];
	    sepm_t := L::separationTableMult([T[1]["lmon"][2]]);
	    doneNMP_t := table(F[i]["elem"] = {} $ i = 1..nops(F));
	    repeat
		h := table("elem" = dom::zero);
		while Q <> [] and dom::iszero(h["elem"]) do
		    g := Q[1];
		    delete Q[1];
		    if not criterion(g) then
			if not contains(g, "mons") then
			    if hr_f then
				g["mons"] := [g["lmon"]];
			    else
				g["mons"] := dom::poly2list(g["elem"]);
			    end_if;
			end_if;
			h := dom::intInvNForm(g, T, sepm_t, L, hr_f);
		    end_if    
		end_while;
		if not dom::iszero(h["elem"]) then
		    h["crit"] := h["lmon"][2];
		    if g["lmon"][2] <> h["lmon"][2] then  
			doneNMP_t[h["elem"]] := {};
			delete doneNMP_t[g["elem"]];
			tmp := [];
			for k from nops(T) downto 1 do
			    if dom::exponentLess(h["lmon"][2], T[k]["lmon"][2]) then
				tmp := append(tmp, T[k]);
				delete T[k];
			    else
				break;
			    end_if;
			end_for;
			if tmp <> [] then
			    Q := listlib::singleMerge(tmp, Q, 
				(l1, l2) -> (DOM::exponentLess(l1["lmon"][2], l2["lmon"][2])));
			end_if;
		    end_if;
		    T := append(T, h);			
		end_if;
		noMore := FALSE;
		repeat 
		    sepm_t := L::separationTableMult([T[i]["lmon"][2] $ i = 1..nops(T)]);
		    sepnm_t := map(sepm_t, l -> ({i $ i = 1..DOM::numVars} minus l));     
		    tmp := sort([[T[i], subsop(T[i]["lmon"][2], j = T[i]["lmon"][2][j] + 1), j] 
			$ j in (sepnm_t[T[i]["lmon"][2]] minus doneNMP_t[T[i]["elem"]])
			$ i = 1..nops(T)], (l1, l2) -> (DOM::exponentLess(l1[2], l2[2])));
		    if (tmp = [] or 
			(Q <> [] and not dom::exponentLess(tmp[1][2], Q[1]["lmon"][2]))) 
		    then
			noMore := TRUE
		    else
			doneNMP_t[tmp[1][1]["elem"]] := 
			    doneNMP_t[tmp[1][1]["elem"]] union {tmp[1][3]};
			g := dom::exponent2monomial(subsop([0 $ dom::numVars], tmp[1][3] = 1)) 
			    * tmp[1][1]["elem"];
			g := table("elem" = g, "lmon" = [dom::lcoeff(g), tmp[1][2]], 
			    "crit" = tmp[1][1]["lmon"][2]);
			if criterion(g) then
			    h := table("elem" = dom::zero);
			else
			    if hr_f then
				g["mons"] := [g["lmon"]];
			    else
				g["mons"] := dom::poly2list(g["elem"]);
			    end_if;
			    h := dom::intInvNForm(g, T, sepm_t, L, hr_f);
			end_if;
			if not dom::iszero(h["elem"]) then
			    doneNMP_t[h["elem"]] := {};
			    if h["lmon"][2] <> g["lmon"][2] then 
				h["crit"] := h["lmon"][2];
			    else  // no reduction
				h["crit"] := g["crit"];
			    end_if;
			    tmp := [];
			    for k from nops(T) downto 1 do
				if dom::exponentLess(h["lmon"][2], T[k]["lmon"][2]) then
				    tmp := append(tmp, T[k]);
				    delete T[k];
				else
				    break;
				end_if;
			    end_for;
			    T := append(T, h);
			    if tmp <> [] then
				Q := listlib::singleMerge(tmp, Q, 
				(l1, l2) -> (DOM::exponentLess(l1["lmon"][2], l2["lmon"][2])));
			    end_if;
			end_if; 
		    end_if;
		until noMore end_repeat;
	    until Q = [] end_repeat;
	    map(T, l -> l["elem"]);
	end_proc;
		
    begin    
    if args(0) <> 1 then error("wrong no of args") end_if;
    if not R::hasProp(Cat::CommutativeRing) then
	error("no commutative ring")
    end_if;
    
end_category;
    
unalias(A_ERR, N_ERR):
