/*++
Dom::DifferentialExpression(DV) -- Domain of general expressions in 
				   differential variables

`Dom::DifferentialExpression(DV)' is an instance of the category
`Cat::DifferentialFunction'.  It implements general expressions in the
differential variables from the domain `DV' (an element of the
category `Cat::DifferentialVariable'). The implementation is based on
the basic type `DOM_EXPR'.

The differential variables are substituted by dummy variables in order
to simplify the handling of the expressions. The conversions are
performed via two hash tables, `Dummy' and `DVar'. New dummies are
created with the local method `makeDummy' which checks whether or not
a variable is already contained in the tables. If not, a new dummy is
generated with prefix `_de_'.


Author: Werner M. Seiler, Uni Karlsruhe/Mannheim
        (wms@ira.uka.de, werner.seiler@math.uni-mannheim.de)
Created: September 30th, 1997
Last update: 	Mon May 21 17:36:10 2001
++*/


alias(REP(x)=extop(x,1),
      VARS(x)=extop(x,2),
      TYPE_ERR=error("wrong type of argument"),
      ARG_ERR=error("wrong number of arguments")):


domain Dom::DifferentialExpression()

    local DVcoerce:DOM_PROC, DVconvto:DOM_PROC, DVexpr:DOM_PROC,
          DVdiff:DOM_PROC, DVindex:DOM_PROC, DVtype:DOM_PROC,
          DVgreater:DOM_PROC, DVrandom:DOM_PROC, 
	  DVnewRep:DOM_PROC, DVrepIndex:DOM_PROC, DVmulti:DOM_PROC, 
	  nn:DOM_INT, mm:DOM_INT, singleMerge:DOM_PROC, arithTypes:DOM_SET,
	  numTypes:DOM_SET, opts:DOM_LIST,
	  Dummy:DOM_TABLE, DVar:DOM_TABLE, 
	  makeDummy:DOM_PROC, elimDummy:DOM_PROC, mapDummy:DOM_PROC,
	  DV:DOM_DOMAIN, DV2:DOM_DOMAIN, DE2:DOM_DOMAIN,
	  DV2coerce:DOM_PROC, DVtest:DOM_PROC, restrictedDV:DOM_BOOL,
	  DV2conv:DOM_PROC, tmp, NORMAL:DOM_PROC, IsZero:DOM_PROC;
	  
    inherits Dom::BaseDomain;
    category Cat::DifferentialFunction(DV), Cat::PartialDifferentialRing,
	     Cat::Field;


    /* ========================= */
    /* Representation, In/Output */
    /* ========================= */

    // Most elements are represented by an element of DOM_EXPR and a set
    // of the differential variables appearing in it. The main exceptions
    // are differential variables and arbitrary functions which are represented
    // by elements of DV and DF, resp., in the first slot. The set in the
    // second slot may contain differential variables which do not explicitly
    // appear in the expression; however, no variable not contained in the
    // set can appear.
      
      
    // The following method is only used for the generation of elements,
    // if one wants to ensure that the second slot contains no redundant
    // differential variables. The checks with `has' are rather expensive.
    generate :=
	proc(x,vars:DOM_SET):dom
	local dv:DV;
	begin
	    case domtype(x)
	    of DV do return(new(dom,makeDummy(x),{x}));
	    of DOM_EXPR do
		for dv in vars do
		    if not(has(x,Dummy[dv])) then
			vars := vars minus {dv};
		    end_if;
		end_for;
		return(new(dom,NORMAL(x),vars));
	    otherwise return(new(dom,NORMAL(x),{}));	// ???
	    end_case;
	end_proc;
	

    // If we are dealing with a restricted domain of differential variables,
    // we must first convert to the corresponding unrestricted domain and
    // then explicitly check whether the restrictions are satisfied.
    convert := if restrictedDV then
	proc(x)
        local tmp, T:DOM_DOMAIN;
	begin
	    T := domtype(x);
	    case T
            of dom do return(x);
            of DV do return(new(dom,makeDummy(x),{x}));
            of DOM_EXPR do return(dom::convertExpr(x));
	    of DOM_POLY do return(dom::convert(expr(x)));
	    of DOM_IDENT do
		tmp := DV2coerce(x);
		if tmp=FAIL then return(new(dom,x,{}))
		elif DVtest(tmp) then
		    tmp := DV2conv(tmp);
		    return(new(dom,makeDummy(tmp),{tmp}));
		else return(FAIL) end_if;
	    otherwise
		if contains(numTypes,T) then
		    return(new(dom,NORMAL(x),{}));
		elif T::hasProp<>FAIL then
		    if T::hasProp(Cat::DifferentialFunction(DV)) then
			return(dom::convert(NORMAL(T::convert_to(x,DOM_EXPR))));
		    elif T::constructor=Dom::ExpressionField then
			return(dom::convert(NORMAL(expr(x))));
		    elif restrictedDV and (T=DV2) then
			if DVtest(x) then
			    x := DV2conv(x);
			    return(new(dom,makeDummy(x),{x}));
			else return(FAIL) end_if;
		    end_if;
		end_if;
		tmp := DV2coerce(x);
		if tmp=FAIL then return(FAIL)
		elif DVtest(tmp) then
		    tmp := DV2conv(tmp);
		    return(new(dom,makeDummy(tmp),{tmp}));
		else return(FAIL) end_if;
            end_case;
	end_proc;
    else
        proc(x)
        local tmp, T:DOM_DOMAIN;
        begin
	    T := domtype(x);
	    case T
            of dom do return(x);
            of DV do return(new(dom,makeDummy(x),{x}));
            of DOM_EXPR do return(dom::convertExpr(x));
	    of DOM_POLY do return(dom::convert(expr(x)));
	    of DOM_IDENT do
		tmp := DVcoerce(x);
		if tmp=FAIL then return(new(dom,x,{}))
		else return(new(dom,makeDummy(tmp),{tmp})) end_if;
	    otherwise
		if contains(numTypes,T) then
		    return(new(dom,NORMAL(x),{}));
		elif T::hasProp<>FAIL then
		    if T::hasProp(Cat::DifferentialFunction(DV)) then
			return(dom::convert(NORMAL(T::convert_to(x,DOM_EXPR))));
		    else return(FAIL) end_if; // ExpressionField ???
		else
		    tmp := DVcoerce(x);
		    if tmp=FAIL then return(FAIL)
		    else return(new(dom,makeDummy(tmp),{tmp})) end_if;
		end_if;
            end_case;
        end_proc;
    end_if;


    // converts recursively elements of DOM_EXPR to dom
    convertExpr :=
	if restrictedDV then
	    proc(x:DOM_EXPR)
	    local convOps:DOM_PROC, dvars:DOM_SET, fail:DOM_BOOL, res;
	    begin
		// auxiliary method for recursively recognizing differential
		// variables and collecting them in `dvars'
		// eliminates all unwanted types
		convOps := proc(x)
		    local tmp;
		    begin
			tmp := DV2coerce(x);
			if tmp<>FAIL then
			    if DVtest(tmp) then
				tmp := DV2conv(tmp);
				dvars := dvars union {tmp};
				return(makeDummy(tmp));
			    else
				fail := TRUE;
				return(FAIL);
			    end_if;
			end_if;
			case domtype(x)
			of DOM_EXPR do 
			    if not(testtype(x,Type::Arithmetical)) then
				fail := TRUE;
				return(FAIL);
			    end_if;
			    return(map(x,convOps));
			of DV do
			    dvars := dvars union {x};
			    return(makeDummy(x));
			of DV2 do
			    if DVtest(x) then
				tmp := DV2conv(x);
				dvars := dvars union {tmp};
				return(makeDummy(x));
			    else
				fail := TRUE;
				return(FAIL);
			    end_if;
			of dom do
			    dvars := dvars union VARS(x);
			    return(REP(x));
			otherwise 
			    if testtype(x,Type::Arithmetical) then return(x)
			    else
				fail := TRUE;
				return(FAIL);
			    end_if;
			end_case;
		    end_proc;
		dvars := {};
		fail := FALSE;
		res := convOps(x);
		if fail then FAIL else new(dom,convOps(NORMAL(x)),dvars) end_if;
	    end_proc;
	else
	    proc(x:DOM_EXPR)
	    local convOps:DOM_PROC, dvars:DOM_SET, fail:DOM_BOOL, res;
	    begin
		// auxiliary method for recursively recognizing differential
		// variables and collecting them in `dvars'
		// eliminates all unwanted types
		convOps := proc(x)
		    local tmp;
		    begin
			tmp := DVcoerce(x);
			if tmp<>FAIL then
			    dvars := dvars union {tmp};
			    return(makeDummy(tmp));
			end_if;
			case domtype(x)
			of DOM_EXPR do 
			    if not(testtype(x,Type::Arithmetical)) then
				fail := TRUE;
				return(FAIL);
			    end_if;
			    return(map(x,convOps));
			of DV do
			    dvars := dvars union {x};
			    return(makeDummy(x));
			of dom do
			    dvars := dvars union VARS(x);
			    return(REP(x));
			otherwise 
			    if testtype(x,Type::Arithmetical) then return(x)
			    else
				fail := TRUE;
				return(FAIL);
			    end_if;
			end_case;
		    end_proc;
		dvars := {};
		fail := FALSE;
		res := convOps(x);
		if fail then FAIL else new(dom,convOps(NORMAL(x)),dvars) end_if;
	    end_proc;
	end_if;


    convert_to :=
        proc(x:dom,T)
        begin
            case T
            of dom do return(x);
            of DOM_EXPR do return(subs(REP(x),DVar));
	    of DOM_POLY do
	    	if args(0)=2 then
		    return(dom::polyDE(x,[op(VARS(x))]))
		else return(dom::polyDE(x,args(3..args(0)))) end_if;
            of DV do
                if domtype(REP(x))=DOM_IDENT and contains(DVar,REP(x)) then 
		    return(DVar[REP(x)]);
                else return(FAIL) end_if;
            of "diff" do
            of "D" do
		return(subs(REP(x),
		            op(map(VARS(x), 
				   (dv -> (Dummy[dv]=DVconvto(dv,T)))))));
            otherwise return(FAIL);
            end_case;
        end_proc;


    // `expr' recursively checks all contained identifiers whether
    // they are dummies.
    expr :=
        proc(x:dom):Type::Arithmetical
        begin
	    subs(REP(x), op(map(VARS(x), (dv -> (Dummy[dv]=DVexpr(dv))))));
	end_proc;


    // Conversion to DOM_POLY; cannot be called `poly' because of 
    // naming conflict with kernel function `poly'.
    polyDE :=
        proc(de:dom):DOM_POLY
	local res:DOM_POLY, vars:DOM_LIST, l:DOM_LIST, tmp;
	begin
	    // conversion to DOM_EXPR, eliminates dummies!
	    tmp := dom::convert_to(de,DOM_EXPR);
	    // conversion to polynomial
            if args(0)=1 then return(poly(tmp,[op(VARS(de))])) end_if;
	    if domtype(args(2))=DOM_LIST then
	        vars := map(args(2),DVcoerce);
		if not(iszero(contains(vars,FAIL))) then TYPE_ERR end_if;
	        res := poly(tmp,vars,args(3..args(0)));
	    else
	        res := poly(tmp,[op(VARS(de))],args(2..args(0)));
	    end_if;
	    // if option `Expr' was given, coefficients must be checked
	    // for differential variables.
	    if contains({args()},hold(Expr)) then
	        l := poly2list(res);
	        l := map(l, 
			 ((c,v) -> 
			  (subsop(c,1=misc::maprec(c[1],
					{DOM_IDENT}=
					    (vv -> mapDummy(vv,DVexpr)),
					PostMap),
			          Unsimplified))), vars);
	        res := poly(l,op(res,2));
	    end_if;
	    res;
	end_proc;


    // only as good as `normal'!
    equal :=
        proc(x1:dom,x2:dom):DOM_BOOL
        begin IsZero(REP(x1)-REP(x2)) end_proc;


    /* ====================== */
    /* Differential Variables */
    /* ====================== */

    // In all of the following methods we must take into account that
    // VARS(x) contains in general also variables which do not appear
    // in REP(x) because of cancellations in arithmetic operations or in
    // differentiations. Thus we must explicitly test all elements of 
    // VARS(x) with has.

    indets :=
        proc(x:dom):DOM_SET
        local res:DOM_SET, opts:DOM_LIST, flag:DOM_BOOL, pos:DOM_INT, dv:DV;
        begin
	    if args(0)>1 then
		opts := [args(2..args(0))];
		pos := contains(opts,hold(DiffVars));
		flag := iszero(pos);
		if not(flag) then delete opts[pos] end_if;
	    else 
		flag := TRUE;
		opts := [];
	    end_if;

	    if flag or opts<>[] then 
		map(indets(REP(x),op(opts)),elimDummy);
	    else
		res := VARS(x);
		for dv in res do
		    if not(has(REP(x),Dummy[dv])) then
			res := res minus {dv};
		    end_if;
		end_for;
		res;
	    end_if;
        end_proc;


    leader :=
        proc(de:dom)
        local vars:DOM_LIST;
        begin
            vars := sort([op(VARS(de))],DVgreater);
	    while vars<>[] do
		if has(REP(de),Dummy[vars[1]]) then return(vars[1]);
		else delete vars[1] end_if;
	    end_while;
            return(FAIL);
        end_proc;


    has :=
	proc(de:dom,x):DOM_BOOL
	begin
	    case domtype(x)
	    of DV do 
		if not(contains(VARS(de),x)) then return(FALSE)
		else return(has(REP(de),Dummy[x])) end_if;
	    of dom do return(has(REP(de),REP(x)));
	    otherwise return(has(REP(de),x));
	    end_case;
	end_proc;


    /* =========== */
    /* Arithmetics */
    /* =========== */

    zero := new(dom,0,{});


    one:= new(dom,1,{});


    iszero :=
        proc(x:dom):DOM_BOOL
        begin IsZero(REP(x)) end_proc;


    isone :=
        proc(x:dom):DOM_BOOL
        begin bool(REP(x)=1) end_proc;


    _plus :=
        proc():dom
	local DOM:DOM_DOMAIN, summ:DOM_LIST, tmp;
        begin
	    if testargs() then
		for tmp in [args()] do
		    if not(contains(_union({dom},arithTypes),
		                    domtype(tmp))) then TYPE_ERR end_if;
		end_for;
	    end_if;
	    summ := [args()];
	    DOM := dom;
	    tmp := _plus(map(op(summ),
			     ((x) -> (if domtype(x)=DOM then REP(x)
				      else x end_if))));
	    if contains(numTypes,domtype(tmp)) then new(dom,NORMAL(tmp),{})
	    else
		new(dom, NORMAL(tmp),
		    _union(map(op(summ),
			       ((x) -> (case domtype(x)
					of DOM do return(VARS(x));
					of DV do return({x});
					otherwise return({}) end_case)))));
	    end_if;
	end_proc;


    _mult :=
        proc():dom
	local DOM:DOM_DOMAIN, facts:DOM_LIST, tmp;
        begin
	    if testargs() then
		for tmp in [args()] do
		    if not(contains(_union({dom},arithTypes),
		                    domtype(tmp))) then TYPE_ERR end_if;
		end_for;
	    end_if;
	    facts := [args()];
	    DOM := dom;
            tmp := _mult(map(op(facts),
			     ((x) -> (if domtype(x)=DOM then REP(x)
				      else x end_if))));
	    if contains(numTypes,domtype(tmp)) then new(dom,NORMAL(tmp),{})
	    else
		new(dom, NORMAL(tmp),
		    _union(map(op(facts),
			       ((x) -> (case domtype(x)
					of DOM do return(VARS(x));
					of DV do return({x});
					otherwise return({}) end_case)))));
	    end_if;
	end_proc;


    _negate :=
        proc(x:dom):dom
        begin new(dom,NORMAL(_negate(REP(x))),VARS(x)) end_proc;


    _invert :=
        proc(x:dom):dom
        begin new(dom,NORMAL(_invert(REP(x))),VARS(x)) end_proc;


    /* ==================== */
    /* Differential Methods */
    /* ==================== */

    // `diff' does not check whether the differentiations eliminate some
    // differential variables. It only checks whether the result is a
    // trivial constant.
    diff :=
	proc(de:dom):dom
	local vars:DOM_LIST, i:DOM_INT, res;
	begin
	    if args(0)=1 then return(de) end_if;
	    vars := [args(2..args(0))];
	    for i from 1 to args(0)-1 do
		if domtype(vars[i])=DV then
		    if not(contains(VARS(de),vars[i])) then
			return(dom::zero);
		    end_if;
		    vars[i] := Dummy[vars[i]];
		end_if;
	    end_for;
	    res := diff(REP(de),op(vars));
	    if contains(numTypes,domtype(res)) then new(dom,NORMAL(res),{})
	    else new(dom,NORMAL(res),VARS(de)) end_if;
	end_proc;
	
	
    // The default implementation of `totalDiff' is overwritten in order to
    // perform all computations directly on the representation.
    totalDiffJMnew :=
        proc(sys:DOM_LIST, mu:DOM_INT, JM)
        local n:DOM_INT, i:DOM_INT, j:DOM_INT, k:DOM_INT, pos:DOM_INT,
	      res:DOM_LIST, t:DOM_LIST, rvars:DOM_LIST, rows:DOM_LIST,
	      vars:DOM_LIST, nvars:DOM_LIST, dvars:DOM_LIST, dummies:DOM_LIST;
        begin
	    // get data from jacobian
            n := (dom::jacobianType)::nrows(JM);
            vars := (dom::jacobianType)::allIndices(JM);
            if vars=[] then
                return(([dom::zero $ n], [[] $ n]));
            end_if;
            rows := (dom::jacobianType)::rows(JM);
	    // select derivatives and generate needed dummies
            nvars := select(vars,
                             ((jv) -> contains({"Deriv","Dep"},DVtype(jv))));
            dvars := map(nvars,DVdiff,mu);
	    dummies := map(dvars,makeDummy);
            vars := [ [] $ n ];
	    // set up prolonged equations
            for i from 1 to n do         // for each equation
		if rows[i]=[] then
		    sys[i] := dom::zero;
		    next;
		end_if;
                rvars := map(rows[i],op,1);
                j := nops(rvars);
                vars[i] := [ 0 $ j ];
                res := [ 0 $ j ];
                j := 0;
		k := 0;
                for t in rows[i] do      // compute derivative
                    case DVtype(t[1])
                    of "Dep" do
                    of "Deriv" do
                        pos := contains(nvars,t[1]);
			k := k+1;
                        res[k] := REP(t[2])*dummies[pos];
                        j := j+1;
                        vars[i][j] := dvars[pos];
                        break;
                    of "Indep" do
                        if DVindex(t[1])=mu then
			    k := k+1;
                            res[k] := REP(t[2]);
                        end_if;
                        break;
                    end_case;
                end_for;
		if iszero(k) then
		    sys[i] := dom::zero;
		    vars[i] := [];
		    next;
		end_if;
		// collect results
                vars[i] := [op(vars[i],1..j)];
                vars[i] := vars[i] . rvars;
		sys[i] := new(dom, _plus(op(res,1..k)), {op(vars[i])});
            end_for;
            (sys,vars);
        end_proc;


    /* ============== */
    /* Simplification */
    /* ============== */

    // The following methods are all lifts of the corresponding kernel
    // functions. As most of them attempt to determine some kind of normal
    // form, the result is computed with the help of `generate', i.e.
    // all differential variables in the set `VARS(x)' are explicitly
    // check for their occurance.
    
    normal :=
        proc(x:dom):dom
	local tmp;
        begin
	    if domtype(REP(x))=DOM_EXPR then
		tmp := normal(REP(x),args(2..args(0)));
		if tmp=REP(x) then return(x)
		elif contains(numTypes,domtype(tmp)) then
		    return(new(dom,tmp,{}));
		else return(dom::generate(tmp,VARS(x))) end_if;
	    else return(x) end_if;
	end_proc;


    simplify :=
        proc(x:dom):dom
	local tmp;
        begin
	    if domtype(REP(x))=DOM_EXPR then
		tmp := simplify(REP(x),args(2..args(0)));
		if tmp=REP(x) then return(x)
		elif contains(numTypes,domtype(tmp)) then
		    return(new(dom,tmp,{}));
		else return(dom::generate(tmp,VARS(x))) end_if;
	    else return(x) end_if;
	end_proc;


    combine :=
	proc(x:dom):dom
	local tmp;
	begin
	    if domtype(REP(x))=DOM_EXPR then
		tmp := combine(REP(x),args(2..args(0)));
		if tmp=REP(x) then return(x)
		else new(dom,tmp,VARS(x)) end_if;
	    else return(x) end_if;
	end_proc;
	
	
    radsimp :=
	proc(x:dom):dom
	local tmp;
	begin
	    if domtype(REP(x))=DOM_EXPR then
		tmp := radsimp(REP(x),args(2..args(0)));
		if tmp=REP(x) then return(x)
		else new(dom,tmp,VARS(x)) end_if;
	    else return(x) end_if;
	end_proc;
	
	
    expand :=
	proc(x:dom):dom
	local tmp;
	begin
	    if domtype(REP(x))=DOM_EXPR then
		tmp := expand(REP(x),args(2..args(0)));
		if tmp=REP(x) then return(x)
		elif contains(numTypes,domtype(tmp)) then
		    return(new(dom,tmp,{}));
		else dom::generate(tmp,VARS(x)) end_if;
	    else return(x) end_if;
	end_proc;
	
	
    rewrite :=
	proc(x:dom):dom
	local tmp;
	begin
	    if domtype(REP(x))=DOM_EXPR then
		tmp := rewrite(REP(x),args(2..args(0)));
		if tmp=REP(x) then return(x)
		else new(dom,tmp,VARS(x)) end_if;
	    else return(x) end_if;
	end_proc;
	
    
    // `eval' is not only a lift of the corresponding kernel function but
    // can also be used to enter an ansatz into given expressions.
    // As this command will often be used interactively, a `coerce' of
    // the second argument is tried, if it is not of the right type.
    
    eval :=
        proc(sys)
	local tmp:DOM_EXPR, sol:DOM_LIST, rep, sysflag:DOM_BOOL, i:DOM_INT, 
	      j:DOM_INT, allvars:DOM_SET, vars:DOM_LIST, dersol:DOM_LIST, 
	      res, slist:DOM_LIST, SDOM:DOM_DOMAIN, f:dom;
        begin
	    // analysis of input
	    if domtype(sys)<>DOM_LIST then
		sysflag := FALSE;
		sys := [sys];
	    else sysflag := TRUE end_if;
	    if args(0)>1 then 		// ansatz given
		sol := args(2);
		if mm=1 and domtype(sol)<>DOM_LIST then sol := [sol] end_if;
		SDOM := domtype(sol[1]);
	    end_if;
	    // check input
	    if testargs() then
	        if {op(map(sys,domtype))}<>{dom} then TYPE_ERR end_if;
		if args(0)>1 then
		    if args(0)>2 then ARG_ERR end_if;
		    if nops(sol)<>mm then
			error("wrong number of functions given");
		    end_if;
		    if SDOM::hasProp=FAIL or 
		       not(SDOM::hasProp(Cat::DifferentialFunction(DV)))
		    then		// try to convert ansatz
			SDOM := Dom::DifferentialExpression(
				    Dom::RestrictedDifferentialVariable(DV,
					    hold(Types)="Indep"));
		        sol := map(sol,SDOM::convert);
			if not(iszero(contains(sol,FAIL))) then 
			    TYPE_ERR 
			end_if;
		    end_if;
		end_if;
	    end_if;

	    if args(0)=1 then		// lift of kernel function `eval'
		res := [ 0 $ nops(sys) ];
		i := 0;
		for f in sys do
		    i := i+1;
		    rep := REP(f);
		    if domtype(rep)=DOM_EXPR then
			tmp := eval(rep);
			if tmp=rep then res[i] := f
			elif contains(numTypes,domtype(tmp)) then
			    res[i] := new(dom,tmp,{});
			else
			    res[i] := dom::generate(tmp,VARS(f));
			end_if;
		    else res[i] := f end_if;
		end_for;
	    elif args(0)=2 then		// evaluation of ansatz
		res := [ 0 $ nops(sys) ];
		// get all contained derivatives
		allvars := _union(op(map(sys,dom::indets,hold(DiffVars))));
		allvars := select(allvars,
			      ((jv) -> not(iszero(contains(["Dep","Deriv"],
							   DVtype(jv))))));
		// compute derivatives via "derivative tree"
		vars := [ 0 $ mm ];
		dersol := [ 0 $ mm ];
		slist := [];
		for j from 1 to mm do
		    vars[j] := [op(select(allvars, 
					  ((jv) -> bool(DVindex(jv)=j))))];
		    if not(iszero(nops(vars[j]))) then
			if DV::notation="repeated" then
			    dersol[j] := SDOM::multiDiff(sol[j],
						     map(vars[j],DVrepIndex));
		        else
			    dersol[j] := SDOM::multiDiff(sol[j],
						     map(vars[j],DVmulti));
		        end_if;
		    	slist := slist . zip(map(vars[j],(v -> Dummy[v])),
			                     dersol[j],_equal);
		    end_if;
		end_for;
		// enter derivatives in expressions
		res := map(sys, (f -> SDOM(subs(REP(f),slist))));
	    else ARG_ERR end_if;
	    if sysflag then res else res[1] end_if;
	end_proc;


    /* ============= */
    /* Miscellaneous */
    /* ============= */

    subs := dom::multiSubs;
    
    
    // Single substitution
    // Again we perform some coercions of the second argument in order to 
    // simplify interactive use.
    subsOne :=
        proc(de:dom,eq:DOM_EXPR):dom
	local y, v, vars:DOM_SET, res;
	begin
	    // analysis of input
	    v := op(eq,1);
	    y := op(eq,2);
	    vars := VARS(de);
	    case domtype(v)
	    of dom do
		v := REP(v);
		break;
	    of DV do
		if not(contains(vars,v)) then return(de) end_if;
		vars := vars minus {v};
		v := Dummy[v];
		break;
	    end_case;
	    // substitution
	    res := REP(de);
	    case domtype(y)
	    of dom do
		vars := VARS(y) union vars;
		res := subs(res,v=REP(y));
		break;
	    of DV do
		vars := vars union {y};
		res := subs(res,v=makeDummy(y));
		break;
	    otherwise
		y := dom::convert(y);
		if y=FAIL then error("illegal substitution")
		else
		    vars := VARS(y) union vars;
		    res := subs(res,v=REP(y));
		end_if;
	    end_case;
	    new(dom,NORMAL(res),vars);
	end_proc;


    // Simple lift of the kernel function `solve';
    // coercion of the second argument performed.
    solve :=
        proc(de:dom,dv)
	local sol, vars:DOM_SET;
	begin
	    // check input
	    dv := DVcoerce(dv);
	    if dv=FAIL then TYPE_ERR end_if;
	    vars := VARS(de);
	    if not(contains(vars,dv)) then return(FAIL) end_if;
	    vars := vars minus {dv};
	    // solve for dv
	    sol := solve(REP(de),Dummy[dv]);
	    // for the time being we ignore piecewise or multi-valued results
	    if domtype(sol)=piecewise then
		sol := op(piecewise::disregardPoints(sol),1);
	    elif nops(sol)<>1 then return(FAIL)
	    else sol := op(sol,1) end_if;
	    // the following is a hack until we better understand all the
	    // different results `solve' may return
	    if not(testtype(sol,Type::Arithmetical)) then return(FAIL) end_if;
	    if contains(numTypes,domtype(sol)) then new(dom,sol,{})
	    else new(dom,NORMAL(sol),vars) end_if;
	end_proc;


    numer :=
	proc(x:dom):dom
	local tmp;
	begin
	    if domtype(REP(x))=DOM_EXPR then
		tmp := NORMAL(numer(REP(x)));
		if tmp=REP(x) then return(x)
		else new(dom,tmp,VARS(x)) end_if;
	    else return(x) end_if;
	end_proc;
	
	
    denom :=
	proc(x:dom):dom
	local tmp;
	begin
	    if domtype(REP(x))=DOM_EXPR then
		tmp := NORMAL(denom(REP(x)));
		if tmp=REP(x) then return(x)
		else new(dom,tmp,VARS(x)) end_if;
	    else return(x) end_if;
	end_proc;
	
	
    random :=
        proc(N=5:DOM_INT):dom
        local i:DOM_INT, vars:DOM_SET;
        begin
	    vars := { DVrandom() $ i=1..random(1..N)() };
	    dom::convert(expr(polylib::randpoly([op(vars)],hold(Expr))));
	end_proc;
	
	
    TeX :=
	proc(x:dom,p:DOM_INT):DOM_STRING
	begin generate::tex(dom::convert_to(x,DOM_EXPR), p) end_proc;


    length := 
	proc(x:dom):DOM_INT
	begin length(REP(x)) end_proc;
    
    
    characteristic := 0;


    // changeIndVars will take as input an element of this domain, 
    // a list of new independent variables, and a list of equations 
    // equating the new variables with the old.
    // (If, for example, the old independent variables are 
    // x and y, the new dependent variables are X and Y, and the change
    // is given by X=a*x + b*y, Y=c*x + d*y, then the third argument
    // would be [X=a*x + b*y, Y=c*x + d*y].)  The old
    // variables in terms of the new will be computed, unless given by
    // the optional argument OldFromNew = ...  
    // The output domain can be given by adding the argument 
    // OutputDomain = ... (although there is a default domain).

    changeIndVars :=
	proc(de, newVars:DOM_LIST, newFromOld:DOM_LIST)
	local ex, dvs, var, newVarTable:DOM_TABLE,
	      newDV:DOM_DOMAIN, i:DOM_INT, outputDomain:DOM_DOMAIN,
	      oldVarTable:DOM_TABLE, ivars:DOM_LIST, dvars:DOM_LIST, 
	      changeVars:DOM_PROC, oldFromNew:DOM_LIST, 
	      optionalArgTable:DOM_TABLE, eqn, depdvs, inddvs;
	begin
	    ivars:= map(DV::indVariables, DV::expr);
	    dvars:= map(DV::depVariables, DV::expr);
	    // Read in the optional arguments
	    optionalArgTable := table();
	    for i from 4 to args(0) do
		optionalArgTable[op(args(i),1)] := op(args(i),2);
	    end_for;
	    // Put the newFromOld in a useful form
	    newVarTable := table();
	    for eqn in newFromOld do
		newVarTable[op(eqn,1)] := op(eqn,2);
	    end_for;
	    // We will probably need oldFromNew.
	    // If it's not given to us, then we will invert newFromOld
	    // and take the first solution.
	    if contains(optionalArgTable, hold(OldFromNew)) then
		oldFromNew := optionalArgTable[hold(OldFromNew)];
	    else
		oldFromNew:= op(solve({op(newFromOld)},{op(ivars)}),1);	    
	    end_if;
	    oldVarTable := table();
	    for eqn in oldFromNew do
		oldVarTable[op(eqn,1)] := op(eqn,2);
	    end_for;
	    // Next, the workhorse function
	    changeVars :=
		proc(v)
		local fn:DOM_EXPR;
		begin
		    case DVtype(v)
		    of "Dep" do
		    of "Deriv" do
			fn:= dvars[DVindex(v)](op(newVars));
			for i in DVrepIndex(v) do
			    fn:= _plus(diff(fn,newVars[j])*
				diff(newVarTable[newVars[j]],ivars[i]) $
				j=1..nn);
			end_for;
			for var in ivars do
			    fn := subs(fn, var=DV(var));
			end_for;
			break;
		    of "Indep" do
			fn := oldVarTable[DVexpr(v)];
			break;
		    end_case;
		    return(fn);
		end_proc; /* changeVars */
	    ex:= REP(de);  
	    ex := subs(ex,DVar); // The expression
	    dvs:= VARS(de); // The differential variables
	    inddvs := select(dvs,x->(DVtype(x)="Indep"));
	    depdvs := select(dvs,x->not(DVtype(x)="Indep"));	    
	    for var in depdvs do
		ex:= subs(ex, var=changeVars(var));
	    end_for;
	    for var in inddvs do
		ex:= subs(ex, var=changeVars(var));
	    end_for;
	    // Now prepare the output domain
	    if contains(optionalArgTable, hold(OutputDomain)) then
		outputDomain := optionalArgTable[OutputDomain]
	    else
		newDV:= Dom::DifferentialVariable(newVars, dvars);	    
		outputDomain := Dom::DifferentialExpression(newDV);
	    end_if;
	    return(outputDomain(ex));
	 end_proc; /* changeIndVars */


    // changeDepVars will take as input an element of this domain, 
    // a list of new dependent variables, and a list of
    // equations equating the old variables with the new,
    // (If, for example, the old dependent variables are 
    // u and v, the new dependent variables are U and V, and the change
    // is given by u=a*U + b*V, v=c*U + d*V, then the third argument
    // would be [u=a*U + b*V, v=c*U + d*V].
    // The output domain can be given by adding the argument 
    // OutputDomain = ... (although there is a default domain).

    changeDepVars :=
	proc(de, newVars:DOM_LIST, oldFromNew:DOM_LIST)
	local ex, dvs, var, oldVarTable:DOM_TABLE, ivars:DOM_LIST,
	      dvars:DOM_LIST, optionalArgTable:DOM_TABLE, eqn, 
	      newDV:DOM_DOMAIN, outputDomain:DOM_DOMAIN, i:DOM_INT, 
	      changeVars:DOM_PROC;
	begin
	    ivars:= map(DV::indVariables, DV::expr);
	    dvars:= map(DV::depVariables, DV::expr);
	    // Read in the optional arguments
	    optionalArgTable := table();
	    for i from 4 to args(0) do
		optionalArgTable[op(args(i),1)] := op(args(i),2);
	    end_for;
	    // Put the oldFromNew in a useful form
	    oldVarTable := table();
	    for eqn in oldFromNew do
		oldVarTable[op(eqn,1)] := op(eqn,2);
	    end_for;
	    // Next, the workhorse function
	    changeVars :=
		proc(v)
		local fn:DOM_EXPR;
		begin
		    case DVtype(v)
		    of "Dep" do
			fn:= oldVarTable[DVexpr(v)];
			break;
		    of "Indep" do
			fn:= DVexpr(v);
			break;
		    of "Deriv" do
			fn:= 
			 oldVarTable[dvars[DVindex(v)]](map(DVrepIndex(v),
						  x->ivars[x]));
			break;
		    end_case;
		    return(fn);
		end_proc; /* changeVars */
	    ex:= REP(de);  
	    ex := subs(ex,DVar); // The expression
	    dvs:= VARS(de); // The differential variables
	    for var in dvs do
		ex:= subs(ex, var=changeVars(var));
	    end_for;
	    // Now prepare the output domain
	    if contains(optionalArgTable, hold(OutputDomain)) then
		outputDomain := optionalArgTable[OutputDomain]
	    else
		newDV:= Dom::DifferentialVariable(ivars,newVars);	    
		outputDomain := Dom::DifferentialExpression(newDV);
	    end_if;
	    return(outputDomain(ex));
	end_proc; /* changeIndVars */


begin

    // analysis of passed arguments
    opts := select([args()], (a -> bool(type(a)="_equal")));
    if iszero(nops(opts)) then		// DV directly given
	DV := args(1);
	if domtype(DV)<>DOM_DOMAIN or DV::hasProp=FAIL or
	   not(DV::hasProp(Cat::DifferentialVariable)) then
	    error("argument must be Cat::DifferentialVariable");
	end_if;
	restrictedDV := bool(DV::constructor=
			     Dom::RestrictedDifferentialVariable);
        tmp := 1;
    else				// options passed
	if op(opts[1],1)<>hold(Vars) then TYPE_ERR end_if;
	if nops(opts)=2 and op(opts[2],1)<>hold(Rest) then TYPE_ERR end_if;
	opts := map(opts, 
		    (a -> if domtype(op(a,2))<> DOM_LIST then 
			      subsop(a,2=[op(a,2)]) 
			  else a end_if));
	DV := Dom::DifferentialVariable(op(op(opts[1],2)));
	if nops(opts)=1 then
	    restrictedDV := FALSE;
	    tmp := 1;
	else
	    DV := Dom::RestrictedDifferentialVariable(DV,op(op(opts[2],2)));
	    restrictedDV := TRUE;
	    tmp := 2;
	end_if;
    end_if;
    case args(0) 
    of tmp do
	NORMAL := id;
	IsZero := iszero;
	break;
    of (tmp + 1) do
	NORMAL := args(tmp + 1);
	if NORMAL = id then
	    IsZero := iszero;
	else
	    IsZero := iszero @ NORMAL;
	end_if;
	break;
    of (tmp + 2) do
	NORMAL := args(tmp + 1);
	IsZero := args(tmp + 2);
	break;
    otherwise
	ARG_ERR;
    end_case;
    
    // restricted differential variables
    if restrictedDV then
	DV2 := DV::Variables;
	DV2coerce := DV2::coerce;
	DV2conv := DV::convDV;
	DVtest := DV::test;
    end_if;

    // methods from DV
    DVcoerce := DV::coerce;
    DVconvto := DV::convert_to;
    DVexpr := DV::expr;
    DVdiff := DV::totalDiff;
    DVindex := DV::index;
    DVtype := DV::vartype;
    DVgreater := ((j1,j2) -> DV::_less(j2,j1));
    DVrandom := DV::random;
    DVrepIndex := DV::repeatedIndex;
    DVmulti := DV::multiIndex;
    DVnewRep := DV::newRepeated;

    // management of dummy variables
    DVar := table();
    Dummy := table();
    // checks whether dv is already in use, otherwise dummy is generated
    makeDummy := proc(dv:DV):DOM_IDENT
		 begin
		     if not(contains(Dummy,dv)) then 
			 Dummy[dv] := genident("_de_");
			 DVar[Dummy[dv]] := dv;
		     end_if;
		     Dummy[dv];
		 end_proc;
    // checks whether identifier is dummy and eliminates it, if yes
    elimDummy := proc(v)
	 	 begin
		     if contains(DVar,v) then DVar[v] else v end_if;
		 end_proc;
    // checks whether identifier is dummy and applies function to it, if yes
    mapDummy := proc(v,f:DOM_PROC)
	 	begin
		    if contains(DVar,v) then 
			f(DVar[v],args(3..args(0)));
		    else v end_if;
	        end_proc;
	
    // other global variables
    nn := DV::numIndVar;
    mm := DV::numDepVar;
    numTypes := {DOM_INT,DOM_COMPLEX,DOM_FLOAT,DOM_RAT};
    arithTypes := _union({DV,DOM_EXPR,DOM_IDENT},numTypes);

end_domain:


unalias(REP,VARS,TYPE_ERR,ARG_ERR):



