// $Date: 2002/02/26 10:12:10 $ $Author: stefanw $ $Revision: 1.62 $ 

/*++
Dom::ImageSet - domain of infinite sets, written as images 
             of sets under mappings 


An ImageSet represents the object {f(x1,..,xn); x1 \in S_1, ..., x_n \in S_n},
                  where S_1,..,S_n are sets and
                  f: S_1 x ... x S_n -> "the universe" is a mapping
                  (i.e. no set containing the range of f needs to be
                   known).

                  Operands:

                  0 - Dom::ImageSet
                  1 - f : DOM_EXPR
                  2 - lv: list of variables
                  3 - ls : list of sets


++*/

domain Dom::ImageSet
  local mapplus, mapmult, mapleftpower;
      
  inherits Dom::BaseDomain;
  category Cat::Set;
    // no axioms 
    
    // entries 


    // slots 

    expr:= M -> extop(M,1);
    
    variables:= M -> extop(M,2);

    sets:= M -> extop(M, 3);

    nvars:= M-> nops(extop(M,2));
    
    // methods: 
   
    new :=
    proc(f, x, S)
      local i:DOM_INT,
      newprop;
    begin
      if args(0)=1 then
         dom::convert(f);
         if %=FAIL then
            error("Cannot convert to type Dom::ImageSet")
         else
            return(%)
         end_if
      end_if;
      if testargs() then
        if args(0)<2 or args(0)>3 then
          error("Wrong number of args")
        end_if;
        if not contains({DOM_IDENT, "_index", DOM_LIST}, type(x)) then
          error("Second argument must be identifier or list")
        end_if;
        if args(0)=3 then
          if type(x)=DOM_LIST and type(S)<>DOM_LIST
            or type(S)=DOM_LIST and type(x)<>DOM_LIST then
            error("Either both or none of the second and third argument ".
                  "must be lists");
          end_if;
          if type(x)=DOM_LIST and nops(x)<>nops(S) then
            error("Both lists must have the same length")
          end_if;
          if type(x)=DOM_LIST and nops(x)<>nops({op(x)}) then
            error("Duplicate identifier")
          end_if
        end_if;
        // test operands of S for being of "Type::Set" ? 
      end_if; // arg-check

      // handle special case

      if type(f) = DOM_IDENT then
        if x=f then
          assert(type(S) <> DOM_LIST);
          return(S)
        end_if;
        if type(x) = DOM_LIST and (i:=contains(x, f)) > 0 then
          return(S[i])
        end_if;
        return({f})
      end_if;

      // prefer a piecewise consisting of image sets over
      // an image set with piecewise first operand
      if type(f) = piecewise then
        return(piecewise::extmap(f, dom::new, x, S))
      end_if;

      // dito, with piecewise defined S

      if type(S) = piecewise then
        return(piecewise::extmap(S, set -> Dom::ImageSet::new(f, x, set)))
      end_if;

      if type(S) = DOM_LIST and nops(S)=1 and type(S[1]) = piecewise then
        // at the moment, do not flatten out if there is more
        // than one operand because this would produce many branches
        return(piecewise::extmap(op(S,1), set -> Dom::ImageSet(f, x, [set])))
      end_if;
      
      if type(x)=DOM_LIST then
        if nops(x)=0 then
          {f}
        else // nops(x) >= 1 
          for i from 1 to nops(x) do
            save x[i];
            newprop:= solvelib::set2prop(S[i], hold(Superset));
            if newprop=property::Null then
              // the only mapping starting from the empty set
              // is the empty map
              return({})
            end_if;
            // attach a new property to x[i]
            eval(hold(assume)(x[i], newprop, _and));
          end_for;
          new(dom, f, x, S)
        end_if
      else // x is not a list
        newprop:= solvelib::set2prop(S, hold(Superset));
        if newprop=property::Null then
          // any image of the empty set is empty
          return({})
        end_if;

        save x;
        eval(hold(assume)(x, newprop, _and));
        new(dom, f, [x], [S])
      end_if
    end_proc;

    convert:=
    proc(x)
    begin
      if type(x)=dom then x else FAIL end_if
    end_proc;
      

    
    /* changevar(S, x, y) - replaces the variable x by y both
       in the expression and in the list of variables */

    // y does *not* implicitly get the same properties as x 
    
    changevar:=
    proc(S:dom, x, y)
    begin
      extsubsop(S, 1=subs(extop(S,1), x=y), 2= subs(extop(S,2), x=y))
    end_proc;


    /* setvar(S, y) - sets the only variable of S to y */
    /* maps to set expressions */
    /* if S is not an ImageSet nor a set expression, nothing happens */

    // note that y gets no properties
    
    setvar:=
    proc(S,y)

    begin
      case type(S)
        of dom do
          if dom::nvars(S)<>1 then
            error("Only allowed for image sets depending on one variable")
          end_if;
          return(dom::changevar(S, dom::variables(S)[1], y))
        of "_union" do
        of "_intersect" do
        of "_minus" do
          return(map(S, dom::setvar, y))
        otherwise
          S
      end_case
    end_proc;



    /* evaluation is done as follows:
       first, bound variables that have a value are substituted by new
       bound variables. After that, the expression and the sets are evaluated
       as usual. */

    evaluate:=
    proc(S: dom)
      local
        j: DOM_INT,
        ausdr,
        l,
        X: DOM_IDENT,
        indices: DOM_LIST,
        vars: DOM_LIST,
        newvars: DOM_LIST,
        sets: DOM_LIST;
    begin
      ausdr:= extop(S, 1);
      indices:=[];
      for j from 1 to nops(extop(S,2)) do
        if extop(S,2)[j] <> eval(extop(S,2))[j] then
          indices:=indices.[j]
        end_if
      end_for;
      vars := extop(S, 2);
      for j in indices do
        X:= genident();
        ausdr:= subs(ausdr, vars[j] = X);
        vars[j]:= X;
      end_for;
      vars:= eval(vars);
      ausdr:= eval(ausdr);
      sets :=  eval(extop(S, 3));


      // remove variables that do not occur
      newvars:={op(vars)} intersect indets(ausdr);
      if nops(newvars) < nops(vars) then
        if nops(newvars)=0 then
          // not an image set, the expression is constant
          return({ausdr})
        end_if;
        // create list of index positions at which the variables
        // have survived
        l:=select([$1..nops(vars)], i-> contains(newvars, vars[i]));
        vars:= [vars[l[i]] $i=1..nops(l) ];
        sets:= [sets[l[i]] $i=1..nops(l)];
        dom::new(ausdr, vars, sets)
      else
        // original state may be kept
        new(dom, ausdr, vars, sets)
      end_if;
      
    end_proc;
        
    

    // homogpointwise - create, for a given a law of composition
    // Operator on the complex numbers, a method that maps each tuple
    // of sets (A_1, ..., A_m) ---
    // where
    // Ai= {f_1(x_1i, x_2i, ..,x_{n_i}i); x_ki in S_ki for 1 <= k <= n_i}
    // to the set
    // { Operator(f_1(x_11, ...,x_{n_1}1), ..., f_m(x_1m, ...,x_{n_m}m));
    //   x_ki in S_ki for all (k,i) with 1<=i<=m and 1<=k <= n_i }
    // (pointwise application of a law of composition)

    homogpointwise:=
    proc(Operator)
      option escape;
      local d;
    begin
      d:=dom;
      
      proc()
        local arglist, vars, i, v, newid;
      
      begin
        arglist:=[args()];
        vars:={};
        for i from 1 to nops(arglist) do
          for v in {op(d::variables(arglist[i]))} intersect vars do
            /* v must be replaced by another variable to avoid
                               name conflicts */
            newid:= genident();
            if property::hasprop(v) then
              eval(hold(assume)(newid, getprop(v)))
            end_if;
            arglist[i]:=d::changevar(arglist[i], v, newid)
          end_for;
          vars:=vars union {op(d::variables(arglist[i]))};
        end_for;
        new(d, Operator(op(map(arglist,d::expr))),
            _concat(op(map(arglist, d::variables))),
            _concat(op(map(arglist, d::sets))))
      end_proc;      
    end_proc;

    homog_plus:=dom::homogpointwise(_plus);
    
    homog_mult:=dom::homogpointwise(_mult); 


// binpointwise(Operator)
/* create a function to compute Operator(imageset, otherset)
   where imageset is a Dom::ImageSet

   imageset = { expression(X1,..,Xn); X_i in S_i }

   that function returns

   { Operator(X, Y);  X in imageset, Y in otherset}
   =
   { Operator(expression(X1, .., Xn), Y); X_i in S_i, Y in otherset }

*/



binpointwise:=
proc(Operator)
  option escape;
  local d;
begin
  d:= dom;
  
  // procedure to be returned

  proc(imageset: Dom::ImageSet, otherset)
    local Y;
  begin
    Y:= genident();
    d::new(Operator(d::expr(imageset), Y), d::variables(imageset).[Y],
           d::sets(imageset).[otherset])
  end_proc;

end_proc;


// it is better to create fixed entries than to make
// binpointweise return one procedure per table entry
pwplus := dom::binpointwise(_plus);
pwmult := dom::binpointwise(_mult);



// bin_power ...


    // bin_union - compute {f(x) ; x \in A} union {g(y); y \in B}
    // this cannot contain more than a few heuristics
    
    bin_union:=
    proc(S, T)
      local SS, f, g, d, x;
    begin
      // do not try to handle more than one variable
      if dom::nvars(S)>1 or dom::nvars(T)>1 then
        return(FAIL)
      end_if;
      
      SS:=dom::setvar(S, dom::variables(T)[1]);
      if dom::expr(SS)=dom::expr(T) then
        // use the rule
        // { f(x); x \in S} union { f(x) ; x \in T } =
        // { f(x): x \in S \cup T }
        return(dom::new(dom::expr(T), dom::variables(T)[1],
                        dom::sets(T)[1] union dom::sets(SS)[1]))
      end_if;


    
      if dom::sets(SS)[1] = Z_ and dom::sets(T)[1]= Z_
        and dom::isLinear(SS) and dom::isLinear(T) then
        x:= dom::variables(SS)[1];
        f:= poly(dom::expr(SS), dom::variables(SS));
        g:= poly(dom::expr(T), dom::variables(T));
        d:= coeff(f, 0) - coeff(g, 0);

        // special case:
        //
        // the union of {a*n + b; n in Z} and {k*a*n + b; n in Z}
        // equals {a*n + b; n in Z}

        if iszero(d) then
          if type(coeff(f,1)/coeff(g,1)) = DOM_INT then
            return(T)
          elif type(coeff(g,1)/ coeff(f,1)) = DOM_INT then
            return(S)
          end_if
        end_if;
        
        // halving stepwidth (special case):
        // the union of
        //
        //  x       x       x       x
        //
        //  and
        //
        //      x       x       x       x
        //
        // equals
        //
        //  x   x   x   x   x   x   x   x
        //
        // that is:
        //  {ax+b; x in Z} union {ax+ (b-a)/2 ;  x\in Z} =
        //  { (a/2)x + b; x \in Z}

        
        if (coeff(f, 1) = coeff(g, 1) or coeff(f, 1)= -coeff(g, 1)) and
          (coeff(f, 1) = 2*d or coeff(f, 1) = -2*d) then   
          return(Dom::ImageSet(d*x + coeff(f, 0), [x], [Z_]))
        end_if        
      end_if;
      
      case S intersect T
        of S do
          return(T)
        of T do
          return(S)
      end_case;    
      
      // else: heuristic did not succeed
      FAIL
    end_proc;

    bin_intersect:=
    proc(S, T)
      local SS, SsetintersectTset, Sset, Tset, success: DOM_BOOL;
    begin
     // do not try to handle more than one variable
      if dom::nvars(S)>1 or dom::nvars(T)>1 then
        return(FAIL)
      end_if;

      // special case
      
      SS:=dom::setvar(S, dom::variables(T)[1]);
      if dom::expr(SS)=dom::expr(T) then
        // in general, the rule
        //  { f(x); x \in S} intersect { f(x) ; x \in T } =
        // { f(x): x \in S intersect T }
        // does not hold!
        // but we check if S \subseteq T or T \subseteq S
        
        Sset:=dom::sets(SS)[1];
        Tset:=dom::sets(T)[1];
        if Sset = Tset then
          SsetintersectTset:= Sset;
          success:= TRUE
        else
          SsetintersectTset:=Sset intersect Tset;
          success:= (%=Sset or %=Tset)
        end_if;
        
        if success then
          return(dom::new(dom::expr(SS), dom::variables(SS)[1],
                          SsetintersectTset))
        end_if
      end_if;

      // special case: is one set contained in the other?
      // let S={f(x); x in S} and T={g(y); y in T}
      
      // if f(x) = g(y) has no solution, the intersection is empty
      // if f(x) = g(y) has a solution for y (whatever x is),
      // then every element of S
      // can be written as an element of T, hence S \subseteq T,
      // and their intersection is S
      
      // if f(x) = g(y) has a solution for x (whatever y is),
      // then every element of T
      // can be written as an element of S
      // hence the intersection is T

      /*
      
      empt:= solvelib::isEmpty(solve(expr(S)=expr(T), dom::variables(T)[1],
                               DontRewriteBySystem));
      if empt=FALSE then
        return(S)
      end_if;
      if not (empt=TRUE) then
        empt:= solvelib::isEmpty(solve(expr(S)=expr(T), dom::variables(S)[1],
                                 DontRewriteBySystem))
      end_if;

      case empt
        of TRUE do
          return({})
        of FALSE do
          return(T)
      end_case;
      // empt = UNKNOWN  
      
*/
      FAIL
    end_proc;

    bin_minus:=
    proc(S, T)
    begin
      // only simple cases can be settled here:
      if S=T then
        return({})
      end_if;
      
      case S intersect T
        of S do
          return({})
        of {} do
          return(S)
      end_case;
      
      FAIL
    end_proc;
      
   /* binary operations, inhomogenous case   */


    inhomog_union:=
    table(
          DOM_SET=
          proc(iset, s)
          begin
            if s={} then
              iset
            else
              FAIL
            end_if
          end_proc
              );


    inhomog_intersect:=
    table(
          DOM_SET=
          proc(iset:Dom::ImageSet, s:DOM_SET)

          begin
            if s={} then
              return({})
            end_if;
          
            if nops(iset::dom::variables(iset))>1 then
              return(FAIL)
            end_if;
            // let iset= {f(x); x \in A }
            // At first, get f^(-1)(s) and intersect with A
            // then apply f
            solvelib::preImage(iset::dom::expr(iset),
                               iset::dom::variables(iset)[1], s)
              intersect iset::dom::sets(iset)[1];
            solvelib::substituteBySet(iset::dom::expr(iset),
                               iset::dom::variables(iset)[1], %)
          end_proc,


          Dom::Interval=
          proc(iset:Dom::ImageSet, iv:Dom::Interval)

          begin
            if nops(iset::dom::variables(iset))>1 then
              return(FAIL)
            end_if;
            // let iset= {f(x); x \in A }
            // At first, get f^(-1)(s) and intersect with A
            // then apply f
            solvelib::preImage(iset::dom::expr(iset),
                               iset::dom::variables(iset)[1], iv)
              intersect iset::dom::sets(iset)[1];
            if type(%)="_intersect" then
              FAIL
            else  
              solvelib::substituteBySet(iset::dom::expr(iset),
                               iset::dom::variables(iset)[1], %)
            end_if
          end_proc,


          DOM_EXPR=
          proc(iset, xpr)
            local l,i, X;
          begin
            case type(xpr)
              of "_union" do
                l:=[op(xpr)];
                for i from 1 to nops(l) do
                  l[i]:=iset intersect l[i];
                  if type(%)="_intersect" then
                    return(FAIL)
                  end_if
                end_for;
                return(_union(op(l)))
              of "_minus" do
                if iset::dom::nvars(iset)>1 then
                  return(FAIL)
                else
                  // delete properties from variable
                  save op(iset::dom::variables(iset));
                  // { f(X); X in S} intersect (A minus B)
                  // get the set of all X in S s.t. f(X) is in A minus B
                  
                  solvelib::preImage(iset::dom::expr(iset),
                                            iset::dom::variables(iset)[1],
                                            xpr)
                      intersect
                                     op(iset::dom::sets(iset),1);
                  if type(%)="_intersect" then
                    return(FAIL)
                  else
                    // apply f
                    return(solvelib::substituteBySet(iset::dom::expr(iset),
                           op(iset::dom::variables(iset)), %))
                  end_if
                end_if
            end_case;
            FAIL
          end_proc
          );


    inhomogleft_minus:=
    table(

          // simple case only
          DOM_SET=
          proc(iset, S)
            begin
              if S={} then
                iset
              else
                FAIL
              end_if
            end_proc

            );

    
    inhomog_plus:=
    table(
          DOM_INT=mapplus,
          DOM_RAT=mapplus,
          DOM_COMPLEX=mapplus,
          DOM_FLOAT=mapplus,
          DOM_EXPR=mapplus,
          DOM_IDENT=mapplus,
          DOM_SET =
          proc(iset, S)
          begin
            _union(iset + x $ x in S)
          end_proc,
          Dom::Interval = dom::pwplus
          );


    inhomog_mult:=
    table(
          DOM_INT=mapmult,
          DOM_RAT=mapmult,
          DOM_COMPLEX=mapmult,
          DOM_FLOAT=mapmult,
          DOM_EXPR=mapmult,
          DOM_IDENT=mapmult,
          DOM_SET =
          proc(iset, S)
          begin
            _union(iset * x $ x in S)
          end_proc,
          Dom::Interval = dom::pwmult
          );
    

    inhomogleft_power:=
    table(
          DOM_INT=mapleftpower,
          DOM_RAT=mapleftpower,
          DOM_COMPLEX=mapleftpower,
          DOM_FLOAT=mapleftpower,
          DOM_EXPR=mapleftpower,
          DOM_IDENT=mapleftpower
          );
  
    

    isEmpty:=
    proc(S:dom)
    begin
      // a set has an empty image iff it is empty itself
      map(dom::sets(S), solvelib::isEmpty);
      _or(op(%))
    end_proc;

    /* simpexIn(a, S) - expression equivalent to a \in S */
    
    simpexIn:=
    proc(a, S: dom)
      local x, eq;
    begin
      if dom::nvars(S) = 1 then
        x:= op(dom::variables(S));
        eq:= expr(S);
        if testtype(eq, Type::PolyExpr([x])) or
          indets(eq) minus Type::ConstantIdents minus {x} = {} then
          not solvelib::isEmpty(solve(eq = a, x)
                                intersect
                                op(dom::sets(S)));
          if % <> UNKNOWN then
            return(%)
          end_if;
        end_if
      end_if;
      hold(_in)(a, S) // not yet implemented, would require an "exists"-
                        // quantifier
    end_proc;

    
    // Union : overloads solvelib::Union
    /* Union(S, x, M) - with S depending on the parameter x,
       return the set of all objects that can be obtained by substituting
       an element of M for x */
    
    Union:=
    proc(S:dom, x:DOM_IDENT, M /*: set */)
    begin
      if has(dom::variables(S), x) then
        error("Not a free parameter")
      end_if;
      // in future: Union(subsBySet(S::dom::expr(S), x, M), x, M)
      simplify(dom::new(dom::expr(S), append(dom::variables(S), x),
               append(dom::sets(S), M)))
    end_proc;

    
    /* map(S,f, x1,...,xn) - computes the set of all
       f(y, x1, ..., xn), y Element of S
    */

    map:=
    proc(S:dom, f)
    begin
      extsubsop(S, 1=f(dom::expr(S), args(3..args(0))))
    end_proc;


    // operations defined by mapping

    normal:= S -> dom::map(S, normal);

    expand:= S-> dom::map(S, expand);

    
    subs:=
    proc(S:dom)
      local substitutions;

    begin
      substitutions:=[args(2..args(0))];
      if indets(substitutions) intersect {op(dom::variables(S))} <> {} then
        error("Cannot substitute bound variables")
      end_if;
      extsubsop(S, 1=subs(extop(S,1), op(substitutions)))
    end_proc;


    // compute all objects that can be obtained by substituting
    // a Dom::ImageSet for x
    // the set of all a(x) with x in {g(y1, ..,yk) ; yi in S_i}
    // is { a(g(y1, ..., yk)) ; y_i in S_i }

    
    substituteBySet:=
    proc(a, x, S:dom)
    begin
      new(dom, subs(a, x=dom::expr(S)), dom::variables(S), dom::sets(S))
    end_proc;

    
    indets:= S -> indets(dom::expr(S)) minus {op(dom::variables(S))};
    
    
    set2prop:=
    proc(S:dom, opt=hold(Exact))
      local newvar, genexpr;
    begin
      // let Xi be the i-th variable //
      // to each Xi, attach the property to be an element of sets[i] //
      // then evaluate
      newvar:= dom::variables(S);
      save op(newvar);
      zip(newvar, map(dom::sets(S), solvelib::set2prop, hold(Superset)),
          assume);
      genexpr:=subs(expr(S), dom::variables(S)[i]=newvar[i] $i=1..dom::nvars(S));
      // questionable, but IV seems to be most special
      property::_getprop(genexpr, property::IV, opt);
      if %<>FAIL and % <> genexpr then
        %
      else
        property::_getprop(genexpr, property::NRange, opt)
      end_if;
      if %<>FAIL and % <> genexpr then
        %
      else
        case opt
          of hold(Exact) do
            return(FAIL)
          of hold(Superset) do
            return(Type::Complex)
          of hold(Subset) do
            return(property::Null)
        end_case
      end_if
    end_proc;


    // evalAt(S, var1=value1, .., vark=valuek)
    // substitutes values for the parameters

    evalAt:=
    proc(S:dom)
      local xpr, varlist, setlist,i, subst, pos;
    begin
      xpr:=dom::expr(S);
      varlist:=dom::variables(S);
      setlist:=dom::sets(S);
      for i from 2 to args(0) do
        if type((subst:=args(i)))<>"_equal" then
          error("Equation expected")
        end_if;
        pos:=contains(varlist, lhs(subst));
        if pos=0 then
          next
        end_if;
        delete varlist[pos];
        delete setlist[pos];
        xpr:=subs(xpr, subst)
      end_for;
      if nops(varlist)=0 then
        xpr
      else  
        new(dom, xpr, varlist, setlist)
      end_if
    end_proc;  
    
    getElement:=
    proc(S:dom)
      local i, aux;
    begin
      // choose one element for each parameter set
      [dom::variables(S)[i]=
      (aux:=solvelib::getElement(dom::sets(S)[i]);
       if aux=FAIL then
         userinfo(10, "Could not find value for parameter ".
                  expr2text(dom::variables(S)[i]));
         aux // should be return(aux) once the kernel bug has been fixed
       else
         aux
       end_if) $i=1..nops(dom::sets(S))];
      // substitute each chosen value into the expression
      if has(%, FAIL) then
        FAIL
      else  
        subs(dom::expr(S), op(%))
      end_if
    end_proc;

    // cannot decide whether the expression really
    // represents an *injective* function
    isFinite:= UNKNOWN; 

    isLinear:=
    proc(S)
    begin
      if testtype(dom::expr(S), Type::PolyExpr(dom::variables(S))) then
        bool(degree(poly(dom::expr(S), dom::variables(S))) = 1)
      else
        FALSE
      end_if
    end_proc;

    simplify:=
    proc(S)
      local newvars, sets, l, newexpressions, v, xpr;
    begin
      sets:= dom::sets(S);
      
      // expand finite sets
      l:=contains(map(dom::sets(S), domtype), DOM_SET);
      if l>0 then
        newvars:=dom::variables(S);
        v:= newvars[l];
        xpr:= S::dom::expr(S);
        newexpressions:= map(sets[l], el-> subs(xpr, v=el));
        delete newvars[l];
        delete sets[l];
        return(_union(op(map(newexpressions,
                             el -> simplify(Dom::ImageSet(el, newvars, sets))
                             ))))
      end_if;
      
      // first simplify the expression
      S:=extsubsop(S, 1=simplify(dom::expr(S)));

      // simplify all sets in the list of sets

      S:= extsubsop(S, 3=map(dom::sets(S), simplify));

      // simplify { x; x in F} to F

      if dom::nvars(S)=1 and op(dom::variables(S),1)=dom::expr(S) then
        return(op(dom::sets(S),1))
      end_if;
      
      // no further heuristisc at the moment
      S
    end_proc;

    float:= S -> extsubsop(S, 1=float(dom::expr(S)), 3=float(dom::sets(S)));
    
    print:=
    proc(x)
      local i;
    begin
      "{ ".expr2text(expr(x))." | ".
      _concat((if i>1 then ", " else " " end_if).
              expr2text(dom::variables(x)[i]). " in ".
              expr2text(dom::sets(x)[i]) $i=1..nops(dom::sets(x)))." }";
    end_proc;

    TeX:=
    proc(x)
    begin
      "\\{ ".generate::TeX(dom::expr(x))." ; ".
      _concat(((if i<>1 then ", " else "" end_if).
               generate::TeX(dom::variables(x)[i])."\\in".
               generate::TeX(dom::sets(x)[i]))
              $i=1..dom::nvars(x)). "\\}"
    end_proc;

    Content:=
    proc(Out, x)
      local i;
    begin
      Out::Cimageset(Out(expr(x)), 
                     Out(hold(_in)(dom::variables(x)[i], dom::sets(x)[i])
                          $ i = 1..nops(dom::sets(x))))
    end_proc;

   // initialization 

  begin
    userinfo(20, "Constructing domain Dom::ImageSet");

    /* local methods */
    
    mapplus:=
    proc(iset, other)
    begin
      iset::dom::map(iset, _plus, other)
    end_proc;

    mapmult:=
    proc(iset, other)
    begin
      iset::dom::map(iset, _mult, other)
    end_proc;

    mapleftpower:=
    proc(iset, other)
    begin
      iset::dom::map(iset, _power, other)
    end_proc;
    
  end_domain:

  



