/*++    -*-MuPAD-*-
An element of MultivariateSeries will consist of
 * a polynomial, using dummy variables (each dummy variable will represent
                                      x-a)
 * a list of variables
 * a list indicating the center of the series
 * an integer (or infinity) indicating the order

The domain is called with no arguments.
An entry of MS:=Dom::MultivariateSeries() can be created by
MS(input), where the input can be of the form
 * an element of MS
 * an expression, a list of the form [var1=a1, var2=a2, ...] and
    an integer containing the order
    (The list can be of the form [var1, var2, ...], in which case the
    center will be [0,0,...], and if the order is omitted, then it will
    default to the global variable ORDER, or infinity, if the expression
    doesn't contain the variables.)
 * a polynomial, a list of variables, a list indicating the center, and
    an integer indicating the order.

The following methods for MS are given (ms, ms1, ms2 are elements of MS):
indets:  MS::indets(ms) will return a list of the variables
point:   MS::point(ms) will return the center of the expansion
order:   MS::order(ms) will return the order of the series
expr:    MS::expr(ms) will return the expression (basically, the series
                      without the order information)
_plus:   ms1 + ms2 will add the two series, if they have the same 
         variables and centers.  (If ms2 is not an element of MS, some
	 conversions will be attempted.)
_mult:   ms1*ms2 will multiply the two series, if they have the same 
         variables and centers.  (If ms2 is not an element of MS, some
	 conversions will be attempted.)
_power:  ms^n will raise the series to the given power
diff:    diff(ms,var) will take the derivative of ms with respect to var

Author: Jay Belanger
        (jay@vh213601.truman.edu)
Created: Friday, 25 May 2001,  2:14 PM CDT
Last Update: Friday, 25 May 2001,  2:14 PM CDT
++*/

domain Dom::MultivariateSeries()
    inherits Dom::BaseDomain;
     // How to create entries
     // The input can be of the form:
     //     1) an element of this domain,
     //  or 2) an expression, a list of the form [var1=a1, var2=a2,...] and
     //                            an integer indicating the order
     //  For case 2), the list can be of the form [var1, var2,...],
     //     indicating the center is [0,0,...]
     //  Also for case 2), if there is no third argument, then the order
     //  will be the global variable ORDER (or infinity, if the expression 
     //  doesn't involve the variables).
     //  Or it can be of the form
     //     3) pol, var, ctr, ord
      new:= 
  	proc():dom
  	local pol, vars, ctr, ord, dummyVars;
 	begin
	    if args(0) = 1 and domtype(args(1)) = dom then
		return(args(1));
	    elif args(0) = 2 or args(0) = 3 then
		// testargs here
		pol := expr(args(1));
		if pol = FAIL then
		    return(FAIL);
		end_if;
		// Get the variables and the center
		if {extop(map(args(2),type))} = {"_equal"} then
		    vars := map(args(2), extop, 1);
		    ctr  := map(args(2), extop, 2);
		else
		    vars := args(2);
		    ctr  := [0$extnops(vars)];
		end_if;
		// Get the order
		if args(0) = 3 then
		    ord := args(3);
		elif has(pol,vars) then
		    ord := ORDER;
		else
		    ord := infinity;
		end_if;
		// Now, get the dummy variables
		dummyVars := [genident()$i=1..extnops(vars)];
		// Put the expression in terms of the dummy variables
		// The dummy variable X represents x-a, so x=X+a
		for i from 1 to extnops(vars) do
		    pol := subs(pol,vars[i]=dummyVars[i]+ctr[i]);
		end_for;
		// Now, put pol in polynomial form
		for i from 1 to extnops(vars) do
		    pol := series(pol,dummyVars[i],ord);
		    pol := expr(pol);
		end_for;
		// Finally, the domain element
		pol := poly(pol,dummyVars);
		pol := dom::truncate(pol, ord);
		return(new(dom, pol, vars, ctr, ord));
	    elif args(0) = 4 then
		return(new(dom, args(1), args(2), args(3), args(4)));
	    else
		error("Wrong number of arguments");
	    end_if;
  	end_proc; /* new */
    
    // How to access information about the entries
    // The polynomial
    dummyPoly :=
	proc(s)
	begin
	    return(extop(s,1));
	end_proc; /* dummyPoly */
    
    // The variables
    indets :=
	proc(s)
	begin
	    return(extop(s,2));
	end_proc; /* vars */
    
    // The center
    point := 
	proc(s)
	begin
	    return(extop(s,3));
	end_proc; /* center */

    // The order
    order := 
	proc(s)
	begin
	    return(extop(s,4));
	end_proc; /* order */

    // The dummy expression
    dummyExpr :=
	proc(s)
	begin
	    return(expr(dom::dummyPoly(s)));
	end_proc; /* expr */

    // The dummy variables
    dummyVars :=
	proc(s)
	begin
	    return(op(dom::dummyPoly(s),2));
	end_proc; /* vars */
    
    // The actual expression
    expr :=
	proc(s)
	begin
	    return(subs(dom::dummyExpr(s),
		zip(dom::dummyVars(s),dom::indets(s),_equal)));
	end_proc; /* expr */
    

    // Now, the operations on the series
    _plus:= 
	proc() 
	local sm, ord, DOM, p, domelements, nondomelements, vars, ctr, 
	      dumVars, fixnondomelements, fixelement;
	begin
	    case args(0)
	    of 1 do return(args());
	    end_case;
	    DOM := dom;
	    domelements:= [select(args(), x->(domtype(x) = DOM))];
	    nondomelements:= [select(args(), x-> (not domtype(x)=DOM))];
	    // I'm assuming that there will be at least one element
	    // of this domain in the list.
	    // The variables, center, etc. of the others will be
	    // compared to the first.
	    vars := dom::indets(domelements[1]);
	    ctr := dom::point(domelements[1]);
	    ord := min(op(map(domelements, x->DOM::order(x))));
	    dumVars := dom::dummyVars(domelements[1]);
	    // What sort of non-dom elements should be considered?
	    fixnondomelements :=
		proc(s)
		begin
		    return(DOM(expr(s), zip(vars,ctr,_equal), ord));
		end_proc; /* fixnondomelements */
	    nondomelements := map(nondomelements, fixnondomelements);
	    if has([nondomelements], FAIL) then
		return(FAIL);
	    end_if;
	    // Do the comparisons
 	    fixelement :=
		proc(s)
		local p;
		begin
		    if (DOM::indets(s) = vars and DOM::point(s)=ctr)
		    then
			p := DOM::dummyExpr(s);
			p := subs(p, zip(DOM::dummyVars(s), dumVars, _equal));
	                return(new(DOM, poly(p, dumVars), vars, ctr, 
			                                   DOM::order(s))); 
		    else
			return(FAIL);
		    end_if;
		end_proc; /* fixelement */
	    domelements := domelements . nondomelements;
	    domelements := map(domelements, fixelement);
	    if has([domelements], FAIL) then
		return(FAIL);
	    end_if;
	    sm := 
		_plus(DOM::dummyPoly(domelements[i]) $ 
					    i=1..extnops(domelements));
	    return(DOM(sm,vars,ctr,ord));
      end_proc;

    _mult:= 
	proc() 
	local pd, ord, DOM, p, domelements, nondomelements, vars, ctr, 
	      dumVars, fixnondomelements, fixelement;
	begin
	    case args(0)
	    of 1 do return(args());
	    end_case;
	    DOM := dom;
	    domelements:= [select(args(), x->(domtype(x) = DOM))];
	    nondomelements:= [select(args(), x-> (not domtype(x)=DOM))];
	    // I'm assuming that there will be at least one element
	    // of this domain in the list.
	    // The variables, center, etc. of the others will be
	    // compared to the first.
	    vars := dom::indets(domelements[1]);
	    ctr := dom::point(domelements[1]);
	    ord := min(op(map(domelements, x->DOM::order(x))));
	    dumVars := dom::dummyVars(domelements[1]);
	    // What sort of non-dom elements should be considered?
	    fixnondomelements :=
		proc(s)
		begin
		    return(DOM(expr(s), zip(vars,ctr,_equal), ord));
		end_proc; /* fixnondomelements */
	    nondomelements := map(nondomelements, fixnondomelements);
	    if has([nondomelements], FAIL) then
		return(FAIL);
	    end_if;
	    // Do the comparisons
 	    fixelement :=
		proc(s)
		local p;
		begin
		    if (DOM::indets(s) = vars and DOM::point(s)=ctr)
		    then
			p := DOM::dummyExpr(s);
			p := subs(p, zip(DOM::dummyVars(s), dumVars, _equal));
	                return(new(DOM, poly(p, dumVars), vars, ctr, 
			                                   DOM::order(s))); 
		    else
			return(FAIL);
		    end_if;
		end_proc; /* fixelement */
	    domelements := domelements . nondomelements;
	    domelements := map(domelements, fixelement);
	    if has([domelements], FAIL) then
		return(FAIL);
	    end_if;
	    pd := 
		_mult(DOM::dummyPoly(domelements[i]) $ 
					    i=1..extnops(domelements));
	    return(DOM(pd,vars,ctr,ord));
      end_proc;

    _power:=
	proc(a,n)
	begin
	    // testargs
	    if not contains({DOM_INT, Dom::Integer}, domtype(n)) 
		    or (args(2) < 0)
	    then 
		error("expected a NonNegativeInteger as exponent");
	    end_if;
	    return(dom(_power(dom::dummyPoly(a),n), dom::indets(a), 
		        dom::point(a), dom::order(a)));
	end_proc;

    seriesdiff :=
	proc(s,x)
	begin
	    if has(dom::indets(s),x) then
		return(dom(diff(dom::dummyPoly(s),
		             dom::dummyVars(s)[contains(dom::indets(s),x)]),
			     dom::indets(s),
			     dom::point(s),
			     dom::order(s) - 1));
	    else
		return(new(dom,diff(dom::dummyPoly(s),x),
		           dom::indets(s),
			   dom::point(s),
			   dom::order(s)));
	    end_if;
	end_proc; /* diff */

    diff :=
	proc(s)
	begin
	    if args(0) = 1 then
		return(s);
	    else
		for i from 2 to args(0) do
		    s := dom::seriesdiff(s, args(i));
		end_for;
		return(s);
	    end_if;
	end_proc; /* diff */

    print :=
	proc(s)
	local i, trm, pmult, reverse, e, v;
	begin
	    i := hold(O)(zip(dom::indets(s),dom::point(s), _subtract),
		                                         dom::order(s));
	    trm := zip(dom::indets(s), dom::point(s), _subtract);
	    // Linear terms are expanded, this should be taken care of.
	    // I'll try this:
	    pmult :=
		proc(a,b)
		begin
		    if a=1 then
			return(b);
		    elif b=1 then
			return(a);
		    else
			return(hold(_mult)(a,b));
		    end_if;
		end_proc; /* pmult */
	    // Some niceties, but this is not a good way of doing it
	    // I want the terms in increasing order, with the terms
	    // with the same order arranged according to the way they
	    // are arranged in the variable list.
	    // Arranging them in increasing order is easy,
	    // I'm sure there's a simple way of doing it in decreasing 
	    // order, but for now I'll reverse everything in sight.
	    reverse :=
		proc(l:DOM_LIST):DOM_LIST
		begin
		    return([l[nops(l)-i+1]$i=1..nops(l)]);
		end_proc; /* rlist */
	    e := dom::dummyPoly(s);
	    v := extop(e,2);
	    v := reverse(v);
	    e := poly(expr(e),v);
	    e := polylib::sortMonomials(e, DegreeOrder);
	    // reverse
	    e := reverse(e);
	    e := map(e,poly2list);
	    e := map(e,extop);
	    e := map(e,li -> [li[1], zip(reverse(trm), li[2], _power)]);
	    e := map(e,li -> [li[1], _mult(op(li[2]))]);
	    e := map(e, li -> pmult(op(li)));
	    return(hold(_plus)(op(e),i));
	    // The following mult is 
	end_proc; /* print */

    // Some utility functions

    // I don't think this is the best way to do it, but it
    // will work for now.
    truncate :=
	proc(pol, n)
	local mon, t;
	begin
	    if n = infinity then
		return(pol);
	    else
		mon:= dom::decreasingOrderList(pol);
		for t in mon do
		    if degree(t) >= n then
			pol := pol - t;
		    else
			break;
		    end_if;
		end_for;
		return(pol);
	    end_if;
	end_proc; /* truncate */
    
    // This next one was stolen from sortmon.mu
    
    // This will take a polynomial and return a list of the monomials 
    // as polynomials, in decreasing order
    decreasingOrderList:=
	proc(pol)
	local o, monomlist, i;
	begin
	    monomlist:=[polylib::nthmonomial(pol, i) $i=1..nterms(pol)];
	    o:=Dom::MonomOrdering(DegLex(extnops(extop(pol,2))));
	    sort(monomlist, (x,y) -> bool(o(degreevec(x), degreevec(y))>0))
	end_proc;

/* body of the domain */

begin
/* Not much to it */
end_domain:
