/*****************************************************************************
  $Id: $
  File:		lib/DOMAINS/DOMAIN/PermutationGroup.mu
  Authors:	Nicolas M. Thiry <nthiery@users.sourceforge.net>
		Original version by Stephan Wehmeier and the MuPAD group
  Taken from:	PerMuVAR, 02/12/2001
  Taken from:	MuPAD 1.4/2.0, 1998-2000
  Created:	12/02/1996
  License:	LGPL
*****************************************************************************/

/**
Dom::PermutationGroup - domain for permutation groups

Dom::PermutationGroup(n, generators [,options])

n	   - positive integer
generators - list of permutations

A permutation of {1..n} is a bijective mapping of {1..n} onto itself.

Dom::PermutationGroup(n,generators <,options>) creates a domain for
the group of permutations of {1..n} generated by generators.

For backward compatibility, if generators is not specified, then the
full symmetric group is created. However, it's better to use
Dom::SymmetricGroup instead.

An element is represented as a domain element with operands:

0 - domain Dom::PermutationGroup(n, generators)
1 - a list [a_1, ... ,a_n ] of n integers, representing the
    permutation that maps i to a_i for each i between 1 and n 

**/

alias(REP(sigma)=extop(sigma,1)):

domain Dom::PermutationGroup(n: Type::PosInt,
                             generators=FAIL: Type::Union(DOM_LIST,
                                                          DOM_FAIL,
                                                          DOM_IDENT
                                                          )
                            )
    // DOM_IDENT allows for the following usage:
    // domain bla
    //    inherits Dom::PermutationGroup(4, toBeDefined)
    //    generators := ...;
    // begin
    // end;

    inherits if Dom::PermutationGroup_GLIP<>FAIL then
                 Dom::PermutationGroup_GLIP
             else 
                 Dom::BaseDomain;
             end_if;
    category Cat::PermutationGroup;
    axiom Ax::canonicalRep;

    info_str := "domain for a permutation group";

    one := new(dom, [ $ 1..n ]);

    convert :=
    proc(x)
    begin
	case domtype(x)
	of dom do 
		return(x);
	of DOM_ARRAY do 
		return(dom::convert([op(x)]));
	of DOM_LIST do
	    case {op(map(x,domtype)) } 
	    of { DOM_INT } do
		if has(map(x, i -> bool(i<= n and i > 0) ), FALSE)
		then
		    error("List contains integers outside the range");
		elif nops(x) <> n then
		    error("List has incorrect length");
		elif nops({ op(x) } ) <> n then
		    error("No permutation (not bijective)");
		else
		    return(new(dom, x));
		end_if;
	    of { DOM_LIST } do
		return(dom::fromCycles(x));
	    otherwise
		error("List should be a list of integers, or a list of list of integers (cycles)");
	    end_case;
// Modifs by NT to handle permutations represented by a table mapping
//	each element of the set to another
// Is this a usefull feature ?
// Should the same be implemented for procedures ?			
	of DOM_TABLE do
	    return(new(dom,
		       [dom::invertSet[x[op(dom::set,i)]] $
			i = 1..dom::card]));
// End of modifs by NT
	end_case;
	FAIL
    end_proc;

    convert_to := 
    proc(x, T)
    begin
	case T
	of DOM_LIST do	
		 return(REP(x));
	of dom do return(x);
	end_case;
	FAIL
    end_proc;

    // returns a list of cycles c_1,..,c_k such that a=c_1*..*c_k; each
    // cycle is written as a list
    cycles :=
    proc(a)

	local l,i,k,cyc,result;
	// l[i] is TRUE iff i is already involved in a cycle

	begin
	l:=[ FALSE $ n ];
	cyc:=[];
	result:=[];
	for i from 1 to n do	
		if not l[i] then
			k:=i; 
			repeat
				l[k]:=TRUE;
				cyc:=append(cyc,k);
				k:=op(REP(a),k);				
			until i=k end_repeat;
			result:=append(result,cyc);
			cyc:=[];
		end_if;
	end_for;
	result
	end_proc;


    // expr: returns permutation as a list

    expr := x -> REP(x);

    print := x -> REP(x);

    _mult := 
    proc(x: dom)
    begin
	if args(0) = 2 then
	    if testargs() and domtype(args(2)) <> dom then
		error("Wrong 2. argument type.");
	    else
		new(dom, [ op(REP(args(1)),op(REP(args(2)),i)) $ i=1..n])
	    end_if
	elif args(0) = 1 then
	    args(1)
	else
	    _mult(args(1..(args(0) div 2)));
	    _mult(args(((args(0) div 2) + 1)..args(0)));
	    if (domtype(%2))::_mult <> FAIL then (domtype(%2))::_mult(%2, %1)
	    elif (domtype(%1))::_mult <> FAIL then (domtype(%1))::_mult(%2, %1)
	    else _mult(%2, %1)
	    end_if
	end_if
    end_proc;

//    "_power" : inherited from category

    _invert := 
    proc(x: dom)
    begin
	new(dom, combinat::permutations::inverse(REP(x)));
    end_proc;

//   "_divide" : defined in Cat::Group

    order :=
    proc(a)
	local c,i;

	begin
	 c:=dom::cycles(a);
	 ilcm(nops(op(c,i)) $ i = 1..nops(c))
	end_proc;
			
    // func_call_inherits is there so that subdomains can use it when
    // for defining their own func_call

    func_call_inherits := 
    proc(sigma: dom, ll)
	local l;
    begin
	l:=context(ll);
	if type(l)=DOM_INT then
	    if testargs() and (l<1 or l>n) then
		error("Permutation is only defined for arguments between 1 and ".
		      expr2text(n));
	    else
		return(op(extop(sigma,1),l))
	    end_if
	elif type(l)=DOM_LIST then
	    if testargs() and nops(l)<>n then
		error("List must have ".expr2text(n)." elements")
	    end_if;
	    dom::actOnList(sigma,l)
	else
	    error("Second argument has wrong type")
	end_if
    end_proc;

    func_call := dom::func_call_inherits;

    TeX := x -> expr2text(REP(x));

    //////////////////////////////////////////////////////////////////////////
    // methods inherited from combinat::permutations
    //////////////////////////////////////////////////////////////////////////

    inversions :=
    proc(sigma: dom) : Type::NonNegInt
    begin
	combinat::permutations::inversions(REP(sigma));
    end_proc;

    sign :=
    proc(sigma: dom) : Type::Integer
    begin
	combinat::permutations::sign(REP(sigma));
    end_proc;

    descent :=
    proc(sigma: dom) : Type::Integer
    begin
	combinat::permutations::descent(REP(sigma));
    end_proc;

    descentPoly := dom::descentPolynomial;

    /**
	descentPolynomial(p:permutation, <z: DOM_IDENT>)
 
	Returns the descent polynomial of p.
	This is the product of all the (z[1]..z[i]), where i are descents of p.
 
	References: Garsia and Stanton 1984
    **/

    descentPolynomial :=
    proc(sigma: dom) : Type::Integer
    begin
	combinat::permutations::descentPolynomial(REP(sigma),args(2..args(0)));
    end_proc;


    // Provided here for backward compatibility only!
    // The result is incorrect for a general permutation group.
    random := (Dom::SymmetricGroup(n))::random;

//    _index := fun((
//	    if args(0) = 1 then op(extop(args(1),1))
//	    else op(extop(args(1),1), args(2)) end_if
//	    )),

    //////////////////////////////////////////////////////////////////////////
    // Stuff related to the action of a permutation group on [1..n]
    //////////////////////////////////////////////////////////////////////////
/*--
    set
    card
    act

    Definition of the default action of a permutation group on [1..n].
    This makes this domain a Cat::PermutationGroup
--*/

    set := [$1..n];

    card := n;

    act :=
    proc(sigma:dom,i:Type::PosInt) 
    begin
	op(REP(sigma),i);
    end_proc;

    actOnList :=
    proc(sigma:dom, l:DOM_LIST)
    local result;					  
    begin
	if testargs() then
	    if nops(l) <> n then
		error("List must have ".expr2text(n)." elements")
	    end_if;
	end_if;
	sigma:=REP(sigma);
	result:=l;
	(result[sigma[i]]:=l[i]) $ i = 1..n; // Faster than a for loop
	result;
    end_proc;

//    actOnListFunc :=
//    proc(g:dom)
//    begin
//	g:=REP(g);


    allElements :=
    (userinfo(5, "Generating all the elements of the group");
     if nops(dom::generators)=0 then
	 [dom([$1..n])];
     else
	 dom::multiplicativeClosure(dom::generators);
     end_if);

    generators :=
    if generators=hold(toBeDefined) then
	hold(toBeDefined)
    else
	map(generators,dom);
    end_if;

    selectDescent :=
    proc(l:Type::ListOf(Type::PosInt))
	local desc;
    begin
	desc:=dom::descent;
	return(select(dom::allElements, p->bool(dom::descent(p) = l)));
    end_proc;

/*--
    fromCycles:

    Return a new permutation from its disjoint cycle representation.
--*/
    fromCycles :=
	proc(cycles: Type::ListOf(Type::ListOf(Type::PosInt)))
	    local c, i, p, known;
	begin
	    if testargs() then
		known:=[FALSE $ n];
		for c in cycles do
		    for i in c do
			if i>n then
			    error("Cycles should consists of integers in {1..n}")
			end_if;
			if known[i] then
			    error("Cycles are not disjoint");
			end_if;
			known[i]:=TRUE;
		    end_for;
		end_for;
	    end_if;
	    p:=[$1..n];
	    for c in cycles do
		for i from 1 to nops(c)-1 do
		    p[c[i]]:=c[i+1];
		end_for;
		p[c[nops(c)]]:=c[1];
	    end_for;
	    new(dom,p);
	end_proc;

begin
    if args(0) < 2 then
        if combinat::warnDeprecated() then
            warning("Dom::PermutationGroup(n) is obsolete.
Please use Dom::SymmetricGroup(n) instead.");
        end_if;
        generators := map((Dom::SymmetricGroup(n))::generators, extop);
    end_if;
end_domain:

unalias(REP):

stdlib::LoadlibState["Dom::PermutationGroup"] := TRUE:
