/*++ $Id: SymmetricGroup.mu,v 1.1 2001/12/03 17:48:16 nthiery Exp $

Dom::SymmetricGroup(n) -- the full symmetric group of {1..n}

n	   - integer >= 1

Elements are represented as in Dom::PermutationGroup(n)

Author:	     Nicolas M. Thiry <nthiery@users.sourceforge.net>
Created:     August 8th, 1999
Last update: $Date: 2001/12/03 17:48:16 $
++*/

domain Dom::SymmetricGroup(n: Type::PosInt)
    inherits Dom::PermutationGroup(n, toBeDefined);
    category Cat::PermutationGroup;
//    Ax::canonicalRep ?

/*--
    size

    Size of the group.
    Please overload if you can, to avoid generating the group if not necessary.
--*/

    size := fact(n);

/*--
    generators

    A list of generators of the group
--*/

    generators := 
    if	 n=1 then [dom([1])];
    elif n=2 then [dom([2,1])];
    else	   [dom([[2,1]]), dom([[$1..n]])]
    end_if;

/*--
    allElements

    List of all the elements of the group
--*/

    allElements := [ new(dom, p) $ p in combinat::permutations(n) ];

/*--
   cycleTypes():

   Count the elements of the group by cycle type.
   (Cf Cat::PermutationGroupModule).

   The conjugacy classes of Sn are indexed by the partitions p of n.
   When g is in this conjugacy class, p is precisely the decreasing list
   of the length of the cycles of g, so it's easy to compute the cycle
   type of g, knowing p. Moreover, the size of the conjugacy class is
   also easy to compute.

   This allows to compute the cycle types really fast, without having
   to generate the whole symmetric group (which is is BIG ...).

   Ref: A. Kerber, Algebraic Combinatoric Via Finite Group Action p 43
--*/

    cycleTypes :=
    proc()
	option remember;
	local t,gen,p;
    begin
	userinfo(3, "cycleTypes: starting computation");
	t:=table();

	gen := combinat::partitions::generator(n);
	while (p:=gen()) <> FAIL do
	    userinfo(5, "working on partition", p);
            // Compute the size of the conjugacy class of Sn indexed by p
            // and the cycle type of a permutation in this conjugacy class
	    t[combinat::partitions::toExp(p,n)]
              := combinat::partitions::conjugacyClassSize(p);
        end_while;
	t;
    end_proc;

    random :=
    proc() : dom
    begin
        new(dom, combinat::permutations::random(n));
    end_proc;

begin
    if testargs() then
	if args(0) <> 1 then error("wrong no of args"); end_if;
	if not testtype(n,DOM_INT) then
	    error("argument must be integer")
	end_if;
	if n < 1 then
	    error("argument must be positive")
	end_if;
    end_if;
end_domain:

stdlib::LoadlibState["Cat::SymmetricGroup"] := TRUE:
