// $Date: 2002/02/12 12:53:11 $  $Author: tonner $  $Revision: 1.26 $ 

/*--
generate::TeX.mu -- TeX-output of expressions

--*/
alias(OP = output::Priority):

proc()
begin
  
/*++
generate::TeX -- return TeX-formatted string for expression

generate::TeX(e,...)

e - expression
++*/

generate::TeX:= proc(e) 
  local i;
begin 
    case args(0)
    of 0 do
    	""; break;
    of 1 do
    	generate::tex(e, OP::Noop); break;
    otherwise
    	_concat(generate::tex(e, OP::Exprseq),
    		(",", generate::tex(args(i), OP::Exprseq)) $ i=2..args(0));
    end_case 
end_proc ;

/*++
generate::TeXdef -- define TeX-string for expression

generate::TeXdef(e, s)

e - expression
s - TeX-string
++*/

generate::TeXdef:= (e, s)->(
    if args(0) <> 2 then error("wrong no of args") end_if;
    if domtype(s) <> DOM_STRING then error("no TeX string") end_if;
    generate::TeXdefs[e]:= s
);

// ------------------------ internas ----------------------- 

// table of user-defined TeX-strings 

generate::TeXdefs:=
table(
      hold(PI)="\\pi", I="i", hold(EULER)="\\gamma",
      op(map([
              hold(Delta), hold(Phi), hold(Gamma), hold(Lambda),
              hold(Pi), hold(Theta), hold(Sigma), hold(Upsilon),
              hold(Omega), hold(Xi), hold(Psi), hold(alpha),
              hold(beta), hold(chi), hold(delta), hold(epsilon),
              hold(varepsilon), hold(phi), hold(varphi), hold(gamma),
              hold(eta), hold(iota), hold(kappa), hold(lambda), hold(mu),
              hold(nu), hold(pi), hold(varpi), hold(theta), hold(vartheta),
              hold(rho), hold(varrho), hold(sigma), hold(varsigma),
              hold(tau), hold(upsilon), hold(omega), hold(xi), hold(psi),
              hold(zeta), hold(partial), hold(aleph), hold(Im), hold(Re),
              hold(wp), 

              hold(arcsin), hold(arccos), hold(arctan),
              hold(sin), hold(sinh), hold(cos), hold(cosh), hold(tan),
              hold(tanh), hold(cot), hold(coth), hold(csc), hold(csch),
              hold(sec), hold(sech)
              ], (x -> (x = "\\".x)))
         )
    ):

/*--
generate::tex -- return TeX-formatted string for expression

generate::tex(e,p)

e - expression
p - priority of enclosing expression
--*/

generate::tex:= (e, p)->(
    if contains(generate::TeXdefs, e) then
	generate::TeXdefs[e]
    else
	slot(domtype(e), "TeX")(args());
        if % = FAIL then
          DOM_IDENT::TeX(args())
        else
           %
        end_if
    end_if;
);

/*--
generate::TeXquote -- quote TeX-string

generate::TeXquote(s,ep,sp)

s - TeX string
ep - priority of enclosing expression
sp - priority of this expression
--*/

generate::TeXquote:= (s,ep,sp)->
    if ep<sp then s else "\\left(".s."\\right)" end_if;

/*--
generate::TeXseq -- create TeX string for sequence

generate::TeXseq(lb,rb,e,...)

lb, rb - strings with left and right brackets
e      - expression
--*/

generate::TeXseq:= proc(lb,rb,e) 
  local i;
begin 
  _concat(
    lb,
    (case args(0)
     of 2 do ""; break;
     of 3 do generate::tex(e, 0); break;
     otherwise
     	generate::tex(e, OP::Exprseq),
     	((",", generate::tex(args(i), OP::Exprseq)) $ i=4..args(0));
     end_case),
    rb) 
end_proc ;

/*--
generate::TeXoperator -- TeX formatting of operator

generate::TeXoperator(sym, ep, tp, oP,...)

sym - operator symbol
ep  - priority of enclosing expression
tp  - priority of this expression
oP  - operands
--*/

generate::TeXoperator:= proc(sym, ep, tp, oP) 
  local i;
begin 
    case args(0)
    of 3 do ""; break;
    of 4 do generate::tex(oP, ep); break;
    otherwise
    	generate::TeXquote(
    	    _concat(generate::tex(oP, tp),
                  (sym, generate::tex(args(i), tp))
    	    	     $ i=5..args(0)),
                         ep, tp)
    end_case
end_proc;

/*--
generate::TeXfun -- TeX formatting function calls

generate::TeXfun(o,p,e,...)

o - TeX string of operator
p - priority of enclosing expression
e - operand expressions
--*/

generate::TeXfun:= proc(o,p,e) 
  local i;
begin 
    o.generate::TeXseq("\\left(", "\\right)", args(i) $ i=3..args(0))
end_proc ;

/* uncomment the following procedure if you prefer f(x) to be
  formatted as 'f x' instead of 'f(x)'... */
/*
generate::TeXfun:= proc(o,p,e) local s, i; begin
    case args(0)
    of 2 do
    	o."\\left(\\right)";
    	break;
    of 3 do
    	s:= generate::tex(e, OP::Fconcat);
    	if substring(s,0,7) = "\\left(" then
    	    o.s
    	else
    	    generate::TeXquote(o." ".s, p, OP::Fconcat)
    	end_if;
    	break;
    otherwise
    	o.generate::TeXseq("\\left(", "\\right)", args(i) $ i=3..args(0))
    end_case
end_proc;
*/

// ------------- TeX formatting of domain elements ------------- 

DOM_INT::TeX:= x -> expr2text(x);
DOM_RAT::TeX:= DOM_INT::TeX;
DOM_BOOL::TeX:= DOM_INT::TeX;
DOM_FAIL::TeX:= DOM_INT::TeX;
DOM_NIL::TeX:= DOM_INT::TeX;
DOM_NULL::TeX:= "";
DOM_SET::TeX:=  x -> generate::TeXseq("\\left\\{", "\\right\\}", op(x));
DOM_LIST::TeX:= x -> generate::TeXseq("\\left[", "\\right]", op(x));
DOM_POLY::TeX:= (x, y) -> generate::tex(expr(x), y);

DOM_COMPLEX::TeX:=
  proc(x, y)
    local imagimary, keepOrder;
    save  _domainOutputMode;
  begin
    if x = I then
      return(generate::TeXquote(I));
    end_if;
    if op(x, 2) = 1 then
      imagimary := I;
    else
      imagimary := hold(_mult)(I, op(x, 2));
    end_if;
    if iszero(op(x, 1)) then
      generate::tex(imagimary, OP::Noop)
    else
      _domainOutputMode := FALSE;
      keepOrder := Pref::keepOrder(Always);
      generate::tex(hold(_plus)(op(x, 1), imagimary), OP::Noop);
      Pref::keepOrder(keepOrder);
      %2
    end_if;
  end_proc:
     
DOM_TABLE::TeX:= proc(x) 
                   local i;
                 begin
                   _concat(
                           "\\begin{eqnarray*}\n",
                           (if nops(x) = 0 then
                              null()
                            else
                              generate::tex(op(x,[1,1]), OP::Relation),
                              " & = & ",
                              generate::tex(op(x,[1,2]),OP::Relation),
                              "\\\\\n",
                              ((generate::tex(op(x,[i,1]),OP::Relation),
                                " & = & ",
                                generate::tex(op(x,[i,2]),OP::Relation),
                                "\\\\\n")
                               $ i=2..nops(x))
                            end_if),
                           "\\end{eqnarray}")
                 end_proc;

DOM_ARRAY::TeX:= proc(a)
                   local n, m, p, i, j;
                 begin
                   if op(a,[0,1]) = 1 then
                     n:= op(a,[0,2,1]);
                     m:= op(a,[0,2,2]);
                     _concat(
                             "\\begin{array}{", ("c" $ (m - n + 1)), "}\n",
                             generate::tex(a[n],OP::Exprseq),
                             ((" & ",
                               generate::tex(a[i]),OP::Exprseq) $ i=(n+1)..m),
                             "\n\\end{array}")
                   elif op(a,[0,1]) = 2 then
                     n:= op(a,[0,3,1]);
                     m:= op(a,[0,3,2]);
                     _concat(
                             "\\begin{array}{", ("c" $ (m - n + 1)), "}\n",
                             ((generate::tex(a[j,n],OP::Exprseq),
                               ((" & ", generate::tex(a[j,i],OP::Exprseq))
                                $ i=(n+1)..m), "\\\\\n")
                              $ j=op(a,[0,2])),
                             "\\end{array}")
                   else
                     p:= proc(x)
                           local i;
                         begin
                           if nops(x) = op(a,[0,1]) then
                             "(".generate::tex(op(x),OP::Relation).") & = & ".
                             generate::tex(a[op(x)],OP::Relation)."\\\\\n"
                           else
                             _concat(p(x.[i]) $ i=op(a,[0,nops(x)+2]))
                           end_if
                         end_proc;
                     "\\begin{eqnarray*}\n".p([])."\\end{eqnarray}"
                   end_if
                 end_proc;

DOM_FLOAT::TeX:= proc(x, y)
                   local p;
                 begin

                   // p removes trailing zeros
                   p:= proc(s)
                         local m;
                       begin
                         if strmatch(s, "\*.\*") then
                           s:= text2list(s, ["."]);
                           m:= s[3];
                           while substring(m, length(m)-1, 1) = "0" do
                             if length(m) = 1 then break end_if;
                             m:= substring(m, 0, length(m)-1)
                           end_while;
                           s[1].".".m
                         else
                           s
                         end_if
                       end_proc;
    
                   x:= expr2text(x);
                   if strmatch(x, "\*e\*") then
                     x:= text2list(x, ["e"]);
                     x[1]:= p(x[1]);
                     generate::TeXquote(_concat(x[1]," 10^",
                                                (if length(x[3]) = 1 then
                                                   x[3]
                                                 else
                                                   "{".x[3]."}"
                                                 end_if)),
                                        y, OP::Mult)
                   else
                     p(x)
                   end_if
                 end_proc;

DOM_STRING::TeX:= proc(x)
                  begin
                    // change _ into \_
                    if strmatch(x, "\*_\*") then
                      x:= _concat(op(subs(text2list(x, ["_"]), "_"="\\_")))
                    end_if;
                    "\\mbox{\"".x."\"}"
                  end_proc;

DOM_IDENT::TeX:= proc(x)
                 begin
                   // change _ into \_
                   x:= expr2text(x);
                   if strmatch(x, "\*_\*") then
                     x:= _concat(op(subs(text2list(x, ["_"]), "_"="\\_")))
                   end_if;
                   if length(x) = 1 then x else "\\mbox{".x."}" end_if;
                 end_proc;

DOM_EXPR::TeX:= proc(x, y)
                  local f;
                begin
                  f:= op(x,0);
                  if domtype(f) = DOM_IDENT then
                    f: f := eval(%);
                    if domtype(f) = DOM_FUNC_ENV and f::TeX <> FAIL then
                      f::TeX(f, x, y)
                    else
                      generate::TeXfun(generate::tex(op(x,0)), y, op(x))
                    end_if
                  else
                    generate::TeXfun(generate::tex(f, OP::Fconcat), y, op(x))
                  end_if
                end_proc;

DOM_DOMAIN::TeX:= proc(x, y)
                  begin
                    if x::Name <> FAIL then
                      if domtype(x::Name) = DOM_STRING then
                        generate::tex(text2expr(x::Name), y)
                      else
                        generate::tex(x::Name, y)
                      end_if
                    elif x::key <> FAIL then
                      if domtype(x::key) = DOM_STRING then
                        generate::tex(text2expr(x::key), y)
                      else
                        generate::tex(x::key, y)
                      end_if
                    else
                      generate::tex(hold(newDomain)(), y)
                    end_if
                  end_proc:

// ------------------- TeX formatting of func-env's ------------------- 
/*--
Each TeX-formatting routine must have the form
	myformatfunc(fenv, expr, prio)
where fenv is the func-env of this routine, expr is the original expression
causing this call (the operator of expr was evaluated to fenv) and prio is
the priority of the enclosing expression.
--*/

_mult::TeX :=
proc( arg1, arg2, arg3 )
  local list, str;
begin
  list := generate::splitProduct(arg2);
  if nops( op(list, 3)) = 0 then
    // no denumerator 
    str := generate::TeXoperator("\\, ", arg3, OP::Mult, op(op(list,2)));
  else
    str := _concat( "\\frac{".
                 generate::TeXoperator("\\, ",arg3, OP::Mult, op(op(list,2))).
                 "}{".
                 generate::TeXoperator("\\, ",arg3, OP::Mult, op(op(list,3))).
                 "}");
  end_if;
  if op(list,1) = -1 then
    generate::TeXquote( "- ".str , arg3, OP::Plus);
  else
    str;
  end_if;
end_proc:

_divide::TeX := 
proc( arg1, arg2, arg3 )
  local str;
begin
  str := _concat( "\\frac{".
               generate::TeXoperator("\\, ",arg3, OP::Mult, op(arg2,1)).
               "}{".
               generate::TeXoperator("\\, ",arg3, OP::Mult, op(arg2,2)).
               "}");
end_proc:

_negate::TeX :=
proc( arg1, arg2, arg3 )
  local str;
begin
  str := generate::TeXoperator(" ",arg3, OP::Plus, op(arg2,1)) ;
  generate::TeXquote( "- ".str , arg3, OP::Plus); 
end_proc:

_subtract::TeX :=
proc( arg1, arg2, arg3 )
  local str;
begin
  str := _concat( 
               generate::TeXoperator(" ",arg3, OP::Plus, op(arg2,1)).
               " - ".
               generate::TeXoperator(" ",arg3, OP::Plus, op(arg2,2)) 
               );
  generate::TeXquote( str , arg3, OP::Plus);
end_proc:

_invert::TeX :=
proc( arg1, arg2, arg3 )
  local str;
begin
  str := _concat( "\\frac{1}{".
               generate::TeXoperator(" ",arg3, OP::Mult, op(arg2,1)).
               "}");
end_proc:

_plus::TeX :=
proc(arg1, arg2, arg3)
  local tmp, s, i, o;
begin
  case nops(arg2)
    of 0 do return("");
        of 1 do return(generate::tex(op(arg2,1), arg3));
  end_case;
  tmp:= generate::sortSum(arg2);
             
  s:= "";   /* generate::tex(op(tmp, 1), OP::Plus); */
  for i from 1 to nops(tmp) do
    o:= op(tmp, i);
    if generate::isNeg(o) then
      s:= s." - ".generate::tex(- o, OP::Plus); next;
    else
      if i > 1 then
        s:= s." + ";
      end_if;
      s := s.generate::tex(o, OP::Plus)
    end_if;
  end_for;
             
  generate::TeXquote(s, arg3, OP::Plus)
end_proc:

_power::TeX :=
proc(arg1, arg2, arg3)
  local s;
begin
  if testtype(op(arg2,2), Type::Real) then
    if op(arg2,2) = -1 then
      return("\\frac{1}{".generate::tex(op(arg2,1), OP::Noop)."}")
    end_if;
    if op(arg2,2) < 0 then
      return("\\frac{1}{".
             _power::TeX(_power,
                         op(arg2,1)^(-op(arg2, 2)), OP::Noop)."}")
    end_if;
    if op(arg2,2) = 1/2 then
      return("\\sqrt{".generate::tex(op(arg2,1), OP::Noop)."}")
    end_if;
    if domtype(op(arg2, 2)) = DOM_RAT then
      if op(arg2, [2,1]) = 1 then
        return("\\sqrt[".expr2text(op(arg2, [2,2]))."]{".
               generate::tex(op(arg2, 1),OP::Noop)."}")
      else
        return("\\sqrt[".expr2text(op(arg2, [2,2]))."]{".
               _power::TeX(_power,
                           op(arg2, 1)^op(arg2, [2,1]), OP::Noop)."}")
      end_if
    end_if
  end_if;
  s:= generate::tex(op(arg2, 2), OP::Noop);
  if length(s) <> 1 then s:= "{".s."}" end_if;
  generate::TeXquote(generate::tex(op(arg2, 1), OP::Power)."^".s, arg3,
                     OP::Power)
end_proc:

diff::TeX :=
proc(arg1, arg2, arg3)
  local res, l, t, x, i;
begin
  l:= [ op(arg2, 2..nops(arg2)) ];
  t:= table();
  for x in l do
    if contains(t, x) then
      t[x]:= t[x] + 1
    else
      t[x]:= 1
    end_if
  end_for;

  t:= _concat(("\\partial ", generate::TeX(op(t,[i,1])^op(t,[i,2])))
              $ i=1..nops(t));

  if nops(l) = 1 then
    res := "\\frac{\\partial"
  else
    res := "\\frac{\\partial^".generate::TeX(nops(l))
  end_if;
  generate::TeXquote(res."}{".t."} ".
                     generate::tex(op(arg2,1), OP::Prime),
                     arg3, OP::Prime)
end_proc:

D::TeX :=
  proc(arg1, arg2, arg3)
  begin
    generate::TeXquote(
                       if nops(arg2) = 1 then
                         "D ".generate::tex(op(arg2,1), OP::Prime)
                       else
                         generate::TeX(hold(_index)(hold(D),
                                                    op(op(arg2,1))))." ".
                         generate::tex(op(arg2,2), OP::Prime)
                       end_if, arg3, OP::Prime);
  end_proc:

int::TeX :=
  proc(arg1, arg2, arg3)
  begin
    generate::TeXquote(
                       if type(op(arg2,2)) = "_equal" then
                         _concat("\\int_{",
                                 generate::tex(op(arg2, [2,2,1]), OP::Noop),
                                 "}^{",
                                 generate::tex(op(arg2, [2,2,2]), OP::Noop),
                                 "} ",
                                 generate::tex(op(arg2, 1), OP::Noop),
                                 " d ",
                                 generate::tex(op(arg2, [2,1]), OP::Fconcat))
                       else
                         _concat("\\int ",
                                 generate::tex(op(arg2, 1), OP::Noop),
                                 " d ",
                                 generate::tex(op(arg2, 2), OP::Fconcat))
                       end_if, arg3, OP::Fconcat);
  end_proc:
    	 
sum::TeX :=
  proc(arg1, arg2, arg3)
  begin
    generate::TeXquote(
                       if type(op(arg2, 2)) = "_equal" then
                         _concat("\\sum_{",
                                 generate::tex(op(arg2,
                                                  [2,1])=op(arg2, [2,2,1]),
                                               OP::Noop),
                                 "}^{",
                                 generate::tex(op(arg2, [2,2,2]), OP::Noop),
                                 "} ",
                                 generate::tex(op(arg2, 1), OP::Plus))
                       else
                         _concat("\\sum_{",
                                 generate::tex(op(arg2, 2), OP::Noop), "} ",
                                 generate::tex(op(arg2, 1), OP::Plus))
                       end_if, arg3, OP::Plus);
  end_proc:

_index::TeX :=
  proc(arg1, arg2, arg3)
    local s, i;
  begin
    s:= generate::TeXseq("", "", op(arg2,i) $ i=2..nops(arg2));
    if length(s) <> 1 then
      s:= "{".s."}"
    end_if;
    generate::TeXquote(generate::tex(op(arg2, 1), OP::Index)."_".s,
                       arg3, OP::Index)
  end_proc:

_not::TeX :=
  (arg1, arg2, arg3) ->
    generate::TeXquote("\\neg ".generate::tex(op(arg2, 1), OP::Not),
                       arg3, OP::Not):

/* remove comment if you prefer e^x instead of exp(x)...
exp::TeX :=
  (arg1, arg2, arg3) ->
    _power::TeX(exp, hold(_power)(hold(e), op(arg2,1)), arg3):
*/

_stmtseq::TeX   :=
  (arg1, arg2, arg3) -> generate::TeXoperator("; ", arg3,
                                              OP::Stmtseq, op(arg2)):
_exprseq::TeX   :=
  (arg1, arg2, arg3) -> generate::TeXoperator(", ", arg3,
                                              OP::Exprseq, op(arg2)):
_or::TeX        :=
  (arg1, arg2, arg3) -> generate::TeXoperator(" \\vee ", arg3, OP::Or,
                                               op(arg2)):
_and::TeX       :=
  (arg1, arg2, arg3) -> generate::TeXoperator(" \\wedge ", arg3, OP::And,
                                              op(arg2)):
_less::TeX      :=
  (arg1, arg2, arg3) -> generate::TeXoperator(" < ", arg3, OP::Relation,
                                              op(arg2)):
_leequal::TeX   :=
  (arg1, arg2, arg3) -> generate::TeXoperator(" \\leq ", arg3, OP::Relation,
                                              op(arg2)):
_equal::TeX     :=
  (arg1, arg2, arg3) -> generate::TeXoperator(" = ", arg3, OP::Relation,
                                              op(arg2)):
_unequal::TeX   :=
  (arg1, arg2, arg3) -> generate::TeXoperator(" \\neq ", arg3, OP::Relation,
                                              op(arg2)):
_range::TeX     :=
  (arg1, arg2, arg3) -> generate::TeXoperator(" \\ldots ", arg3, OP::Range,
                                               op(arg2)):
_mod::TeX       :=
  (arg1, arg2, arg3) -> generate::TeXoperator(" \\mbox{mod} ", arg3, OP::Mod,
                                              op(args(2))):
_div::TeX       :=
  (arg1, arg2, arg3) -> generate::TeXoperator(" \\mbox{div} ", arg3, OP::Mod,
                                              op(arg2)):
_union::TeX     :=
  (arg1, arg2, arg3) -> generate::TeXoperator(" \\cup ", arg3, OP::Union,
                                              op(arg2)):
_minus::TeX     :=
  (arg1, arg2, arg3) -> generate::TeXoperator(" \\setminus ", arg3, OP::Minus,
                                              op(arg2)):
_intersect::TeX :=
  (arg1, arg2, arg3) -> generate::TeXoperator(" \\cap ", arg3, OP::Intersect,
                                              op(arg2)):
_fconcat::TeX   :=
  (arg1, arg2, arg3) -> generate::TeXoperator(" \\circ ", arg3, OP::Fconcat,
                                              op(arg2)):
_concat::TeX    :=
  (arg1, arg2, arg3) -> generate::TeXoperator(" \\cdot ", arg3, OP::Concat,
                                               op(arg2)):

end_proc():

unalias(OP):

// end of file 
