// $Date: 2001/04/26 12:41:27 $ $Author: stefanw $ $Revision: 1.1 $


/*

groebner::leftBasicSet(G <, ord>)

G - groebner basis: a list of polynomials
ord - (optional) term ordering


This method returns the so-called left basic set of the ideal
generated by G, represented as a list. 

Let G be a basis of an ideal over R[x_1, ..,x_m], where R is some ring.
[x_i1, ..., x_in] is the left basic set of G iff
no leading term of G is a polynomial in x_i1,..,x_in and if, for every
0 <= j<= n-1, k=i_(j+1) is the smallest index k such that no leading term of G is a polynomial in [x_1,..,x_ij, x_k].


The algorithm returns FAIL if G contains 1.

Source: Kredel/Weispfennig, Computing dimension and independent sets
for Polynomial Ideals, Journal of Symbolic Computation 6 (1988), pp. 231-247 


*/


groebner::leftBasicSet:=
proc(G: DOM_LIST, ord = null())
  
  local S: DOM_LIST,  // result to be returned
        Sx: DOM_LIST, // auxiliary variable to store S union {x}
        U: DOM_LIST,  // list of indets of the polynomial ring
        HT: DOM_LIST, // list of leading terms
        i: DOM_INT;   // loop variable

begin
  if testargs() then
    if nops(G) = 0 then
      error("List must be nonempty")
    end_if;
    
    if map(G, domtype) <> [DOM_POLY $nops(G)] then
      error("List must consist of polynomials")
    end_if;

    if nops({op(map(G, op, 2))}) <> 1 then
      error("All polynomials must have the same list of variables")
    end_if;

    
    if nops({op(map(G, op, 3))}) <> 1 then
      error("All polynomials must have the same coefficient ring")
    end_if;

    // test monomial ordering ?!

  end_if; // argument testing


  if contains(map(G, degree), 0) > 0 then
    return(FAIL)
  end_if;

  S:= [];

  U:= op(G, [1, 2]); // list of variables
  assert(type(U) = DOM_LIST);

  
  
  HT := map(G, lterm, ord);

  
  for i from 1 to nops(U) do
   
    // choose x \in U
    Sx:= append(S, U[i]); // S = S union {x}
   
    // we have to check whether some element of HT
    // is entirely made up of S union U[i]

    if nops(select(HT, lt -> indets(expr(lt)) minus {op(Sx)} = {})) = 0 then
      // no element of HT can be made up of variables in Sx
      S:= Sx;
      // else: nothing to do
    end_if;

  end_for;

  
  S
end_proc:

