// $Id: gamma.mu,v 1.10 2001/10/24 13:55:57 ccr Exp $

// Evaluation of the gamma function, following
// C. Lanzcos, "A Precision Approximation of the Gamma Function",
// J. SIAM Numerical Analysis, Series B, Vol. 1, 1964, pp. 86-96

// Apart from the rd*-functions, the whole thing is valid for complex
// numbers with Re>0, too.  Reflection holds for complex arguments.

// Here, precomputed coefficients are stored.
// The first element is DIGITS used, the second one is
// the value of `&gamma;` for the coefficients and the
// third is a list of coefficients computed (in interval 
// arithmetics).  After that, an error estimate and a
// numerical approximation of the zero of psi(x) are stored.
DOM_INTERVAL::_precomputed_gamma_coeffs := [0, 0, [], 0...0, 1.46]:

// coefficients of chebycheff polynomials
DOM_INTERVAL::_cheby_coeffs:=[]:

DOM_INTERVAL::_compute_gamma_coeffs := proc(g)
local k_max, i, rho, a, dopfak, eta, min_of_gamma, x;
begin
  if DIGITS <= DOM_INTERVAL::_precomputed_gamma_coeffs[1]
     and g = DOM_INTERVAL::_precomputed_gamma_coeffs[2] then return(); end;

  // TODO: compute required k_max.
  // maybe this must be decided dynamically
  k_max := 60;
  x := genident("blabla");
  for i from nops(DOM_INTERVAL::_cheby_coeffs) to k_max do
    sysassign(DOM_INTERVAL::_cheby_coeffs, DOM_INTERVAL::_cheby_coeffs
      . [revert([coeff(orthpoly::chebyshev1(2*i,x))])]);
  end_for;

  rho := [0 $ k_max];

  dopfak := [1$(k_max+1)];
  for i from 2 to k_max do
    dopfak[i+1] := dopfak[i]*(2*i-1);
  end_for;
  
  for i from 1 to k_max do
    rho[i] := _plus((DOM_INTERVAL::_cheby_coeffs[i][a+1] * 
        // F(a), needed exactly once
        hull(sqrt(2)/PI)*
           dopfak[a+1]*hull(sqrt(PI)/2^a) /
           hull(a+g+1/2)^(a+1/2) * DOM_INTERVAL::exp(a+g+1/2)
        )$a=0..(i-1));
    // If a coefficient contains 0,
    // everything from there on is dominated by rounding errors
    if 0 in rho[i] then break; end;
  end_for;

  rho:=[op(rho,1..i)];

// The formula uses rho_0/2 only.  Simplify the code:
  rho[1] := rho[1]/2;

// claimed to be a quality measure
// note that in the paper, the factor PI/2 is
// not part of eta.
  eta := hull(exp(g)/sqrt(2));
  eta := (_plus(((-1)^(i+1)*rho[i])$i=1..nops(rho)) 
           - eta)/eta * hull(PI/2);

  eta := (-1...1) * eta;

  min_of_gamma := op(numeric::solve(psi(x)=0, x=1.4..1.5));

  sysassign(DOM_INTERVAL::_precomputed_gamma_coeffs,
        [DIGITS, g, rho, eta, min_of_gamma]);

end_proc:

// This function is used with x > 0.
DOM_INTERVAL::_internal_Gamma := proc(x, rounding, gamma)
local res, i, s1, s2, s, rho, rnd, ii;
begin
// this is a crude "seems to work" calculation
  if args(0) < 3 then gamma := 3+ceil(DIGITS/10); end_if;

  DOM_INTERVAL::_compute_gamma_coeffs(gamma);

  if iszero(x) then return(RD_INF); end_if;
  if x = RD_INF then return(RD_INF); end_if;

  // shifting the input down by 1 makes the formulas much less messy.
  x := rdplus(x, -1, rounding);

  if rounding = -1 then
    // lower bound
    res := lhs(DOM_INTERVAL::_precomputed_gamma_coeffs[3][1]);
    i := 1;
    s := 1;
    for i from 1 to nops(DOM_INTERVAL::_precomputed_gamma_coeffs[3])-1 do
      s1 := 1;
      s2 := 1;
      rho := DOM_INTERVAL::_precomputed_gamma_coeffs[3][i+1];
      if rho<0 then
        rnd := 1;
      else
        rnd := -1;
      end_if;
      for ii from 1 to i do
        s1 := rdmult(s1, x+ii, -rnd);
        s2 := rdmult(s2, x-ii+1, rnd);
      end_for;
      s := rho * rddivide(s2, s1, rnd);
      res := res +s;
    end_for;
// use the error bound
    res := res + (res * DOM_INTERVAL::_precomputed_gamma_coeffs[4]);
    res := (hull(sqrt(2*PI))
            *(x+gamma+1/2)^hull(x+1/2)
            *exp(hull(-x-gamma-1/2))
            *res);
  else
    // upper bound
    res := rhs(DOM_INTERVAL::_precomputed_gamma_coeffs[3][1]);
    i := 1;
    s := 1;
    for i from 1 to nops(DOM_INTERVAL::_precomputed_gamma_coeffs[3])-1 do
      s1 := 1;
      s2 := 1;
      rho := DOM_INTERVAL::_precomputed_gamma_coeffs[3][i+1];
      if rho<0 then
        rnd := -1;
      else
        rnd := 1;
      end_if;
      for ii from 1 to i do
        s1 := rdmult(s1, x+ii, -rnd);
        s2 := rdmult(s2, x-ii+1, rnd);
      end_for;
      s := rho * rddivide(s2, s1, rnd);
      res := res + s;
    end_for;
// use the error bound
    res := res + (res * DOM_INTERVAL::_precomputed_gamma_coeffs[4]);
    res := (hull(sqrt(2*PI))
            *(x+gamma+1/2)^hull(x+1/2)
            *exp(hull(-x-gamma-1/2))
            *res);
  end_if;

  return(res);

end_proc:

DOM_INTERVAL::gamma := proc(iv)
local t1, t2;
begin
  if iv::dom <> DOM_INTERVAL then
     if iv = {} then return(iv); end_if;
     t1 := interval(iv);
     if type(t1) = DOM_INTERVAL then
        iv := t1;
     else
        return(FAIL);
     end_if;
  end_if;

  if op(iv,0) = hold(_union) then
     return(_union(map(op(iv), DOM_INTERVAL::gamma)));
  end_if;

  if op(iv,0) = FAIL then // complex interval
     error("only real intervals permitted");
  end_if;

  // Is one of the poles inside the interval?
  if lhs(iv) < 0 then
     if lhs(iv) = RD_NINF or rhs(iv) > ceil(lhs(iv)) then
        return(RD_NINF...RD_INF);
     end_if;
     // No, but the interval contains (only) negative numbers.
     // use reflection formula
     t1 := DOM_INTERVAL::sin(hull(PI)*iv);
     t2 := DOM_INTERVAL::gamma(1-iv);
     return ((PI/rdmult(rhs(t1), rhs(t2), 1))
         ... (PI/rdmult(lhs(t1), lhs(t2), -1)));
  end_if;

  t1 := DOM_INTERVAL::_internal_Gamma(lhs(iv), -1);
  t2 := DOM_INTERVAL::_internal_Gamma(rhs(iv), -1);

  if DOM_INTERVAL::_precomputed_gamma_coeffs[5] in iv then
     // minimum of gamma >= 0.885601.
     // TODO: Implement better approximation
     t1 := t1...0.885601;
  end_if;

  t1 ... t2;

end_proc:
