//------   file  det.mu  25.9.97    ---------------------
/*  Help file:

 det - determinant of a matrix

 Call:

 numeric::det( A <, Symbolic> <,hard_soft> <, MinorExpansion> )
                                                                   
 Parameter:

  A         -- a  square real or complex matrix of domtype DOM_ARRAY
               or of category Cat::Matrix
  hard_soft -- either 'Hard','HardwareFloats' or 'Soft','SoftwareFloats'
               If DIGITS <=15, the default is HardwareFloats
               If DIGITS  >15, the default is SoftwareFloats

                                                                   
Synopsis:

  numeric::det(A); returns the determinant of A. Non-float numerical
  entries of the matrix are evaluated to floats in the course of the
  computation. The algorithm is based on QR factorization using
  Householder transformations.

  Conversion to floats can be suppressed by the option Symbolic. In
  this case a symbolic LU factorization is used, non-numeric symbolic
  entries are admissible. However, the presence of such entries will
  slow down the computation drastically.

  With the option MinorExpansion recursive minor expansion along the
  first row is used. This option is useful for small dense matrices 
  with symbolic entries as well as for large sparse matrices.

  Examples: 

>> A:=array(1..3,1..3,[ [1,1,I] , [1,exp(1),I] , [1,2,2] ]):

>> numeric::det(A);
                        3.436563656 - 1.718281828 I

>> A:=array(1..3,1..3,[ [1/x,1,I/(2+x)] , [1,1,1] , [1,x,1] ]):     

>> numeric::det(A,Symbolic);
                                                 2
                        2 - (1 + I) x - (1 - I) x
                        --------------------------
                                x (x + 2)
>> numeric::det(A,MinorExpansion);

                           1     I      I x
                           - - ----- + ----- - 1
                           x   x + 2   x + 2

See also:  linalg::det

------------------------------------------------------------------*/

numeric::det:=proc(A)
local n, m, useHardwareFloats, HardwareFloatsRequested, f,
      symbolic, use_normal, method, B, i, j, k, result,
      real, h, s, ss, tmp, pivindex, subdet;
begin
   if args(0) < 1 then 
       error("expecting at least one argument") 
   end_if;

  //------------------
  // Check the matrix 
  //------------------
  if domtype(args(1)) = DOM_ARRAY then
     [n, m]:= [op(A, [0, 2, 2]), op(A, [0, 3, 2])];
  elif A::dom::hasProp(Cat::Matrix)=TRUE then
     [n, m]:= A::dom::matdim(A):
     if A::dom::constructor = Dom::SparseMatrix then
          // nothing to do: the scilab link accepts
          // objects of type Dom::Matrix(any ring)
     elif A::dom <> Dom::Matrix() then
          // convert to Dom::Matrix(), because the
          // scilab link just accepts Dom::Matrix(..)
          A:= (Dom::Matrix())::create(extop(A, 3));
     end_if;
  else
     error("expecting an array or a matrix of category 'Cat::Matrix'"):
  end_if;

  if m <> n then
     error("expecting a square matrix");
  end_if;
  //---------------------------------------------
  // Check the options
  //---------------------------------------------
  // set defaults
  method:= 2;        // floating point method
  if DIGITS <= 15 then
       useHardwareFloats:= TRUE;
  else useHardwareFloats:= FALSE;
  end_if;
  symbolic:= FALSE;
  HardwareFloatsRequested:= FALSE;

  for i from 2 to args(0) do
      case args(i)
      of Symbolic do method:= 1: // symbolic mode
                     symbolic:= TRUE;
                     useHardwareFloats:= FALSE;
                     break;
      of Hard do
      of HardwareFloats do
                     if method = 1 then
                        warning("ignoring the option 'HardwareFloats' ".
                                "because of the option 'Symbolic'"):
                        method:= 1;
                        useHardwareFloats:= FALSE:
                     else
                        useHardwareFloats:= TRUE:
                        HardwareFloatsRequested:= TRUE;
                     end_if;
                     break;
      of Soft do
      of SoftwareFloats do
                     useHardwareFloats:= FALSE:
                     break;
      of MinorExpansion do
                     method:=3;  // minor expansion 
                     useHardwareFloats:= FALSE;
                     break;
      otherwise
         error("unknown option ".expr2text(args(i)));
      end_case;
  end_for;

  //--------------------------------
  // ----- use HardwareFloats ------
  //--------------------------------
  case useHardwareFloats // use case because we want to use break below
  of TRUE do
     if DIGITS > 15 then
        if HardwareFloatsRequested then
           warning("the current precision goal DIGITS = ".expr2text(DIGITS).
                   " cannot be achieved with the requested 'HardwareFloats'.".
                   " Using 'SoftwareFloats' instead.");
        end_if;
        break;
     end_if;
     if not numeric::startHardwareFloats() then
        userinfo(1, "cannot start the hardware floats interface"):
        if HardwareFloatsRequested then
           warning("cannot start the 'HardwareFloats' interface, ".
                   "using 'SoftwareFloats' instead");
        end_if;
        break;
     end_if;

     userinfo(1, "using hardware floats"):

     //------------------------------------------------------
     // Problem: scilab::svd does not accept sparse matrices.
     // Workaround: we send a program to scilab that converts
     // the sparse matrix to a full matrix and then computes
     // the singular values:
     //------------------------------------------------------

     if A::dom::constructor = Dom::SparseMatrix then
        f:= scilab::func(["x", "dummy"], ["y"], ["y=det(full(x))"]);
        B:= f(A, 1, Quiet);
     else
        B:= scilab::det(A, Quiet);
     end_if;

     if has(B, FAIL) then
        userinfo(1, "hardware floats failed"):
        if HardwareFloatsRequested then
            warning("'HardwareFloats' failed, using 'SoftwareFloats' instead"):
        end_if;
        break;
     end_if;

     //------------------------------
     //-------- return --------------
     //------------------------------

     return(B);
  end_case; // of useHardwareFloats = TRUE

  //---------------------------------
  //---- use SoftwareFloats ---------
  //---------------------------------

  userinfo(1, "using software floats"):
  // convert to domtype DOM_ARRAY for speed
  if A::dom::hasProp(Cat::Matrix)=TRUE then
     A:= expr(A);
  end_if;
  if domtype(A) <> DOM_ARRAY then
     error("the matrix must be an array or of category 'Cat::Matrix'");
  end_if:

  if method = 2 then
    B:= map(A, float):
    if {op(map(B, domtype))} minus {DOM_FLOAT, DOM_COMPLEX} <> {} then
      warning("symbolic entry found, switching to symbolic mode");
      method := 1;
      use_normal := TRUE
    else
      A:= B;
      if ({op(map(A,(iszero@Im)))} = {TRUE}) then  // is matrix real ? 
           real:= TRUE;
      else real:= FALSE;
      end_if:
   end_if;
  end_if;

  use_normal:= FALSE;
  if method=1 then
    if (indets(A)<>{})
    or has({op(map(A,domtype))},DOM_NIL)
    or has({op(map(A,domtype))},DOM_EXPR) then
       use_normal:= TRUE
    end_if;
  end_if;

  if method=1 then
    userinfo(1,"symbolic mode, no pivoting unless necessary");
  end_if;
  if method=2 then
    userinfo(1,"numerical mode, using column pivoting");
  end_if;

  //--------------------------------------------------
  if method=1 then
     userinfo(1,"computing LU factorization");
     // enforce full evaluation
     A:= map(A, eval): 
     if (indets(A)<>{}) 
        or has({op(map(A,domtype))},DOM_NIL)
        or has({op(map(A,domtype))},DOM_EXPR)
        then use_normal:=TRUE else use_normal:=FALSE;end_if;
     if use_normal then A[1,1]:=normal(A[1,1]); end_if; 
     result:=1;
     for j from 1 to n-1 do
       if iszero(A[j,j]) then
           for pivindex from j+1 to n-1 do
               if not iszero(A[pivindex,j]) then break; end_if;
           end_for;
        else pivindex:=j;
        end_if;
        if pivindex<>j then
           result:=-result;
          (tmp:=A[j,k];A[j,k]:=A[pivindex,k];A[pivindex,k]:=tmp;) $ k=1..n;
        end_if;
        if iszero(A[j,j]) then return(0) end_if;
        result:=result*A[j,j];
        if use_normal
        then for i from j+1 to n do
                 tmp:=A[i,j]/A[j,j];
                 if not iszero(tmp) then
                 (A[i,k]:=normal(A[i,k]-tmp*A[j,k])) $ k=j+1..n;
                 end_if;
             end_for;
        else for i from j+1 to n do
                 tmp:=A[i,j]/A[j,j];
                 if not iszero(tmp) then
                 (A[i,k]:=A[i,k]-tmp*A[j,k]) $ k=j+1..n;
                 end_if;
             end_for;
        end_if;
     end_for;
     result:=result*A[n,n]:
     if use_normal then return(normal(result)):
                   else return(result):
     end_if;
  end_if; //method=1

  //--------------------------------------------------
  if method=2 then            
     userinfo(1,"computing QR factorization by Householder transformations");
     result:=1;
     for j from 1 to n-1 do
          // Householder step from the left to obtain upper triangular form. 
          s:= _plus(specfunc::abs(A[i,j])^2 $ i=j+1..n);
          if iszero(s) then result:= result*A[j,j];
          else ss:=s+specfunc::abs(A[j,j])^2;
               s:=specfunc::sqrt(ss):
               h := 1 / (ss + specfunc::abs(A[j,j]) * s);
               if not iszero(A[j,j]) then s:=sign(A[j,j])*s;end_if:
               A[j,j]:=A[j,j]+s; result:=result*s;
               for i from j+1 to n do
                   if real then   s:=_plus(A[k,j] *A[k,i]$ k=j..n) ;
                   else s:=_plus(conjugate(A[k,j])*A[k,i]$ k=j..n) ;
                   end_if;
                   s:=s*h;
                   if not iszero(s) then
                     (A[k,i]:=A[k,i]-s*A[k,j]) $ k=j+1..n ;
                   end_if;
               end_for;
          end_if;
     end_for;
     return(result*A[n,n]);
  end_if; //method=2

  //--------------------------------------------------
  if method=3 then
     userinfo(1,"using minor expansion along first row");
     // enforce full evaluation
     A:= map(A, eval): 
     if not symbolic then
        A:= map(A, float);
     end_if;
  
     // define local subroutine  subdet = minor 
  
     subdet:=proc(r,c,n) local result,i; option remember;
     begin
        if n=3 then return(A[r[1],c[1]]*A[r[2],c[2]]*A[r[3],c[3]]
                         + A[r[1],c[2]]*A[r[3],c[1]]*A[r[2],c[3]]
                         + A[r[2],c[1]]*A[r[1],c[3]]*A[r[3],c[2]]
                         - A[r[1],c[1]]*A[r[2],c[3]]*A[r[3],c[2]]
                         - A[r[1],c[2]]*A[r[2],c[1]]*A[r[3],c[3]]
                         - A[r[1],c[3]]*A[r[2],c[2]]*A[r[3],c[1]]
                          ) end_if;
        if n=2 then return(A[r[1],c[1]]*A[r[2],c[2]]
                          -A[r[1],c[2]]*A[r[2],c[1]]) end_if;
        if n=1 then return(A[r[1],c[1]]) end_if;
        result:=0;
        for i from 1 to n do
           if not iszero(A[r[i],c[1]]) then
              result:=result+(-1)^(1+i)*A[r[i],c[1]]
              *subdet(subsop(r,i=null()),subsop(c,1=null()),n-1);
           end_if;
        end_for;
        result;
    end_proc; //subdet

    //-------------------------------------------------------
    // find ordering c=[c[1],..,c[n]] of the columns s.th. 
    //    number of zeroes in column c[1]                  
    // >= number of zeroes in column c[2]                  
    // >=   ...                                            
    // Then expand recursively along column c[1] ...       
    //
    //  for j from 1 to n do                        
    //    z:=0;                                     
    //    for i from 1 to n do                      
    //       if iszero(A[i,j]) then z:=z+1;end_if;  
    //    end_for;                                  
    //    zeroes[j]:=z;                             
    //  end_for;                                    
    //                                              
    //  c:=[i $ i=1..n ];                     
    /*  s:=1;  */ // (s=signum of column permutation) 
    //                                              
    //  for j from 2 to n do                        
    //    z:=zeroes[j];                             
    //    for i from j downto 2 do                  
    //       if zeroes[i-1]>=z then break end_if;   
    //       zeroes[i]:=zeroes[i-1];c[i]:=c[i-1];   
    //    end_for;                                  
    //    zeroes[i]:=z; c[i]:=j; s:=(-1)^(j-i)*s;   
    //  end_for;                                    
    //                                              
    //  return(s*subdet([i$ i=1..n],c,n));    
    //-------------------------------------------------------
  
    result:= subdet([i$ i=1..n],[j$ j=1..n],n);
    delete subdet;
    return(result);
  end_if; //method=3
  //--------------------------------------------------

end_proc: //det
