/* W. Oevel, 17.2.98  */
/* 29.10.01: Scilab eingebaut */
/*-------------------------------------------------------------------------------*/
/* numeric::fMatrix - functional calculus for matrices                           */
/*                                                                               */
/* Aufruf:                                                                       */
/*                                                                               */
/* 1) numeric::fMatrix(f, A, param1, param2, .., hard_soft, ReturnType = d)      */
/*                                                                               */
/* Parameter:                                                                    */
/*   f                  : procedure, f(a, param1, param2, ...) with a in C       */
/*                                   must produce scalar values                  */
/*   A                  : diagonalizable numerical square matrix                 */
/*                        (array or Cat::Matrix)                                 */
/*   param1, param2, .. : numerical or symbolical parameters                     */
/*   hard_soft          : Hard, HardwareFloat, Soft, SoftwareFloats              */
/*   d                  : DOM_ARRAY, Dom::Matrix() or Dom::SparseMatrix()        */
/*                                                                               */
/* Rueckgabewert:                                                                */
/*   f(A, param1, param2,..) = array, Dom::Matrix() or Dom::SparseMatrix()       */
/*                                                                               */
/* Beschreibung:                                                                 */
/*                                                                               */
/*   Die Matrix A wird diagonalisiert: A = X diag(lambda1,lambda2,..) X^(-1).    */
/*   Das Ergebnis wird per                                                       */
/*   f(A,p1,p2,..) = X diag(f(lambda1,p1,p2,..),f(lambda2,p1,p2,..),..) X^(-1)   */
/*   berechnet.                                                                  */
/*   Achtung: dies funktioniert nicht fuer beliebiges A! Nichtdiagonalisierbare  */
/*   Matrizen fuehren zu einer Fehlermeldung, bzw. es kommt zu einer grossen     */
/*   Verstaerkung von Rundungsfehlern, wenn die Nichtdiagonalisierbarkeit        */
/*   numerisch nicht ermittelt werden kann.                                      */
/*   Diese Methode ist aber FAIL safe fuer hermitesche oder schief-hermitesche   */
/*   Matrizen A.                                                                 */
/*                                                                               */
/* Examples:                                                                     */
/*  >> A:= array(1..2,1..2,[[0,1],[-1,0]]):                                      */
/*  >> B:= numeric::fMatrix(sqrt,A);                                             */
/*                                                                               */
/*                   +-                             -+                           */
/*                   |   0.7071067811, 0.7071067811  |                           */
/*                   |                               |                           */
/*                   |  -0.7071067812, 0.7071067811  |                           */
/*                   +-                             -+                           */
/*                                                                               */
/*  >> numeric::fMatrix(_power, B, 2);                                           */
/*                                                                               */
/*    +-                                                                  -+     */
/*    |  3.833233541e-20 - 2.439454887e-19 I, 10.0e-1 - 5.421010862e-20 I  |     */
/*    |                                                                    |     */
/*    |     - 10.0e-1 + 2.710505431e-20 I,          4.065758146e-20        |     */
/*    +-                                                                  -+     */
/*                                                                               */
/*  >> numeric::fMatrix(exp@_mult, A, t);                                        */
/*                                                                               */
/*    +-                                                                    -+   */
/*    |    0.5 exp(-I t) + 0.5 exp(I t),   0.5 I exp(-I t) - 0.5 I exp(I t)  |   */
/*    |                                                                      |   */
/*    |  0.5 I exp(I t) - 0.5 I exp(-I t),   0.5 exp(-I t) + 0.5 exp(I t)    |   */
/*    +-                                                                    -+   */
/*                                                                               */
/*-------------------------------------------------------------------------------*/

numeric::fMatrix:=proc(f,A)
local Args, useHardwareFloats, HardwareFloatsRequested, 
      SoftwareFloatsRequested, request, returnType, scif,
      B, m, n, d, dd, X, dummy, i, j, k;
begin
  if args(0) < 2 then error("not enough arguments") end_if;

  //----------------------------------------------
  // Check the matrix and set default return types
  //----------------------------------------------
  if domtype(A) = DOM_ARRAY then
     [n, m]:= [op(A, [0, 2, 2]), op(A, [0, 3, 2])];
     returnType:= DOM_ARRAY;
  elif A::dom::hasProp(Cat::Matrix)=TRUE then
     [n, m]:= A::dom::matdim(A):
     returnType:= Dom::Matrix();
     if A::dom::constructor = Dom::SparseMatrix then
          returnType:= Dom::SparseMatrix();
     elif A::dom <> Dom::Matrix() then
          A:= (Dom::Matrix())::create(extop(A, 3));
     end_if;
  else
     error("expecting an array or a matrix of category 'Cat::Matrix'"):
  end_if;

  if m <> n then
     error("expecting a square matrix");
  end_if;

  //---------------------------------------------
  // Check the options
  //---------------------------------------------
  // set defaults

  if DIGITS <= 15 then
       useHardwareFloats:= TRUE;
  else useHardwareFloats:= FALSE;
  end_if;
  HardwareFloatsRequested:= FALSE;
  SoftwareFloatsRequested:= FALSE;
  request:= null();

  // Args is the list of further parameters for f
  // and the options for numeric::fMatrix.
  // Search for the options and eliminate them
  // from the list:
  Args:= [args(i) $ i = 3..args(0)];
  for i from nops(Args) downto 1 do
      case Args[i]
      of Hard do
      of HardwareFloats do
                     HardwareFloatsRequested:= TRUE;
                     request:= HardwareFloats;
                     useHardwareFloats:= TRUE:
                     delete Args[i];
                     break;
      of Soft do
      of SoftwareFloats do 
                     SoftwareFloatsRequested:= TRUE;
                     request:= SoftwareFloats;
                     useHardwareFloats:= FALSE:
                     delete Args[i];
                     break;
      otherwise
         if type(Args[i]) = "_equal" and
             lhs(Args[i]) = ReturnType then
              returnType:= rhs(Args[i]);
              if {returnType} minus {DOM_ARRAY, Dom::Matrix(), Dom::SparseMatrix()}
                 <> {} then
                 error("illegal return type ".expr2text(Args[i]).
                       " specified. Choose between DOM_ARRAY, ".
                       "Dom::Matrix() or Dom::SparseMatrix()"):
              end_if;
              delete Args[i];
         end_if;
      end_case;
  end_for:

  //------------------------------------------------------------
  // Trivial case: 1 x 1 matrices
  //------------------------------------------------------------

  if n=1 then 
     A := array(1..1,1..1,[[f(float(A[1,1]), op(Args))]]);
  else

  //------------------------------------------------------------
  // Step 1: compute the diagonalization
  //------------------------------------------------------------

  userinfo(1, "diagonalizing matrix");
  // If request = null(), numeric::eigenvectors decides on 
  // its own whether to use HardwareFloats or SoftwareFloats

  [d, X, dummy]:= numeric::eigenvectors(
              A, request, NoErrors, ReturnType = Dom::Matrix());

  userinfo(1, "the eigenvalues are ".expr2text(d));
  userinfo(1, "applying the function to the eigenvalues:");

  //------------------------------------------------------------
  // Step 2: apply the function to the eigenvalues
  //------------------------------------------------------------

  d:=map(d, f, op(Args)):
  userinfo(1, expr2text(d));

  //------------------------------------------------------------
  // Step 3:  result = X*diag(f(eigenvalues))*X^(-1)
  //------------------------------------------------------------

  //--------------------------------
  // ----- use HardwareFloats ------
  //--------------------------------
  // Shall X*d*X^(-1) be computed by HardwareFloats?

  case useHardwareFloats // use case because we want to use break below
  of TRUE do
     if DIGITS > 15 then
        if HardwareFloatsRequested then
           warning("the current precision goal DIGITS = ".expr2text(DIGITS).
                   " cannot be achieved with the requested 'HardwareFloats'.".
                   " Using 'SoftwareFloats' instead.");
        end_if;
        break;
     end_if;

     if not numeric::startHardwareFloats() then
        userinfo(1, "cannot start the hardware floats interface"):
        if HardwareFloatsRequested then
           warning("cannot start the 'HardwareFloats' interface, ".
                   "using 'SoftwareFloats' instead");
        end_if;
        break;
     end_if;

     userinfo(1, "using hardware floats"):

     // convert d to a row vector. scilab::diag(d)
     // reshapes it to a diagonal matrix.
     dd:= array(1..1, 1..n, [d]):
     scif:= scilab::func(["d", "X"], 
                         ["fA"],
                         ["fA=X*diag(d)*inv(X)"]);
     B:= scif(dd, X, ReturnType = [returnType],  Quiet);

     if has(B, FAIL) then
        userinfo(1, "hardware floats failed"):
        if HardwareFloatsRequested then
            warning("'HardwareFloats' failed, using 'SoftwareFloats' instead"):
        end_if;
        break;
     end_if;

     //------------------------------
     //-------- return --------------
     //------------------------------

     return(B);
  end_case; // of useHardwareFloats = TRUE

  //---------------------------------
  //---- use SoftwareFloats ---------
  //---------------------------------

  userinfo(1, "inverting eigenvector matrix");
  A:= numeric::inverse(X, request, ReturnType = Dom::Matrix()):
  for i from 1 to n do
    (A[i,j]:=d[i]*A[i,j]) $ j=1..n;
  end_for;
  userinfo(1, "building result");
  A:= array(1..n,1..n,[[_plus(X[i,k]*A[k,j]$k=1..n)$j=1..n]$i=1..n]);

  end_if; // if n=1

  //-------------------------
  //------- return ----------
  //-------------------------
  case returnType
  of Dom::SparseMatrix() do
       return(returnType(A));
  of Dom::Matrix() do
       return(returnType::create(A));
  of DOM_ARRAY do
       return(A);
  otherwise
       error("unexpected return type"):
  end_case;

end_proc:
