
//------   file  factorCh.mu   29.9.97  ---------------------
/*  Help file:

 factorCholesky - numerical Cholesky factorization of a matrix

 Call:

 numeric::factorCholesky( A <, hard_soft> <,Symmetric> <, Symbolic>
                          <, NoCheck> <, ReturnType = typ )
                                                                   
 Parameter:

  A         -- a square real or complex matrix of domtype DOM_ARRAY
               or Cat::Matrix
  hard_soft -- either 'Hard','HardwareFloats' or 'Soft','SoftwareFloats'
               If DIGITS <=15, the default is HardwareFloats
               If DIGITS  >15, the default is SoftwareFloats

  typ       -- either DOM_ARRAY or Dom::Matrix() or Dom::SparseMatrix()
                                                                   
Synopsis:

  numeric::factorCholesky(A); returns the lower triangular
  Cholesky factor L of the factorization A=LL^H, where L^H 
  denotes the Hermitean transpose of L. The result is of 
  domtype DOM_ARRAY. It is normalized such that the diagonal
  of L is real and positive.

  numeric::factorCholesky(A, Symmetric); returns the lower triangular
  Cholesky factor L of the factorization A=LL^T, where L^T 
  denotes the usual transpose of L. The result is of 
  domtype DOM_ARRAY. 
                                                                  
  Non-float numerical entries are admissible, they will be 
  evaluated to floats. Conversion to floats may be suppressed
  by using the option Symbolic.
 
  If the matrix is not Hermitean or not positive definite 
  and the option <Symmetric> is not used, then
  the computation will terminate with an error message.
  Checking of these properties can be suppressed with the option
  NoCheck. This option should be used, whenever the matrix contains 
  indeterminates. If numeric symbolic entries such as PI, sqrt(2) etc.
  are present, then these properties are checked numerically.

  Examples: 

>> A := array( 1..2 , 1..2 , [ [1,2*I],[-2*I,5] ] ):

>> numeric::factorCholesky(A);                                

                            +-              -+
                            |    1.0,    0   |
                            |                |
                            |  - 2.0 I, 1.0  |
                            +-              -+

>> numeric::factorCholesky(A,Symbolic);

                              +-          -+
                              |    1,   0  |
                              |            |
                              |  - 2 I, 1  |
                              +-          -+

>> A := array( 1..2 , 1..2 , [ [-1 ,1],[1,1] ] ):

>> numeric::factorCholesky(A);                                

Error: matrix is not positive definite within working precision [nume\
ric::factorCholesky]

>> numeric::factorCholesky(A, Symmetric);                                

                     +-                        -+
                     |   I,          0          |
                     |                          |
                     |  - I, 1.414213562 + 0 I  |
                     +-                        -+

See also: linalg::factorCholesky, numeric::factorLU , numeric::factorQR 
                                                                  */
//-----------------------------------------------------------------
numeric::factorCholesky:=proc(A)
local useHardwareFloats, HardwareFloatsRequested, returnType, f,
      symbolic_mode, symmetricDecomp, warning_given, do_check,
      m, n, real,i,j,k,B;
begin
  if args(0) < 1 then error("expecting at least one argument") end_if;

  //----------------------------------------------
  // Check the matrix and set default return types
  //----------------------------------------------
  if domtype(args(1)) = DOM_ARRAY then
     [n, m]:= [op(A, [0, 2, 2]), op(A, [0, 3, 2])];
     returnType:= DOM_ARRAY;
  elif A::dom::hasProp(Cat::Matrix)=TRUE then
     [n, m]:= A::dom::matdim(A):
     returnType:= Dom::Matrix();
     if A::dom::constructor = Dom::SparseMatrix then
          returnType:= Dom::SparseMatrix();
     elif A::dom <> Dom::Matrix() then
          A:= (Dom::Matrix())::create(extop(A, 3));
     end_if;
  else
     error("expecting an array or a matrix of category 'Cat::Matrix'"):
  end_if;

  if m <> n then
     error("expecting a square matrix");
  end_if;

  //---------------------------------------------
  // Check the options
  //---------------------------------------------
  // set defaults
  if DIGITS <= 15 then
       useHardwareFloats:= TRUE;
  else useHardwareFloats:= FALSE;
  end_if;
  HardwareFloatsRequested:= FALSE;
  do_check:=TRUE; 
  symbolic_mode:=FALSE; 
  symmetricDecomp:= FALSE;

  for i from 2 to args(0) do
      case args(i)
      of Symbolic do symbolic_mode := TRUE;
                     useHardwareFloats:= FALSE;
                     break;
      of Hard do
      of HardwareFloats do
                     if symbolic_mode then
                        warning("ignoring the option 'HardwareFloats' ".
                                "because of the option 'Symbolic'"):
                        useHardwareFloats:= FALSE:
                     else
                        useHardwareFloats:= TRUE:
                        HardwareFloatsRequested:= TRUE;
                     end_if;
                     break;
      of Soft do
      of SoftwareFloats do
                     useHardwareFloats:= FALSE:
                     break;
      of Symmetric do
                     symmetricDecomp:= TRUE;  
                     break;
      of NoCheck do do_check := FALSE; 
                     break;
      otherwise
         if type(args(i)) = "_equal" and
            lhs(args(i)) = ReturnType then
              returnType:= rhs(args(i));
              if {returnType} minus {DOM_ARRAY, Dom::Matrix(), Dom::SparseMatrix()}
                 <> {} then
                 error("illegal return type ".expr2text(args(i)).
                       " specified. Choose between DOM_ARRAY, ".
                       "Dom::Matrix() or Dom::SparseMatrix()"):
              end_if;
         else error("unknown option ".expr2text(args(i)));
         end_if;
      end_case;
  end_for:

  //--------------------------------
  // ----- use HardwareFloats ------
  //--------------------------------
  case useHardwareFloats // use case because we want to use break below
  of TRUE do
     if DIGITS > 15 then
        if HardwareFloatsRequested then
           warning("the current precision goal DIGITS = ".expr2text(DIGITS).
                   " cannot be achieved with the requested 'HardwareFloats'.".
                   " Using 'SoftwareFloats' instead.");
        end_if;
        break;
     end_if;

     if not numeric::startHardwareFloats() then
        userinfo(1, "cannot start the hardware floats interface"):
        if HardwareFloatsRequested then
           warning("cannot start the 'HardwareFloats' interface, ".
                   "using 'SoftwareFloats' instead");
        end_if;
        break;
     end_if;

     userinfo(1, "using hardware floats"):

     //-----------------------------------------------------
     // scilab uses the data in the upper triangular part of
     // the matrix, MuPAD uses the lower triangular part.
     // Pass the transpose of A to scilab::chol.
     // Scilab computes the upper triangular Cholesky factor,
     // MuPAD wants the lower triangular one. Hence,
     // Return the Hermitean transpose of the scilab result:
     //-----------------------------------------------------
     if A::dom::constructor = Dom::SparseMatrix then
        f:= scilab::func(["A", "dummy"], 
                         ["L"], 
                         ["L = chol(full(A)')';"]);
     else
        f:= scilab::func(["A", "dummy"], 
                         ["L"], 
                         ["L = chol(A')';"]);
     end_if;
     B:= f(A, 1, ReturnType = [returnType],  Quiet);

     if has(B, FAIL) then
        userinfo(1, "hardware floats failed"):
        if HardwareFloatsRequested then
            warning("'HardwareFloats' failed, using 'SoftwareFloats' instead"):
        end_if;
        break;
     end_if;

     //------------------------------
     //-------- return --------------
     //------------------------------

     return(B);
  end_case; // of useHardwareFloats = TRUE

  //---------------------------------
  //---- use SoftwareFloats ---------
  //---------------------------------

  userinfo(1, "using software floats"):

  // convert to domtype DOM_ARRAY for speed 
  if A::dom::hasProp(Cat::Matrix)=TRUE then 
     A:= expr(A) 
  end_if; 

  if domtype(A) <> DOM_ARRAY then
     error("the matrix must be an array or of category 'Cat::Matrix'");
  end_if:

  if symbolic_mode and do_check and
       (  (indets(A)<>{}) or has({op(map(A,domtype))}, DOM_NIL) )
  then do_check:=FALSE; 
       warning("symbolic indeterminate found, switching to option 'NoCheck'");
  end_if;

  if not symbolic_mode then
    B:=map(A,float);
    if ({op(map(B,domtype))} minus {DOM_FLOAT,DOM_COMPLEX}) <> {}
      then symbolic_mode:=TRUE;
           do_check:=FALSE;
           warning("symbolic entry found, switching to options 'Symbolic', 'NoCheck'");
      else A:=B;
    end_if:
  end_if;
  delete B;
  A:=subs(A);

  if symmetricDecomp
  then 
     for i from 1 to n do
       for j from 1 to i-1 do
          if A[i,j]<>A[j,i] then error("the matrix is not symmetric") end_if;
       end_for;
     end_for;
     for i from 1 to n do
       (A[i,j]:=(A[i,j]-_plus(A[i,k]*A[j,k] $ k=1..j-1))/A[j,j]) $ j=1..i-1;
        A[i,i]:=A[i,i]-_plus(A[i,k]^2 $ k=1..i-1);
        A[i,i]:=(A[i,i])^(1/2);
       if symbolic_mode then
            (A[i,j]:= 0) $ j=i+1..n;
       else (A[i,j]:= float(0)) $ j=i+1..n;
       end_if;
        if i<n and iszero(A[i,i]) then 
            error("a regular symmetric factorization does not exist") 
        end_if;
     end_for;
  else 
     real:=TRUE;
     if do_check then
       for i from 1 to n do
         for j from 1 to i-1 do
           if iszero(Im(A[i,j]))
              then if A[i,j]<>A[j,i] then error("the matrix is not hermitean") end_if;
              else if A[i,j]<>conjugate(A[j,i]) then error("the matrix is not hermitean") end_if:
                   real:= FALSE;
           end_if;
         end_for;
         if not iszero(Im(A[i,i]))then error("the matrix is not hermitean"); end_if;
       end_for;
     end_if;
     warning_given:= FALSE;

     for i from 1 to n do
       if real
       then (A[i,j]:= (A[i,j]-_plus(A[i,k]*A[j,k] $ k=1..j-1))/A[j,j])
                          $ j=1..i-1;
       else (A[i,j]:= (A[i,j]-_plus(A[i,k]*conjugate(A[j,k]) $ k=1..j-1)
                      )/A[j,j]) $ j=1..i-1;
       end_if;
       if symbolic_mode
         then A[i,i]:= A[i,i] - _plus(abs(A[i,k])^2 $ k=1..i-1);
         else A[i,i]:= A[i,i] - _plus(specfunc::abs(A[i,k])^2 $ k=1..i-1);
       end_if;
       if do_check then
         if traperror(bool(A[i,i] <= 0))<>0 
         then if not warning_given then
                warning("positiveness checked numerically");
                warning_given:=TRUE;
              end_if;
              if float(A[i,i]) <= 0 then
                 error("the matrix is not positive definite within working precision");
              end_if;
         else if A[i,i] <= 0 then
                 if symbolic_mode
                  then error("the matrix is not positive definite");
                  else error("the matrix is not positive definite within working precision");
                 end_if;
              end_if;
         end_if;
       end_if;
       A[i,i]:=(A[i,i])^(1/2);
       if symbolic_mode then
            (A[i,j]:= 0) $ j=i+1..n;
       else (A[i,j]:= float(0)) $ j=i+1..n;
       end_if;
     end_for;
  end_if;

  //-------------------------
  //------- return ----------
  //-------------------------
  case returnType
  of Dom::SparseMatrix() do
       return(returnType(A));
  of Dom::Matrix() do
       return(returnType::create(A));
  of DOM_ARRAY do
       return(A);
  otherwise
       error("unexpected return type"):
  end_case;

end_proc:

