
//------   file  factorLU.mu  ---------------------
/*  Help file:

 factorLU - LU factorization of a matrix

 Call:

 numeric::factorLU( A <, Symbolic>  <, hard_soft> <, ReturnType = typ )
 )
                                                                   
 Parameter:

  A         --  an m x n matrix (DOM_ARRAY or category Cat::Matrix)
  hard_soft -- either 'Hard','HardwareFloats' or 'Soft','SoftwareFloats'
               If DIGITS <=15, the default is HardwareFloats
               If DIGITS  >15, the default is SoftwareFloats

  typ       -- either DOM_ARRAY or Dom::Matrix() or Dom::SparseMatrix()

Synopsis:

  The LU factorization of a real/complex matrix is PA=LU, where
  L is lower triangular, normalized to 1 along the diagonal,
  and R is upper triangular.  P is a permutation matrix 
  corresponding to row exchanges.

  numeric::factorLU(A) returns the list [L,U,p], where L and R are
  of domtype DOM_ARRAY. The list of integers p=[ p[1],p[2], .. ]
  represents the row exchanges in pivoting steps, i.e., B = PA = LU, 
  where B[i,j] = A[p[i],j]. Multiplication y=Px of the permutation
  matrix P with a vector x is realized by y[i]:=x[p[i]].
                                                                  
  Non-float numerical entries are allowed in the matrix, they
  will be evaluated to floats in the course of the computation.
  Non-numeric symbolic entries will lead to an error, unless
  the option Symbolic is specified.

  The option Symbolic prevents conversion of the matrix entries
  to floats. No row exchanges will be performed unless necessary.

  Examples: 

>> A:=array(1..3,1..3,[[1,2,3],[2,4,6],[4,8,9]]):
>> ([L,U,p]):=numeric::factorLU(A);

          -- +-            -+  +-                -+            --
          |  |    1,  0, 0  |  |  4.0, 8.0,  9.0  |             |
          |  |              |  |                  |             |
          |  |   0.5, 1, 0  |, |   0,   0,   1.5  |, [3, 2, 1]  |
          |  |              |  |                  |             |
          |  |  0.25, 0, 1  |  |   0,   0,  0.75  |             |
          -- +-            -+  +-                -+            --

>> A:= array(1..3,1..2,[[1,2],[2,3],[3,4]]):
>> ([L,U,p]):=numeric::factorLU(A);

 -- +-                      -+  +-                   -+            --
 |  |        1,       0,  0  |  |  3.0,      4.0      |             |
 |  |                        |  |                     |             |
 |  |  0.3333333333,  1,  0  |, |   0,  0.6666666666  |, [3, 1, 2]  |
 |  |                        |  |                     |             |
 |  |  0.6666666666, 0.5, 1  |  |   0,        0       |             |
 -- +-                      -+  +-                   -+            --

See also:  linalg::factorLU, numeric::factorQR, numeric::factorCholesky
                                                                  */
//-----------------------------------------------------------------
numeric::factorLU:=proc(A)
local  useHardwareFloats, HardwareFloatsRequested, 
       returnType, result, f,
       m, n, method, B, i, j, k, tmp, pivot, pivindex, rownorm, L, p;
begin

   if args(0) < 1 then error("expecting at least one argument") end_if;

  //----------------------------------------------
  // Check the matrix and set default return types
  //----------------------------------------------
  if domtype(args(1)) = DOM_ARRAY then
     [m, n]:= [op(A, [0, 2, 2]), op(A, [0, 3, 2])];
     returnType:= DOM_ARRAY;
  elif A::dom::hasProp(Cat::Matrix)=TRUE then
     [m, n]:= A::dom::matdim(A):
     returnType:= Dom::Matrix();
     if A::dom::constructor = Dom::SparseMatrix then
          returnType:= Dom::SparseMatrix();
     elif A::dom <> Dom::Matrix() then
          A:= (Dom::Matrix())::create(extop(A, 3));
     end_if;
  else
     error("expecting an array or a matrix of category 'Cat::Matrix'"):
  end_if;

  //---------------------------------------------
  // Check the options
  //---------------------------------------------
  // set defaults
  method:= 2;        // floating point method
  if DIGITS <= 15 then
       useHardwareFloats:= TRUE;
  else useHardwareFloats:= FALSE;
  end_if;
  HardwareFloatsRequested:= FALSE;

  for i from 2 to args(0) do
      case args(i)
      of Symbolic do method:= 1:
                     useHardwareFloats:= FALSE;
                     break;
      of Hard do
      of HardwareFloats do
                     if method = 1 then
                        warning("ignoring the option 'HardwareFloats' ".
                                "because of the option 'Symbolic'"):
                        useHardwareFloats:= FALSE:
                     else
                        useHardwareFloats:= TRUE:
                        HardwareFloatsRequested:= TRUE;
                     end_if;
                     break;
      of Soft do
      of SoftwareFloats do
                     useHardwareFloats:= FALSE:
                     break;
      otherwise
         if type(args(i)) = "_equal" and
            lhs(args(i)) = ReturnType then
              returnType:= rhs(args(i));
              if {returnType} minus {DOM_ARRAY, Dom::Matrix(), Dom::SparseMatrix()}
                 <> {} then
                 error("illegal return type ".expr2text(args(i)).
                       " specified. Choose between DOM_ARRAY, ".
                       "Dom::Matrix() or Dom::SparseMatrix()"):
              end_if;
         else error("unknown option ".expr2text(args(i)));
         end_if;
      end_case;
  end_for:

  //--------------------------------
  // ----- use HardwareFloats ------
  //--------------------------------
  case useHardwareFloats // use case because we want to use break below
  of TRUE do
     if DIGITS > 15 then
        if HardwareFloatsRequested then
           warning("the current precision goal DIGITS = ".expr2text(DIGITS).
                   " cannot be achieved with the requested 'HardwareFloats'.".
                   " Using 'SoftwareFloats' instead.");
        end_if;
        break;
     end_if;

     if m <> n then
        if HardwareFloatsRequested then
           warning("with 'HardwareFloats', a LU factorization can be computed ".
                   "only for square matrices. Using 'SoftwareFloats' instead'.");
        end_if;
        break;
     end_if;

     if not numeric::startHardwareFloats() then
        userinfo(1, "cannot start the hardware floats interface"):
        if HardwareFloatsRequested then
           warning("cannot start the 'HardwareFloats' interface, ".
                   "using 'SoftwareFloats' instead");
        end_if;
        break;
     end_if;

     userinfo(1, "using hardware floats"):

     //-----------------------------------------------------
     //-----------------------------------------------------
     if A::dom::constructor = Dom::SparseMatrix then
        f:= scilab::func(["A", "dummy"],
                         ["L", "U", "p"],
                         ["[L, U, P] = lu(full(A));",
                          "[p, v, mn] = spget(sparse(P));"
                         ]);
     else
        f:= scilab::func(["A", "dummy"],
                         ["L", "U", "p"],
                         ["[L, U, P] = lu(A);",
                          "[p, v, mn] = spget(sparse(P));"
                         ]);
     end_if;
     result:= [f(A, 1, ReturnType = [returnType, returnType, Dom::Matrix()],  Quiet)];

     if has(result, FAIL) then
        userinfo(1, "hardware floats failed"):
        if HardwareFloatsRequested then
            warning("'HardwareFloats' failed, using 'SoftwareFloats' instead"):
        end_if;
        break;
     end_if;

     // the permutation matrix result[3] was returned as a
     // Dom::Matrix() with two columns. Convert to a list:
     result[3]:= map(result[3], round);
     p:= [0 $ m];
     for i from 1 to m do
        p[result[3][i, 1]]:= result[3][i, 2];
     end_for;
     result[3]:= p;

     //------------------------------
     //-------- return --------------
     //------------------------------

     return(result);

  end_case; // of useHardwareFloats = TRUE

  //---------------------------------
  //---- use SoftwareFloats ---------
  //---------------------------------

   userinfo(1, "using software floats"):

   // convert to domtype DOM_ARRAY for speed 
   if A::dom::hasProp(Cat::Matrix)=TRUE then 
      A:= expr(A) 
   end_if; 
   if domtype(A) <> DOM_ARRAY then
      error("the matrix must be an array or of category 'Cat::Matrix'");
   end_if:
   if op(A,[0,1])<>2 then error("the argument is not a matrix");end_if;

   if method=2 then            
      B:=map(A, float):
      if {op(map(B,domtype))} minus {DOM_FLOAT, DOM_COMPLEX} <> {} then 
         warning("symbolic entry found, switching to option Symbolic");
         method:=1;
      else A:=B;
      end_if:
   end_if;

   p:=[i $ i=1..m]: // initialize list to store row permutations 

   if method=1 then
      userinfo(1,"symbolic mode, avoiding pivoting if possible ");
   end_if;
   if method=2 then            
      userinfo(1,"using column pivoting");
   end_if;

   for j from 1 to min(m-1,n) do
     if method=1 then
        for pivindex from j to m do
           if not iszero(A[pivindex,j]) then break; end_if;
        end_for;
        if pivindex > m then next end_if; // no nontrivial pivot element 
     end_if;
     if method=2 then
        pivot:=0;
        pivindex:=j;
        for i from j to m do 
            if iszero(A[i,j]) then next end_if;
            if j=n
            then if (tmp:= specfunc::abs(A[i,j]))>pivot
                    then pivot:= tmp; pivindex:= i;
                 end_if;
            else rownorm:=max(specfunc::abs(A[i,k]) $ k=j+1..n );
                 if not iszero(rownorm)
                 then specfunc::abs(A[i,j])/rownorm;
                      if pivot<last(1) then pivot:=last(1); pivindex:=i; end_if;
                 else pivindex:= i; break;
                 end_if;
            end_if;
        end_for;
     end_if;

     if pivindex<>j then
       (tmp:=A[j,k]; A[j,k]:=A[pivindex,k];A[pivindex,k]:=tmp;) $ k=1..n;
       tmp:=p[j];p[j]:=p[pivindex];p[pivindex]:=tmp;
     end_if;
     if not iszero(A[j,j]) then
        for i from j+1 to m do
            A[i,j]:=A[i,j]/A[j,j]:
            if not iszero(A[i,j]) then
              (A[i,k]:=A[i,k]-A[i,j]*A[j,k]) $ k=j+1..n;
            end_if;
        end_for;
     end_if;
   end_for;

   /* LU data are stored in lower and upper part of A. Extract L and U: */
   if method = 1 then
        L:=array(1..m,1..m,[[0$m]$m]); 
        (L[i,i]:= 1) $ i=1..m;
   else L:=array(1..m,1..m,[[float(0)$m]$m]); 
        (L[i,i]:= float(1)) $ i=1..m;
   end_if;
   
   for j from 1 to min(m-1, n) do
     (L[i,j]:= A[i,j]) $ i=j+1..m;
   end_for:
   for j from 1 to n do
      if method = 1 then
           (A[i,j]:= 0) $ i=j+1..m 
      else (A[i,j]:= float(0)) $ i=j+1..m 
      end_if;
   end_for:


  //-------------------------
  //------- return ----------
  //-------------------------
  case returnType
  of Dom::SparseMatrix() do
       return([returnType(L), returnType(A), p]);
  of Dom::Matrix() do
       return([returnType::create(L), returnType::create(A), p]);
  of DOM_ARRAY do
       return([L, A, p]);
  otherwise
       error("unexpected return type"):
  end_case;
end_proc:
