
//------   file  inverse.mu   26.9.97         ---------------------
//-----------------------------------------------------------------
/*  Help file:

 inverse - inverse of a numerical matrix

 Call:

 numeric::inverse( A <, Symbolic> <, hard_soft> <, ReturnType = typ>)
                                                                   
 Parameter:

  A         --  a  square matrix (DOM_ARRAY or category Cat::Matrix)
  hard_soft -- either 'Hard','HardwareFloats' or 'Soft','SoftwareFloats'
               If DIGITS <=15, the default is HardwareFloats
               If DIGITS  >15, the default is SoftwareFloats
  typ       -- either DOM_ARRAY or Dom::Matrix() or Dom::SparseMatrix()
  NoWarning -- suppress warning when falling back to the symbolic
               mode because of symbolic objects

Returns: the inverse of A as an object of type 'typ'.
         If typ is not specified:
         the inverse of an array A is returned as an array. 
         The inverse of a Dom::SparseMatrix(any Ring)
         is returned as a Dom::SparseMatrix().
         The inverse of other matrix objects (Dom::Matrix(any Ring),
         Dom::SquareMatrix(..), Dom::MatrixGroup(..)
         are returned as Dom::Matrix()-objects.
                                                                   
Synopsis:

  numeric::inverse(A); returns the inverse of A computed by
  Gauss elimination with partial pivoting. The result type
  depends on the type of the input matrix A or the
  requested return type.
                                                                  
  Non-float numerical entries are admissible, they will be 
  evaluated to floats in the course of the computation.

  The option Symbolic prevents conversion of the matrix entries
  to floats. No pivoting will occur unless necessary. With this
  option non-numeric symbolic entries are admissible. However, 
  their presence will slow down the computation drastically.

  Examples: 

>> A := array( 1..2 , 1..2 , [ [1,2],[3,4] ] ):

>> numeric::inverse(A);
                             +-            -+
                             |  -2.0,  1.0  |
                             |              |
                             |   1.5, -0.5  |
                             +-            -+

>> numeric::inverse(A,Symbolic);

                              +-           -+
                              |   -2,   1   |
                              |             |
                              |  3/2, -1/2  |
                              +-           -+
                                                                  */
//-----------------------------------------------------------------
numeric::inverse:=proc(A)
local B, useHardwareFloats, HardwareFloatsRequested, f,
      m, n, method, nowarning,
      returnType, use_normal,i,j,k,tmp,pivot,pivindex,
      rownorm, density,nontriv,pA,pB;
begin
  if args(0) <1 then error("expecting at least one argument") end_if;

  //----------------------------------------------
  // Check the matrix and set default return types
  //----------------------------------------------
  if domtype(args(1)) = DOM_ARRAY then
     [n, m]:= [op(A, [0, 2, 2]), op(A, [0, 3, 2])];
     returnType:= DOM_ARRAY;
  elif A::dom::hasProp(Cat::Matrix)=TRUE then
     [n, m]:= A::dom::matdim(A):
     returnType:= Dom::Matrix();
     if A::dom::constructor = Dom::SparseMatrix then
          returnType:= Dom::SparseMatrix();
     elif A::dom <> Dom::Matrix() then
          A:= (Dom::Matrix())::create(extop(A, 3));
     end_if;
  else
     error("expecting an array or a matrix of category 'Cat::Matrix'"):
  end_if;

  if m <> n then 
     error("expecting a square matrix"); 
  end_if;

  //---------------------------------------------
  // Check the options
  //---------------------------------------------
  // set defaults
  method:= 2;        // floating point method
  if DIGITS <= 15 then
       useHardwareFloats:= TRUE; 
  else useHardwareFloats:= FALSE; 
  end_if;
  HardwareFloatsRequested:= FALSE;
  nowarning:= FALSE;

  for i from 2 to args(0) do
      case args(i)
      of Symbolic do method:= 1:
                     useHardwareFloats:= FALSE;
                     break;
      of Hard do
      of HardwareFloats do
                     HardwareFloatsRequested:= TRUE;
                     if method = 1 then
                        warning("ignoring the option 'HardwareFloats' ".
                                "because of the option 'Symbolic'"):
                        method:= 1;
                        useHardwareFloats:= FALSE: 
                     else
                        useHardwareFloats:= TRUE: 
                     end_if;
                     break;
      of Soft do
      of SoftwareFloats do  
                     useHardwareFloats:= FALSE:
                     break;
      of NoWarning do
                     nowarning:= TRUE;
                     break;
      otherwise
         if type(args(i)) = "_equal" and
            lhs(args(i)) = ReturnType then
              returnType:= rhs(args(i));
              if {returnType} minus {DOM_ARRAY, Dom::Matrix(), Dom::SparseMatrix()} 
                 <> {} then
                 error("illegal return type ".expr2text(args(i)). 
                       " specified. Choose between DOM_ARRAY, ".
                       "Dom::Matrix() or Dom::SparseMatrix()"):
              end_if;
         else error("unknown option ".expr2text(args(i)));
         end_if;
      end_case;
  end_for:

  //--------------------------------
  // ----- use HardwareFloats ------
  //--------------------------------
  case useHardwareFloats // use case because we want to use break below
  of TRUE do
     if DIGITS > 15 then
        if HardwareFloatsRequested then
           warning("the current precision goal DIGITS = ".expr2text(DIGITS).
                   " cannot be achieved with the requested 'HardwareFloats'.".
                   " Using 'SoftwareFloats' instead.");
        end_if;
        break;
     end_if;

     if not numeric::startHardwareFloats() then
        userinfo(1, "cannot start the hardware floats interface"):
        if HardwareFloatsRequested then
           warning("cannot start the 'HardwareFloats' interface, ".
                   "using 'SoftwareFloats' instead");
        end_if;
        break;
     end_if;

     userinfo(1, "using hardware floats"):

     //------------------------------------------------------
     // Problem: scilab::svd does not accept sparse matrices.
     // Workaround: we send a program to scilab that converts
     // the sparse matrix to a full matrix and then computes
     // the singular values:
     //------------------------------------------------------

     if A::dom::constructor = Dom::SparseMatrix then
        f:= scilab::func(["x", "dummy"], ["y"], ["y=inv(full(x))"]);
        B:= f(A, 1, ReturnType = [returnType],  Quiet);
     else
        B:= scilab::inv(A, ReturnType = returnType, Quiet);
     end_if;

     if has(B, FAIL) then
        userinfo(1, "hardware floats failed"):
        if HardwareFloatsRequested then
            warning("'HardwareFloats' failed, using 'SoftwareFloats' instead"):
        end_if;
        break;
     end_if;

     //------------------------------
     //-------- return --------------
     //------------------------------
     
     return(B);
  end_case; // of useHardwareFloats = TRUE

  //---------------------------------
  //---- use SoftwareFloats ---------
  //---------------------------------

  userinfo(1, "using software floats"):
  // convert to domtype DOM_ARRAY for speed 
  if A::dom::hasProp(Cat::Matrix)=TRUE then  
     A:= expr(A);
  end_if;
  if domtype(A) <> DOM_ARRAY then
     error("the matrix must be an array or of category 'Cat::Matrix'");
  end_if:

  if method = 2 then
    B:= map(A, float):
    if {op(map(B, domtype))} minus {DOM_FLOAT, DOM_COMPLEX} <> {} then
      if nowarning <> TRUE then
         warning("symbolic entry found, switching to symbolic mode");
      end_if;
      method := 1;
      use_normal := TRUE
    else 
      A:= B;
    end_if;
  end_if;

  use_normal:=FALSE;
  if method=1 then
    if (indets(A)<>{}) 
    or has({op(map(A,domtype))},DOM_NIL)
    or has({op(map(A,domtype))},DOM_EXPR) then 
       use_normal:=TRUE
    end_if;
  end_if;

  if method=1 then
    userinfo(1,"symbolic mode, no pivoting unless necessary");
  end_if;
  if method=2 then 
    userinfo(1,"using column pivoting"); 
  end_if;

  B:=array(1..n,1..n,[[0$n]$n]); // initialize inverse as 
  (B[i,i]:=1) $ i=1..n;    // identity matrix       

  density:=0; // compute how many of the n(n+1)/2 elements in upper    
              // factor U of the LU decomposition of A are non-trivial 
  for j from 1 to n-1 do
  // search for pivot element 
    if method=1 then
      if iszero(A[j,j]) then
         for pivindex from j+1 to n-1 do
             if not iszero(A[pivindex,j]) then break; end_if;
         end_for;
      else pivindex:=j;
      end_if;
    end_if;
    if method=2 then
      pivot:=0; pivindex:=j;
      for i from j to n do 
          rownorm:=max(specfunc::abs(A[i,k]) $ k=j+1..n );
          if not iszero(rownorm) then
             specfunc::abs(A[i,j])/rownorm;
             if pivot<last(1) then pivot:=last(1); pivindex:=i; end_if;
          else pivindex:=i;break;
          end_if;
      end_for;
    end_if;
    if pivindex<>j then
      (tmp:=A[j,k];A[j,k]:=A[pivindex,k];A[pivindex,k]:=tmp;) $ k=1..n;
      (tmp:=B[j,k];B[j,k]:=B[pivindex,k];B[pivindex,k]:=tmp;) $ k=1..n;
    end_if;
    if iszero(A[j,j]) then return(FAIL) end_if;

    // check sparsity of j.th row to reduce cost of elimination step 
    // sparsity of A: 
    pA:=[0$n-j]; // list to store indices of non-trivial entries of jth row in A 
    nontriv:=1;
    for k from j+1 to n do 
        if not iszero(A[j,k]) then pA[nontriv]:=k;nontriv:=nontriv+1;end_if;
    end_for;
    density:=density+nontriv-1;
    // sparsity of B: 
    nontriv:=1;
    pB:=[0$n]; // list to store indices of non-trivial entries of jth row in B 
    for k from 1 to n do 
        if not iszero(B[j,k]) then pB[nontriv]:=k;nontriv:=nontriv+1;end_if;
    end_for;

    // Now do the elimination. 
    // Only need to touch elements in rows pA[1],pA[2],.. and pB[1],pB[2],.. 
    for i from j+1 to n do
      if iszero(A[i,j]) then next end_if;
      tmp:=A[i,j]/A[j,j];
      if use_normal then 
         for k from 1 to n-j do
           if pA[k]=0 then break else A[i,pA[k]]:=normal(A[i,pA[k]]-tmp*A[j,pA[k]]);end_if;
         end_for;
         for k from 1 to n do
           if pB[k]=0 then break else B[i,pB[k]]:=normal(B[i,pB[k]]-tmp*B[j,pB[k]]);end_if;
         end_for;
      else 
         for k from 1 to n-j do
           if pA[k]=0 then break else A[i,pA[k]]:=A[i,pA[k]]-tmp*A[j,pA[k]];end_if;
         end_for;
         for k from 1 to n do
           if pB[k]=0 then break else B[i,pB[k]]:=B[i,pB[k]]-tmp*B[j,pB[k]];end_if;
         end_for;
      end_if;
    end_for;
  end_for;
  if iszero(A[n,n]) then 
     return(FAIL)
  end_if;

  // Now A is upper triangular. Backsubstitution: 
  if density> 0.4*n^2
  then // original version, does not make use of sparsity in upper triangular A 
       for j from 1 to n do
         for i from n downto 1 do
           B[i,j]:=(B[i,j]-_plus(A[i,k]*B[k,j] $ k=i+1..n))/A[i,i] ;
           if use_normal then B[i,j]:=normal(B[i,j]) end_if;
         end_for;
       end_for;
  else // try to use sparsity 
       for i from n downto 1 do
         for k from i+1 to n do
           if not iszero(A[i,k]) then  
              (B[i,j]:=B[i,j]-A[i,k]*B[k,j];) $ j=1..n;
           end_if;
         end_for;
         (B[i,j]:=B[i,j]/A[i,i];) $ j=1..n;
         if use_normal then (B[i,j]:=normal(B[i,j]);)$ j=1..n; end_if;
       end_for;
  end_if;

  //-------------------------
  //------- return ----------
  //-------------------------
  case returnType
  of Dom::SparseMatrix() do
       return(returnType(B));
  of Dom::Matrix() do
       return(returnType::create(B));
  of DOM_ARRAY do
       return(B);
  otherwise
       error("unexpected return type"):
  end_case;
end_proc:
