// $Date: 2001/12/09 14:13:17 $  $Author: walter $   $Revision: 1.23 $ 
// 
// W.Oevel, 01/04/98 

//------   file  matlinsolve.mu   1.4.98  ------------------------
//-------------------  W. Oevel  ---------------------------------
/*  Help file:

 matlinsolve - solve a system of linear equations numerically

 Call:

 numeric::matlinsolve(A, B 
                      <, hard_soft> 
                      <, Symbolic>, 
                      <, ShowAssumptions>
                      <, NoWarning>
                      <, ReturnType = d>
                     )
                                                                   
 Parameter:

  A          --  an m x n matrix (DOM_ARRAY or category Cat::Matrix)
  B          --  an m x p matrix (DOM_ARRAY or category Cat::Matrix)
                 For p=1 the vector B may also be a 1-dimensional array 
                 or a list of length m.
  hard_soft  --  Soft, SoftwareFloats, Hard, HardwareFloats
  Symbolic   --  do not convert input data to floats
  NoWarning  --  no warning is given if matlinsolve switches to
                 the 'symbolic mode' due to symbolic parameters
  d          --  return the result as matrices of domain type d,
                 with d = DOM_ARRAY, Dom::Matrix(), Dom::SparseMatrix()
                                                                   
Synopsis:

  numeric::matlinsolve(A,B,<Symbolic>); 
      returns the list [X,Kernelbasis].

  numeric::matlinsolve(A,B,<Symbolic>, ShowAssumptions);
      returns the list [X,Kernelbasis,Constraints, PivotAssumptions].

  (*) X is a special solution of the linear system A*X=B. It is
      an n x p array.
  (*) Kernelbasis represents the general solution of the homogeneous
      linear system A*X=0. It is an n x d array,
      where d is the dimension of the kernel of A. The columns of 
      Kernelbasis form a basis of the kernel. If the kernel is empty,
      then the returned value of Kernelbasis is the null vector,
      represented by array(1..n,1..1,[[0]$n]).
  (*) Constraints is a list of equations for symbolic parameters in A and B,
      which are necessary and sufficient to grant solvability of 
      the linear system A*X=B. If no such contraints arise, then the
      returned value of Constraints is the empty list.
  (*) PivotAssumptions is a list of inequalities [piv1 <>0, piv2<>0, ...].
      These assumptions were implicitly used by matlinsolve to compute
      the solution and the kernel.

  By default, all entries in A and B are converted to floats. Non-numeric
  symbolic entries in A are not admissible, unless the option Symbolic
  is used. In the symbolic mode no conversion to floats will occur. We note 
  that the presence of symbolic parameters in the matrix A will slow down 
  the computation drastically.

  Due to roundoff errors, some or all basis vectors in the kernel of A may be
  missed in the numeric mode.

  matlinsolve returns [FAIL,NIL], if no solution of A*X=B exists. If B contains
  symbolic parameters, then Constraints is the set of linear equations
  for the entries of B needed to grant solvability.

  matlinsolve uses Gaussian Elimination with row pivoting. Internally
  a sparse representation of the matrices is used.
  
  Examples: 

>> A:=array(1..2,1..2,[[1,2],[3,4]]): B:=[3,7]:
>> numeric::matlinsolve(A,B);

                       -- +-     -+  +-   -+ --
                       |  |  1.0  |  |  0  |  |
                       |  |       |, |     |  |
                       |  |  1.0  |  |  0  |  |
                       -- +-     -+  +-   -+ --

>> numeric::matlinsolve(A,B,Symbolic);
 
                    -- +-   -+  +-   -+ --
                    |  |  1  |  |  0  |  |
                    |  |     |, |     |  |
                    |  |  1  |  |  0  |  |
                    -- +-   -+  +-   -+ --

>> A:=array(1..3,1..3,[ [1,1,0],[0,1,1],[0,0,1]]):
>> B:=array(1..3,1..3,[ [1,0,0],[0,1,0],[0,0,1]]):
>> InverseOfA:=numeric::matlinsolve(A,B,Symbolic)[1];

                              +-           -+
                              |  1, -1,  1  |
                              |             |
                              |  0,  1, -1  |
                              |             |
                              |  0,  0,  1  |
                              +-           -+

>> A:=Dom::Matrix()(3,3,[[2,2,3],[1,1,2],[3,3,5]]):
>> B:=Dom::Matrix()(3,1,[sin(y)^2,cos(y)^2,x]):
>> ([X,Kernel,Constraints,PivAssupmt]):=
       numeric::matlinsolve(A,B,Symbolic,ShowAssumptions);

-- +-                 -+
|  |          2        |  +-    -+
|  |  5 sin(y)  - 3 x  |  |  -1  |
|  |                   |  |      |         2             2
|  |         0         |, |   1  |, [cos(y)  - x + sin(y)  = 0]
|  |                   |  |      |
|  |                2  |  |   0  |
|  |  2 x - 3 sin(y)   |  +-    -+
-- +-                 -+

        --
         |
         |
         |
   , []  |
         |
         |
         |
        --

>> check_the_result:= A*Dom::Matrix()(X)-B,
                      A*Dom::Matrix()(Kernel); 

                  +-                       -+
                  |           0             |  +-   -+
                  |                         |  |  0  |
                  |            2         2  |  |     |
                  |  x - cos(y)  - sin(y)   |, |  0  |
                  |                         |  |     |
                  |           0             |  |  0  |
                  +-                       -+  +-   -+

 See also:  linalg::matlinsolve, linsolve, numeric::linsolve
                                                                  */
//-----------------------------------------------------------------

numeric::matlinsolve:= proc(A,b)
local returnType, rb, useHardwareFloats, HardwareFloatsRequested, f,
      m, neqs,n,method,coefftypes,use_normal,i,j,tmp,B,cb,check,
      Amap, bmap, Aop, L, list, showassumptions, PivotAssumptions,dowarn,
      x,X,X0,Xvars,y,nTerms,koeff,np2,np2mj,p,kdim,Kernel,
      piv,pivot,pivindex,rownorm,constraints,
      lowerbands,jpl,minlength, count, result, triv, kernindex;
begin
  if args(0)<2 then error("expecting at least two arguments"); end_if;
  if args(0)>7 then error("expecting no more than seven arguments"); end_if;

  //----------------------------------------------
  // Check the matrix and set default return types
  //----------------------------------------------
  if domtype(A) = DOM_ARRAY then
     if op(A, [0, 1]) <> 2 then
        error("first argument: expecting a 2-dimensional array");
     end_if;
     [neqs, n]:= [op(A, [0, 2, 2]), op(A, [0, 3, 2])];
     returnType:= DOM_ARRAY;
  elif A::dom::hasProp(Cat::Matrix)=TRUE then
     [neqs, n]:= A::dom::matdim(A):
     returnType:= Dom::Matrix();
     if A::dom::constructor = Dom::SparseMatrix then
          returnType:= Dom::SparseMatrix();
     elif A::dom <> Dom::Matrix() then
          A:= (Dom::Matrix())::create(extop(A, 3));
     end_if;
  else
     error("first argument: expecting an array or ".
           "a matrix of category 'Cat::Matrix'"):
  end_if;

  if domtype(b) = DOM_ARRAY then
       case op(b, [0, 1])
       of 1 do 
          [rb, cb]:= [op(b, [0, 2, 2]), 1];
          b:=array(1 .. rb,1..1,[[b[i]]$ i=1..rb]);
          break;
       of 2 do
          [rb, cb]:= [op(b, [0, 2, 2]), op(b, [0, 3, 2])];
          break;
       otherwise
          error("second argument: expecting a 1-dimensional array ".
                "(a vector) or a 2-dimensional array (a matrix)");
       end;
  elif domtype(b) = DOM_LIST then
       [rb, cb]:= [nops(b), 1];
       b:=array(1 .. rb, 1..1,[[b[i]]$ i=1..rb]);
  elif b::dom::hasProp(Cat::Matrix) = TRUE then
       [rb, cb]:= b::dom::matdim(b):
       if b::dom::constructor <> Dom::SparseMatrix and
          b::dom <> Dom::Matrix() then
            b:= (Dom::Matrix())::create(extop(b, 3));
       end_if;
  else error("second argument: expecting a list, an array, or ".
             "a matrix of category 'Cat::Matrix'"):
  end_if;

  if neqs <> rb  then 
     error("dimensions of matrices/vectors not compatible"); 
  end_if;

  //---------------------------------------------
  // Check the options
  //---------------------------------------------
  // set defaults
  method:= 2;        // floating point method
  showassumptions:= FALSE; // default
  dowarn:= TRUE; // default
  if DIGITS <= 15 then
       useHardwareFloats:= TRUE;
  else useHardwareFloats:= FALSE;
  end_if;
  HardwareFloatsRequested:= FALSE;

  for i from 3 to args(0) do
      case args(i)
      of Hard do
      of HardwareFloats do
                     HardwareFloatsRequested:= TRUE;
                     if method = 1 then
                        warning("ignoring the option 'HardwareFloats' ".
                                "because of the option 'Symbolic'"):
                        method:= 1; // symbolic method
                        useHardwareFloats:= FALSE:
                     else
                        useHardwareFloats:= TRUE:
                     end_if;
                     break;
      of Soft do
      of SoftwareFloats do
                     useHardwareFloats:= FALSE:
                     break;
      of Symbolic do method:=1; // symbolic method
                     useHardwareFloats:= FALSE;
                     break;
      of ShowAssumptions do
                     useHardwareFloats:= FALSE;
                     showassumptions:= TRUE; 
                     break;
      of NoWarning do 
                     dowarn:= FALSE;
                     break;
      otherwise
         if type(args(i)) = "_equal" and
            lhs(args(i)) = ReturnType then
              returnType:= rhs(args(i));
              if not has({DOM_ARRAY, Dom::Matrix(), Dom::SparseMatrix()}, returnType) then
                 error("illegal return type ".expr2text(args(i)).
                       " specified. Choose between DOM_ARRAY, ".
                       "Dom::Matrix() or Dom::SparseMatrix()"):
              end_if;
         else error("unknown option ".expr2text(args(i)));
         end_if;
      end_case;
  end_for:

  //--------------------------------
  // ----- use HardwareFloats ------
  //--------------------------------
  case useHardwareFloats // use case because we want to use break below
  of TRUE do
     if DIGITS > 15 then
        if HardwareFloatsRequested then
           warning("the current precision goal DIGITS = ".expr2text(DIGITS).
                   " cannot be achieved with the requested 'HardwareFloats'.".
                   " Using 'SoftwareFloats' instead.");
        end_if;
        break;
     end_if;

     if not numeric::startHardwareFloats() then
        userinfo(1, "cannot start the hardware floats interface"):
        if HardwareFloatsRequested then
           warning("cannot start the 'HardwareFloats' interface, ".
                   "using 'SoftwareFloats' instead");
        end_if;
        break;
     end_if;

     userinfo(1, "trying hardware floats"):

     /*-----------------------------------------------------------
     Call scilab via numeric::linsolveSci{Sparse,Dense}:
     -----------------------------------------------------------*/

     if A::dom::hasProp(Cat::Matrix)=TRUE and
        A::dom::constructor = Dom::SparseMatrix then

        f:= scilab::func(
                    ["A", "b"],
                    ["x", "kern", "kernindices", "info"], 
                    numeric::linsolveSciSparse
                    ):
     else
        f:= scilab::func(
                    ["A", "b"],
                    ["x", "kern", "kernindices", "info"], 
                    numeric::linsolveSciDense
                    ):
     end_if;

     // B =  [special solution, kernel, kernelindices, info]
     B:= [f(A, b, ReturnType = [returnType, returnType, Default, Default], Quiet)];

     //--------------------------------------
     // post-processing
     //--------------------------------------

     if has(B, FAIL) then
        userinfo(1, "hardware floats failed"):
        if HardwareFloatsRequested then
            warning("'HardwareFloats' failed, using 'SoftwareFloats' instead"):
        end_if;
        break;
     end_if;

     if B[4] = "FAIL" then
        //break;             // try with MuPAD 
        return([FAIL, NIL]); // or exit (i.e., trust scilab) ??
     end_if; 

     // An empty kernel is returned as the string B[4] = "EmptyKernel".
     // Convert it to the exact integer 0 (this is
     // matlinsolve's way to indicate an empty kernel)

     case B[4]
     of "OK" do
         delete B[4], B[3];
         break;
     of "EmptyKernel" do
         B[2]:= 0; 
         delete B[4], B[3];
         break;
     end_case;

     //------------------------------
     //-------- return --------------
     //------------------------------

     return(B);
  end_case; // of useHardwareFloats = TRUE

  //---------------------------------
  //---- use SoftwareFloats ---------
  //---------------------------------

  userinfo(1, "not using hardware floats"):

  if A::dom::hasProp(Cat::Matrix)=TRUE and
     A::dom::constructor = Dom::SparseMatrix then
       Amap:= A::dom::mapNonZeroes;
       Aop := A::dom::nonZeroOperands;
  else Amap:= map;
       Aop := op;
  end_if;

  // convert dense matrices to domtype DOM_ARRAY for speed 
  if A::dom::hasProp(Cat::Matrix)=TRUE then 
     if A::dom::constructor = Dom::SparseMatrix then
        if A::dom <> Dom::SparseMatrix() then
           A:= Amap(A, expr):
           A:= (Dom::SparseMatrix())::create(extop(A));
        end_if;
     else A:= expr(A)
     end_if;
  end_if; 

  if b::dom::hasProp(Cat::Matrix)=TRUE and
     b::dom::constructor = Dom::SparseMatrix then
       bmap:= b::dom::mapNonZeroes;
  else bmap:= map;
  end_if;

  // convert dense rhs to domtype DOM_ARRAY for speed 
  if b::dom::hasProp(Cat::Matrix)=TRUE then
     if b::dom::constructor = Dom::SparseMatrix then
        if b::dom <> Dom::SparseMatrix() then
           b:= bmap(b, expr):
           b:= (Dom::SparseMatrix())::create(extop(b));
        end_if;
     else b:= expr(b);
          if not contains({1,2},op(b,[0,1])) then 
             error("second argument: wrong format");
          end_if;
     end_if;
  end_if;

  // force evaluation of A and b (both are arrays now) 
  A:= Amap(A, eval):
  b:= bmap(b, eval);

  use_normal:=FALSE;  // default 

  if method=2 then
    tmp:= Amap(A, float):
    B:= bmap(b, float):
    coefftypes := map({Aop(tmp)}, domtype);
    if coefftypes // union {op(map(B,domtype))} 
        minus {DOM_FLOAT,DOM_COMPLEX,DOM_INT} <> {} then
        if dowarn then
          warning("symbolic coefficient found, switching to symbolic mode");
        end_if;
        method:=1;B:=NIL;
    else A:=tmp; b:=B; B:=NIL:
    end_if;  
  end_if;

  use_normal := FALSE;
  if method=1 then
    coefftypes := map({op(A)}, domtype);
    if coefftypes minus {DOM_FLOAT,DOM_COMPLEX,DOM_INT,DOM_RAT} <> {}
      then use_normal:=TRUE
    end_if;
  end_if;

  userinfo(10,
           if method = 2 then
                "using numerical mode"
           elif use_normal then
                "using symbolic mode with normalization"
           else "using symbolic mode without normalization"
           end_if
          );

  if method=1 then
    userinfo(1,"symbolic mode: no pivoting unless necessary");
  end_if;
  if method=2 then userinfo(1,"numerical mode: using column pivoting"); end_if;

  // -- initialization done --- 

  kdim:=0;  // dimension of kernel 
  np2:=n+2;
  lowerbands:=0:       // count sudiagonal bands and store in lowerbands: 
  x:=genident("x"):    // free identifiers as unknown in polynomials to   
  y:=genident("y"):    // avoid conflicts with global variables           

  //-----------------------------------------------
  // convert A from polys = columns to polys = rows.
  // Use p as container. Store the j-th element in p[j]
  // as the coefficient of x^(n+2-j) (j=1.. n )
  //-----------------------------------------------
  m:= neqs; // dimension(A) = m x n
  // table p = container for the rows
  // Initialize as a table
  p := table();
  if A::dom::hasProp(Cat::Matrix)=TRUE and
     A::dom::constructor = Dom::SparseMatrix then
      // In analogy to (Dom::SparseMatrix())::transpose:
      // For sparse matrices the version using L:= poly2list(x[j])
      // is faster than the version using coeff(x[i], j).
      // For dense matrices it is the other way round.
      // Regard a matrix as 'reasonably sparse', if at most
      // half its entries are nonzero:

      if A::dom::nonZeroes(A) < m*n/2
      then //sparse case
        A:= extop(A, 3);
        for j from 1 to n do
          L:= poly2list(A[j]);
          for list in L do
            // fill the rows
            p[list[2]][j]:= [list[1], np2 - j];
            if list[2] - j >lowerbands then 
               lowerbands:=  list[2] - j; 
            end_if;
          end_for;
        end_for;
        // So far, the row p[i] is a table created implicitly by
        // p[list[2]][j]:= [list[1], np2 - j];
        // Convert the rows to polynomials.
        for i from 1 to m do
          if type(p[i]) = "_index" then
             // p[i] was not implictly created. This row is empty
             p[i] := poly(0, [x]);
          else
             // convert table p[i] to a list p[i]:
             p[i]:= map([op(p[i])], op, 2);
             // convert list p[i] to a poly p[i]:
             p[i]:= poly(p[i], [x]);
             if use_normal then p[i]:=mapcoeffs(p[i],normal);end_if;
          end_if;
        end_for;
      else //dense case
        // build up row for row to avoid doubling of memory
        // due to doubly nested sequences
        A:= extop(A, 3);
        for i from 1 to m do
            p[i] := poly([[coeff(A[j],i), np2-j] $j=1..n],[x]);
            if use_normal then p[i]:=mapcoeffs(p[i],normal);end_if;
        end_for;
        lowerbands:= m - 1;
      end_if;
  else // A is a DOM_ARRAY
      for i from 1 to neqs do // convert equations to univariate polynomials 
        tmp:= 0:
        for j from 1 to n do  // x[j] -> x^(n+2-j),j=1.. n 
           if not iszero(A[i,j]) then 
                  tmp:=tmp+A[i,j]*x^(np2-j): 
                  if i-j>lowerbands then lowerbands:=i-j; end_if;
           end_if;
        end_for;
        p[i]:=poly(tmp,[x]):
        if use_normal then p[i]:=mapcoeffs(p[i],normal);end_if;
      end_for;
  end_if;

  //-----------------------------------------------
  // convert b from polys = columns to polys = rows.
  // Use B as container. Store the j-th element in B[j]
  // as the coefficient of y^(j-1) (j=1.. n )
  //-----------------------------------------------
  // B = container for the rows of the rhs
  B := table();
  if b::dom::hasProp(Cat::Matrix)=TRUE and
     b::dom::constructor = Dom::SparseMatrix then
      if b::dom::nonZeroes(b) < m*cb/2
      then //sparse case
        b:= extop(b, 3);
        for j from 1 to cb do
          L:= poly2list(b[j]);
          for list in L do
            // fill the rows
            B[list[2]][j]:= [list[1], j-1];
          end_for;
        end_for;
        // So far, the row B[i] is an implicitly created table.
        // Convert the rows to polynomials.
        for i from 1 to m do
          if type(B[i]) = "_index" then
             // B[i] was not implictly created. This row is empty
             B[i] := poly(0, [y]);
          else
             // convert table B[i] to a list B[i]:
             B[i]:= map([op(B[i])], op, 2);
             B[i]:= poly(B[i], [y]);
          end_if;
        end_for;
      else //dense case
        // build up row for row to avoid doubling of memory
        // due to doubly nested sequences
        b:= extop(b, 3);
        for i from 1 to m do
            B[i] := poly([[coeff(b[j],i), j-1] $j=1..cb],[y]);
        end_for;
      end_if;
  else // b is a DOM_ARRAY
      for i from 1 to neqs do  // convert equations to univariate polynomials 
        tmp:= b[i,1];
        for j from 2 to cb do           // store rows of b as poly in y 
           if not iszero(b[i,j]) then tmp:=tmp+b[i,j]*y^(j-1); end_if;
        end_for;
        B[i]:=poly(tmp,[y]):
      end_for;
  end_if;

  //----------------------------------------------------
  // Row compression: get rid of all trivial equations
  //----------------------------------------------------
  count:= 0;
  m:= neqs;
  for i from 1 to neqs do
      if iszero(p[i]) and iszero(B[i]) then
         m:= m - 1;
      else
         count:= count + 1;
         p[count]:= p[i];
         B[count]:= B[i];
      end_if;
  end_for;
  for i from m+1 to neqs do
      delete p[i], B[i];
  end_for;
  neqs:= m;

  //----------------------------------------------------
  // check = utility to check rhs of equations 0 = b:
  //----------------------------------------------------
  constraints:= {};

  check:=proc(b) // routine to check right hand side of equations 0=b 
  local koeff, tmp;
  begin //b:=subs(b,float(0)=0); 
        if iszero(b) then return(null()) end_if;
        if showassumptions=FALSE then return(FAIL); end_if;
        // Now showassumptions = TRUE and iszero(b) = FALSE.
        // Investigabe, whether b has symbolic parameters.
        // If b is numeric, then return FAIL (the system is not solvable),
        // otherwise insert contraint 0=b into constraints.
        koeff:=[coeff(b)]:     // b is a polynomial in y 
        for tmp in koeff do    
          if not iszero(tmp) then
            if hastype(tmp,{DOM_EXPR,DOM_IDENT,"_index"})
               then constraints:=constraints union {numer(tmp)};
               else return(FAIL):
            end_if;
          end_if;
        end_for;
        return(b);
  end_proc;

 //  Conversion of matrices to sparse univariate poly representation done 

  PivotAssumptions:= {};
  triv:= table();
  kernindex:= table();

  //------------------------------
  // Start Gauss-Elimination 
  //------------------------------
   
  //for j from 1 to n do // eliminate column j
  j:= 0;
  while j < min(neqs, n) do
    j:= j + 1;
    np2mj:=np2-j;
    jpl:=min(j+lowerbands,neqs):

   // ------------------- search for pivot element ------------------ 
   // ------- initialization: collect piv[j],..,piv[jpl] -------- 

    pivot:=0;pivindex:=j; minlength:=RD_INF;
    (piv[i]:= 0;) $ i=j..max(neqs,n):
    /* store element j from p[i] = i.th row in piv[i] and
      truncate p[i] to become the remainder of the row */
    for i from j to jpl do
        piv[i]:= coeff(p[i],np2mj);
        // Make sure, vanishing piv[i] are identified as 0:
        if use_normal and j=1 and not iszero(piv[i])
           then piv[i]:=normal(piv[i]);
                mapcoeffs(p[i], normal);
        end_if;
        /* Must cut piv[i] from p[i], even if iszero(piv[i]). Note that
          piv[i]=float(0)<>0 might happen. */
        if not iszero(piv[i]) then p[i]:=lmonomial(p[i],Rem)[2]: end_if;
    end_for;

    // ------- search for best pivot element----------------- 
    // piv[j],..,piv[jpl] are candidates for the pivot elements.
    for i from j to jpl do

        if method=1 then  /*symbolic mode: take the "simplest" nonsymbolic
                           element<>0 as pivot */
           if (not iszero(piv[i])) then
             // look for non-symbolic Pivot elements only 
             if indets(piv[i],PolyExpr)={} then
               if iszero(p[i])
               then /*remaining row is trivial. This piv[i] is the ideal
                     candidate for a pivot element: accept it! */
                    pivindex:=i; break;   //stop pivot search 
               else /*remaining row is nontrivial. Look for rows
                     with pivot element of smallest complexity */
                    length(piv[i]):
                    if last(1)<minlength
                       then minlength:=last(1); pivindex:=i:
                    end_if;
               end_if;
             end_if;
           end_if;
        end_if;

        if method=2 then //numeric mode: do row pivoting 
           if not iszero(piv[i]) then
              rownorm:=max(op(map([coeff(p[i])],specfunc::abs))):
              if iszero(rownorm) //remaining row is trivial 
              then pivindex:=i; break;  //stop pivot search 
              else specfunc::abs(piv[i])/rownorm:
                   if pivot<last(1) then pivot:=last(1);pivindex:=i;end_if;
              end_if;
           end_if;
        end_if;

    end_for;

    /* so far we tried to avoid a symbolic pivot element. If no such
      pivot element exists, we have to start again. This time we must
      accept symbolic elements. */
    if method=1 and iszero(piv[pivindex]) then
      pivot:=0; pivindex:=j; minlength:=RD_INF;
      for i from j to jpl do
        if not iszero(piv[i]) then
          if iszero(p[i]) then pivindex:=i; break;
          else length(piv[i]):
               if last(1)<minlength
                  then minlength:=last(1); pivindex:=i:
               end_if;
          end_if;
        end_if;
      end_for;
    end_if;

    //------------------------------------------------
    // optimal Pivot element found, now exchange rows 
    //------------------------------------------------

    if j<>pivindex then
      ([p[j],p[pivindex]]):=[p[pivindex],p[j]]:
      ([B[j],B[pivindex]]):=[B[pivindex],B[j]]:
      ([piv[j],piv[pivindex]]):=[piv[pivindex],piv[j]]:
    end_if;

/* ----- if symbolic pivot must be used, this is assumed to be <>0.
        Store this assumption in PivotAssumptions: 
--------*/

    if method=1 and showassumptions=TRUE and
       indets(piv[j],PolyExpr)<>{} then
       tmp:= numer(piv[j]);
       if indets(float(tmp),PolyExpr)<>{} then
          PivotAssumptions:= PivotAssumptions union {tmp};
       end_if;
    end_if;

// ----- check if non-trivial pivot element was found ------- 
    if iszero(piv[j]) then
         kdim:= kdim + 1;
         triv[j]:= TRUE;
         kernindex[j]:= kdim;

         // Entzerrungsschritt: fuege eine weitere Gleichung ein 
         // j may have become larger than the number of
         // input equations, i.e., p[j] and B[j] have not
         // been set before:

//!!!!!!!!!!!!!!!1
         if not(iszero(p[j]) and iszero(B[j])) then
//!!!!!!!!!!!!!!!1
            neqs:=neqs+1; 
            piv[neqs]:=0;
            p[neqs]:=p[j]:
            B[neqs]:=B[j];
            // the number of lower bands was increased 
            lowerbands:=max(lowerbands,neqs-j-1);
            jpl:=min(j+lowerbands,neqs):
//!!!!!!!!!!!!!!!1
         end_if;
//!!!!!!!!!!!!!!!1
    else // do the elimination. Column j is stored in piv[j],..,piv[neqs]
         for i from j+1 to jpl do
             if not iszero(piv[i]) then
                tmp:=piv[i]/piv[j]:
                case tmp
                of  1 do  p[i]:=p[i]-p[j];B[i]:=B[i]-B[j];break;
                of -1 do  p[i]:=p[i]+p[j];B[i]:=B[i]+B[j];break;
                otherwise p[i]:=p[i]-mapcoeffs(p[j],_mult,tmp):
                          B[i]:=B[i]-mapcoeffs(B[j],_mult,tmp):
                end_case;
                if use_normal then 
                    p[i]:=mapcoeffs(p[i],normal):
                    B[i]:=mapcoeffs(B[i],normal):
                end_if;
             end_if;
         end_for:
    end_if;
  end_while; // while j < min(neqs, n)
  //end_for; // for j = 1..n do

  //-------------------------------------------------
  // we now have neqs equations for n unknowns, where 
  // the first n equations are upper triangular. 
  // Check that the remaining equations are trivial
  //-------------------------------------------------

  for j from n+1 to neqs do
    if check(B[j])=FAIL then 
      if showassumptions=FALSE then 
           return([FAIL, NIL]);
      else return([FAIL, NIL, [], []]);
      end_if;
    end_if;
  end_for;
  constraints:=[op(constraints)];

  //-----------------------------
  // Fill in trivial equations to
  // obtain a quadratic scheme:
  //-----------------------------
  for j from neqs + 1 to n do
      triv[j]:= TRUE;
      piv[j]:= 0:
      p[j]:= poly(0, [x]):
      B[j]:= poly(0, [y]):
      kdim:= kdim + 1;
      kernindex[j]:= kdim;
  end_for;

/*--------------------
  // this is a little utility needed during
  // the design of this function. It checks
  // that above no unexpected indexed calls to
  // rows occured that where not properly
  // initialized
// if nops(select(p, hastype, DOM_TABLE)) <> 0 then 
//    warning("p has DOM_TABLE"): 
// end_if;
// if nops(select(B, hastype, DOM_TABLE)) <> 0 then 
//    warning("B has DOM_TABLE"): end_if;
// if nops(select(p, hastype, DOM_TABLE)) <> 0 or
//    nops(select(B, hastype, DOM_TABLE)) <> 0 then
//   print("-------------"): 
//   print(p, B);
//   error(" p or B has DOM_TABLE"): 
// end_if;
------------------------*/

  //--------------------------------------------------------
  // Backsubstitution:
  // Now the system is in upper triangular form. The diagonal
  // is stored in piv[1],..,piv[n]
  //--------------------------------------------------------

  // convert univariate polynomials to expressions
  // for fast sparse backsubstition:  
  // coeff*x^(n+2-i) <-> coeff*x[i] 
  if triv[n] = TRUE then
       X[n]:=y^(cb-1+kernindex[n]);
  else X[n]:= op(B[n],1)/piv[n];
       if use_normal then X[n]:=normal(X[n]): end_if:
  end_if;
  for i from n-1 downto 1 do
     if triv[i] = TRUE then
          X[i]:=y^(cb-1+kernindex[i]);
     else // pick out only non-trivial coefficients of the poly:
          nTerms:=nterms(p[i]):
          Xvars:=[X[np2-op(nthterm(p[i],j),[1,2])]$ j=1..nTerms]:
          koeff:=[coeff(p[i])]:
          p[i]:=_plus( koeff[j]*Xvars[j] $ j=1..nTerms):
          if piv[i]=1 
            then X[i]:=(op(B[i],1)-p[i]);
            else X[i]:=(op(B[i],1)-p[i])/piv[i]:
          end_if;
     end_if;
  end_for;

  //-------------------------------------------------------------
  // The backsubstituion is finished.
  // Now, X[i] = polynomial expression in y. We need to pick out the
  // various y powers. Reconvert expressions to polys in y for
  // faster access via coeff. 
  // Do not normalize X here, because poly would undo normal   
  //-------------------------------------------------------------

  for i from 1 to n do 
      X[i]:=poly(X[i], [y]);
  end_for;

  //-------------------------------------------------------------
  // The polynomials X[i] store the special solution X0[i, j]
  // (i=1..n, j=1..cb) as the coefficients of y^0, .. , y^{cb -1}
  // and the kernel elements as the coefficients of
  // y^cb, .. ,y^cb + kdim -1
  /* Dense extraction:
     X0:=array(1..n,1..cb,[[coeff(X[i],j-1)$ j=1..cb]$ i=1..n]);
     Kernel:=array(1..n,1..kdim,
           [[coeff(X[i],cb-1+j)$ j=1..kdim]$ i=1..n]);
  */
  //-------------------------------------------------------------

  // sparse extraction:
  if returnType = Dom::SparseMatrix() then
       // use a table (sparse structure!) to write into
       // when extracting the coeffs of the row polys.
       // Further down, the table is passed to
       // (Dom::SparseMatrix())::new for conversion
       // to a sparse matrix
       X0:= table();
       if kdim > 0 then
            Kernel:= table();
       else Kernel:= 0; // array(1..n,1..1,[[0]$n]) 
       end_if;
  else // if arrays or Dom::Matrices are to be returned,
       // it is the cheapest to create arrays right here
       // Note that array -> Dom::Matrix is extremely
       // cheap via (Dom::Matrix())::create)
       X0:=array(1..n,1..cb, [[0$cb]$n]);
       if kdim > 0 then
            Kernel:=array(1..n,1..kdim, [[0$kdim]$n]);
       else Kernel:= 0; // array(1..n,1..1,[[0]$n]) 
       end_if;
  end_if;

  // sparse extraction:
  for i from 1 to n do
     L:= poly2list(X[i]);
     for list in L do
       if list[2] < cb then
            X0[i, list[2] + 1]:= list[1];
       else Kernel[i, list[2] + 1 - cb]:= list[1];
      end_if;
     end_for; 
  end_for;
  if method = 2 and kdim > 0 then
     // for kdim = 0, return 0, not float(0)!
     Kernel:= map(Kernel, float);
  end_if;
  X:= X0:

  if method = 2 then
     X:= map(X, float);
  end_if;
   
  //----------------------------------
  // That's it. The special solution X
  // and the kernel 'Kernel' are extracted.
  // Final simplifications: 
  //----------------------------------
  if use_normal then 
     X:=map(X,normal); 
     Kernel:=map(Kernel,normal);
  end_if:
   
  constraints:=map(constraints,
                   proc(x) begin
                    if iszero(x) then null() else x end_if
                   end_proc);

  if showassumptions=FALSE then 
       result:= [X, Kernel];
  else constraints:=map([op(constraints)],
             proc() begin args(1)=0 end_proc);
       PivotAssumptions:=map([op(PivotAssumptions)],
             proc() begin args(1)<>0 end_proc);
       result:= [X, Kernel, constraints, PivotAssumptions];
  end_if;

  //-------------------------
  //------- return ----------
  //-------------------------
  case returnType
  of Dom::SparseMatrix() do
       if result[1] <> FAIL then
          result[1]:= returnType(n, cb, result[1]);
       end_if;
       if result[2] <> 0 and result[2] <> NIL then
          result[2]:= returnType(n, kdim, result[2]);
       end_if;
       break;
  of Dom::Matrix() do
       if result[1] <> FAIL then
          result[1]:= returnType::create(result[1]);
       end_if;
       if result[2] <> 0 and result[2] <> NIL then
          result[2]:= returnType::create(result[2]);
       end_if;
       break;
  of DOM_ARRAY do
       // nothing to do
       break;
  otherwise
       error("unexpected return type"):
  end_case;

  return(result):
end_proc:
