// 
// $Date: 2001/07/17 12:14:32 $  $Author: ccr $   $Revision: 1.23 $ 
// 

/*++

  numeric::realroots(p(x), x=a..b, <eps>, <Merge> )

  p(x) : some arithmetical expression in x
  x    : DOM_IDENT or indexed identifier
  a, b : real numerical constants (not +- infinity)
  eps  : positive numerical constant >= 10^(-DIGITS)
  Merge: do merge resulting intervals whatever the length
         of the result is.

  Returns a list of intervals (=lists), in which real roots of the
  expression p(x) may exist. All intervals are inside [a,b]. All 
  intervals have length < eps. The default is eps = 0.01.

  Warning: if p(x) is identical to zero, then the entire
  search interval is returned. If the result violates
  b-a < eps, then a warning is given;

  Also equations p(x): lhs(x)=rhs(x) are permitted as 1st argument.

  It is garanteed that no real roots exist outside the intervals.
  However, it is not guaranteed that each interval does indeed
  contain a root.

>> numeric::realroots((x^2 - 1)^2,x=-6..6,0.01);

                      [[-1.0, -1.0], [1.0, 1.0]]

>> numeric::realroots(5*x-20*x^3+16*x^5, x=-6..6,0.01);



    [0.5859375, 0.58984375], [0.9453125, 0.953125]]

>> numeric::realroots(sin(x),x=-4..4, 0.1 );

   [[-3.146875, -3.0859375], [-0.1, 0.0], [0.0, 0.1], [3.0859375, 3.146875]]

++*/

numeric::realroots := proc(eq, var)
local p, x, a, b, ab, eps, OKTypes, doMerge, Arg;
begin
   if args(0)<2 then error("expecting at least two arguments") end_if;
   if args(0)>4 then error("expecting at most four arguments") end_if;
   p:= eval(subs(eq, float(0)=0));
   if type(p)="_equal" then p:= op(p,1)-op(p,2) end_if;
   if not testtype(p, Type::Arithmetical) and 
      not domtype(p)=DOM_POLY then
      error("the first argument must be an equation ".
            "or of domain type DOM_POLY ".
            "or of type 'Type::Arithmetical'")
   end_if;
   if type(var)<>"_equal" or type(op(var,2))<>"_range"
      then error("second argument is not of the form x=a..b"); 
   end_if;
   x:= op(var, 1);
   if domtype(x) <> DOM_IDENT and type(x) <> "_index" then
      error("in the second argument 'x = a..b', the unknown x ".
            "must be an identifier or an indexed identifier")
   end_if;
   if numeric::indets(p) minus {x} <> {} then
      error("the first argument must be a univariate expression or equation in ".expr2text(x)) 
   end_if; 
   // Do not (!!!) convert bounds of search interval to floats.
   // Otherwise there is danger of missing roots due to roundoff!!
   // Example: numeric::realroots(sin(PI*x),x=-1.0..1.0,0.1);
   a:= op(var, [2,1]);
   if domtype(float(a))<>DOM_FLOAT then 
      error("illegal symbolic or complex left bound a in 2nd argument x = a..b") 
   end_if;
   b:= op(var, [2,2]);
   if domtype(float(b))<>DOM_FLOAT then 
      error("illegal symbolic or complex right bound b in 2nd argument x = a..b") 
   end_if;
   if float(b)<float(a) 
      then error("illegal search interval x = a..b with a>b");
   end_if;

   // search for optional arguments:
   doMerge:= FALSE; // default
   eps:= 0.01;      // default
   for Arg in [args(3..args(0))] do
       if Arg = hold(Merge)
          then doMerge:= TRUE; break;
       end_if;
       eps:= float(Arg);
       if domtype(eps)<>DOM_FLOAT
          then error("illegal precision goal")
       end_if;
       if eps<0.999*10^(-DIGITS) 
          then error("precision goal too small. It must be >= 10^(-DIGITS)")
       end_if;
   end_for;
   if iszero(p) then 
      if specfunc::abs( float(b)-float(a))>eps then
         warning("zero function, returning search interval"); 
      end_if;
      return([[float(a),float(b)]])
   end_if;

   //Treat polynomial expressions via polylib::realroots.
   //This is rigorous and also much faster.
   if domtype(p)=DOM_POLY or testtype(p, Type::PolyExpr(x))
   then p:=polylib::realroots(p, eps);
        // intersect isolating intervals from polylib::realroots
        // with the search interval [a,b] of numeric::realroots
        OKTypes:= {DOM_INT,DOM_RAT,DOM_FLOAT}:
        if {domtype(a)} minus OKTypes<>{} then a:=float(a):end_if;
        if {domtype(b)} minus OKTypes<>{} then b:=float(b):end_if;
        p:=map(p,
               proc(x) begin  //x[1], x[2] are DOM_INT or DOM_RAT
                 if x[1]<a and x[2]< a then return(null()) end_if; 
                 if x[1]<a and x[2]>=a and x[2]<=b then return([a, x[2]]) end_if; 
                 if x[1]<a and x[2]> b then return([a, b]) end_if; 
                 if x[2]>b and x[1]> b then return(null()) end_if; 
                 if x[2]>b and x[1]>=a and x[1]<=b then return([x[1],b]) end_if; 
               //if x[2]>b and x[2]< a then return([a, b]) end_if; 
                 return(x);
               end_proc);
        p:= map(p, map, float);
   else //treat non-polynomial expressions via interval arithmetic.
        // First, convert constant expressions in p into intervals:
        p:= interval(p):
        // Enlarge the search interval a bit via interval arithmetic.
        ab:= a ... b;
        // Pass the interval to the routine doing the actual work
        p:= [numeric::realroots_(p, x, ab, eps)];
   end_if;

   if doMerge then
      p := _union(op(p));
      if op(p,0) = hold(_union) then
         p := [op(p)];
      else
         p := [p];
      end_if;
   end_if;

   return(map(p, x->[op(x)]));

//   // Now merge intervals with common bounds.
//   a:=1; 
//   while a < nops(p) do
//      for b from a+1 to nops(p) do
//         if op(p[b-1],2) <> op(p[b],1) then break end_if
//      end_for;
//      if doMerge or float(op(p[b-1],2)-op(p[a],1))<=1.0001*eps
//      then
//         p[a]:=[op(p[a],1), op(p[b-1],2)];
//         for b from a+1 to b-1 do delete p[a+1] end_for;
//      end_if;
//      a:=a+1
//   end_while;

//   return(p);
end_proc:

//-----------------------------------------------------
// recursive utility realroots_. If 0 is not contained in the
// image p(ab), then ab is eliminated from the search
// range. If 0 is contained in p(ab), then bisectioning is
// used to investigate the 2 halfs of the interval. The search
// stops when DOM_INTERVAL::width(ab) < eps:
//-----------------------------------------------------

numeric::realroots_ := proc(p, x, ab, eps)
   local image, containszero, c;
begin
   userinfo(1, "investigating interval ".expr2text(ab));
   image:=eval(subs(p, x = ab));

   if not testtype(image, DOM_INTERVAL) then
       error("cannot handle the expression ".expr2text(p).
             "  via interval arithmetic");
   end_if:

   containszero:= bool(0 in image);
   if containszero = FALSE
   then // The interval ab definitely does not contain a root,
        // so eliminate this interval from the range that may 
        // contain roots:

        return(null());

   else // If 0 is in the image interval, then we really do not know
        // whether the interval [a, b] contains a root: interval arithmetic
        // may have enlarged the image interval p([a, b]) 'artificially' 
        // to contain zero. So, investigate further:

        if DOM_INTERVAL::width(ab) < eps then return(ab) end_if;
        c := DOM_INTERVAL::center(ab);
        return(numeric::realroots_(p, x, subsop(ab, 2=c), eps),
               numeric::realroots_(p, x, subsop(ab, 1=c), eps));
   end_if;
end_proc:

// end of file 
