/**
 *   MuPlotML:  Domain which writes MuPAD plot primitives into a
 *              file;  the output is XML code conform with the
 *              muplotml_{23}d.dtd 
 */

alias(MuPlotML = plot::MuPlotML):

MuPlotML := newDomain("MuPlotML"):
MuPlotML::create_dom:=hold(MuPlotML):
MuPlotML::OutputFile := 0:
delete dom::xbmin:
delete dom::xbmax:
delete dom::ybmin:
delete dom::ybmax:
/**
 *   MuPlotML::openOutputFile :  this function opens an output file.
 *
 *     All following calls of MuPlotML::new will write into this file
 *     until it is closed with MuPlotML::closeOutputFile.
 *
 *   Argument :     file : DOM_STRING  - the name of the file 
 *                  If the string is empty an temporary file is opened.

 *   Return value : TRUE   iff the file was successfully opened
 *                  FALSE  iff the file could not be opened
 */
MuPlotML::openOutputFile :=
proc(file : DOM_STRING) : DOM_BOOL
  local fd;
begin
  if file = "" then
  	fd := fopen(Text, TempFile, Write);
	else
  	fd := fopen(Text, file, Write);
	end;
  if fd <> FAIL then
    dom::OutputFile := fd;
    return(TRUE)
  end;
  return(FALSE)
end:


/**
 *   MuPlotML::closeOutputFile :  closes the file opened by
 *                                MuPlotML::openOutputFile.
 *
 *   Argument :      -
 *
 *   Return value :  (0   -- (which is the new value of MuPlotML::OutputFile))
 *                   Testwise: the name of the just closed file.
 */
MuPlotML::closeOutputFile :=
proc()
  local a;
begin
  if dom::OutputFile <> 0 then
    a := fname(dom::OutputFile);
     fclose(dom::OutputFile)
     else
     a:=0
  end;
  dom::OutputFile := 0:
  a
end:



/**
 *   MuPlotML::new :  write scene
 *
 *   Arguments :   obj  -- a plot primitive
 *
 *   Return value : TRUE  -- print successful
 *                  FALSE -- an error occurred
 */
MuPlotML::new :=
proc(obj)
begin
  if domtype(obj) = plot::Scene then
  // xml header
    dom::pr("<?xml version=\"1.0\" encoding=\"iso-8859-1\" ?>");
    dom::newline();
  end;
  obj::dom::MuPlotML(dom, obj);

  if domtype(obj) = plot::Scene then
    dom::newline();
  end
end:

/**
 *   MuPlotML::printElement :  write one XML element
 *
 *   Arguments :    str    -- the nam of the element
 *                  attrib -- all attributes in a table
 *                  prfunc -- a procedure accepting one domain (MuPlotML)
 *                            which prints all entries 
 *
 *   Return value :  null()
 */

MuPlotML::printElement :=
proc(str, attrib, prfunc=null())
begin
  dom::pr("<", str);
  dom::printAttributes(attrib);
  if prfunc <> null() then
    dom::pr(">");
    dom::IndentDepth := dom::IndentDepth+1:
    dom::newline();
    prfunc(dom);
    dom::IndentDepth := dom::IndentDepth-1:
    dom::newline("</", str, ">");
  else
    dom::pr(" />");
  end;
end:

MuPlotML::printAttributes :=
proc(attrib : DOM_TABLE)
  local i, attrName, attrVal;
begin
  dom::IndentDepth := dom::IndentDepth+2:
  for i from 1 to nops(attrib) do
    attrName := op(attrib, [i, 1]);
    attrVal  := op(attrib, [i, 2]);

    /*  In this case switch we translate MuPAD option names into
        MuPlotML attribut names and values into the proper format. */
    case attrName
      of hold(Key) do
        break;
        
      of hold(Arrows) do
        attrName := hold(AxisArrows);
        break;
        
      of hold(Axes) do
        attrName := hold(AxisVisible);
        attrVal := if attrVal = hold(None) then TRUE else FALSE end;
        break;
        
      of hold(BackGround) do // convert color
        attrName := hold(Backgnd);
        attrVal := dom::colStr(attrVal);
        break;
        
      of hold(Color) do
        attrVal := dom::colStr(op(attrVal, 2));
        break;

      of hold(FontFamily) do  // all fine
        break;

      of hold(FontSize) do   
        // maybe we have to translate the value
        break;
        
      of hold(FontStyle) do  // all fine
        break;
        
      of hold(ForeGround) do // convert color
        attrVal := dom::colStr(attrVal);
        break;

      of hold(GridLines) do
        // is ignored, but combined with GridLinesStyle
        next;

      of hold(GridLinesColor) do // convert color
        attrName := hold(GridLineColor);
        attrVal := dom::colStr(attrVal);
        break;

      of hold(GridLinesStyle) do
        if attrib[hold(GridLines)] = hold(None) then
          attrName := hold(GridStyle);
          attrVal := hold(Invisible);
        else
          attrName := hold(GridStyle);
          attrVal := if attrVal = hold(SolidLines) then hold(Solid)
                     else hold(Dashed) end;
        end;
        break;
        
      of hold(GridLinesWidth) do  // all fine
        attrName := hold(GridLineWidth);
        break;
        
      of hold(LineStyle) do
        attrVal := if attrVal = hold(SolidLines) then hold(Solid)
                   else hold(Dashed) end;
        break;
        
      of hold(LineWidth) do
        break;
        
      of hold(PointStyle) do
        case attrVal
          of hold(Circles) do
            dom::newline();
            dom::pr(" PointFilled=\"False\"");
            break;
          of hold(Squares) do
            dom::newline();
            dom::pr(" PointFilled=\"False\"");
            break;
          of hold(FilledCircles) do
            dom::newline();
            dom::pr(" PointFilled=\"True\"");
            attrVal := hold(Circles);
            break;
          of hold(FilledSquares) do
            dom::newline();
            dom::pr(" PointFilled=\"True\"");
            attrVal := hold(Squares);
            break;
          otherwise
            error("invalid PointStyle")
        end;
        break;
        
      of hold(PointWidth) do
        attrName := hold(PointSize);
        break;
        
      of hold(Scaling) do
        attrName := hold(ConstrainScales);
        attrVal := if attrVal = hold(Constrained) then TRUE else FALSE end:
        break;
        
      of hold(Style) do
        if attrVal = [hold(Impulses)] then
          attrName := hold(Impulses);
          attrVal := TRUE
        else
          next
        end;
        break;
       
      of hold(Bounds) do
	break;
	of hold(Plotvar) do
	break;
      of hold(Delta) do
        break;
      of hold(Origin) do
        break;
      of hold(Ticks) do
      of hold(Title) do
      of hold(TitlePosition) do
      
        next;

        // Scale attributes
      of hold(id) do
      of hold(Orient) do
      of hold(Type) do
      of hold(Pos) do
      of hold(ScaleId) do
      //of hold(Bounds) do
      of hold(GridNum) do
      of hold(Smoothness) do
        break;
        
        // these options should never be here
        // =>  fall through to error
      of hold(AxesOrigin) do
      of hold(AxesScaling) do
      of hold(AxesInFront) do
      of hold(Discont) do        // internal
      of hold(PlotDevice) do
      of hold(RealValuesOnly) do // internal

      of hold(CameraPoint) do   // 3d option
      of hold(Filled) do        // Ellipse2d
      of hold(FocalPoint) do    // 3d option

      of hold(Labeling) do
      of hold(Labels) do
      of hold(Smoothness) do
      otherwise
        error("Unknown plot option: ".attrName)
   end_case;

    // TRUE and FALSE have to be repesented by "True" and "False"
    if attrVal = TRUE then
      attrVal := "True"
    elif attrVal = FALSE then
      attrVal := "False"
    end_if;
    
    if i mod 2 = 0 then dom::newline(); end;
    dom::pr(" ", attrName, "=\"", attrVal,"\"")
  end_for;
  dom::IndentDepth := dom::IndentDepth-2:
end:


// create an option table from the given table tab containing only
// those indices given in sel
MuPlotML::selectOptions :=
proc(tab : DOM_TABLE, sel : DOM_SET) : DOM_TABLE
begin
  select(tab, x -> contains(sel, op(x,1)))
end_proc:
  


/***   Utilities   ***/

MuPlotML::pr := () -> fprint(NoNL, dom::OutputFile, args()):

// current IndentDepth
MuPlotML::IndentDepth := 0;
MuPlotML::Indent := "  ";
  
MuPlotML::newline :=
proc()
  local i;
begin
  fprint(NoNL, dom::OutputFile, "\n", dom::Indent $ i = 1..dom::IndentDepth, args());
end:


MuPlotML::GridLines:=proc(d,grid=0)
local i;
begin
   if grid <> 0 then
      if domtype(grid) = DOM_LIST then
      d::pr("<Grid1d>"):
      d::pr(op(grid,1)):
      for i from 2 to nops(grid) do
          d::pr(",".expr2text(op(grid,i))):
      end:
      d::pr("</Grid1d>"):
      end: 
      if domtype(grid) = DOM_EXPR then
      d::printElement("Grid1d", table(hold(Origin)=0, hold(Delta) = expr2text(op(grid,2)))):
      end:
   end:
end:

MuPlotML::Title := proc(d,t)
begin
//d::printElement("Title",table(),d->(d::pr(t))):
d::pr("<Title>"):
d::pr(t):
d::pr("</Title>"):
end:

MuPlotML::scaleFunc1 := proc(d,sc)
begin
d::printElement("Scale",  table(hold(id) = 1,
                                                     hold(Orient) = hold(X),
                                                     hold(Type) = hold(Lin),
                                                     hold(Pos) = hold(Origin),
						     hold(Bounds) = (d::xbmin,",",d::xbmax)),d->(if d::xTitle <> FAIL then d::Title(d,d::xTitle) end_if, d::GridLines(d,op(sc::GridLines,1)))):
end:

MuPlotML::scaleFunc2 := proc(d,sc)
begin
d::printElement("Scale",  table(hold(id) = 2,
                                                     hold(Orient) = hold(Y),
                                                     hold(Type) = hold(Lin),
						     hold(Pos) = hold(Origin),
                                                     hold(Bounds) = (d::ybmin,",",d::ybmax)),d->(if d::yTitle <> FAIL then d::Title(d,d::yTitle) end_if, d::GridLines(d,op(sc::GridLines,2)))):
end:

MuPlotML::xdigit:=
proc(d)
begin
  case d
    of 10 do return("a");
    of 11 do return("b");
    of 12 do return("c");
    of 13 do return("d");
    of 14 do return("e");
    of 15 do return("f");
  end_case;
  expr2text(d)
end_proc:

    // return sequence (d1,d2) for number 0 <= d < 256 where
    // d1, d2 are strings with the hex digits of d
MuPlotML::hex2str :=
proc(d)
begin
  dom::xdigit(d div 16), dom::xdigit(d mod 16)
end:

    // return sequence of strings for printing the HTML color rep of
    // the MuPAD color c (a list with three numbers in [0,1])
MuPlotML::colStr :=
proc(c)
begin
  "#",
  dom::hex2str(floor(op(c,1)*255.0)),
  dom::hex2str(floor(op(c,2)*255.0)),
  dom::hex2str(floor(op(c,3)*255.0))
end_proc:

MuPlotML::SetXBounds :=
proc(min,max)
begin
  
  dom::xbmin := min:
  dom::xbmax := max:
end_proc:

MuPlotML::SetYBounds :=
proc(min,max)
begin
  
  dom::ybmin := min:
  dom::ybmax := max:
end_proc:


// Writes a very short xml-output to inform
// the 2D Plotter about a fatal error.
// The Plotter should show a dialogbox with the
// errmessage
MuPlotML::writeError :=
proc(errmessage)
begin
dom::pr("<?xml version=\"1.0\" encoding=\"iso-8859-1\" ?>");
    dom::newline();
    dom::pr("<Error>");
    dom::pr(errmessage);
    dom::pr("</Error>");
end_proc:
