// $Date: 2003/01/28 15:44:23 $  $Author: olli $   $Revision: 1.34.2.1 $
// igel 
// completely rewritten by frankp (21.09.2000)

/* -------------------------------------------------------------
    Curve2d -- The graphical primitive for two-dimensional curves

    Syntax:

    Curve2d([x_t, y_t], t = tr <, op1, op2, ...>)

    x_t, y_t        : arithmetical expressions in t
    t               : DOM_IDENT
    tr              : either form 'a..b' or list of two arithmetical expressions.
    op1, op2, ...   : plot2d options (equations of form: keyword = AnyType)

    Examples:

    >> plot::Curve2d( [sin(t), cos(t)], t = 0..2*PI )

    >> plot::Curve2d( [sin(t), cos(t)], t = 0..2*PI, Color = RGB::Blue )

----------------------------------------------------------------*/
alias( DEFAULTOPTIONS = plot::Curve2d::defaultOptions ):
alias( OPTIONNAMES = plot::Curve2d::optionNames ):

plot::Curve2d := newDomain("plot::Curve2d"):
plot::Curve2d::create_dom:=hold(plot::Curve2d):

plot::Curve2d::info:= "plot::Curve2d -- graphical primitive for two-dimensional curves":
plot::Curve2d::type:= "graphprim":
plot::Curve2d::dimension:= () -> 2:

plot::Curve2d::new:= proc(f, r)
    local i, opt, range, options, curve;
begin
    if args(0) < 2 then 
        error("expecting at least two arguments");
    end_if:
    
    //check type of the first two arguments:
    if not testtype(f, Type::ListOf(Type::Arithmetical, 2, 2)) then
        error("1st argument: expecting a list of two arithmetical expressions")
    elif (range := plot::conv2range(r)) = FAIL then
        error("2nd argument: expecting form 'ident = a..b')")
    end_if;

    options := DEFAULTOPTIONS;  // default options (a table)
    opt := [args(3..args(0))];  // user options (a list)
    // check user options and override the default options:
    for i from 1 to nops(opt) do
        if (r:= plot::Curve2d::checkOption(opt[i]))[1] = FALSE then
            error(r[2])
        else
            options[r[2]]:= r[3]
        end_if
    end_for;

    curve:= newDomain( expr2text(genident("Curve2d")) );

    // attributes:
    curve::term           := f;
    curve::range          := range;
    curve::refreshPlotdata:= TRUE;
    curve::plotdata       := [];
    curve::options        := options;

    // return the new curve:
    return( new(plot::Curve2d, curve) )
end_proc:

plot::Curve2d::copy := proc(c)
begin
    return(extsubsop(c, 1=newDomain(genident("Curve2d"), extop(c, 1))))
end_proc:

plot::Curve2d::modify := proc(c)
    local m, o, options;
begin
    m:= plot::Curve2d::copy(c);
    options:= [args(2..args(0))];

    ( slot(m, expr2text(op(o, 1)), op(o, 2)) ) $ o in options;
   
    m::refreshPlotdata:= TRUE;
    return(m)
end_proc:

plot::Curve2d::getPlotdata := proc(o)
local e;
begin
    e:= extop(o,1);
    if e::refreshPlotdata = FALSE then return( e::plotdata ) end_if;

    plot::Curve2d::buildObjects(o);
    e::refreshPlotdata:= FALSE;
    e::plotdata:= map(e::objects, oo -> op(plot::Curve2d::getPlotdataObject(oo)))
end_proc:

plot::Curve2d::buildObjects := proc(f)
    local c, col, x, e, i, intervals, objects, tmp;
begin
    e:= extop(f, 1);

    c:= e::term;
    x:= plot::range2list(e::range);

    i:= stdlib::ShowPlotWarning;
    sysassign(stdlib::ShowPlotWarning, FALSE);

    if type(e::options[hold(Color)]) = "_index" then 
        col := null()
    else
       col := Color = e::options[hold(Color)]
    end_if;

    intervals:= [plotlib::check_args(
        hold(Discont) = e::options[hold(Discont)], 
        hold(RealValuesOnly) = e::options[hold(RealValuesOnly)],
        [hold(Mode) = hold(Curve), c, x, col]
    )];
        
    sysassign(stdlib::ShowPlotWarning, i);

    // Result is of the form [sceneoptions, [Mode=..., ...]<, [Mode=...,...] ...>].
    // Select only those sublists with 'Mode=Curve':
    intervals:= select( intervals,l -> domtype(l) = DOM_LIST and l[1] = hold(Mode = Curve) );
    if nops(intervals) > 0 and (col := select(intervals[1], has, Color)) <> [] then
       col := col[1]
    end_if;
    // Select the entries 'x=[a,b]', i.e., the third element of these lists:
    intervals:= map( intervals, l -> l[3] );

    if intervals <> [] then
        if col <> [] then
           objects := [plot::Curve2d::modify(f, hold(range) = intervals[1], col)];
	else
	   objects := [plot::Curve2d::modify(f, hold(range) = intervals[1])];
	end_if;
        delete intervals[1];
        for i in intervals do
            tmp := plot::Curve2d::modify(f, hold(range) = i, hold(Title) = "");
            if col <> [] then
               tmp := plot::Curve2d::modify(tmp, col);
	    end_if;
            objects := append( objects, tmp );
        end_for
    else
        objects:= [f]
    end_if;

    e::objects:= objects
end_proc:

plot::Curve2d::getPlotdataObject := proc(f)
    local e, x, c, opt;
begin
    e:= extop(f, 1);

    c:= e::term;
    x:= plot::range2list(e::range);

    // remove specific plot::Curve2d options:
    opt:= e::options;
    delete opt[hold(RealValuesOnly)], opt[hold(Discont)];
    // convert opt into a list:
    opt:= [op(opt)];

    // create the plot data:
    e::plotdata:= [
        [hold(Mode) = hold(Curve), c, x] . opt
    ]
end_proc:

plot::Curve2d::print := proc(c)
begin
    c:= extop(c,1);
    hold(plot::Curve2d)(c::term,c::range)
end_proc:

plot::Curve2d::op := proc(c, n)
begin
    if args(0) < 1 or args(0) > 2 then
        error("invalid number of arguments")
    end_if;   

    c:= extop(c,1);
    c:= [c::term, c::range];
    if args(0) = 1 then op(c) else op(c, n) end_if
end_proc:

plot::Curve2d::nops := c -> 2:

plot::Curve2d::_index := proc(c:plot::Curve2d,n:Type::PosInt)
begin
    if args(0) <> 2 then
        error("expecting two arguments")
    end_if;

    c:= extop(c,1);
    c:= [c::term, c::range];
    if n > nops(c) then
        error("index out of range")
    else
        return( c[n] )
    end_if
end_proc:

// Note: Do not return an error message in the case of an invalid call,
//       because then the object would be destroyed! Hence a warning 
//       message is issued, and the call of this method has no effect 
//       on the object.
plot::Curve2d::set_index := proc(c:plot::Curve2d, n:Type::PosInt, v)
    local e;
begin
    if args(0) <> 3 then
        warning("wrong number of arguments; assignment ignored");
        return(c)
    end_if; 

    e:= extop(c,1);
    case n
    of 1 do
        if v <> e::term then 
            if not testtype(v,Type::ListOf(Type::Arithmetical,2,2)) then
                warning("3rd argument: expecting a list of two arithmetical expressions; assignment ignored");
                return(c)
            end_if;
            e::refreshPlotdata:= TRUE;
            e::term:= v;
        end_if;
        return( extsubsop(c,1=e) )
    of 2 do
        if v <> c::range then
            if (v:= plot::conv2range(v)) = FAIL then
                warning("3rd argument: expecting form 'ident = a..b'; assignment ignored");
                return(c)
            end_if;
            e::refreshPlotdata:= TRUE;
            e::range:= v;
        end_if;
        return( extsubsop(c,1=e) )
    otherwise
        warning("index out of range; assignment ignored");
        return(c)
    end_case;
end_proc:

// Note: Do not return an error message in the case of an invalid call,
//       because then the object would be destroyed! Hence a warning 
//       message is issued, and the call of this method has no effect 
//       on the object.
plot::Curve2d::slot:= proc(o,s,v=null())
    local e, t, _s;
begin
    if args(0) = 3 then
        // slot assignment:
        e:= extop(o,1);

        // first handle all user-accessable attributes:
        case s 
        of "refreshPlotdata" do
            if v <> TRUE and v <> FALSE then
                warning("attribute 'refreshPlotdata': expecting either 'TRUE' or 'FALSE'; assignment ignored");
                return( o )
            else
                e::refreshPlotdata:= v;
                return( extsubsop(o,1=e) )
            end_if
        of "range" do
            if (v:= plot::conv2range(v)) = FAIL then
                warning("attribute 'range': expecting form 'ident = a..b'; assignment ignored");
                return( o )
            else
                e::range:= v;
                e::refreshPlotdata:= TRUE;
                return( extsubsop(o,1=e) )
            end_if
        of "term" do
            if not testtype(v,Type::ListOf(Type::Arithmetical, 2, 2)) then
                warning("attribute 'term': expecting a list of two arithmetical expressions; assignment ignored");
                return( o )
            else
                e::term:= v;
                e::refreshPlotdata:= TRUE;
                return( extsubsop(o,1=e) )
            end_if
        of "options" do
            if domtype(v) <> DOM_TABLE then
                warning("attribute 'options': expecting a table of plot options; assignment ignored");
                return( o )
            else
                e::options:= v;
                e::refreshPlotdata:= TRUE;
                return( extsubsop(o,1=e) )
            end_if
        end_case;

        // an option given?
        t:= plot::Curve2d::checkOption(text2expr(s) = v);
        if t[1] = FALSE then
            warning( t[2]."; assignment ignored" );
            return(o)
        else
            e::options[t[2]]:= t[3];
            e::refreshPlotdata:= TRUE;
            return( extsubsop(o,1=e) )
        end_if
    else
        // slot access:
        e:= extop(o,1);
        if (t:= slot(e,s)) <> FAIL then
            return(t)
        else
            _s:= text2expr(s);
            if not contains(OPTIONNAMES,_s) then
                error("unknown attribute or option '".s."'")
            elif type((t:= e::options[_s])) <> "_index" then
                return(t)
            else
                return(FAIL)
            end_if
        end_if
    end_if
end_proc:

// undocumented!
plot::Curve2d::diff := proc(c)
    local cc;
begin
    if args(0) = 1 then return(c) end_if;

    cc:= plot::Curve2d::copy(c);
    cc::term:= map(cc::term, diff, args(2..args(0)));
    return( cc )
end_proc:

plot::Curve2d::checkOption := proc(opt)
    local value;
begin
    if type(opt) <> "_equal" then
        return( [FALSE,"options must have the form 'OptionName = value'"] )
    end_if;

    value:= op(opt,2); opt:= op(opt,1);
    if not contains( OPTIONNAMES,opt ) then
        return( [FALSE,"unknown option name '".expr2text(opt)."'"] )
    else
        // First, handle special cases for certain options:
        // Color : Beside the standard format 'Color = [{Height,Flat},[r,g,b]]',
        //         allow 'Color = [r,g,b]' which corresponds to '[Flat,[r,g,b]]'.
        case opt
        of hold(Color) do
            if (value:= plot::getColorFormat(value)) = FAIL then
                return( [FALSE,"option 'Color': invalid color specification"] )
            end_if
        end_case;

        return( [TRUE,opt,value] )
    end_if
end_proc:

plot::Curve2d::MuPlotML :=
proc(MuPlotML, obj)
  local options, prfunc, functionData, plotOptions;
begin
  options := obj::options;
  options[hold(Key)] := obj::key;
  plotOptions := MuPlotML::selectOptions(options, {hold(RealValuesOnly), hold(Discont)});
  delete options[RealValuesOnly];
  delete options[Discont];
  delete options[Grid];

  functionData :=
  proc(d)
    local i, j, parts, opts, u, xu, yu, lbu, ubu, nu, nsu, cfun,
          du, x, y, pr, ticks, range;
  begin
    u:= op(obj, [2,1]);               // parameter u
    xu:= op(obj, [1,1]);              // x(u)
    yu:= op(obj, [1,2]);              // y(u)
    lbu:= float(op(obj, [2,2,1]));    // lower bound of u
    ubu:= float(op(obj, [2,2,2]));           // upper bound of u
    opts:= obj::options;
    nu:= if contains(opts, Grid) then op(opts[Grid],1) else 20 end;
    nsu:= if contains(opts, Smoothness) then op(opts[Smoothness],1) else 0 end;
    cfun:= if contains(opts, Color) and op(opts[Color], 1) = hold(Function)
             then op(opts[Color], 2) else FAIL end;
    parts := plotlib::check_args(op(plotOptions), [Mode=Curve, [xu, yu], u=[lbu, ubu]]);

    if domtype(xu) <> DOM_PROC then xu:= fp::unapply(xu, u) end;
    if domtype(yu) <> DOM_PROC then yu:= fp::unapply(yu, u) end;

    ticks := op(parts, 1..2);
/*    print();
    print();
    print("functionData:", ticks, nops(parts));
    print();
    print(); */
    for j from 3 to nops(parts) do
      range := op(parts, [j, 3, 2]);
      lbu := op(range, 1);
      ubu := op(range, 2);
      nu := op(parts, [j, 4, 2,1]);
/*   print();
    print();
    print("functionData:", op(parts, j));
    print();
    print(); */
      if j > 3 then MuPlotML::newline(); end;
      d::printElement("FunctionData",
                    table(hold(Bounds) =
                          expr2text(lbu, ubu),
                          hold(GridNum) = nu,
                          hold(Smoothness) = nsu),
                    d -> (
    nu:= if contains(opts, Grid) then op(opts[Grid],1) else 20 end;
                          nu:= nu - 1;
                          nu:= nu + nu*nsu;
                          du:= (ubu - lbu)/nu;
                          pr := d::pr;
                          if cfun = FAIL then
                            for i from 1 to nu do
                              pr("<P2>", float(xu(lbu)), ",",
                                         float(yu(lbu)), "</P2>");
                              lbu:= lbu+du;
                              pr := d::newline
                            end_for;
                          else
                            for i from 1 to nu do
                              x:= float(xu(lbu));
                              y:= float(yu(lbu));
                              pr("<P2 C=", d::colStr(cfun(x, y, lbu)),
                                         ">", x, ",", y, "</P2>");
                              lbu:= lbu+du;
                              pr := d::newline
                            end_for
                          end_if;
                          // calc last point exactly at ubu
                          x:= float(xu(ubu));
                          y:= float(yu(ubu));
                          d::newline("<P2",
                               (if cfun <> FAIL then
                                  (" C=", d::colStr(cfun(x, y, ubu)))
                                else null() end),
                               ">", x, ",", y, "</P2>")
                    )):
      end_for
    end_proc:
  
  prfunc := d -> d::pr(op(obj));
  
  MuPlotML::printElement("Curve2d", options, functionData):
end:

OPTIONNAMES:= {
  // options for 2-dimensional objects:
  hold(Color),hold(LineStyle),hold(LineWidth),hold(PointStyle),
  hold(PointWidth),hold(Title),hold(TitlePosition),
  // options for 2d-curves:
  hold(Grid),hold(Smoothness),hold(Style),
  // options for plot::Curve2d:
  hold(Discont),hold(RealValuesOnly)
}:

DEFAULTOPTIONS:= table():
// Default options for objects of the domain plot::Curve2d:
DEFAULTOPTIONS[hold(Color)]         := [hold(Flat),RGB::Red]: // needed to give every part of a curve graph the same color!
DEFAULTOPTIONS[hold(Grid)]          := [100]:
DEFAULTOPTIONS[hold(Discont)]       := TRUE:
DEFAULTOPTIONS[hold(RealValuesOnly)]:= TRUE:

