// $Date: 2003/01/28 15:44:24 $  $Author: olli $   $Revision: 1.36.2.1 $
// igel
// completely rewritten by frankp (21.09.2000)

/* -----------------------------------------------------------
    Function2d -- The graphical primitive for two-dimensional 
                  function plots

    Syntax:

    Function2d(f_x, x = xr <, y = yr > <, op1, op2, ...>)

    f_x             : arithmetical expression (depending on x)
    x, y            : DOM_IDENT
    xr, yr          : either form 'a..b' or list of two arithmetical expressions.
    op1, op2, ...   : plot2d options, see plot2d and plot3d for details.

    Example:

    >> plot::Function2d( sin(t), t = 0..2*PI )

    >> plot::Function2d( sin(t), t = 0..2*PI, Color = RGB::Blue )

----------------------------------------------------------------*/
alias( DEFAULTOPTIONS = plot::Function2d::defaultOptions ):
alias( OPTIONNAMES = plot::Function2d::optionNames ):

plot::Function2d := newDomain("plot::Function2d"):
plot::Function2d::create_dom:=hold(plot::Function2d):
plot::Function2d::info:= "plot::Function2d -- graphical primitive for two-dimensional function graphs":
plot::Function2d::type:= "graphprim":
plot::Function2d::dimension:= () -> 2:

plot::Function2d::new:= proc(f, x, r)
    local i, y, options, opt, function;
begin
    case args(0) 
    of 0 do
        error("no argument given");
    of 1 do
        error("expecting at least two arguments");
    of 2 do
        if not (testtype(f, Type::Arithmetical) or testtype(f, piecewise)) then
            error("1st argument: expecting an arithmetical expression");
        end_if:
        if (x:= plot::conv2range(x)) = FAIL then
            error("2nd argument: expecting form 'ident = a..b'");
        end_if;
        y:= hold(Automatic);
        opt:= []; // user specified options
        break
    otherwise
        if not (testtype(f, Type::Arithmetical) or testtype(f, piecewise)) then
            error("1st argument: expecting an arithmetical expression");
        end_if:
        if (x:= plot::conv2range(x)) = FAIL then
            error("2nd argument: expecting form 'ident = a..b'");
        end_if;
        if (y:= plot::conv2range(r)) = FAIL then
            y:= hold(Automatic);
            opt:= [args(3..args(0))]  // user specified options
        elif op(y,1) = op(x,1) then
            error("range for ".expr2text(op(x,1))." specified twice")
        else
            y:= op(y,2); // keep only the interval, not the parameter
            opt:= [args(4..args(0))]  // user specified options
        end_if
    end_case;

    options := DEFAULTOPTIONS; // default options (a table)
    // check user options and override the default options:
    for i from 1 to nops(opt) do
        if (r:= plot::Function2d::checkOption(opt[i]))[1] = FALSE then
            error(r[2])
        else
            options[r[2]]:= r[3]
        end_if
    end_for;
                
    function:= newDomain( expr2text(genident("Function2d")) );                
 
    // attributes:
    function::term           := f;
    function::range          := x;
    function::yrange         := y;
    function::refreshPlotdata:= TRUE;
    function::plotdata       := [];
    function::options        := options;
    function::objects        := [];

    return( new(plot::Function2d, function) )
end_proc:

plot::Function2d::copy := proc(f)
begin
    return(extsubsop(f, 1=newDomain(genident("Function2d"), extop(f, 1))));
end_proc:

// undocumented!
plot::Function2d::diff := proc(f)
    local c;
begin
    if args(0) = 1 then return(f) end_if;

    c:= plot::Function2d::copy(f);
    c::term := diff(c::term, args(2..args(0)));
    c::refreshPlotdata:= TRUE;
    return( c )
end_proc:

plot::Function2d::modify := proc(f)
    local o, m, options;
begin
    m:= plot::Function2d::copy(f);
    options:= [args(2..args(0))];

    ( slot(m, expr2text(op(o, 1)), op(o, 2)) ) $ o in options;
    
    m::refreshPlotdata:= TRUE;
    return(m)
end_proc:

plot::Function2d::getPlotdata := proc(o)
    local e;
begin
    e:= extop(o,1);
    if e::refreshPlotdata = FALSE then return( e::plotdata ) end_if;

    plot::Function2d::buildObjects(o);
    e::refreshPlotdata:= FALSE;
    e::plotdata:= map(e::objects, oo -> op(plot::Function2d::buildPlotdataObject(oo)));
end_proc:

plot::Function2d::buildObjects := proc(f)
    local col, e, x, y, c, X, Y, tmp, vBox, intervals, objects, discont, realValuesOnly, grid;
begin
    e:= extop(f,1);
    x:= plot::range2list(e::range);
    c:= [op(x,1),e::term];

    // default values for Discont and RealValuesOnly:
    if type((discont:= e::options[hold(Discont)])) = "_index" then discont:= TRUE end_if;
    if type((realValuesOnly:= e::options[hold(RealValuesOnly)])) = "_index" then realValuesOnly:= TRUE end_if;
    if type(e::options[hold(Color)]) = "_index" then 
        col := null()
    else
       col := Color = e::options[hold(Color)]
    end_if;

    intervals:= [];
    y:= e::yrange;
    if y = hold(Automatic) then
         tmp:= stdlib::ShowPlotWarning;
         sysassign(stdlib::ShowPlotWarning, FALSE);

         intervals := [plotlib::check_args(
             hold(Discont) = discont, hold(RealValuesOnly) = realValuesOnly, 
             [hold(Mode) = hold(Curve), c, x, col]
         )];

         sysassign(stdlib::ShowPlotWarning, tmp);

         // Result is of the form [sceneoptions, [Mode=..., ...]<, [Mode=...,...] ...>].
         // Select only those sublists with 'Mode=Curve':
         intervals:= select( intervals,l -> domtype(l) = DOM_LIST and l[1] = hold(Mode = Curve) );
         if nops(intervals) > 0 and (col := select(intervals[1], has, Color)) <> [] then
            col := col[1]
         end_if;
         // Select the entries 'x=[a,b]', i.e., the third element of these lists:
         intervals:= map( intervals, l -> l[3] );
     else
         // extract the grid value for the clipping routine:
         if type((grid:= e::options[hold(Grid)])) <> "_index" then 
             grid:= hold(Grid) = grid
         else
             grid:= null()
         end_if;

         X := float(op(x, 2)); // range: x = [a,b]
         Y := float([op(y)]);  // yrange: c..d -> [float(c),float(d)]
         // arrange box for clip2d:
         vBox := [X[1], Y[1], X[2], Y[2]];
         intervals := [plotlib::clip2d(hold(Curve), [hold(Mode) = hold(Curve), c, x, grid, col], 
            vBox, 
            1,              // object number
            TRUE,           // clip
            realValuesOnly, // RealValuesOnly
            discont,        // Discont
            [FALSE,FALSE]   // LogXY
         )];
         if nops(intervals) > 0 and (col := select(intervals[1], has, Color)) <> [] then
            col := col[1]
         end_if;
         // select the entries 'x=[a,b]', i.e., the 3rd element of these lists:
         intervals:= map( intervals, l -> l[3] )
    end_if:

    if nops(intervals) > 0 then
        if col <> [] then
           objects := [plot::Function2d::modify(f, hold(range) = intervals[1], col)];
	else
	   objects := [plot::Function2d::modify(f, hold(range) = intervals[1])];
	end_if;
        delete intervals[1];
        for x in intervals do
            tmp := plot::Function2d::modify(f, hold(range) = x, hold(Title) = "");
            if col <> [] then
               tmp := plot::Function2d::modify(tmp, col);
	    end_if;
            objects := append( objects, tmp );
        end_for;
    else
        objects := [f]
    end_if;

    e::objects:= objects
end_proc:

plot::Function2d::buildPlotdataObject := proc(f)
    local e, r, c, options;
begin
    e:= extop(f, 1);
    if e::refreshPlotdata = FALSE then return( e::plotdata ) end_if;

    c:= e::term;
    r:= plot::range2list(e::range);

    // remove specific Function2d options and options with value NIL:
    options:= e::options;
    delete options[hold(RealValuesOnly)],options[hold(Discont)];

    // create the plot data:
    e::plotdata:= [
        [hold(Mode) = hold(Curve), [op(r,1),c], r] . [op(options)]
    ]
end_proc:

plot::Function2d::print:= proc(f)
begin
    f:= extop(f,1);
    hold(plot::Function2d)(f::term,f::range)
end_proc:

plot::Function2d::op := proc(c, n)
begin
    if args(0) < 1 or args(0) > 2 then
        error("invalid number of arguments")
    end_if;

    c:= extop(c,1);
    c:= [c::term,c::range,c::yrange];
    if args(0) = 1 then op(c) else op(c, n) end_if
end_proc:

plot::Function2d::nops:= c -> 3:

plot::Function2d::_index:= proc(c:plot::Function2d,n:Type::PosInt)
begin
    if args(0) <> 2 then
        error("expecting two arguments")
    end_if;

    c:= extop(c,1);
    c:= [c::term,c::range,c::yrange];
    if n > nops(c) then
        error("index out of range")
    else
        return( c[n] )
    end_if
end_proc:

// Note: Do not return an error message in the case of an invalid call,
//       because then the object would be destroyed! Hence a warning 
//       message is issued, and the call of this method has no effect 
//       on the object.
plot::Function2d::set_index:= proc(c:plot::Function2d, n:Type::PosInt, v)
    local e;
begin
    if args(0) <> 3 then
        warning("wrong number of arguments; assignment ignored");
        return(c)
    end_if; 

    e:= extop(c,1);
    case n
    of 1 do
        if v <> e::term then 
            if not testtype(v,Type::Arithmetical) then
                warning("3rd argument: expecting an arithmetical expression; assignment ignored");
                return(c)
            end_if;
            e::refreshPlotdata:= TRUE;
            e::term:= v;
        end_if;
        return( extsubsop(c,1=e) )
    of 2 do
        if v <> e::range then 
            if (v:= plot::conv2range(v)) = FAIL then
                warning("3rd argument: expecting form 'ident = a..b'; assignment ignored");
                return(c)
            end_if;            
            e::refreshPlotdata:= TRUE;
            e::range:= v;
        end_if;
        return( extsubsop(c,1=e) )
    of 3 do
        if v <> e::yrange then 
            if v <> hold(Automatic) and not testtype(v,Type::Arithmetical..Type::Arithmetical) then
                if not testtype(v,Type::ListOf(Type::Arithmetical,2,2)) then
                    warning("3rd argument: expecting form 'ymin..ymax'; assignment ignored");
                    return(c)
                else
                    // convert range into the form ymin..ymax:
                    v:= op(v,1)..op(v,2)
                end_if
            end_if;   
            e::refreshPlotdata:= TRUE;
            e::yrange:= v;
        end_if;
        return( extsubsop(c,1=e) )
    otherwise
        warning("index out of range; assignment ignored");
        return(c)
    end_case
end_proc:

// Note: Do not return an error message in the case of an invalid call,
//       because then the object would be destroyed! Hence a warning 
//       message is issued, and the call of this method has no effect 
//       on the object.
plot::Function2d::slot:= proc(o,s,v=null())
    local e, t, _s;
begin
    if args(0) = 3 then
        // slot assignment:
        e:= extop(o,1);

        // first handle all user-accessable attributes:
        case s 
        of "refreshPlotdata" do
            if v <> TRUE and v <> FALSE then
                warning("attribute 'refreshPlotdata': expecting either 'TRUE' or 'FALSE', assignment ignored");
                return( o )
            else
                e::refreshPlotdata:= v;
                return( extsubsop(o,1=e) )
            end_if
        of "range" do
            if (v:= plot::conv2range(v)) = FAIL then
                warning("attribute 'range': expecting form 'x = a..b'; assignment ignored");
                return( o )
            else
                e::range:= v;
                e::refreshPlotdata:= TRUE;
                return( extsubsop(o,1=e) )
            end_if
        of "yrange" do
            if v <> hold(Automatic) and not testtype(v,Type::Arithmetical..Type::Arithmetical) then
                if not testtype(v,Type::ListOf(Type::Arithmetical,2,2)) then
                    warning("attribute 'yrange': expecting either 'Automatic', or the form 'c..d'; assignment ignored");
                    return( o )
                else
                    // convert range into the form ymin..ymax:
                    v:= op(v,1)..op(v,2)
                end_if
            end_if;   

            e::yrange:= v;
            e::refreshPlotdata:= TRUE;
            return( extsubsop(o,1=e) )
        of "term" do
            if not testtype(v,Type::Arithmetical) then
                warning("attribute 'term': expecting an arithmetical expression; assignment ignored");
                return( o )
            else
                e::term:= v;
                e::refreshPlotdata:= TRUE;
                return( extsubsop(o,1=e) )
            end_if
        of "options" do
            if domtype(v) <> DOM_TABLE then
                warning("attribute 'options': expecting a table of plot options; assignment ignored");
                return( o )
            else
                e::options:= v;
                e::refreshPlotdata:= TRUE;
                return( extsubsop(o,1=e) )
            end_if
        end_case;

        // an option given?
        t:= plot::Function2d::checkOption(text2expr(s) = v);
        if t[1] = FALSE then
            warning( t[2]."; assignment ignored" );
            return(o)
        else
            e::options[t[2]]:= t[3];
            e::refreshPlotdata:= TRUE;
            return( extsubsop(o,1=e) )
        end_if
    else
        // slot access:
        e:= extop(o,1);
        if (t:= slot(e,s)) <> FAIL then
            return(t)
        else
            _s:= text2expr(s);
            if not contains(OPTIONNAMES,_s) then
                error("unknown attribute or option '".s."'")
            elif type((t:= e::options[_s])) <> "_index" then
                return(t)
            else
                return(FAIL)
            end_if
        end_if
    end_if
end_proc:

plot::Function2d::checkOption := proc(opt)
    local value;
begin
    if type(opt) <> "_equal" then
        return( [FALSE,"options must have the form 'OptionName = value'"] )
    end_if;

    value:= op(opt,2); opt:= op(opt,1);
    if not contains( OPTIONNAMES,opt ) then
        return( [FALSE,"unknown option name '".expr2text(opt)."'"] )
    else
        // First, handle special cases for certain options:
        // Color : Beside the standard format 'Color = [{Height,Flat},[r,g,b]]',
        //         allow 'Color = [r,g,b]' which corresponds to '[Flat,[r,g,b]]'.
        case opt
        of hold(Color) do
            if (value:= plot::getColorFormat(value)) = FAIL then
                return( [FALSE,"option 'Color': invalid color specification"] )
            end_if
        end_case;

        return( [TRUE,opt,value] )
    end_if
end_proc:


plot::Function2d::MuPlotML :=
proc(MuPlotML, obj)
  local options, prfunc, functionData, plotOptions;
begin
  options := obj::options;
  options[hold(Key)] := obj::key;
  plotOptions := MuPlotML::selectOptions(options, {hold(RealValuesOnly), hold(Discont)});
  delete options[RealValuesOnly];
  delete options[Discont];
  delete options[Grid];

  functionData :=
  proc(d)
    local i, j, parts, opts, u, xu, yu, lbu, ubu, nu, nsu, cfun,
          du, x, y, pr, ticks, range, ymin, ymax;
  begin
    u:= op(obj, [2,1]);               // parameter u
    xu:= u;              // x(u)
    yu:= op(obj, 1);              // y(u)
    lbu:= float(op(obj, [2,2,1]));    // lower bound of u
    ubu:= float(op(obj, [2,2,2]));           // upper bound of u
    ymin:= float(op(obj, [3,1]));           // lower bound of y
    ymax:= float(op(obj, [3,2]));           // upper bound of y
    if domtype(ymin) <> DOM_FLOAT then ymin:=-1000000.0: ymax:=1000000.0 end:
    opts:= obj::options;
    nu:= if contains(opts, Grid) then op(opts[Grid],1) else 20 end;
    nsu:= if contains(opts, Smoothness) then op(opts[Smoothness],1) else 0 end;
    cfun:= if contains(opts, Color) and op(opts[Color], 1) = hold(Function)
             then op(opts[Color], 2) else FAIL end;
    parts := plotlib::check_args(op(plotOptions), [Mode=Curve, [xu, yu], xu=[lbu, ubu]]);

    if domtype(xu) <> DOM_PROC then xu:= fp::unapply(xu, u) end;
    if domtype(yu) <> DOM_PROC then yu:= fp::unapply(yu, u) end;

    ticks := op(parts, 1..2);
/*    print();
    print();
    print("functionData:", ticks, nops(parts));
    print();
    print(); */
    for j from 3 to nops(parts) do
      range := op(parts, [j, 3, 2]);
      lbu := op(range, 1);
      ubu := op(range, 2);
      nu := op(parts, [j, 4, 2,1]);
/*   print();
    print();
    print("functionData:", op(parts, j));
    print();
    print(); */
      if j > 3 then MuPlotML::newline(); end;
      d::printElement("FunctionData",
                    table(hold(Bounds) =
                          expr2text(lbu, ubu),
                          hold(GridNum) = nu,
                          hold(Smoothness) = nsu),
                    d -> (
    nu:= if contains(opts, Grid) then op(opts[Grid],1) else 20 end;
                          nu:= nu - 1;
                          nu:= nu + nu*nsu;
                          du:= (ubu - lbu)/nu;
                          pr := d::pr;
                          if cfun = FAIL then
                            for i from 1 to nu do
			      y:= float(yu(lbu)):
			      if y >= ymin and y<=ymax then
                              pr("<P2>", float(xu(lbu)), ",",
                                         y, "</P2>");
                              end:
			      lbu:= lbu+du;
                              pr := d::newline
			      
                            end_for;
                          else
                            for i from 1 to nu do
                              x:= float(xu(lbu));
                              y:= float(yu(lbu));
                              pr("<P2 C=", d::colStr(cfun(x, y, lbu)),
                                         ">", x, ",", y, "</P2>");
                              lbu:= lbu+du;
                              pr := d::newline
                            end_for
                          end_if;
                          // calc last point exactly at ubu
                          x:= float(xu(ubu));
                          y:= float(yu(ubu));
                          if y<ymax and y>ymin then
			  d::newline("<P2",
                               (if cfun <> FAIL then
                                  (" C=", d::colStr(cfun(x, y, ubu)))
                                else null() end),
                               ">", x, ",", y, "</P2>") end:
                    )):
      end_for
    end_proc:
  
  prfunc := d -> d::pr(op(obj));
  
  MuPlotML::printElement("Function", options, functionData):
end:


OPTIONNAMES:= {
  // options for 2-dimensional objects:
  hold(Color),hold(LineStyle),hold(LineWidth),hold(PointStyle),
  hold(PointWidth),hold(Title),hold(TitlePosition),
  // options for 2d-curves:
  hold(Grid),hold(Smoothness),hold(Style),
  // options for plot::Function2d:
  hold(Discont),hold(RealValuesOnly)
}:

DEFAULTOPTIONS:= table():
// Default options for objects of the domain plot::Function2d:
DEFAULTOPTIONS[hold(Color)]         := [hold(Flat),RGB::Red]: // needed to give every part of a function graph the same color!
DEFAULTOPTIONS[hold(Grid)]          := [100]:
DEFAULTOPTIONS[hold(Discont)]       := TRUE:
DEFAULTOPTIONS[hold(RealValuesOnly)]:= TRUE:

