// $Date: 2003/01/28 15:44:24 $  $Author: olli $   $Revision: 1.52.2.1 $
// igel
// completely rewritten by frankp (21.09.2000)

/* -----------------------------------------------------------
    Scene -- the graphical primitive for two- and three-dimensional graphical scenes

    Syntax:

    Scene(o1 <, o2, ...> <, op1, op2, ...>)

    o1, o2, ...  : graphical objects 
    op1, op2, ...: plot2d or plot3d options, see plot2d and plot3d 
                   for details.

    Example:
    >> plot::Scene( plot::Point(1,2), plot::Curve2d( [sin(x), cos(x)], x = 0..2*PI ) )

    >> plot::Scene( 
         plot::Point(1,2), plot::Curve2d( [sin(x), cos(x)], x = 0..2*PI ), 
         Scaling = Constrained 
       )
----------------------------------------------------------------*/ 
alias( DEFAULTOPTIONS2D = plot::Scene::defaultOptions2d ):
alias( OPTIONNAMES2D = plot::Scene::optionNames2d ):
alias( DEFAULTOPTIONS3D = plot::Scene::defaultOptions3d ):
alias( OPTIONNAMES3D = plot::Scene::optionNames3d ):

plot::Scene:= newDomain("plot::Scene"):
plot::Scene::create_dom:=hold(plot::Scene):
plot::Scene::info:= "plot::Scene -- graphical primitive for graphical scenes":
plot::Scene::dimension:= s -> (extop(s,1))::dimension:

plot::Scene::new:= proc()
    local o, i, opt, objects, options, scene, dimen;
begin
    if args(0) = 0 then
        error("no argument given");
    end_if;

    objects:= split( [args()],x -> type(x) = "graphprim" );
    opt:= objects[2];

    if (objects:= objects[1]) = [] then
        error("expecting a graphical primitive, i.e., an object of type \"graphprim\"")
    end_if;

    dimen:= {op(map(objects, e -> e::dom::dimension(e)))};
    if nops(dimen) > 1 then
        error("objects have different dimensions");
    else 
        dimen:= op(dimen,1)
    end_if;
    
    // default options (a table)
    if dimen = 2 then options := DEFAULTOPTIONS2D else options := DEFAULTOPTIONS3D end_if;

    //----------------------------------------------------
    // Comment added by Walter, 16.2.02
    // Some objects such as Function2d carry around
    // scene-options  (e.g., Discont = TRUE/FALSE or
    // RealValuesOnly = TRUE/FALSE).
    // One should allow the objects to overwrite plot::Scene::defaultOptions2d/3d
    // by passing these scene options to opt (opt contains the scene
    // options that the user passed to Scene::new).
    // In the following code, only the 'object options' 
    // Discont and RealValuesOnly are interpreted as
    // 'scene options'. However, objects should be able to
    // specify preferred scene options of any kind..
    // E.g., piechart etc. should be able to specify
    // viewing parameters (otherwise, scene objects must
    // be returned rather than graphpims).
    //
    // Presently, I do not activate the code below because setting the
    // scene option Discont = TRUE would double the runtime
    // (Note, that Function2d uses plotlib::checkargs(Discont = TRUE)
    // etc. to prepare the plot data. Then these data are passed
    // plot2d with Discont = FALSE to avoid the overhead of using
    // Dicont = TRUE again.

    /* --------------------
    for obj in objects do
        objopts:= obj::options:
        for objopt in objopts do
            if has(objopt, Discont) or
               has(objopt, RealValuesOnly) 
            then
               opt:= append(opt, objopt);
            end_if;
        end_for:
    end_for;
    -------------------- */
    //----------------------------------------------------

    // check user specified options and override the default options:
    for i from 1 to nops(opt) do
        if dimen = 2 then
            if (o:= plot::Scene::checkOption2d(opt[i]))[1] = FALSE then
                error(o[2])
            else
                options[o[2]]:= o[3]
            end_if
        elif (o:= plot::Scene::checkOption3d(opt[i]))[1] = FALSE then
            error(o[2])
        else
            options[o[2]]:= o[3]
        end_if;
    end_for;

    scene:= newDomain( expr2text(genident("Scene")) );

    // attributes:
    scene::plotdata       := [];
    scene::options        := options;
    scene::objects        := objects;
    scene::dimension      := dimen;

    return( new(plot::Scene, scene) )
end_proc:

plot::Scene::copy := proc(s)
    local l, o;
begin
    o:= extop(s,1);
    l:= map( o::objects, x -> x::dom::copy(x) );
    o:= newDomain(genident("Scene"), o);
    o::objects:= l;
    return( extsubsop(s, 1=o) )
end_proc:

plot::Scene::modify := proc(s)
    local o, m, options;
begin
    m:= plot::Scene::copy(s);
    options:= [args(2..args(0))];

    ( slot(m, expr2text(op(o, 1)), op(o, 2)) ) $ o in options;
    
    return(m)
end_proc:

/*--
    Note: getPlotdata of graphical primitives must
    return the list [[Mode=...,...]<, [Mode=...]]
--*/
plot::Scene::getPlotdata := proc(s)
    local e, plotdata;
begin
    e:= extop(s, 1);
    plotdata:= map( e::objects, o -> op(o::dom::getPlotdata(o)) );
    if plotdata = [] then 
        // the empty plot
        e::plotdata:= []
    else
        e::plotdata:= [ op(e::options),op(plotdata) ]
    end_if
end_proc:

plot::Scene::print := () -> hold(plot::Scene)():

plot::Scene::expose := s -> hold(plot::Scene)(op(map((extop(s,1))::objects,expose))):
        
plot::Scene::op := proc(s, n)
begin
    if args(0) < 1 or args(0) > 2 then
        error("invalid number of arguments")
    end_if;

    s:= (extop(s,1))::objects;
    if args(0) = 1 then op(s) else op(s, n) end_if
end_proc:

plot::Scene::nops := s -> nops(s::objects):

plot::Scene::_index := proc(s:plot::Scene,n:Type::PosInt)
begin
    if args(0) <> 2 then
        error("expecting two arguments")
    end_if;
    s:= (extop(s,1))::objects;
    if n > nops(s) then
        error("index out of range")
    else
        return( s[n] )
    end_if
end_proc:

// Note: Do not return an error message in the case of an invalid call,
//       because then the object would be destroyed! Hence a warning 
//       message is issued, and the call of this method has no effect 
//       on the object.
plot::Scene::set_index := proc(s:plot::Scene, n:Type::PosInt, v)
    local l;
begin
    if args(0) <> 3 then
        warning("wrong number of arguments; assignment ignored");
        return(s)
    elif type(v) <> "graphprim" then
        warning("3rd argument: expecting an object of type \"graphprim\"; assignment ignored");
        return(s)
    end_if; 
 
    l:= (extop(s,1))::objects;
    if n > nops(l) then
        warning("index out of range; assignment ignored");
        return(s)
    else
        s::objects:= subsop( l,n=v );
        return( s )
    end_if
end_proc:

// Note: Do not return an error message in the case of an invalid call,
//       because then the object would be destroyed! Hence a warning 
//       message is issued, and the call of this method has no effect 
//       on the object.
plot::Scene::slot:= proc(o,s,v=null())
    local e, t, _s;
begin
    if args(0) = 3 then
        // slot assignment:
        e:= extop(o,1);

        // first handle all user-accessable attributes:
        case s 
        of "objects" do
            if domtype(v) <> DOM_LIST and map({op(v)},type) <> {"graphprim"} then
                warning("attribute 'objects': expecting a list of objects of type \"graphprim\"; assignment ignored");
                return( o )
            elif map({op(v)}, x -> x::dom::dimension(x)) <> {e::dimension} then
                warning("attribute 'objects': objects have invalid dimensions; assignment ignored");
                return( o )
            else
                e::objects:= v;
                return( extsubsop(o,1=e) )
            end_if
        of "options" do
            if domtype(v) <> DOM_TABLE then
                warning("attribute 'options': expecting a table of plot options; assignment ignored");
                return( o )
            else
                e::options:= v;
                return( extsubsop(o,1=e) )
            end_if
        end_case;

        // an option given?
        if e::dimension = 2 then 
            t:= plot::Scene::checkOption2d(text2expr(s) = v);
        else
            t:= plot::Scene::checkOption3d(text2expr(s) = v);
        end_if;
        if t[1] = FALSE then
            warning( t[2]."; assignment ignored" );
            return(o)
        else
            e::options[t[2]]:= t[3];
            return( extsubsop(o,1=e) )
        end_if
    else
        // slot access:
        e:= extop(o,1);
        if (t:= slot(e,s)) <> FAIL then
            return(t)
        else
            _s:= text2expr(s);
            if e::dimension = 2 then
                if not contains(OPTIONNAMES2D,_s) then
                    error("unknown attribute or option '".s."'")
                end_if
            elif not contains(OPTIONNAMES3D,_s) then
                error("unknown attribute or option '".s."'")
            end_if;

            if type((t:= e::options[_s])) <> "_index" then
                return(t)
            else
                return(FAIL)
            end_if
        end_if
    end_if
end_proc:

plot::Scene::checkOption2d := proc(opt)
    local value;
begin
    if type(opt) <> "_equal" then
        return( [FALSE,"options must have the form 'OptionName = value'"] )
    end_if;

    value:= op(opt,2); opt:= op(opt,1);
    if not contains( OPTIONNAMES2D,opt ) then
        return( [FALSE,"unknown option name '".opt."'"] )
    else
        // First, handle special cases for certain options:
        // Color : Beside the standard format 'Color = [{Height,Flat},[r,g,b]]',
        //         allow 'Color = [r,g,b]' which corresponds to '[Flat,[r,g,b]]'.
        if opt = hold(Color) then
            if (value:= plot::getColorFormat(value)) = FAIL then
                return( [FALSE,"option 'Color': invalid color specification"] )
            end_if;
        end_if;

        return( [TRUE,opt,value] )
    end_if
end_proc:

plot::Scene::checkOption3d := proc(opt)
    local value;
begin
    if type(opt) <> "_equal" then
        return( [FALSE,"options must have the form 'option = value'"] )
    end_if;

    value:= op(opt,2); opt:= op(opt,1);
    if not contains( OPTIONNAMES3D,opt ) then
        return( [FALSE,"unknown option name '".expr2text(opt)."'"] )
    else
        // First, handle special cases for certain options:
        // Color : Beside the standard format 'Color = [{Height,Flat},[r,g,b]]',
        //         allow 'Color = [r,g,b]' which corresponds to '[Flat,[r,g,b]]'.
        if opt = hold(Color) then
            if (value:= plot::getColorFormat(value)) = FAIL then
                return( [FALSE,"option 'Color': invalid color specification"] )
            end_if;
        end_if;

        return( [TRUE,opt,value] )
    end_if
end_proc:

// 
plot::Scene::MuPlotML :=
proc(MuPlotML, obj)
  local options, diagram, sceneRow, sceneAttr, Frame, scaleRef, 
        curveSprite, functionSprite, theSprite, funcExpression,
	params, paramName;
begin
  options := obj::options;
  options[hold(Key)] := obj::key;
  
  if obj::dom::dimension(obj) = 2 then
    // there exist no MuPAD representations for Scene2d and SceneRow,
    // so we fake them here

   hold(Labeling), hold(Labels), 
   hold(ViewingBox):

	if op(obj::ViewingBox,[1,1]) <> Automatic then
	MuPlotML::SetXBounds(op(obj::ViewingBox,[1,1]),op(obj::ViewingBox,[1,2])):
	end;

	if op(obj::ViewingBox,[2,1]) <> Automatic then
	MuPlotML::SetYBounds(op(obj::ViewingBox,[2,1]),op(obj::ViewingBox,[2,2])):
	end;

    // select the Scene2d attributes
    sceneAttr := MuPlotML::selectOptions(options,
                                         {hold(Key), hold(LineStyle), hold(LineWidth),hold(ForeGround), hold(BackGround), hold(FontFamily),
   hold(FontSize), hold(FontStyle), hold(PointStyle), hold(PointWidth),hold(Ticks), hold(Axes), hold(GridLinesWidth), hold(GridLinesColor), hold(GridLinesStyle),hold(Arrows), hold(Scaling)});

    options := MuPlotML::selectOptions(options,
                                       {hold(Title), hold(TitlePosition)}):

    scaleRef := proc(i) option escape;
                begin
                  muplot -> muplot::printElement("ScaleRef",
                                                 table(hold(ScaleId)=i))
                end;
         
	paramName := proc(d,i)
		local nam, varnum, tnam;
                begin
		nam:=expr2text(op(d::paramlist,[i,1])):
		tnam:=stringlib::subs(nam,"&"="&amp;"):
                  d::pr("<Name>".tnam."</Name>"): 
		  d::newline("<Expr>".expr2text(op(d::paramlist,[i,2]))."</Expr>" ):
		  d::newline("<Expr>".expr2text(op(d::paramlist,[i,3]))."</Expr>" ):
		  d::newline("<Expr>".expr2text(op(d::paramlist,[i,4]))."</Expr>" ):
		  if nops(d::paramlist[i])>5 then
		  	varnum:=d::paramlist[i][7]:
		  	d::newline("<Expr>".expr2text(op(d::paramlist,[i,5]))."</Expr>" ):
			d::newline("<Expr>".expr2text(if varnum = 1 then d::umin else d::vmin end)."</Expr>" ):
			d::newline("<Expr>".expr2text(if varnum = 1 then d::umax else d::vmax end)."</Expr>" ):
			d::newline("<Expr>".expr2text(op(d::paramlist,[i,6]))."</Expr>" ):
			d::newline("<Expr>".expr2text(varnum)."</Expr>" ):
		  else
		  d::newline("<Expr>".expr2text(slot(d,nam))."</Expr>" ):	
		  d::newline("<Expr>".op(d::paramlist,[i,5])."</Expr>" ):
		  end_if:
                  null()
                end;

   params :=
   proc(d)
     option escape;
   begin
     proc(muplot)
       local i,v;
     begin
       for i from 1 to nops(d::paramlist) do
        v:= nops(d::paramlist[i]):
	if v > 5 then 
	muplot::printElement("Param",
                table(hold(Bounds)=
                      expr2text(float(op(d::paramlist,[i,2]))).",".
                      expr2text(float(op(d::paramlist,[i,3]))), hold(Plotvar) = d::paramlist[i][7]),
                d->(paramName(d,i)(d))):
	else
	muplot::printElement("Param",
                table(hold(Bounds)=
                      expr2text(float(op(d::paramlist,[i,2]))).",".
                      expr2text(float(op(d::paramlist,[i,3])))),
                d->(paramName(d,i)(d))):
         end_if:
	d::newline():
	 end_for
     end;
   end;

    funcExpression := proc(d) option escape;
    		begin
		if d::paramlist <> FAIL then
			muplot -> muplot::printElement("FunctionExpression",table(),d->(params(d)(d)))
			end:
		end;

    functionSprite := muplot ->
                      muplot::printElement("FunctionSprite", table(),
                                           d -> (funcExpression(d)(d): map([op(obj)], d::new))):
    
    curveSprite := muplot ->
                      muplot::printElement("Curve2dSprite", table(),
                                           d -> map([op(obj)], d::new)):
    
    if(domtype(op(obj,1))=plot::Function2d) then theSprite:= functionSprite else theSprite:=curveSprite end;


    Frame := muplot ->
             muplot::printElement("Frame", table() ,
                                  d -> (scaleRef(1)(d);
                                        d::newline();
                                        scaleRef(2)(d);
                                        d::newline();
                                        functionSprite(d))):
                                  
    diagram := muplot ->
               muplot::printElement("Diagram", options,
                                    d -> (d::scaleFunc1(d,obj);
                                          d::newline();
                                          d::scaleFunc2(d,obj);
                                          d::newline();
                                          if d::isFix <> FAIL then d::pr("<IsFix />") end:
					  Frame(d))):
    
    sceneRow := muplot ->
                muplot::printElement("SceneRow", table(), diagram):

    MuPlotML::printElement("Scene2d", sceneAttr, sceneRow);
  MuPlotML::xbmin:=FAIL:
  MuPlotML::xbmax:=FAIL:
  MuPlotML::ybmin:=FAIL:
  MuPlotML::ybmax:=FAIL:
    
   else
    // 3d-Scenes are not yet supported
    error("Not yet implemented.")
  end
end:


OPTIONNAMES2D:= {
   hold(Arrows), hold(Axes), hold(AxesOrigin), hold(AxesScaling), hold(AxesInFront),
   hold(BackGround), hold(Discont), hold(DiscontLines), hold(DiscontLinesWidth),
   hold(DiscontLinesStyle), hold(DiscontLinesColor), hold(FontFamily),
   hold(FontSize), hold(FontStyle), hold(ForeGround), hold(GridLines),
   hold(GridLinesWidth), hold(GridLinesColor), hold(GridLinesStyle),
   hold(Hatch), hold(Labeling), hold(Labels), hold(LineStyle), hold(LineWidth),
   hold(RealValuesOnly), hold(PlotDevice), hold(PointStyle), hold(PointWidth),
   hold(Scaling), hold(Ticks), hold(Title), hold(TitlePosition),
   hold(ViewingBox)
}:

DEFAULTOPTIONS2D:= table():
// Default options for objects of the domain plot::Scene:
DEFAULTOPTIONS2D[hold(Arrows)]         := FALSE:
DEFAULTOPTIONS2D[hold(Axes)]           := hold(Origin):
DEFAULTOPTIONS2D[hold(AxesOrigin)]     := hold(Automatic):
DEFAULTOPTIONS2D[hold(AxesScaling)]    := [hold(Lin),hold(Lin)]:
DEFAULTOPTIONS2D[hold(AxesInFront)]    := FALSE:
DEFAULTOPTIONS2D[hold(BackGround)]     := RGB::White:
DEFAULTOPTIONS2D[hold(Discont)]        := FALSE:
DEFAULTOPTIONS2D[hold(ForeGround)]     := RGB::Black:
DEFAULTOPTIONS2D[hold(GridLines)]      := hold(None):
DEFAULTOPTIONS2D[hold(GridLinesWidth)] := 5:
DEFAULTOPTIONS2D[hold(GridLinesColor)] := RGB::Gray:
DEFAULTOPTIONS2D[hold(GridLinesStyle)] := hold(DashedLines):
DEFAULTOPTIONS2D[hold(Labeling)]       := TRUE:
DEFAULTOPTIONS2D[hold(LineStyle)]      := hold(SolidLines):
DEFAULTOPTIONS2D[hold(LineWidth)]      := 1:
DEFAULTOPTIONS2D[hold(RealValuesOnly)] := FALSE:
DEFAULTOPTIONS2D[hold(PointStyle)]     := hold(FilledSquares):
DEFAULTOPTIONS2D[hold(PointWidth)]     := 30:
DEFAULTOPTIONS2D[hold(Scaling)]        := hold(UnConstrained):
DEFAULTOPTIONS2D[hold(Ticks)]          := hold(Automatic):
DEFAULTOPTIONS2D[hold(TitlePosition)]  := hold(Above):
DEFAULTOPTIONS2D[hold(ViewingBox)]     := hold(Automatic):

OPTIONNAMES3D:= {
   hold(Arrows), hold(Axes), hold(AxesOrigin), hold(AxesScaling),
   hold(BackGround), hold(CameraPoint), hold(FocalPoint),
   hold(FontFamily), hold(FontSize), hold(FontStyle), hold(ForeGround), 
   hold(Labeling), hold(Labels), hold(LineStyle), hold(LineWidth),
   hold(PointStyle), hold(PointWidth), hold(PlotDevice),
   hold(Scaling), hold(Ticks), hold(Title), hold(TitlePosition),
   hold(ViewingBox)
}:
DEFAULTOPTIONS3D:= table():
// All options for 3-dimensional graphical scenes (a value NIL means that 
// the corresponding option will not be taken into account when building 
// the plotdata):
DEFAULTOPTIONS3D[hold(Arrows)]         := FALSE:
DEFAULTOPTIONS3D[hold(Axes)]           := hold(Box):
DEFAULTOPTIONS3D[hold(AxesOrigin)]     := hold(Automatic):
DEFAULTOPTIONS3D[hold(AxesScaling)]    := [hold(Lin),hold(Lin),hold(Lin)]:
DEFAULTOPTIONS3D[hold(BackGround)]     := RGB::White:
DEFAULTOPTIONS3D[hold(CameraPoint)]    := hold(Automatic):
DEFAULTOPTIONS3D[hold(FocalPoint)]     := hold(Automatic):
DEFAULTOPTIONS3D[hold(ForeGround)]     := RGB::Black:
DEFAULTOPTIONS3D[hold(Labeling)]       := TRUE:
DEFAULTOPTIONS3D[hold(LineStyle)]      := hold(SolidLines):
DEFAULTOPTIONS3D[hold(LineWidth)]      := 1:
DEFAULTOPTIONS3D[hold(PointStyle)]     := hold(FilledSquares):
DEFAULTOPTIONS3D[hold(PointWidth)]     := 30:
DEFAULTOPTIONS3D[hold(Scaling)]        := hold(UnConstrained):
DEFAULTOPTIONS3D[hold(Ticks)]          := hold(Automatic):
DEFAULTOPTIONS3D[hold(TitlePosition)]  := hold(Above):
DEFAULTOPTIONS3D[hold(ViewingBox)]     := hold(Automatic):

