/* -------------------------------------------------------------------
  The function boxplot gives a plot summarizing the data.
  Box plots are quite useful to compare similar data sets. 
  For example, one can quickly compare the monthly mid-day temperature 
  over the period of one year by producing one boxplot per month, 
  and displaying them side by side.
   
   A Boxplot contains a rectangle with the following properties: 
   - a lower bound showing the first quartile,  
   - an upper bound showing the third quartile;  
   - a central line showing the median,        
    - Two lines extending from the central box of maximal length 3/2 the
      interquartile range but not extending past the range of the data; 
   - Outliers, points that lie outside the extent of the previous elements. 

    Call: boxplot(data,<options>)

    data: List of Lists, Matrix or Array(one- or two-dimensional)

    Options: 
      DrawMode = Vertical/Horizontal (Default=Vertical)
      Notched  = TRUE/FALSE/Float between 0 and 0.5 (Default=FALSE)
                 Using the parameter format=notched will create a box plot 
                 with one additional feature. The sides of the box are indented, 
                 or notched, at the median line. This notch has a maximal width 
                 of 3.14 times the interquartile range divided by the square 
                 root of the total weight of the data. The notch is constrained 
                 inside the first and third quartiles.
                               
              -> Notched-box plots can be used to determine if two random samples 
                 were drawn from the same population. Overlapping notches
                 indicate that the data sets have the same distribution.
                 However, overlapping notches are no \emph{rigorous} criterion 
                 that the data sets are indeed identically distributed.

      Shift    = DOM_INT (Default=0)
                 The Plot will be shifted in x-direction.
      LineWidth = DOM_INT (Default=1)
      Merge     = Merges two or three Plots into one diagram, while one list is plotted 
                  against the others in 2- or 3 dimensional diagrams.
      Colors    = DOM_LIST of Type Color (Default=RGB::ColorList)
                  Centers = A list of int-values which specify 
                  the x-centers for each of the boxes. The number 
                  of elements in Centers must be equal to the number of lists.
      Widths    = A list of int-values which specify the widths for each of the boxes. 
                  The numbers of  elements in Widths must be equal to the number of lists.
*/
plot::boxplot:= proc(l)
local big, bigY, bigZ, lineWidth, myLine, myPoint, notch, notchDown, 
      columnIndices, numberOfColumns,
      notchIndent, notchUp, outRange, plotMerge, mergecolor, range, rangeY,
      boxList, checkOptions, color, dim, options, pw, t, v, 
      zaehl, i, j, rangez, small, smallY, smallZ, temp, x, y, z, shift,
      stretch, notchIndentU, centers, widths, outRangeY, outRangeZ, 
      x25, x50, x75, y25, y50, y75, z25, z50, z75;
begin
//-------------------------------------------------------------------------
//      myLine and myPoint are used to draw x/y swapped lines/points
//-------------------------------------------------------------------------
myLine:=proc(tausche, start, ende) //unnormal, start, ende ,opt
  local opt;  
  begin
   opt := args(4..args(0));
      if (tausche=Horizontal) and nops(opt)>0 then 
           return (plot::line([start[2],start[1]],[ende[2],ende[1]], op(opt)) );
      else return (plot::line([start[1],start[2]],[ende[1],ende[2]], op(opt)) );
      end_if;                 
   end_proc;

myPoint:=proc(tausche, start, ende)
   local opt;
   begin      
         opt := args(4..args(0));
      if (tausche=Horizontal) and nops(opt)>0 then 
          return (plot::Point( ende,start, op(opt) )) 
      end_if;
      if (tausche=Horizontal) then 
         return (plot::Point( ende,start )) 
      end_if;
      if nops(opt)>0 then 
         return (plot::Point( start,ende, op(opt) )) 
      end_if;
      return (plot::Point( start,ende))                        
   end_proc;

// -----------------------------------------------
// --- checks, if the options are ok and ---------
// --- extracts them to local variables  ---------
// -----------------------------------------------

checkOptions:= proc()
name boxplot::checkOptions;
local p;
begin
 // Copy one Option to P
 for i from 1 to nops(options) do
   p:= options[i];
   if p = hold(Merge) then
      p:= hold(Merge) = TRUE;
   end_if;
   if p = hold(Notched) then
      p:= hold(Notched) = TRUE;
   end_if;
   if type(p) <> "_equal" then 
      error("specification of the form 'option = value' expected")
   end_if;
     case op(p,1)
     of hold(Colors) do
        color:= op(p,2);
        if not testtype(color, Type::ListOf(DOM_LIST)) then
           //keine ListOfRGBs, sondern nur eine einzelner RGB-Wert:
           color:=[color];         
        end_if;
        // now, color = [RGB1, RGB2, ...]
        if domtype(op(color,1) ) <> DOM_LIST then 
           error("option 'Colors': expecting a single RGB color or a list of RGB colors"):
        end_if;
        //Teste, ob RGB-Spezifikationen ok sind:
        if _or( (nops(op(color, j)) <> 3) $ j=1..nops(color)) then
           error("Illegal RGB color specification");
        end_if;
        break;
     of hold(Centers) do
        centers:= op(p,2);
        if domtype(centers) <> (DOM_LIST)  then
            error("option 'Centers': list with ".expr2text(nops(l))." values expected");
        end_if;
        if nops(centers)<>nops(l) then 
           error("option 'Centers': list with ".expr2text(nops(l))." values expected");
        end_if;
        centers:= map(centers, float);
        break;
     of hold(Widths) do
        widths:= op(p,2);
        if domtype(widths) <> (DOM_LIST)  then
           error("option 'Widths': list with ".expr2text(nops(l))." values expected");
        end_if;
        if nops(widths)<>nops(l) then 
           error("option 'Widths': list with ".expr2text(nops(l))." values expected");
        end_if;
        widths:= map(widths, float);
        for j from 1 to nops(widths) do
          if (widths[j] < 0) then 
             error("option 'Widths': only positive allowed"); 
          end_if;
        end_for:
        break;
     of hold(LineWidth) do
        lineWidth:= op(p,2);
        if (domtype(lineWidth) <> (DOM_INT)) or (lineWidth<0) then
            error("option 'LineWidth': positive integer expected")
        end_if;
        break
     of hold(Shift) do
        shift:= float(op(p,2));
        if domtype(shift) <> DOM_FLOAT then
           error("option 'Shift': real value expected")
        end_if;
        shift:=shift+0.5;
        break
     of hold(Merge) do
        plotMerge:= op(p, 2):
        if not contains ({TRUE, FALSE},plotMerge) then 
           error("option 'Merge': Expected TRUE/FALSE")
        end_if;
        if plotMerge and (not contains ( {2,3},nops(l) )) then 
           error("Need 2 or 3 Lists when Option 'Merge' is checked") 
        end_if;
        break;
     of hold(Notched) do 
        t:= float(op(p,2));
        if domtype(t) <> DOM_FLOAT then
          if not contains ({TRUE, FALSE},t) then 
             error("option 'Notched': Expected TRUE/FALSE or a value between 0 and 0.5")
           else if t then notchIndent:=0.2 end_if; //Default-Value
           end_if;
        else
          if (t > 1/2 or t < 0) then
              error("option 'Notched': Expected TRUE/FALSE or a value between 0 and 0.5") 
          end_if;
          notchIndent:=t;
        end_if;
        break
     of hold(DrawMode) do
        v:=op(p,2);
        if (v <> Vertical and v <> Horizontal) then
          error("option 'DrawMode': 'Vertical' or 'Horizontal' expected")
        end_if;            
        break
     otherwise 
        v:=op(p,1); 
        error("Option '".expr2text(v)."' not known");
     end_case;
  end_for;
  return (TRUE);
 end_proc:

// ------------------------------End Check-----------------------------

   if args(0) < 1 then
      error("No data provided")
   end_if;

   //-----------------------------
   // set defaults for the options
   //-----------------------------
   color := [op(RGB::ColorList, 1..10)]; // default color list
   lineWidth:=12;
   shift:=0.5;
   centers:=[];
   widths:=[];
   stretch:=0.5;
   plotMerge:=FALSE;
   t := []; // Titles
   v := Vertical; // DrawMode=Vertical
   pw := 40; // PointWidth
   notchIndent:=0; // Zwischen 0 und 0.5
   notchIndentU:=0;

   //--------------------------------------------
   // Check the first argument:
   //--------------------------------------------
   if not (
      testtype(l, Type::ListOf(DOM_LIST)) or
      domtype(l) = DOM_ARRAY or
      domtype(l) = stats::sample or
      domtype(l) = DOM_LIST  or
      l::dom::hasProp(Cat::Matrix) = TRUE
      ) then
       error("1st argument: list or list of lists, array, matrix, or stats::sample expected")
   end_if;

   //--------------------------------------------
   // process boxplot(List1, options): 
   // convert to boxplot([List1], options)
   //--------------------------------------------
   if domtype(l) = DOM_LIST and
      domtype(l[1])<>DOM_LIST then 
      l:=[l];
   end_if;

   //--------------------------------------------
   // process boxplot(List1, List2, .., options): 
   // convert to boxplot([List1, List2, ..], options)
   //--------------------------------------------
   options:= [];
   if domtype(l) <> stats::sample then
     // stats::sample is special, because column indices
     // may turn up in the arguments
     for i from 2 to args(0) do
        if type(args(i)) = "_equal" or
                args(i) = hold(Merge) or
                args(i) = hold(Notched) then
           // first option found
           options:= [args(i .. args(0))];
           break;
         end_if;
         if domtype(args(i)) = DOM_LIST then
             l:= append(l, args(i)); 
         else
            error("Unknown option: received ".expr2text(args(i))); 
         end_if;
     end_for;
  end_if;

   //--------------------------------------------
   // Matrix, Array --> List of Lists :
   //--------------------------------------------
   if l::dom::hasProp( Cat::Matrix ) = TRUE then
      l:= expr(l): // Matrix -> Array
   end_if;

   //--------------------------------------------
   // process data array: convert to list of lists
   // Warning: transpose the array (we need a list
   // columns, not a list of rows)!
   //--------------------------------------------
   if domtype(l) = DOM_ARRAY then
      // Array -> List of Lists
      dim := op(l, [0,1]): // Dimension of Array
      if dim = 2 then
         l:=[ [l[j,i] $ j = op(l, [0,2,1])..op(l, [0,2,2])] 
                      $ i = op(l, [0,3,1])..op(l, [0,3,2])];
      elif dim = 1 then 
         // 1-dimensionaler Array in eine ListOfList
         l:= [ [op(l)] ]:
      else
          error("1st argument: the data array must be one or two-dimensional")
      end_if;
   end_if;

   // ---------------------------
   // process statistical samples
   // ---------------------------

   if domtype(l) = stats::sample then
      columnIndices:= []:
      // check if column indices are provided by the user:
      if args(0) > 1 and domtype(args(2)) = DOM_LIST 
      then
        // the column indices are specified by a list
        columnIndices:= args(2);
        numberOfColumns:= nops(columnIndices);
        options := [args(3..args(0))]; 
      else
        i:= 1; // initialize: if args(0) = 1, the
               // following loop will not be entered
        for i from 2 to args(0) do
          if domtype(args(i)) = DOM_INT then
             columnIndices:= append(columnIndices, args(i));
          else
             break;
          end_if;
        end_for;
        numberOfColumns:= nops(columnIndices);

        // If no break occurred, then i = args(0) + 1;
        // and there is no option.
        // If no break occurred, then i is the first 
        // argument that is an option.
        // options may be empty if i > args(0): 
        options := [args(i..args(0))]; 
      end_if;
      if numberOfColumns = 0 then
         // No columns were specified by the
         // user. Extract *all* columns of l.
         // Check the number of elements in the
         // first row extop(l, 1)[1]:
         numberOfColumns:= nops(extop(l, 1)[1]);
         columnIndices:= [i $ i = 1 .. numberOfColumns];
      end_if;
      l:= [stats::getdata(
            testargs(),     // double check the data
            "all_data",     // accept all sorts of numerical data
                            // as well as strings
            numberOfColumns,// number of data columns to be read
            l,              // the sample
            columnIndices   // the indices of the columns to be read
           )];
      // a string returned by stats::getdata indicates an error:
      if domtype(l) = DOM_STRING then 
         error(l);
      end_if;
   
      // check for string columns in l:
      for i from 1 to nops(l) do
        if domtype(l[i][1]) = DOM_STRING then
           l[i]:= NIL;
        end_if;
      end_for:
      if has(l, NIL) then 
         l:= subs(l, NIL = null);
      end_if;
      if nops(l) = 0 then
         error("No numerical data found");
      end_if;
   end_if; // of domtype(l) = stats::sample

   //---------------------------------------------
   // ---> Now: the data l is a list of lists ----
   //---------------------------------------------
   l := float(l);
   if (_union (map({op(l[i])},domtype )$i=1..nops(l)) minus {DOM_FLOAT}) <> {} then 
       error("Some data elements are not numerical"); 
   end_if;  

   //--------------------------------------------------
   // Check if the options passed by the user are valid
   //--------------------------------------------------
   checkOptions(); // Check all options

   //--------------------------------------------------
   // Now, the work starts.
   //--------------------------------------------------
   boxList:=[];   
   notchIndentU:=notchIndent;
   for zaehl from 1 to nops(l) do
      x:=stats::empiricalQuantile(l[zaehl]);
      x25:=x(0.25);
      x50:=x(0.5);
      x75:=x(0.75);
      outRange:=[];
      range:=1.5*(x75-x25);
      // Select Points, which lie outside of the range
      outRange:=select(l[zaehl], y -> (y > x75 + range) or ( y < x25 - range )  ); 
      
      // Select Points, which lie outside of our box, but within the range
      temp:= select(l[zaehl],(y-> (y > x25-range) and (y < x25)));
      if nops(temp)<1 then
           //now max()/min() won't work
           small:=x25;
      else small:=min(op(temp)) end_if;
      temp:=select(l[zaehl],(y -> (y > x75) and (y < x75+ range )));

      if nops(temp)<1 then 
           big:=x75 
      else big:=max(op(temp))
      end_if;
      
      notch:=float(PI*( (range/1.5) / (nops(l[zaehl]))^(1/2) ));
       
      notchUp:= x50+(notch/2);
      notchDown:=x50-(notch/2);
       
      if notchUp>x75 then notchUp:=x75 end_if; 
      if notchDown<x25 then notchDown:=x25 end_if; 


      if plotMerge then 
         // ------------------
         // 2-dimensional
         // ------------------

         // do the same as for the first dimension 

         y:=stats::empiricalQuantile(l[zaehl+1]);
         y25:=y(0.25);
         y50:=y(0.50);
         y75:=y(0.75);
        
         rangeY:=1.5*(y75-y25);
         // Select Points, which lie outside of our box, but within the range
         temp:= select(l[zaehl+1],(g->(g>y25-rangeY) and (g<y25)));
         if nops(temp)<1 then 
            //now max()/min() won't work
            smallY:=y25 
         else smallY:=min(op(temp)) 
         end_if;
         temp:=select(l[zaehl+1],(g->(g>y75) and (g<y75+rangeY)));
         // Select Points, which lie outside of the range
         outRangeY:=select(l[zaehl+1], y -> (y > y75 + rangeY) or ( y < y25 - rangeY )  ); 
         if nops(temp)<1 then 
              bigY:=y75 
         else bigY:=max(op(temp)) 
         end_if;

         // Sebastian:
         // mergecolor:= color[(zaehl-1 mod nops(color) )+1];
         // Walter:
         // Fix the color: otherwise, how do you want to
         // document the color scheme?
         mergecolor:= color[1];

         if nops(l)=2 then
            boxList:=append(boxList, 
                
            plot::line( //Kleinstes Element innerhalb der XRange
                  [x50,y75],[x50,smallY],
                  hold(Color)=mergecolor,
                  LineWidth=lineWidth
            ),
            plot::line( 
                  [x50-(range/5),smallY],[x50+(range/5),smallY],
                  hold(Color)=mergecolor,
                  LineWidth=lineWidth
            ),
            plot::line( //Grtes Element innerhalb der xRange
                  [x50,y75],[x50,bigY],
                  hold(Color)=mergecolor,
                  LineWidth=lineWidth
            ),
            plot::line( 
                  [x50-(range/5),bigY],[x50+(range/5),bigY],
                  hold(Color)=mergecolor,
                  LineWidth=lineWidth
            ),
            plot::line( //Kleinstes Element innerhalb der yRange
                  [x75,y50],[small,y50],
                  hold(Color)=mergecolor,
                  LineWidth=lineWidth
                  ),
            plot::line( 
                  [small,y50-(rangeY/5)],[small,y50+(rangeY/5)],
                  hold(Color)=mergecolor,
                  LineWidth=lineWidth
            ),
            plot::line( //grtes Element innerhalb der yRange
                  [x75,y50],[big,y50],
                  hold(Color)=mergecolor,
                  LineWidth=lineWidth
            ),
            plot::line( 
                  [big,y50-(rangeY/5)],[big,y50+(rangeY/5)],
                  hold(Color)=mergecolor,
                  LineWidth=lineWidth
            ),
            plot::Polygon(
                  plot::Point( x25,y25),    
                  plot::Point( x25,y75),
                  plot::Point( x75,y75),
                  plot::Point( x75,y25),
                  plot::Point( x25,y25),   
                  // If we need more colors than speecified, start again the cycle
                  hold(Color)=mergecolor,
                        LineWidth=lineWidth,
                  Closed=FALSE
            ),
            eval(myPoint( //Ausreisser
                   v, 
                   x50,outRangeY[i],
                   [PointWidth=50,
                   hold(Color)=mergecolor,
                    PointStyle=FilledCircles]
                  )) $ i=1..nops(outRangeY),
            eval(myPoint( //Ausreisser
                  v, 
                  outRange[i],y50,
                  [PointWidth=50,
                  hold(Color)=mergecolor,
                   PointStyle=FilledCircles]
                  )) $ i=1..nops(outRange)    
            );   
            return( plot::Group(op(boxList)));      
         end_if; // of 2-dim Merge

         //-------------------------------------------
         // 3 Dimensionen
         // --------------------------------------------
   
         z:=stats::empiricalQuantile(l[zaehl+2]);
         z25:=z(0.25);
         z50:=z(0.50);
         z75:=z(0.75);
         // color:=append(color, RGB::Blue);
         // Compute absolute Data
         
         rangez:=1.5*(z75-z25);
         
         temp:= select(l[zaehl+2],(g->(g>z25-rangez) and (g<z25)));
         if nops(temp)<1 then 
            smallZ:=z25 //Evtl. + 1/10 Range o.., um auf jeden Fall eine untere Begrenzung zu haben
         else smallZ:=min(op(temp)) 
         end_if;
      
         temp:=select(l[zaehl+2],(g->(g>z75) and (g<z75+rangez)));
         if nops(temp)<1 then 
              bigZ:=z75 
         else bigZ:=max(op(temp)) 
         end_if;

         outRangeZ:=[];
         outRangeZ:=select(l[zaehl+2], y -> (y > z75 + rangez) or ( y < z25 - rangez )  ); 

         // Sebastian:
         // mergecolor:= color[(zaehl-1 mod nops(color) )+2];
         // Walter:
         // Fix the color: otherwise, how do you want to
         // document the color scheme?
         mergecolor:= color[1];

         boxList:=append(boxList, 
                  plot::line( 
                     [x50,y50,z25],[x50,y50,bigZ],
                     hold(Color)=mergecolor,
                     LineWidth=lineWidth
                  ),
                  plot::line( 
                     [x50,y50,z25],[x50,y50,smallZ],
                     hold(Color)=mergecolor,
                     LineWidth=lineWidth
                  ),
                  plot::line( 
                     [x50,y25,z50],[x50,bigY,z50],
                     hold(Color)=mergecolor,
                     LineWidth=lineWidth
                  ),
                  plot::line( 
                     [x50,y25,z50],[x50,smallY,z50],
                     hold(Color)=mergecolor,
                     LineWidth=lineWidth
                  ),
                  plot::line( 
                     [x25,y50,z50],[big,y50,z50],
                     hold(Color)=mergecolor,
                     LineWidth=lineWidth
                  ),
                  plot::line( 
                     [x25,y50,z50],[small,y50,z50],
                     hold(Color)=mergecolor,
                     LineWidth=lineWidth
                  ),
                  plot::Polygon(
                     plot::Point( x25,y25,z25),    
                     plot::Point( x25,y25,z75),    
                     plot::Point( x75,y25,z75),    
                     plot::Point( x75,y25,z25),    
                     plot::Point( x25,y25,z25),    
                     plot::Point( x25,y75,z25),    
                     plot::Point( x25,y75,z25),       
                     plot::Point( x25,y75,z75),    
                     plot::Point( x25,y25,z75),
                     plot::Point( x25,y75,z75),
                     plot::Point( x75,y75,z75), 
                     plot::Point( x75,y25,z75),
                     plot::Point( x75,y75,z75),
                     plot::Point( x75,y75,z25), 
                     plot::Point( x75,y25,z25),
                     plot::Point( x75,y75,z25),
                     plot::Point( x25,y75,z25),                 
                     hold(Color)=mergecolor,
                           LineWidth=lineWidth,
                     Closed=FALSE
                  ) ,
                  plot::Point( //Ausreisser
                     [outRange[i],y50,z50],
                     PointWidth=50,
                     hold(Color)=mergecolor,
                      PointStyle=FilledCircles
                  ) $ i=1..nops(outRange),
                  plot::Point( //Ausreisser
                     [x50,outRangeY[i],z50],
                     PointWidth=50,
                     hold(Color)=mergecolor,
                      PointStyle=FilledCircles
                   ) $ i=1..nops(outRangeY),
                   plot::Point( //Ausreisser
                     [x50,y50,outRangeZ[i]],
                     PointWidth=50,
                     hold(Color)=mergecolor,
                      PointStyle=FilledCircles
                    ) $ i=1..nops(outRangeZ)
                  );   
         return (plot::Group(op(boxList))):   
      end_if; // if plotMerge

      // "Normal"
      // Default Values fr Center und Width
      if nops(centers)>0 then shift:=float(centers[zaehl]-2*zaehl+2); end_if;
      if nops(widths)>0 then 
                stretch:=float(widths[zaehl]/2); 
                notchIndentU:=float(notchIndent*widths[zaehl]);
      end_if;
      
      boxList:=append(boxList, 
          myLine( //Kleinstes Element innerhalb der Range
                 v,
                 [2*zaehl-2-(stretch/2)+shift,small],[2*zaehl-2+(stretch/2)+shift,small],
                 [hold(Color)=color[(zaehl-1 mod nops(color) )+1],LineWidth=lineWidth]
          ),
          myLine( 
                 v,
                 [2*zaehl-2+shift,small],[2*zaehl-2+shift,x25],
                 [hold(Color)=color[(zaehl-1 mod nops(color) )+1],LineWidth=lineWidth]
          ),
          myLine( // Grtes Element innerhalb der Range 
                 v,
                 [2*zaehl-2-(stretch/2)+shift,big],[2*zaehl-2+(stretch/2)+shift,big],
                 [hold(Color)=color[(zaehl-1 mod nops(color) )+1],LineWidth=lineWidth]
                 ),
          myLine( 
                 v,
                 [2*zaehl-2+shift,big],[2*zaehl-2+shift,x75],
                 [hold(Color)=color[(zaehl-1 mod nops(color) )+1],LineWidth=lineWidth]
          ),
          //Rectangle
          plot::Polygon(
             myPoint(v,2*zaehl-2-stretch+shift,x25),    
             myPoint(v,2*zaehl-2+shift-stretch,notchDown),
             myPoint(v,2*zaehl-2+notchIndentU-stretch+shift,x50),   
             myPoint(v,2*zaehl-2-notchIndentU+stretch+shift,x50),
             myPoint(v,2*zaehl-2+notchIndentU-stretch+shift,x50),      
             myPoint(v,2*zaehl-2+shift-stretch,notchUp),
             myPoint(v,2*zaehl-2+shift-stretch,x75),
             myPoint(v,2*zaehl-2+shift+stretch, x75),
             myPoint(v,2*zaehl-2+shift+stretch,notchUp),
             myPoint(v,2*zaehl-2-notchIndentU+shift+stretch,x50),
             myPoint(v,2*zaehl-2+shift+stretch,notchDown),
             myPoint(v,2*zaehl-2+shift+stretch,x25),
             myPoint(v,2*zaehl-2+shift-stretch,x25),            
             // Falls weniger Farben als Boxen starte von vorne.
             hold(Color)=color[(zaehl-1 mod nops(color) )+1],
                   LineWidth=lineWidth,
                   Closed=FALSE
          ),
          eval(myPoint( //Ausreisser
               v, 
               2*zaehl-2+shift,outRange[i],
               [PointWidth=50,
                hold(Color)=color[(zaehl-1 mod nops(color) )+1],
                PointStyle=FilledCircles]
               )) $ i=1..nops(outRange)
         ); // end of append
   end_for; // end of for zaehl from 1 to nops(l) do
   return ( plot::Group(op(boxList))):   
 end_proc: /*of main*/
