// $Date: 2002/03/03 18:57:59 $  $Author: walter $   $Revision: 1.21 $

/* Comments:

    1) Diagrammtypen:
       Points      : Nur Datenpunkte werden gezeichnet 
                     Points ist fuer 2D und 3D implementiert.
       Lines       : gerade Verbindungslinien zwischen den
                     Datenpunkten
       LinesPoints : Zeichnen der Verbindungslinien und der 
                     Datenpunkte
       Splines     : Verbindungs der Datenpunkten durch kubische
                     Splines
       SplinesPoints: Zeichnen der Verbindungskurven mittels
                      kubischer Splines. Zusaetzlich die Datenpunkte

    2) Die Achsenbeschriftung wurde vollstaendig ueberarbeitet. 

    3) Bzgl. Points, Lines, Splines, LinesPoints und SplinesPoints
       koennen nun auch nicht-equidistante Datenpunkte gezeichnet werden. 
       Ausserdem kann die Anzahl der Datenpunkte in den Datenlisten 
       unterschiedlich sein.
       Auch kann anstelle einer Liste ein array angegeben werden.
       Wird ein Objekt vom Typ Dom::Matrix angegeben, so wird
       dieses immer als 2D-Array interpretiert, unabh. davon,
       ob es sich u.U. nur um einen Spalten- oder Zeilen-Vektor
       handelt, d.h. nur der Diagrammtyp Points steht derzeit fuer
       Dom::Matrix zur Verfuegung.
*/

/*++
    plot::data  --  Plots data in different styles

    Call:
    data(data <, Style = style> 
              <, Colors = c>,
              <, Titles = t>,
              <, options>)

    Parameters:
      data:  real numerical values:
             - a list such as [y1, y2, ..], 
             - a list of lists such as [[y11, y12, ..], [y21, y22, ..]]
             - a list of lists of lists
                 such as [[[x11, y11], [x12, y12], ..], 
                          [[x12, y21], [x22, y22], ..], ...]
             - an array(1..n, [y1, y2, ..]),
             - an array(1..m, 1..n, [[y11, y12, ..], [y21, y22, ..], ..]),
             - a matrix of Cat::Matrix

             Each sublist represents a data sample.
             The data samples may be of different lengths.

    Options:
      Style = style:
               style =  Points, Lines, LinesPoints, 
                        Splines or SplinesPoints
      Color  = c: 
      Colors = [c1, c2, ..]: 
               c, c1, c2, ..  = RGB colors. The number of colors 
               do not have to match the number of data samples: 
               the color list is used cyclically
         Remark: also Color = [c1,c2,..] and Colors=c is accepted
      Titles = t:
               t = a list of strings
               The number of titles do not have to match the
               number of data samples: it is padded with 
               empty strings
      options: scene options


    Return Value:
    An object of domain type plot::Scene

    Examples:
    >> export(plot): export(RGB):
    >> plot(data([[5,10,24,-3],[6,5,2,18],[19,45,12,-10]],
                 Colors=[Red,Green,Blue], Style = Lines ))
    >> plot(data([[[1,5],[2,10],[4,24],[7,-3]], 
                  [[0,6],[1,5],[2,2],[3,18]],
                  [[-2,19],[1,45],[2,12],[7,-10]]], 
                 Colors=[Red,Green,Blue], Style = LinesPoints))
    >> plot(data([[5,10,24,-3], [6,5,2,18], [19,45,12,-10]],
                 Colors=[Red,Green,Blue], Style = Splines))
    >> plot(data([[[1, 5], [2,10], [2.1,24], [3, -3]], 
                  [[1, 6], [2, 5], [2.5, 2], [3, 18]], 
                  [[2,19], [3,45], [PI, 12], [5,-10]]],
                 Colors=[Red,Green,Blue], Style = Splines))
    >> plot(data([[5,10,24,-3], [6,5,2,18], [19,45,12,-10]],
                 Colors=[Red,Green,Blue], Style = SplinesPoints))
------------------------------------------------------------------------*/

plot::data:= proc(l)
local dim, range, i, j, c, t, p, 
      tspec, cspec, sspec, options, 
      style, implementedStyles;
begin
    if args(0) < 1 then
        error("expecting diagram type and a list of data")
    end_if;
    if l::dom::hasProp(Cat::Matrix) = TRUE then
       l:= expr(l);
    end_if; 
    if domtype(l) = DOM_ARRAY then 
       dim:= op(l, [0, 1]);
       for i from 1 to dim do
           range[i]:= op(l, [0, i+1]):
       end_for;
       if dim = 1 then
          l:= [op(l)];
       elif dim = 2 then
          l:= [[l[i, j] $ j=op(l, [0, 3, 1]) .. op(l, [0, 3, 2])] 
                        $ i=op(l, [0, 2, 1]) .. op(l, [0, 2, 2])]; 
       else
          error("Expecting a 1- or 2-dimensional data matrix/array");
       end_if;
    end_if;

    if domtype(l) <> DOM_LIST then 
       error("expecting a list of data");
    end_if;
    if nops(l) = 0 then
       error("empty data list")
    end_if;

    //-------------------------------------------------
    // the data list l may be a list of lists or just
    // a list of numerical data. In the latter case,
    // convert l = [data1, data2, ..] to a list of
    // lists l = [ [data1, data2, ..] ]
    //-------------------------------------------------
    if domtype(l[1]) <> DOM_LIST then
       l:= [l]
    end_if;

/*  the following conversion is done in inside
    data_Lines and data_Splines:

    //-------------------------------------------------
    // convert [[y11, y12, ..], [y21, y22, ..], ..] to
    //         [[[1, y11], [2, y12], ..], [[1, y21], [2, y22], ..]] 
    //-------------------------------------------------
    l:= plot::data_convert(l);
*/

    //-------------------------------------------------
    // convert input to floats (note that float
    // is automatically mapped to sublists)
    //-------------------------------------------------
    l := float(l);

    //-------------------------------------------------
    // get options:
    //-------------------------------------------------
    c:= [op(RGB::ColorList, 1..10)]; // default color list
    t:= []; // Titles
    style:= LinesPoints; // Default
    implementedStyles:=
        [Points, Lines, Splines, LinesPoints, SplinesPoints]:
    tspec:= cspec:= sspec:= FALSE;

    options:= [args(2..args(0))];
    for i from 1 to nops(options) do
        if type( (p:= options[i]) ) <> "_equal" then
            error("specification of the form 'option = value' expected")
        end_if;
        case op(p,1)
        of hold(Style) do
            style:= op(p, 2);
            if contains(implementedStyles, style) = 0 then
               error("Option Style: the available styles are ".
                      expr2text(op(implementedStyles)).". Received: ".
                      expr2text(style));
            end_if;
            options[i]:= FAIL;
            sspec:= TRUE;
            break
        of hold(Color) do 
        of hold(Colors) do 
            c:= op(p,2);
            if domtype(c) <> DOM_LIST then
               error("option 'Colors': an RGB color or a list of ".nops(l)." RGB colors expected")
            end_if;
            if nops(c) = 3 and domtype(c[1]) <> DOM_LIST then
               // specification Colors = color
               c:= [c];
            end_if;
            // Expecting Colors = [color1, color2, ...]
            c:= float(c);
            if map({op(c)}, testtype, Type::ListOf(DOM_FLOAT, 3)) <> {TRUE} then
               error("option 'Colors': an RGB color or ".
                     "a list of ".nops(l)." RGB colors expected. Received: Colors = "
                     .expr2text(c))
            end_if;
            options[i]:= FAIL;
            cspec:= TRUE;
            break
        of hold(Titles) do 
            t:= op(p,2);
            if not testtype(t,Type::ListOf(DOM_STRING)) then
                error("option 'Titles': list with ".nops(l)." strings")
            end_if;
            options[i]:= FAIL;
            tspec:= TRUE;
            break;
        end_case;

        if {cspec, tspec, sspec} = {TRUE} then break end_if;
    end_for;
    options:= subs( options,FAIL=null() );

    /*----------------------------
    //Walter 28.2.02: The following testargs() is obsolete:
    //Now, data lists of different lengths are accepted.
    if testargs() then
       if testtype(l, Type::ListOf(Type::ListOf(Type::Real,1),1)) then
          if _or(nops(l[i])<>nops(l[i+1])$i=1..nops(l)-1) then
              error("all lists must have the same number of elements")
          end_if
       elif testtype(l, Type::ListOf(Type::ListOf(DOM_LIST),1)) then
          if _or((nops(l[i][j])<>nops(l[i][j+1])$j=1..nops(l[i])-1)$i=1..nops(l)) then
            error("all lists must have the same number of elements")
          end_if
       else 
          error("2nd argument: a list of lists expected")
       end_if;
    end_if;
    ---------------------------- */

    //-------------
    // do the work
    //-------------
    case style
    of Points do
        return( plot::data_Lines(l, c, t, 1, options) )
    of Lines do
        return( plot::data_Lines(l, c, t, 2, options) )
    of LinesPoints do
        return( plot::data_Lines(l, c, t, 3, options) )
    of Splines do
        return( plot::data_Splines(l, c, t, 1, options) )
    of SplinesPoints do
        return( plot::data_Splines(l, c, t, 2, options) )
    otherwise
        error("unknown style")
    end_case
end_proc: 


/*------------------------------------------------------------
    dataplot_Lines
------------------------------------------------------------*/
plot::data_Lines:= proc(data, c, t, typ, options=[])
name plot::data::Lines;
local i, j, m, objs, scene_default, titlepositions;
begin
    // typ: 1 -> Points, 2 -> Lines, 3 -> LinesPoints 
    m:= nops(data); 
    if domtype(data[1][1]) = DOM_LIST then
      case nops(data[1][1])
      of 2 do
          break;
      of 3 do
          if typ <> 1 then
              error("only diagram style 'Points' available for three-dimensional lists")
          else
              // set options:
              scene_default:= 
                  [hold(Axes) = hold(Box), 
                   hold(Scaling) = hold(UnConstrained),
                   hold(Labeling) = TRUE, 
                   hold(Ticks) = plot::data_ticks1(data),
                   hold(Title) = ""];
              if options <> [] then
                  if has(options,hold(Axes)) then scene_default[1]:= FAIL end_if;
                  if has(options,hold(Scaling)) then scene_default[2]:= FAIL end_if;
                  if has(options,hold(Labeling)) then scene_default[3]:= FAIL end_if;
                  if has(options,hold(Ticks)) then scene_default[4]:= FAIL end_if;
                  if has(options,hold(Title)) then scene_default[5]:= FAIL end_if;
                  options:= subs( scene_default, FAIL=null()) . options;
              else
                  options:= scene_default
              end_if;

              t:= plot::data_set_titles1(t, m);
              return(plot::Scene(
                  plot::data_point3d_list1(data, t, c), op(options)
              ));
          end_if;
      otherwise
          error("lists of two or three numbers expected")
      end_case
    end_if;

    t:= plot::data_set_titles1(t, m);

    if type(data[1][1]) <> DOM_LIST then
        // [y1,..., yN] -> [[y1],...[N, yN]]
        data := [ [ [i, data[j][i]] $ i=1..nops(data[j])] $ j=1..m ]
    end_if;

    // data = [ [ [x_1_1, y_1_1], ..., [x_1_m1, y_1_n1] ], ... , ]

    // typ: 1 -> Points, 2 -> Lines, 3 -> LinesPoints 
    objs:= null(); 
    if (typ = 2 or typ = 3) then
       titlepositions:= plot::data_getTitlePositions(data, t, c):
       for j from 1 to m do
         if nops(data[j]) > 1 then
           // there are more than 1 point in the i-direction
           // to be connected via lines
           objs := objs,
             plot::Polygon(
               (plot::Point(data[j][ i ][1], data[j][ i ][2]),
                plot::Point(data[j][i+1][1], data[j][i+1][2])
               ) $ i=1..nops(data[j])-1, 
               hold(Color) = c[((j-1) mod nops(c)) + 1],
               hold(Title) = t[j], 
               hold(TitlePosition) = [5*j/nops(data), titlepositions[j]]
             );
          else 
            objs:= objs, 
              plot::Pointlist(                     
                  plot::Point(data[j][i][1],data[j][i][2], 
                              hold(Color)=c[((j-1) mod nops(c)) + 1]
                             ) $ i = 1..nops(data[j]), 
                  hold(Title) = t[j], 
                  hold(TitlePosition) = [5*j/nops(data), titlepositions[j]]
               // hold(PointWidth) = 40, 
               // hold(PointStyle) = hold(FilledCircles)
              );
         end_if;
       end_for;
    end_if;

    if typ = 1 then
        objs := objs, plot::data_point2d_list1(data, t, c) ;
    end_if;
    if typ = 3 then
        // the titles were set above. Pass an emtpy title list
        // to avoid duplication of titles.
        objs := objs, plot::data_point2d_list1(data, ["" $ m], c) ;
    end_if;
    if nops([objs]) = 0 then
        objs := objs, plot::data_point2d_list1(data, t, c) ;
    end_if;

    // set options:
    scene_default:= 
        // Ticks=min(max(2,n-2),10)
        [hold(Axes) = hold(Corner),
         hold(Scaling) = hold(UnConstrained),
         hold(Labeling) = TRUE,
         hold(Ticks) = plot::data_ticks1(data),
         hold(Labels) = ["",""],
         hold(Title) = ""];
    if options <> [] then
        if has(options,hold(Axes)) then scene_default[1]:= FAIL end_if;
        if has(options,hold(Scaling)) then scene_default[2]:= FAIL end_if;
        if has(options,hold(Labeling)) then scene_default[3]:= FAIL end_if;
        if has(options,hold(Ticks)) then scene_default[4]:= FAIL end_if;
        if has(options,hold(Labels)) then scene_default[5]:= FAIL end_if;
        if has(options,hold(Title)) then scene_default[6]:= FAIL end_if;
        options:= subs(scene_default, FAIL=null()) . options;
    else
        options:= scene_default
    end_if;

    return( plot::Scene(objs, op(options)) )
end_proc:

/*------------------------------------------------------------
    dataplot_Splines

    data may be 
        [[y_1_1, .. , y_1_N1], .., [y_M_1, ... , y_M_NM]

    (equidistand w.r.t. x) or

        [ [[x_1_1, y_1_1],...,[x_1_N1, y_1_N1]],
           ... , 
          [[x_M_1, y_M_1],...,[x_M_NM, y_M_NM]] 
        ]   
    (user specified points (x.i.j, y.i.j))
------------------------------------------------------------*/
plot::data_Splines:= proc(data, c, t, typ, options=[])
    name plot::data::Splines;
    local i, j, m, objs, x, xmax, xmin, ymax, ymin, SplineFunc, 
          scene_default, titlepositions;
begin
    m:= nops(data); 

    if domtype(data[1][1]) <> DOM_LIST then
        // [y1,..., yN] -> [[1,y1],...[N, yN]]
        data := [ [ [i, data[j][i]] $ i=1..nops(data[j]) ] $ j=1..m ]
    end_if;

    t := plot::data_set_titles1(t, m);
    x:= genident();
    objs:= null():
    for j from 1 to m do 
      xmin[j]:= float(min(data[j][i][1] $ i=1..nops(data[j]))):
      xmax[j]:= float(max(data[j][i][1] $ i=1..nops(data[j]))):
      SplineFunc[j]:= numeric::cubicSpline(op(data[j])):
      titlepositions:= plot::data_getTitlePositions(data, t, c):
      if xmin[j] < xmax[j] then
         objs := objs,
           plot::Curve2d([x, hold(fp::apply)(SplineFunc[j],x)], 
                         x=[xmin[j],xmax[j]], 
                         hold(Color)=[hold(Flat),c[((j-1) mod nops(c)) + 1]],
                         hold(Title)=t[j],
                         hold(TitlePosition) = [5*j/nops(data), titlepositions[j]],
                         hold(Discont) = FALSE
                        );
       else
         // if xmin[j] = xmax[j], plotlib::clip2d_Curve complains.
         // Plot a single point instead of a spline curve
         objs:= objs, 
           plot::Pointlist(                     
               plot::Point(data[j][i][1],data[j][i][2], 
                           hold(Color)=c[((j-1) mod nops(c)) + 1]
                          ) $ i = 1..nops(data[j]), 
               hold(Title) = t[j], 
               hold(TitlePosition) = [5*j/nops(data), titlepositions[j]]
           );
      end_if;
    end_for;

    if typ = 2 then
       // SplinesPoints. We still need to add the points.
       // Since the titles were set above, pass a list of
       // emtpy strings as titles to avoid duplication:
        objs := objs, plot::data_point2d_list1(data, ["" $ m], c)       
    end_if;

    // Folgende 2 Zeilen nur interessant, falls kleinster
    // und groesster y-Wert ermittelt werden soll.
    [ymin, ymax] := plot::data_curve_min_max1(SplineFunc, xmin, xmax, 100);

    // set options:
    scene_default:= 
        [hold(Axes) = hold(Corner), 
         hold(Scaling) = hold(UnConstrained),
         hold(Labeling) = TRUE,
         hold(Labels) = ["",""],
         hold(Ticks) = plot::data_ticks1(data, ymin, ymax),
         hold(Title) = ""];
    if options <> [] then
        if has(options,hold(Axes)) then scene_default[1]:= FAIL end_if;
        if has(options,hold(Scaling)) then scene_default[2]:= FAIL end_if;
        if has(options,hold(Labeling)) then scene_default[3]:= FAIL end_if;
        if has(options,hold(Labels)) then scene_default[4]:= FAIL end_if;
        if has(options,hold(Ticks)) then scene_default[5]:= FAIL end_if;
        if has(options,hold(Title)) then scene_default[6]:= FAIL end_if;
        options:= subs(scene_default, FAIL=null()) . options;
    else
        options:= scene_default
    end_if;

    return( plot::Scene(objs, op(options)) )
end_proc:

/*-----------------------------------------------------------*/
/*-------------------- Utilities ----------------------------*/
/*-----------------------------------------------------------*/
// If there are more data samples than titles, 
// fill the titles list with empty strings:
plot::data_set_titles1:= proc(t, n)
name plot::data;
local m;
begin
    m:= nops(t);
    if m < n then t := t.["" $ i = 1..n-m] end_if;
    return(t)
end_proc:

/*-----------------------------------------------------------*/
plot::data_ticks1:= proc(l, ymin=FAIL, ymax=FAIL)
    name plot::data;
    local i, tx, ty, tz;
begin
    if type(l[1][1]) = DOM_LIST then
        tx := [ l[i][j][1] $ j = 1..nops(l[i]) $ i = 1..nops(l) ];
        tx := plot::data_xyzticks1(tx);
        ty := [ l[i][j][2] $ j = 1..nops(l[i]) $ i = 1..nops(l) ];
        ty := plot::data_xyzticks1(ty, ymin, ymax);
        if nops(l[1][1]) = 2 then
            return([ tx, ty ])
        else
            tz := [ l[i][j][3] $ j = 1..nops(l[i]) $ i = 1..nops(l) ];
            tz := plot::data_xyzticks1(tz);
            return([ tx, ty, tz ])
        end_if
    else 
        ty := [ op(l[i]) $ i = 1..nops(l) ];
        return([ 0, plot::data_xyzticks1( ty ) ])
    end_if
end_proc:

/*-----------------------------------------------------------*/
plot::data_xyzticks1:= proc(l, ymin=FAIL, ymax=FAIL)
    name plot::data;
    local d, i, j, n;
begin
    // Convert ty to a set to remove multiple entries
    // and re-convert into a list
    l:= sort( [op({op(l)})] );  
    n:= nops(l);
    // Maximal tick marks: 40 (Internal upper bound!)
    if n > 40 then
       // 10 (arbitrary) tick marks 
       return( 10 )
    end_if;
    if type(ymin) = DOM_FLOAT then
        d:= (ymax-ymin)/20.0
    else
        d:= (l[n]-l[1])/20.0
    end_if;
    j:= l[1];
    for i from 2 to n do
        if l[i] - j < d then
            l[i]:= l[i] = ""
        else
            j:= l[i]
        end_if
    end_for;

    return(l)
end_proc:

/*-----------------------------------------------------------*/
plot::data_curve_min_max1 := proc(p, a, b, g)
    name plot::data;
    local j, ma, mi, v, x, dx;
begin
    mi:= infinity; ma:= -infinity;
    for j from 1 to nops(p) do
        dx:= (b[j]-a[j])/g;
        if iszero(dx) then dx:= 1: end_if;
        for x from a[j] to b[j] step dx do
            v := p[j](float(x));
            if v < mi then 
                mi := v
            elif v > ma then
                ma := v
            end_if
        end_for
    end_for;
    return( [mi, ma] )
end_proc:

/*-----------------------------------------------------------*/
// The following utilities returns a list of y-coordinates for 
// the titles of the data samples between 0 (top) and 10 (bottom).
// It expects
// data = [[[x11,y11],..,[x1.n1,y1.n2]], .. , [[x.m.1, y.m.1],..,[x.m.nm,y.m.nm]]]
// t = ["title1", "title2", .., "title.m"]

plot::data_getTitlePositions := proc(data, t)
name plot::data;
local titlepositions, ymax, ymin, j, maxy, miny;
begin
    titlepositions:= [4.0 $ nops(data)];
    if {op(t)} = {""} then
       // if only empty titles are set, return a dummy
       // of valid title positions:
       return(titlepositions);
    end_if;
    // determine heuristic title positions
    ymax:= [0 $ nops(data)];
    ymin:= [0 $ nops(data)];
    for j from 1 to nops(data) do
      ymax[j]:= max(data[j][i][2] $ i=1..nops(data[j]));
      ymin[j]:= min(data[j][i][2] $ i=1..nops(data[j]));
    end_for;
    maxy:= max(op(ymax)):
    miny:= min(op(ymin)):

    if specfunc::abs(maxy - miny) <= 
          10^(-DIGITS)* max(specfunc::abs(maxy), specfunc::abs(miny))
    then // all data have the same value. Place all titles
         // at the same y-position 4.0 (from the range 0..10).
         titlepositions:= [4.0 $ nops(data)];
    else 
         // Position title[j] at the convex combination
         // 0.3*ymin[j] + 0.7*ymax[j]. This is between 
         // ymin[j] and ymax[j], but closer to the max.
         // Scale this to the range 0..10, where 0 is the
         // top of the picture.
         for j from 1 to nops(data) do
           // Here, we can rely on miny < maxy:
           titlepositions[j]:= 10* (maxy - (0.3*ymin[j] + 0.7*ymax[j]))/(maxy - miny);
           // The values above should be between 0 and 10.
           // As an additional safety measure (e.g., against
           // numerical roundoff), clip the values to the range 0 .. 10:
           if titlepositions[j] < 0 then
              titlepositions[j]:= 0:
           end_if;
           if titlepositions[j] > 10 then
              titlepositions[j]:= 10:
            end_if;
         end_for;
    end_if;
    return(titlepositions);
end_proc:

/*-----------------------------------------------------------*/
plot::data_point2d_list1 := proc(data, t, c)
name plot::data;
local i, j, titlepositions;
begin
    titlepositions:= plot::data_getTitlePositions(data, t, c):
    plot::Pointlist(                     
        plot::Point(data[j][i][1],data[j][i][2],
                    hold(Color)=c[((j-1) mod nops(c)) + 1]
                   ) $ i = 1..nops(data[j]), 
        hold(Title) = t[j], 
        hold(TitlePosition) = [5*j/nops(data), titlepositions[j]]
    ) $ j=1..nops(data)
end_proc:

/*-----------------------------------------------------------*/
plot::data_point3d_list1 := proc(data, t, c)
name plot::data;
local i, j, titlepositions;
begin
    titlepositions:= plot::data_getTitlePositions(data, t, c):
    return(plot::Pointlist(
        plot::Point(data[j][i][1],data[j][i][2],data[j][i][3],
                    hold(Color) = c[((j-1) mod nops(c)) + 1]
                   ) $ i = 1..nops(data[j]),
        hold(Title) = t[j], 
        hold(TitlePosition) = [5*j/nops(data), titlepositions[j]]
    ) $ j = 1..nops(data))
end_proc:

/*-----------------------------------------------------------*/
// convert [y1, y2, ..] to [[1, y1], [2, y2], ..]
plot::data_convert := proc(data)
name plot::data;
local j, check;
begin
  // -------- utility ------------
  check:= proc(data)
  local a, b, dim, i, j, nl, t;
  begin
    if domtype(data) = DOM_ARRAY and (dim := op(data,[0,1])) > 2 then 
       error("one- or two-dimensional array or matrix expected")
    else return( data )
    end_if;
    // Walter, 3.3.02: the following code cannot be reached. 
    (a[i] := op(data, [0, i+1, 1])) $ i=1..dim;
    (b[i] := op(data, [0, i+1, 2])) $ i=1..dim;
    nl := [];
    if dim = 1 then
      for i from a[1] to b[1] do
        if testtype(data[i], Type::Real) then
           nl := append(nl, [i, data[i]])
        end_if;
      end_for;
    else // dim = 2
      for i from a[1] to b[1] do
        for j from a[2] to b[2] do
          if testtype(data[i,j], Type::Real) then
            nl := append(nl, [i, j, data[i,j]])
          end_if
        end_for
      end_for
    end_if;
    return(nl)
  end_proc:

  // -------- main ------------
  if type(data) = DOM_LIST then
       [ check(data[j]) $ j = 1..nops(data) ]
  else [ check(data) ]
  end_if;
end_proc:
