// $Date: 2001/10/19 20:57:20 $  $Author: walter $   $Revision: 1.6 $

/*++
Author: W. Oevel

    inequality - plot inequalities

Call(s):

    plot::inequality(
             ineq, 
             x = left..right, 
             y = bottom..top 
             <, Grid = [nx, ny] >
             <, Colors = [col1, col2, col3]>,
             <, ShowGrid>
             <, opt1, opt2, ...> 
    );

Parameters:

    ineq             -- a single inequality f(x,y) < g(x, y)
                        in 2 indeterminates or a set or list 
                        [f1(x,y) < g1(x,y), f2(x,y) < g1(x,y), ..]
                        of such inequalities 
    x, y             -- the indeterminates in ineq
                        identifier or indexed identifier
    left, right,     -- numerical real values
    bottom, top 
    nx, ny           -- positive integers
    col1, col2, col3 -- color names (RGB values). Also
                        'Transparent' is accepted as a color name.
    opt1, opt2, ..   -- plot option(s) accepted by plot::Group

ReturnValue:  an object of type plot::Group

Synopsis:

    inequality serves for displaying points (x,y) in the rectangle 
    Q = [left,right] x [bottom,top] satisfying the inequalities

      f1(x,y) < g1(x,y) and f1(x,y) < g1(x,y) and  ...      (1)

    specified by ineq.

    The rectangle Q is divided into 2^nx times 2^ny subrectangles.
    The default value for nx, ny is 6.

    A subrectangle is displayed with the color col1, if all its points
    (x,y) satisfy all inequalities (1). 

    A subrectangle is displayed with the color col3, if there is at
    least on inequality f(x,y) <= g(x,y) in ineq such that all points
    in the subrectangle satisfy f(x,y) > g(x, y).
    Consequently, it is guaranteed that points of this color violate (1).

    The remaining subrectangles are displayed with the color col2. 
    They cover the boundary of the region defined by (1).

    The default colors are Colors = [RGB::Green, RGB::Yellow, RGB::Red].

    The color 'Transparent' corresponds to the background color.
    It is highly recommended to specify 'Transparent' instead of
    the current background color, because this increases the
    efficiency of the algorithm.

    With ShowGrid, the subrectangles become visible: the polygons
    are rendered with their border lines.

    The plot is rendered via plot(plot::inequality(...)).

    Interval arithmetic is used to check the inequalities. 
    The input functions must be suitable for this kind of 
    arithmetic, In particular, the calls 
       f(left ... right, bottom ... top) - g(left ... right, bottom ... top) 
    must produce valid intervals.

    Warning: due to a bug in the present renderer, coordinate axes
             ticks and tick marks are hidden by the polygons.
             Use Axes = Box or Axes = Corner to avoid this problem.

Examples:

    >> plot(plot::inequality(x^2 + y^2 <  1, x = -1..1, y = -1..1));
    >> plot(plot::inequality(x^2 + y^2 <= 1, x = -1..1, y = -1..1));
    >> plot(plot::inequality(x^2 + y^2 >  1, x = -1..1, y = -1..1));
    >> plot(plot::inequality(x^2 + y^2 >= 1, x = -1..1, y = -1..1));

    >> plot(plot::inequality([y < cos(x), y > - cos(x)],
            x = -PI..PI, y = -2..2, Grid = [5, 5]));

    // old syntax in 2.0:
    >> plot(plot::inequality([(x,y)->cos(x)-y, (x,y)->y+cos(x)],
                              -PI..PI, -2..2, 5);

    >> reset();
       plot( plot::inequality([y < x^2 , x*sin(x) < cos(y^2)], 
                   x = -2..2, y = -1..1, Grid = [7, 7],
                   ShowGrid,
                   Colors = [RGB::Red, RGB::Black, RGB::White]),
             Axes = Corner)

    // old syntax in 2.0:
    >> plot(plot::inequality([(x,y)->y^2-x, (x,y)->cos(y^2)-x*sin(x)],
                              -2..2, -1..1, 7,
                   Colors = [RGB::Red, RGB::Black, RGB::White]))


--------------------------------------------------------------------------*/

plot::inequality:= proc(f, rx, ry)
local x, y, left, right, bottom, top, i, opt, g, 
      nx, ny, PColor, ZColor, NColor, showgrid, c;
begin

    if args(0) < 3 then error("expecting at least 3 arguments") end_if;

    //--------------------------
    // get variables and ranges:
    //--------------------------
    if type(rx) <> "_equal" then
       error("expecting a specification such as 'x = left .. right' as ".
             "second argument"):
    end_if;
    x:= op(rx, 1):
    if domtype(x) <> DOM_IDENT and type(x) <> "_index" then
       error("no variable specified in the second argument");
    end_if;
    if type(op(rx, 2)) <> "_range" then
       error("expecting a range specification '".expr2text(x).
             " = left .. right', received ". expr2text(rx));
    end_if;
    left:= op(rx, [2, 1]):
    right:= op(rx, [2, 2]):
    if domtype(float(left)) <> DOM_FLOAT then
       error("expecting a numerical real number as left bound for ".
             expr2text(x).", received ".expr2text(op(rx, [2, 1])));
    end_if;
    if domtype(float(right)) <> DOM_FLOAT then
       error("expecting a numerical real number as right bound for ".
             expr2text(x).", received ".expr2text(op(rx, [2, 2])));
    end_if;
    //--------------------------------
    if type(ry) <> "_equal" then
       error("expecting a specification such as 'y = bottom .. top' as ".
             "third argument"):
    end_if;
    y:= op(ry, 1):
    if domtype(y) <> DOM_IDENT and type(y) <> "_index" then
       error("no variable specified in the third argument");
    end_if;
    if type(op(ry, 2)) <> "_range" then
       error("expecting a range specification '".expr2text(y).
             " = bottom .. top, received ". expr2text(ry));
    end_if;
    bottom:= op(ry, [2, 1]):
    top:= op(ry, [2, 2]):
    if domtype(float(bottom)) <> DOM_FLOAT then
       error("expecting a numerical real number as left bound for ".
             expr2text(y).", received ".expr2text(op(ry, [2, 1])));
    end_if;
    if domtype(float(top)) <> DOM_FLOAT then
       error("expecting a numerical real number as right bound for ".
             expr2text(y).", received ".expr2text(op(ry, [2, 2])));
    end_if;

    //--------------------------
    // get the inequalities
    //--------------------------

    // convert a set of inequalities to a list
    if domtype(f) = DOM_SET then
       f:= [op(f)];
    end_if;
    // convert a plain expression to a list 
    if domtype(f) <> DOM_LIST then
       f:= [f]:
    end_if;

    // check and convert inequalities to an expression
    for i from 1 to nops(f) do
        g:= f[i];
        if not (type(g) = "_less" or 
                type(g) = "_leequal" or
                type(g) = "_equal"
               ) then
           error("expecting an inequality 'f(x, y) < g(x, y)' ".
                 "or 'f(x, y) <= g(x, y)', received ".
                 expr2text(g));
        end_if:
        f[i] := rhs(g) - lhs(g);
    end_for:

    //------------------------------------
    // 'hullify' constant expressions in f
    //------------------------------------
    f:= map(f, interval):

    //-------------------
    // set default values 
    //-------------------
    [nx, ny]:= [6, 6]; 
    PColor:=RGB::Green: 
    ZColor:=RGB::Yellow: 
    NColor:=RGB::Red:
    showgrid:= FALSE;

    //--------------------------
    // override default values
    // by optional arguments: 
    //--------------------------
    for i from 4 to args(0) do
        opt:= args(i);
        if type(opt) = "_equal" and
           lhs(opt) = Grid and 
           type(rhs(opt)) = DOM_LIST and
           nops(rhs(opt)) = 2 then
           [nx, ny]:= rhs(opt);
           if (domtype(nx) <> DOM_INT or nx < 2) or
              (domtype(ny) <> DOM_INT or ny < 2) then
              error("expecting a grid specification 'Grid = [n, m]' with ".
                    "integers n >= 2, m >= 2, received ".expr2text(opt));
           end_if;
           next;
        end_if;
        if type(opt) = "_equal" and
           lhs(opt) =  Colors and 
           type(rhs(opt)) = DOM_LIST and
           nops(rhs(opt)) = 3 then
           [PColor, ZColor, NColor]:= rhs(opt);
           if (PColor <> hold(Transparent) and 
                  (domtype(PColor) <> DOM_LIST or nops(PColor) <> 3)) or
              (ZColor <> hold(Transparent) and 
                  (domtype(ZColor) <> DOM_LIST or nops(ZColor) <> 3)) or
              (NColor <> hold(Transparent) and 
                  (domtype(NColor) <> DOM_LIST or nops(NColor) <> 3)) then
              error("expecting a color specification ".
                    "'Colors = [color1, color2, color3]', ".
                    "received ".expr2text(opt));
           end_if;
           next;
        end_if;
        if opt = ShowGrid then
           // make the polygons render their boundaries
           showgrid:= TRUE;
           next;
        end_if;
        error("unknown option, received ".expr2text(opt));
    end_for;

    // the following c consists either of a color
    // (an RGB color or 'Transparent') (if the entire
    //  rectangle should have this color)
    // or of a sequence of polygons
    // or of an empty sequence.
    // The last case arise only if 'Transparent' is
    // specified as one of the colors.
    c:= plot::inequality_main(
              f, x, y, left ... right, bottom ... top, 
              nx, ny, PColor, ZColor, NColor, 1, 
              Closed = showgrid):

    if c <> null() and type(c) <> "_exprseq" then
       // only a single rectangle with the color c is returned.
       // We still need to convert it to 2 polygons:
       c := plot::inequality_plotdata(
             c, left...right, bottom...top, Closed = showgrid);
    end_if;

    if c = null() then
       // we do not really want an emtpy plot without
       // axes. Enforce the rendering of the axes by
       // creating an artificial box consisting of the
       // corners of the requested viewing area:
       return(plot::Group(
                plot::Point(left, bottom, Color = RGB::White),
                plot::Point(right, bottom, Color = RGB::White),
                plot::Point(right, top, Color = RGB::White),
                plot::Point(left, top, Color = RGB::White)
       ));
    end_if;

    // return the plot data (a long sequence of polygons)
    return(plot::Group(c));

end_proc:

//--------------------------------------------
// utility: generate the 2 polygons that build
// up the rectangle left_right, bottom_top:
//--------------------------------------------
plot::inequality_plotdata:= proc(
      color,      // RGB color
      left_right, // DOM_INTERVAL
      bottom_top, // DOM_INTERVAL
      showgrid    // Close = TRUE/FALSE
)
local left, right, bottom, top;
begin
  left:= lhs(left_right);
  right:= rhs(left_right);
  bottom:= lhs(bottom_top);
  top:= rhs(bottom_top);
  if color = hold(Transparent) then
     if showgrid = (Closed = FALSE) then
          return(null());
     else return(
           plot::Polygon(
             plot::Point(left,  bottom),
             plot::Point(right, bottom),
             plot::Point(right, top),
             showgrid, // Closed = TRUE
             Filled = FALSE,
             Color = RGB::Black
           ),
           plot::Polygon(
             plot::Point(left, bottom),
             plot::Point(left, top),
             plot::Point(right,top),
             showgrid, // Closed = TRUE
             Filled = FALSE,
             Color = RGB::Black
           )
          );
     end_if;
  end_if;
  return(plot::Polygon(
          plot::Point(left,  bottom),
          plot::Point(right, bottom),
          plot::Point(right, top),
          showgrid, // = null() or Closed = TRUE
          Filled = TRUE, Color = color),
         plot::Polygon(
          plot::Point(left, bottom),
          plot::Point(left, top),
          plot::Point(right,top),
          showgrid, // Closed = TRUE/FALSE
          Filled = TRUE, Color = color)
        );
end_proc:

//-------------------------------------------------------
// plot::inequality_main returns plotdata for plot::inequality
// It works recursively:
// 1)  via interval arithmetic, it check
//     whether there is a conclusive answer 
//         'all inequalities are satisfied' --> PColor'
//     or
//         'at least on inequality is violated' --> NColor
// 
// 2) If there is no conclusive answer, it splits the
//    rectangle into 2 halves. If the two inspected
//    subrectangles have the same color, this color
//    is returned. Otherwise, the sequence of plotdta
//    of both subrectangles are created and returned
//    as a sequence.
//
// This strategy serves for creating plotdata only
// if it cannot be helped. Note that the handling of
// the plot data is the most expensive part, so try
// to avoid creating plot data whenever possible. 
//-------------------------------------------------------
plot::inequality_main:= proc(
    f,          // list of inequalities
    x, y,       // symbolic names of the unknowns
    left_right, // x-coordinates of the rectangle: DOM_INTERVAL
    bottom_top, // y-ccordinates of the rectangle: DOM_INTERVAL
    nx, ny,     // recursive depth of the subdivisions until stop
    PColor,     // color for subrectangles satisfying the inequalities
    ZColor,     // color for subrectangles for which there is no
                // conclusive answer
    NColor,     // color for subrectangles violating at least one
                // of the ineqalities
    dir,        // direction of the next subdivision: +1 or -1
    showgrid    // Closed = TRUE or Closed = FALSE
)
local left, right, bottom, top, 
      LO, lo, hi, g, IVg, mx, my, 
      c1, c2, tc1, tc2;
begin

  //----------------------------------------------
  // dir = -1: next subdivision in the x-direction
  // dir =  1: next subdivision in the y-direction
  // change the direction with each recursive step:
  //----------------------------------------------
  dir:= -dir;

  left:= lhs(left_right):
  right:= rhs(left_right):
  bottom:= lhs(bottom_top):
  top:= rhs(bottom_top):

  LO:= 1; // initialize LO =  minimum of the minima 
          // of g(left_right, bottom_top) with g 
          // running through all inequalities in f

  for g in f do

    //-----------------------------------------
    // intervall evaluation of the expression g
    //-----------------------------------------
    IVg:= hull(eval(subs(g, [x = left_right, y = bottom_top])));

    if not iszero(Im(IVg)) then
       // nonreal complex values found.
       // This may be because of overestimation, so we should not
       // raise an error here. This rectangle is 'don't know': it
       // will be subdivided for further inspection down below.
       next;
    end_if;

    hi:= rhs(IVg): // max of g(left_right, bottom_top)

    if hi < 0 then
        //---------------------------------------------------------
        // For one of the functions g in f we know that g(x, y) < 0
        // for all (x, y) in this rectangle. We may conclude that
        // all points in this rectangle violate the requested condition.
        // Indicate this negative result by NColor.
        //---------------------------------------------------------
        return(NColor);
    end_if;

    lo:= lhs(IVg); // min of g(left_right, bottom_top)
    if lo > 0 then
        //-----------------------------------------------------
        // The inequality g(x, y) > 0 in f is satisfied
        // for all (x, y) in this rectangle. We still need to
        // investigate the other inequalities in f before we
        // have a conclusive answer for this rectangle.
        // However, we may eliminate the inequality q 
        // from the refined search below, because we 
        // already know that we will get a positive
        // answer for this g.
        //-----------------------------------------------------
        f:= subs(f, g = null());
        if nops(f) = 0 then 
           break;
        end_if;
    end_if;
    // Presently we cannot use min(RD_NINF, lo):
    // LO:= min(LO, lo);
    if lo < LO then
       LO:= lo;
    end_if;
  end_for;

  if LO > 0 then 
     //---------------------------------------------------
     // There is a conclusive answer: all functions g in f
     // satisfy g(x, y) > 0 for all x, y in this rectangle.
     // Indicate this positive result by PColor.
     //---------------------------------------------------
    return(PColor);
  end_if;

  //-------------------------------------
  // At this point, we have no conclusive
  // answer for this rectangle. We need 
  // to subdivide and have a closer look.
  //-------------------------------------

  //-------------------------------------------
  // Stop if the requested precision is reached.
  // Indicate by ZColor that no conclusive
  // answer was obtained.
  //-------------------------------------------
  if nx = 0 and ny = 0 then
     return(ZColor);
  end_if;

  //-------------------------
  // Subdivide in x direction
  //-------------------------
  if dir = -1 and nx > 0 then
     mx:= DOM_INTERVAL::center(left_right); 
     left:= subsop(left_right, 2 = mx):
     right:= subsop(left_right, 1 = mx):
     c1:= plot::inequality_main(f, x, y, left,  bottom_top, nx - 1, ny,
                            PColor, ZColor, NColor, dir, showgrid);
     c2:= plot::inequality_main(f, x, y, right, bottom_top, nx - 1, ny,
                            PColor, ZColor, NColor, dir, showgrid);

     // c1 = c2 can only happen if c1 = c2 = (RGB color = DOM_LIST
     // or 'Transparent').
     // Return this color instead of concrete plot data to minimize
     // the amount of plot data that are generated.

     if c1 = c2 then return(c1); end_if;

     tc1:= type(c1):
     tc2:= type(c2):

     // c1, c2 can either be either lists (= RGB color specifications),
     // or 'Transparent' or a sequence of plotdata

     if tc1 <> "_exprseq" and tc2 = "_exprseq" then
        // need to create plotdata of color c1 for the left subrectangle.
        // c2 consists of the plotdata of the right subrectangle
        return(plot::inequality_plotdata(c1, left, bottom_top, showgrid), c2);
     end_if;

     if tc1 = "_exprseq" and tc2 <> "_exprseq" then
        // need to create plotdata of color c2 for the right subrectangle.
        // c1 consists of the plotdata of the left subrectangle
        return(c1, plot::inequality_plotdata(c2, right, bottom_top, showgrid));
     end_if;

     if tc1 <> "_exprseq" and tc2 <> "_exprseq" then
        // Both subrectangles are unicolored, but of different colors.
        // Need to create the plot data for both subrectangles
        return(plot::inequality_plotdata(c1, left, bottom_top, showgrid),
               plot::inequality_plotdata(c2, right, bottom_top, showgrid));
     end_if;

     // both c1, c2 are the plotdata of the subrectangles:
     return(c1, c2);
  end_if;

  //----------------------------
  // Subdivide in y direction.
  // See the previous for details
  //----------------------------
  if dir = 1 and ny > 0 then
     my:= DOM_INTERVAL::center(bottom_top);
     bottom:= subsop(bottom_top, 2 = my):
     top:= subsop(bottom_top, 1 = my):
     c1:= plot::inequality_main(f, x, y, left_right, bottom, nx, ny - 1,
                           PColor, ZColor, NColor, dir, showgrid); 
     c2:= plot::inequality_main(f, x, y, left_right, top,    nx, ny - 1,
                           PColor, ZColor, NColor, dir, showgrid);
     if c1 = c2 then return(c1); end_if;
     tc1:= type(c1):
     tc2:= type(c2):
     if tc1 <> "_exprseq" and tc2 = "_exprseq" then
        return(plot::inequality_plotdata(c1, left_right, bottom, showgrid), c2);
     end_if;
     if tc1 = "_exprseq" and tc2 <> "_exprseq" then
        return(c1, plot::inequality_plotdata(c2, left_right, top, showgrid));
     end_if;
     if tc1 <> "_exprseq" and tc2 <> "_exprseq" then
       return(plot::inequality_plotdata(c1, left_right, bottom, showgrid),
              plot::inequality_plotdata(c2, left_right, top, showgrid));          
     end_if;
     return(c1, c2);
  end_if;

/* 
  //---------------------------------------------
  // Alternatively: subdivide in both directions.
  // Note that the expensive part is not the interval
  // evaluation of the rectangles. The composition of
  // the graphical data is much more costly!
  // Subdivision in both directions creates 
  // 4 subrectangles, each producing graphical data.
  // Subdivision in one direction creates only
  // 2 subrectangles that produce graphical data. Since
  // most rectangles return // a uniform color (no need
  // for further subdivisions), the version with 
  // subdivisions in only one directions produces only
  // about half the number of unicolored rectangles than
  // produced by subdivision in both directions.
  // Hence, in most cases, the following version is about
  // twice as expensive because about twice the graphical
  // data are produced.
  //---------------------------------------
  // dir would not be used in this strategy
  //---------------------------------------
   return(
     plot::inequality_main(f, x, y, left,  bottom, nx - 1, ny - 1,
                           PColor, ZColor, NColor, dir, showgrid), 
     plot::inequality_main(f, x, y, left,  top,    nx - 1, ny - 1,
                           PColor, ZColor, NColor, dir, showgrid), 
     plot::inequality_main(f, x, y, right, bottom, nx - 1, ny - 1,
                           PColor, ZColor, NColor, dir, showgrid), 
     plot::inequality_main(f, x, y, right, top,    nx - 1, ny - 1,
                           PColor, ZColor, NColor, dir, showgrid)
   );
*/

  // error("should not arrive here");
  // Indeed, we should not arrive here.
  // Still, just as a safety measure:
  return(ZColor);
end_proc:
