/*++
Author: W. Oevel 
 3. 9. 2000: first implementation
26. 2. 2001: option Grid = n for Style = Splines added


plot::ode - plot the numerical solution of an ODE

Intro:  the ODE  dY/dt = f(t,Y), Y(t0)=Y0 is solved
        by numeric::odesolve. 
        A mesh of sample points Y0=Y(t0), Y1=Y(t1), .. is
        generated. 

        These sample points (or components of these points)
        can be plotted as curves or points

Calls:
  plot::ode(f, [t0,t1,..], Y0, 
            <method,> <Stepsize = h,> <RelativeError = tol,> 
            [G1 <, Style = style1> <, Color = c1> <, Grid = n1>],
            [G2 <, Style = style2> <, Color = c2> <, Grid = n2>],
            ...)
  plot::ode([t0,t1,..], f, Y0, 
            <method,> <Stepsize = h,> <RelativeError = tol,> 
            [G1 <, Style = style1> <, Color = c1> <, Grid = n1>],
            [G2 <, Style = style2> <, Color = c2> <, Grid = n2>],
            ...)

Parameters:
    f          - a procedure (t, Y) -> f(t,Y) representing
                 the rhs of the ODE dY/dt = f(t, Y)
    t0, t1, .. - sample points for the "time" parameter: real numerical values
    Y0         - the initial condition: a list of numerical values
    method            - see numeric::odesolve
    Stepsize = tol    - see numeric::odesolve
    RelativeError = h - see numeric::odesolve

    G1, G2, .. - the 'generators of plot data':
                 procedures (t, Y) -> F(t, Y) providing
                 the numerical data that are to be plotted
                 (see the details below).
                 These functions must accept a float t
                 and a list Y of floats. The list Y
                 represents the solutions vector Y(t)
                 of the ODE.
                 The functions G1, G2, .. must produce lists
                 with numerical values. All lists must be
                 either of length 2 (this produces a 2D plot)
                 or of length 3 (this produces a 3D plot).
    style1, style2, ..
               - the style in which the solution curve is plotted:
                 either 'Points' (points only),
                 or 'Lines' (curve with straight line segments)
                 or 'Splines' (curve via cubic spline interpolation)
    c1, c2, .. - RGB colors
    n1, n2, .. - positive integers. 
                 Besides Grid = n, also Grid = [n] is accepted.
                 The option Grid = n only has an
                 effect for generators with Style = Splines.
                 A spline curve is displayed by a total of
                 n-1 straight line seqments connecting n points
                 The default value is Grid = 5*(m-1) + 1,
                 where m = nops([t0, t1, ...]).
                 (i.e., for an equidistant time mesh t0, t1, ..., 
                 the spline solution is plotted using 4 points
                 between two mesh points of the numerical solution
                 of the ode).
                 For an equidistant time mesh t0, t1, ...,  
                 n - 1 should be an integer multiple of m - 1.
                 Otherwise, the rendered image of the spline curve
                 might not pass through the mesh points of the ode
                 solution.

Return Value: an object of type plot::Group

Details:
   - A mesh of sample points Y0=Y(t0), Y1=Y(t1), .. is
     generated via numeric::odesolve. 
     The functions G1, G2 etc. are called at these points,
     each generating a data sequence 
     (G1(t0, Y0), G1(t1, Y1), ...),
     (G2(t0, Y0), G2(t1, Y1), ...).
     Each of these sequences is regarded as a sequence
     of mesh points in R^2 (if G1, G2, .. produce lists with 2 elements)
     or in R^3 (if G1, G2, .. produce lists with 3 elements)
     Each sequence is regarded a curve, which is plotted in the
     specified style and the specified color.
     E.g.:
        G1 := (t, Y) -> [t, Y[1]]    --> x-axis == t
                                         y-axis == Y[1]
        G2 := (t, Y) -> [Y[1], Y[2]] --> x-axis == Y[1]
                                         y-axis == Y[2]
        G2 := (t, Y) -> [t, h(Y)]    --> x-axis == t
                                         y-axis == some function h(Y)

   - A combination of lines and sample points can be generated
     by using the same 'generator' twice with different styles,
     e.g.,
     plot::ode([t0,..], f, Y0, [G1, Style = Points], [G1, Style = Lines])
   - For a spline interpolation, the mesh points t0, t1, ... must
     be monotonically increasing.
     For Style = Points or Style = Lines the mesh points can be
     arbitrary.

Example1:
  We consider Y = [y1, y2, y3], dY/dt = [y2*y3-y2^2*y1, ..], Y(0)=[1,1,1]:
  We plot the components Y[2] and Y[3] against t as a 3D-curve:

>> obj:= plot::ode(
         [i $ i=0..4],
         (t, Y) -> [Y[2]*Y[3] - Y[2]^2*Y[1],
	            Y[2]*Y[1]^2 + Y[3]^2, 
		    -Y[1]*Y[2] - Y[2]*Y[3]],
	 [1, 1, 1], 
         [(t, Y) -> [t,Y[2], Y[3]], Style = Points, Color = RGB::Red],
	 [(t, Y) -> [t,Y[2], Y[3]], Style = Lines, Color = RGB::Green],
         [(t, Y) -> [t,Y[2], Y[3]], Style = Splines, Color = RGB::Blue]
        ):
>> obj::PointWidth:= 50:
>> plot(obj, Scaling = UnConstrained):

Example2: We plot the solution of dy/dt = f(t,y) together with
          a vector field plot of the right hand side of the ODE.
>> f:= (t, y) -> t*sin(t + y^2) - y:
>> p1 := plot::vectorfield([1, f(t, y)], t = 1..3, y = 0..1,
                           Grid = [10, 5], Color = RGB::Black):
>> p2 := plot::ode([i/3 $ i=0..12], (t,Y) -> [f(t, Y[1])], [0],
                [(t, Y) -> [t, Y[1]], Style = Points, Color = RGB::Red],
                [(t, Y) -> [t, Y[1]], Style = Splines, Color = RGB::Blue]
	       ):
>> plot(p1, p2, Scaling = Constrained, 
              GridLines = [Steps = 0.25, Steps = 0.5],
              Ticks = [Steps = 1.0, Steps = 0.5]):
++*/
plot::ode:= proc(f, t, Y0)
  local odesolveoptions, GG, dummy, N,
        F, Styles, Colors, Grids, AllStyles, 
        i, j, k, n, nn, Y, primitive,
        dimension, grid, XX, YY, ZZ, tt;
begin
  if args(0) < 4 then 
     error("expecting at least 4 arguments: ".
           "(procedure, [timemesh], [initial values], [plot generator])" 
          ) 
  end_if:

  // allow ode(t, f, Y0) as well as ode(f, t, Y0):
  if domtype(f) = DOM_LIST then
     [f, t]:= [t, f];
  end_if;

  // ----  checking input parameters "odesolve input"  ------------
  if type(t) <> DOM_LIST then
     error("expecting a list of sample points as first argument") 
  end_if;
  if {op(map(t, domtype@float))} <> {DOM_FLOAT} then
     error("cannot convert the sample points to real floats")
  end_if:

  [GG, odesolveoptions, dummy]:= 
      split([args(4..args(0))], x -> bool(domtype(x) = DOM_LIST));
  odesolveoptions:= op(odesolveoptions);


  // ----  checking input parameters "plot data generators" ---
  N:= nops(GG): // the number of curves to be plotted
  if N = 0 then
     error("need a procedure to generate plot data")
  end_if:

  // collect data generators, styles, colors, and grid parameters:
  F:= [0$N]:
  Styles := [Lines $ N]:
  Colors := [RGB::Black $ N]:
  Grids  := [NIL $ N]:
  AllStyles := [Points, Lines, Splines]:
  for j from 1 to N do
     F[j]:= GG[j][1];
     if domtype(F[j]) <> DOM_PROC then
        error("expecting a procedure to generate plot data"):
     end_if;
     for k from 2 to nops(GG[j]) do
         if type(GG[j][k]) <> "_equal" then
            error("expecting an equation 'Style = style' or 'Color = color',". 
                  " got '".expr2text(GG[j][k])."'");
         end_if;
         case op(GG[j][k], 1) 
         of Style do
            Styles[j]:= op(GG[j][k], 2);
            if {Styles[j]} minus {op(AllStyles)} <> {} then
               nn:= nops(AllStyles):
               error("unknown plot style '".expr2text(Styles[j])."'; ".
                     "available styles are ".
                     _concat(("'". expr2text(op(AllStyles,j)) . "', ")
                                  $ j = 1..nn-1,
                              "and '". expr2text(op(AllStyles,nn)) . "'"
                            )
                     );
            end_if:
            break;
         of Color do
            Colors[j]:= op(GG[j][k], 2);
            if domtype(Colors[j]) <> DOM_LIST then
               error("illegal color specification");
            end_if:
            break;
         of Grid do
            Grids[j] := op(GG[j][k], 2);
            // Accept 'both Grid = n and 'Grid = [n]'.
            if domtype(Grids[j]) = DOM_INT then
               Grids[j]:= [Grids[j]]
            end_if;
            if domtype(Grids[j]) <> DOM_LIST or
               nops(Grids[j]) <> 1 or
               domtype(op(Grids[j])) <> DOM_INT or
               op(Grids[j]) < 2
               then
               error("illegal Grid specification. ".
                     "Expecting 'Grid = n' or ".
                     "'Grid = [n]' with some integer n > 1.");
            end_if:
            break;
         otherwise
            error("expecting arguments of the form " .
                  "'[procedure, Style = style, Color = RGB color]'"):
         end_case;
     end_for:
  end_for:

  // --- the work starts ------------

  // --- solve the ode: ---
  n:= nops(t);  // the number of sample points
  if n < 2 then 
     error("the first argument must be a list with at least 2 sample points")
  end_if;

  Y:= [[t[1], Y0], 0 $ n-1]:  // initialize container for sample points
  userinfo(2, "starting to solve the ode"):
  for i from 2 to n do
     userinfo(3, "computing ".expr2text(i)."-th sample point:");
     Y[i]:= [t[i], numeric::odesolve(t[i-1]..t[i], f, Y[i-1][2], odesolveoptions)];
     userinfo(3, "--- time: ".expr2text(t[i]) . ", solution: ". expr2text(Y[i][2]));
  end_for:
  userinfo(2, "the ode is solved"):
  // --- the ode is solved. ----

  // --- generate plot data ----
  userinfo(2, "starting to generate plot data"):
  primitive:= [0 $ N];  // initialize container for primitives
  for j from 1 to N do  // the j-th curve generated
                        // by F[j] with Styles[j], Color[j]
     case Styles[j]
     of Points do
        primitive[j]:= plot::Pointlist(
                 plot::Point(F[j](op(Y[i]))) $ i=1..n,
                 Color = Colors[j]);
        break;
     of Lines do
        primitive[j]:= plot::Polygon(
                 plot::Point(F[j](op(Y[i]))) $ i=1..n,
                 Color = Colors[j]);
        break;
     of Splines do
        dimension:= nops(F[j](op(Y[1]))):
        if Grids[j] <> NIL then
             grid := Grids[j]
        else grid := [5*(n-1) + 1];  // default
        end_if;
        case dimension
        of 2 do
           XX:= numeric::cubicSpline([t[i], F[j](op(Y[i]))[1]] $ i = 1..n):
           YY:= numeric::cubicSpline([t[i], F[j](op(Y[i]))[2]] $ i = 1..n):
           tt:= genident("t"):
           primitive[j]:= plot::Curve2d(
	         [XX(tt), YY(tt)],
                 tt = t[1]..t[n],
		 Color = Colors[j],
		 Title = "",
                 Discont = FALSE,
		 Grid = grid);
           break;
        of 3 do
           XX:= numeric::cubicSpline([t[i], F[j](op(Y[i]))[1]] $ i = 1..n):
           YY:= numeric::cubicSpline([t[i], F[j](op(Y[i]))[2]] $ i = 1..n):
           ZZ:= numeric::cubicSpline([t[i], F[j](op(Y[i]))[3]] $ i = 1..n):
           tt:= genident("t"):
           primitive[j]:= plot::Curve3d(
	         [XX(tt), YY(tt), ZZ(tt)],
                 tt = t[1]..t[n], 
		 Color = Colors[j], 
		 Title = "", 
                 // Discont = FALSE,// presently, plot3d does not support Discont
		 Grid = grid);
           break;
        otherwise error("generators of plot data must produce lists with " .
                        "2 or 3 elements"):
        end_case;
        break;
     otherwise
        error("unknown style"):
     end_case;
  end_for;

  plot::Group(primitive[j] $ j = 1..N);

end_proc:
