/*-------------------------------------------------------------------

faclib::knapsack -- combine modular factors via the 
                    van Hoeij Algorithm

Introduction -- faclib::knapsack(f,mf,p,a) will try to combine the 
                modular factors in mf of f which were determined 
                up to accuracy p^a to find the proper rational 
                factors of f.
 
Call -- faclib::knapsack(f, mf, p, a <, GetBound>)

Parameters -- f  - a monic squarefree, univariate polynomial
                   with integer coefficients.
              mf - a list of modular factors
              p  - a prime number
              a  - a positive integer - the accuracy up to which 
                   the modular factors are determined, i.e.
                   f = _mult(op(mf)) mod p^a

Options: -- GetBound - if GetBound is specified, faclib::knapsack
                       will return an integer a such that
                       p^a will be a sufficiently high power of
                       p for the algorithm to run. It is called
                       by faclift::plift to determine how far to lift.

Return value -- either FAIL, which will mean that additional 
                lifting is necessary or a list L of polynomials
                over Expr such that f = _mult(op(L)).
                
Description -- for a detailed description see 
               Marc van Hoeij: "Factoring Polynomials and the 
               Knapsack problem", TCS, to appear.
                              
               It is used only if there are many (> 16) modular
               factors left after the lifting process.
               
               The algorithm employs lattice reduction on
               truncated traces of the modular factors.
               Empirically it is much faster than the original
               LLL approach and -- for more than about 12 factors
               -- the Zassenhaus factorcombination.
               
               It factors S-D-polynomials reasonably fast, but 
               should be used only if there are 'many' modular
               factors left. It is called from faclib::plift.
               
Author: Peter Horn, hornp@upb.de, ver. 0.9b, Sept. 2001
-------------------------------------------------------------------*/

faclib::knapsack := proc(f : DOM_POLY, 
                        mf : DOM_LIST, 
                        p : DOM_INT, 
                        a : DOM_INT) 
local N, d, D, n, C, M, st, r, k, tmp, rf, need_ab, new_f,
     B, L, Lred, Lgso, i, j, ai, bi, T, bnd, 
     min_cutsize, cut_inc, min_traces, trace_inc,
     rfs, had_mystery, ltranspose;
begin
    userinfo(1, "Knapsack-algorithm called...");
    userinfo(9, "With ".expr2text(nops(mf))." modular factors, p^a=".
        expr2text(p)."^".expr2text(a)." deg f=".expr2text(degree(f)));
    
    //-----------------------------------------------------
    // Further checks on the arguments
    //-----------------------------------------------------
    if testargs() then
        // not yet implemented
    end_if;
    
    //-----------------------------------------------------
    // At first we will do some initializations
    //-----------------------------------------------------
    // needed functions
    ltranspose := A -> [map(A,_index,i) $ i=1..nops(A[1])];
    // needed because of crude lllint-syntax

    // other values, mainly independent of f
    new_f := TRUE; // need initialization of N, D,...
    need_ab := TRUE; // need to calculate new cut-borders
    n := nops(mf); // the number of modular factors
    rfs := 0; // the number of identified rational factors
    rf := []; // the already found rational factors 

    // Constants
    min_cutsize := max(ceil(log(p,2^n)),5); // ai-bi is initially...
    cut_inc := 2; // how much ai-bi is to be increased
    min_traces := 2; // look at at least so many traces
    trace_inc := 2; // how many traces are to be added if necessary

    //-----------------------------------------------------
    // We only have to calculate a lifting-bound and exit.
    //-----------------------------------------------------
    if args(0) = 5 and args(5) = hold(GetBound) then
        // faclib::knapsack was called only to get the information
        // how far to lift.
        bnd := polylib::rootbound(f);
        N := degree(f);
        D := N/2;
        d := max(D, min_traces+5*trace_inc);
        i := ceil(log(p,(bnd^(d))*N));
        // p^i is now bigger than N*alpha^(N/2) for
        // any root alpha of f
        j := i + (min_cutsize+3*cut_inc);
        // j-i now is min_cutsize+3*cut_inc, which should suffice
        userinfo(1,"calculated sufficiently high power: ".p."^".j);
        return(j);
    end_if;
    
    //-----------------------------------------------------
    // the main-loop, in any ending case the routine will
    // be left via a return, therefore I use the infinite
    // loop.
    //-----------------------------------------------------
        
    while TRUE do    
    
        //-----------------------------------------------------
        // if necessary calculate all constants that depend
        // vitally on f.
        //-----------------------------------------------------
        if new_f then
            N := degree(f);
            D := (N div 2); // If we look at D traces everything will 
                // be mathematically safe.
            n := nops(mf); // The number of modular factors
            bnd := ceil(polylib::rootbound(f));    
                // all roots are absolutely less than bnd
            had_mystery := FALSE; // no strange things happened yet.

            userinfo(9,"calculated rootbound for f: ".expr2text(bnd));
                
            // Remark on B and r:
            // It may happen in the process that we find that
            // in the reduced lattice there are r M-short rows 
            // which do not correspond to factors. Then we will
            // continue with the lattice 'created' from these
            // r rows which are coded by B.
            // Consequently we start with r=n, B=[{1},..,{1}]
            r := n; 
            B := [{i} $ i=1..r]; // any factor is coded by one 
            need_ab := TRUE; // need new cut-bounds
        end_if; // got new polynomial f

        d := min(D,min_traces); // anyway, we will look only at these traces

        // Remark on C and M:
        // The C is responsible for a proper factor being 'shorter'
        // than the product of two proper factors.
        // All proper factors are 'shorter' than M.
        // they need to be computed in every pass, because d may change.
        C := ceil(sqrt(d*n)/2);
        M := float(sqrt(C^2*n + d*(n/2)^2));
            
        //-----------------------------------------------------
        // calculate the borders for the cuts.
        // For the sake of safety we will calculate these for
        // all D traces, though we might use far fewer.
        //-----------------------------------------------------
        if need_ab then
            st := time();
            ai := bi := [];
            for i from 1 to D do
                bi := bi.[ceil(log(p,(bnd^i)*N))]; // the lower cut-border
                ai := ai.[bi[i]+min_cutsize]; // the upper cut-border
                if ai[i] >= a then // oops, need re-lifting
                    if i <= d then
                        userinfo(9,"More Lifting necessary, aborting.");
                        return(FAIL);
                    elif i = d then
                        userinfo(9,"More Lifting may become necessary,".
                            " continuing anyway.");
                    end_if;
                end_if; // ai[i] >= a
            end_for; // i=1..D
            userinfo(9,"ai and bi calculated in time ".expr2text(time()-st));
        end_if;
        userinfo(9,"Variables: D=".expr2text(D)." d=".expr2text(d).
            " C=".expr2text(C)." M=".expr2text(M)." min_cutsize="
            .expr2text(min_cutsize));
                
        new_f := need_ab := FALSE;
    
        //-----------------------------------------------------
        // calculate the truncated traces T:
        //
        //            (Tr_j(f[i]) - (Tr_j(f[i]) mods p^bi[i]))
        // T[i][j] = ------------------------------------------
        //                            p^bi[i]
        //                                 mods p^(ai[i]-bi[i])
        // for i=1..n, j=1..D
        //-----------------------------------------------------
        st := time();
        T := [0 $ n];
        for i from 1 to n do
            // calculate traces modulo p^ai[j]
            tmp := [polylib::trace(
                poly(mf[i], op(f,2), IntMod(p^ai[j])) , j) $ j=1..d];
            // make the LEFT-sided cut
            tmp := [(tmp[j]-mods(tmp[j], p^bi[j]))/p^bi[j] 
                                    $ j=1..d];
            T[i] := tmp;
        end_for; // i=1..d
        userinfo (9,"T calculated in time ".expr2text(time()-st));
                
        //-----------------------------------------------------
        // Now set up the basis for the lattice L, it initially
        // looks like this:
        //
        // C            T[1][1]  T[1][2]  ..  T[1][d]
        //    C         T[2][1]  T[2][2]  ..  T[2][d]
        //       ..        ..       ..    ..     ..
        //          C   T[n][1]  T[n][2]  ..  T[n][d]
        //              p^(a1-b1)
        //                       p^(a2-b2)
        //                                ..
        //                                    p^(ad-bd)    
        //
        // if e.g. n=6, d=2, B=[{1,2,4}, {3}, {5,6}] it 
        // will be:
        //
        // C  C     C        T[1,1]+T[2,1]+T[4,1]  T[1,2]+T[2,2]+T[4,2]         
        //       C           T[3,1]                T[3,2]
        //             C  C  T[5,1]+T[6,1]         T[5,2]+T[6,2]
        //                   p^(a1-b1)
        //                                         p^(a2-b2)
        //
        // (All empty fields represent zeros.)
        //-----------------------------------------------------
        st := time();
        L := [([_if(j in B[i], C, 0) $ j=1..n]. 
            [_plus(T[k][j] $ k in B[i]) $ j=1..d]) $i=1..r,
            [_if(j-n=i, p^(ai[i]-bi[i]), 0) $ j=1..n+d]
            $ i=1..d];    
        userinfo (9,"L calculated in time ".expr2text(time()-st));
        
        //-----------------------------------------------------
        // let LLL reduce the basis L
        //-----------------------------------------------------
        st := time();
        // here we need two calls of ltranspose -- ask the 
        // designers of lllint why!!!
        if traperror((Lred := 
                ltranspose(op(lllint(ltranspose(L),All),2)))) <> 0 then
            print(Unquoted, "Hint: If you encounter the following error frequently, you should think\n".
              "      about increasing the initial PARI-stacksize with the option\n".
              "      -a SizeInBytes on Unix or via the appropriate GUI-command.");
            lasterror();
        end_if;
        userinfo (9, "Reduced basis L in time  ".expr2text(time()-st));
        
        //-----------------------------------------------------
        // orthogonalize basis and calculate the norm-vector,
        // then look which rows of Lred span the space of all
        // M-short vectors.
        // Remark: faclib::myorthog will return ONLY a vector 
        // containing the norms.
        //-----------------------------------------------------
        st := time();
        Lgso := faclib::myorthog(Lred,hold(Float)); // GS-orthogonalize reduced basis
        // now get a vector with the row-norm
        // Lgso := map(Lgso, x->
        //    sqrt(_plus(x[i]^2 $ i=1..nops(x))));
        userinfo (9, "Orthogonalized basis Lred in time "
            .expr2text(time()-st));
        // now look which rows are shorter than M
        // we check for M*1.1 to be sure that float-errors
        // are dealt with.
        r := nops(Lgso);
        while _lazy_and(r>0, Lgso[r]>M*1.1) do r := r-1 end_while;
        userinfo(9,"There are r=".expr2text(r)." 'short' rows"); 

        //-----------------------------------------------------
        // now we will evaluate the results of the lattice-reduction
        // and -- if necessary -- construct the new B
        //-----------------------------------------------------
        if r=0 then
            if had_mystery then
                userinfo(1, "That's mysterious, no short rows! ".
                     "Aborting.");
                return(FAIL);
            else
                userinfo(1, "That's mysterious, no short rows! ".
                     "One more try.");
                had_mystery := TRUE; // it once went wrong.
                r := n;
                B := [{i} $ i=1..n ];
            end_if;
            
        //-----------------------------------------------------
        // the polynomial is irreducible
        //-----------------------------------------------------
        elif r=1 then
            userinfo(9,"The (remaining) polynomial is irreducible");
            rf := rf.[poly(_mult(mf[i] $ i=1..n),Expr)];
            return (rf);

        //-----------------------------------------------------
        // we now know that there are at most r rational factors.
        // from the construction we know that r<n/3 since all 3-products
        // where tested during the lifting-process
        //-----------------------------------------------------
        elif r>floor(n/3) then
            //-----------------------------------------------------
            // Ooops -- obviously we put too little information
            // into the lattice -- we need to 
            // increase the information put into the lattice.
            //-----------------------------------------------------
            userinfo (9, "In the GSO-reduced basis there are ".expr2text(r).
                " short vectors -- need to increase the a's!");    
            min_cutsize := min_cutsize + cut_inc;
            min_traces := min_traces + trace_inc;
            need_ab := TRUE;
            // Anyway, these values need to be restored.
            r := n; 
            B := [{i} $ i=1..r];
            st := time();    
        else

            //-----------------------------------------------------
            // OK, there were several short vectors, i.e. some
            // possibly rational factors
            //-----------------------------------------------------
            // find the factor-coding defined by the
            // 'short' rows.
            B := [ FAIL $ r ]; 
            for i from 1 to r do
                if {op(map(Lred[i],abs),1..n)} = {0, C} then
                    B[i] := select({$ 1..n}, k -> (Lred[i][k] <> 0));
                end_if;
            end_for;
            B := select (B, _unequal, FAIL);
            r := nops(B);
            // check whether the B do partition {1..n}
            if r = 0 then
                // these B are unusable
                userinfo(9, "Result is unusable, increasing Information");
                r := n;
                B := [{i} $ i=1..n];
                min_cutsize := min_cutsize + cut_inc;
                min_traces := min_traces + trace_inc;
                need_ab := TRUE;
            else
                // now check, which of these are true factors
                for j from 1 to r do
                    // construct the product coded by B[j]
                    rfs := rfs+1;
                    tmp := poly(_mult(mf[i] $ i in B[j]),Expr);
                    if tmp <> FAIL then // there were no 'used' factors.

                        rf := rf.[tmp];
                        tmp := divide(f,rf[rfs],Exact); // trial division

                        if tmp <> FAIL then

                            // Yippie, found rational factor
                            f := tmp; //use the new f
                            new_f := TRUE;

                            // remove the 'used' modular factors
                            // we replace them by FAIL and strip them
                            // away in a next step.
                            for i in B[j] do
                                mf[i] := FAIL; 
                            end_for; // i in B[j]
                            B[j] := FAIL;
                        else // it was NOT a true factor
                            delete (rf[rfs]);
                            rfs := rfs-1;
                        end_if; // found true factor
                    end_if; // 'used' factors?
                end_for; // i=1..r, walk over the possibly rational factors
                
                //-----------------------------------------------------
                // Now remove the 'used' modular factors and modify B
                // in the appropriate way
                // Return if all modular factors are 'used up'
                //-----------------------------------------------------
                mf := select(mf, _unequal, FAIL);
                B := select(B, _unequal, FAIL);
                    
                // Now look whether there is still work to be done
                if nops(B) = 0 then // there's nothing left to do.
                    userinfo(9,"Knapsack done, returning ".expr2text(rfs)
                        ." factors");
                    return (rf);
                end_if;
                
                //-----------------------------------------------------
                // if there were non-true factors increase d to get more
                // information into the lattice
                //-----------------------------------------------------
                min_traces := min_traces + trace_inc;
            end_if; // B is not a partition.
        end_if; // r = ...
    end_while; // TRUE
    
    assert(FALSE);
    // no return here since the while-loop will be exited via return.
    
end_proc: //  faclib::knapsack

/*-------------------------------------------------------------------

faclib::myorthog -- compute the Gram-Schmidt-orthogonalization and
                    return the norm of each vector.
 
Call -- faclib::myorthog(S <, Float>)

Parameters -- S  - a list of lists representing a matrix

Options -- Float - when Float is specified, the GSO will be computed
           using floats.

Return value -- a list contraining the Euclidean lengths of the
                vectors created by the GSO.
                
Description -- This function is intended ONLY for use in 
               faclib::knapsack.
               
Author: Walter Oevel, Peter Horn , hornp@upb.de, ver. 1.0, Sept. 2001
-------------------------------------------------------------------*/

faclib::myorthog := proc(S) // S = a list of n sublists
local n, m, OS, j, k, ss, s, tmp, normsquares, term, v;
save X;
begin
     if args(0) = 2 and args(2) = hold(Float) then
        S:= map(S, map, float);
     end_if;
     delete (X);
     n := nops(S);
     m := nops(S[1]); // the length of the vectors (=lists)

     // initialize container 'OS' = list of polynomials
     OS := [poly(0, [X]) $ n]; 

     // initialize container 'normsquares' to store the
     // squares norms of the orthogonolized vectors
     normsquares:= [ 0 $ n]; 

     // Gram-Schmidt of input lists S.1, S.2, .. , S.n:
     // b.j = S.j - sum(<S.j, b.k>/<b.k,b.k>*b.k,  k=1..j-1)
     // We use polynomials OS.k = b.k / <b.k, b.k> for speed
     // to compute the linear combinations of the vectors.
     // The polynomial OS.k and the list b.k are kept simultaneously.
     // The orthogonalized b.k is again stored in S.k!
     // Thus, the recursion is
     // a) OS.j = convert S.j (= list) to poly
     // b) OS.j = OS.j - sum(<S.j, S.k>*OS.k,  k=1..j-1)
     // c) S.j = convert OS.j (= poly) to list
     // d) OS.j = OS.j / <S.j, S.j> 

     OS[1]:= poly([[S[1][k], k] $ k = 1..m], [hold(_X)]):
     normsquares[1]:=  _plus(op(map(S[1], _power, 2)));
     if not iszero(normsquares[1]) then
        OS[1]:= mapcoeffs(OS[1], _mult, 1/normsquares[1]);
     end_if;

     for j from 2 to n do
       // convert list S[j] to polynomial S[j]
       OS[j]:= poly([[S[j][k], k] $ k = 1..m], [hold(_X)]):
       ss:= OS[j]: // ss = poly = S.j - sum(<S.j, b.k>/<b.k,b.k>*b.k, k=1..j-1)
       for k from 1 to j - 1 do
             if iszero(normsquares[k]) then next end_if;
             // scalar product
             s:= _plus(op(zip(S[j], S[k], _mult)));
             if not iszero(s) then
                 ss:= ss - mapcoeffs(OS[k],_mult,s );
             end_if
       end_for;
       OS[j] := ss;
       // reconvert poly to list
       v:= [0 $ m]: 
       tmp:= poly2list(OS[j]):
       for term in tmp do
           v[term[2]]:= term[1];
       end_for;
       S[j]:= v;
       normsquares[j]:=  _plus(op(map(S[j], _power, 2)));
       if not iszero(normsquares[j]) then
          OS[j]:= mapcoeffs(OS[j], _divide, normsquares[j]);
       end_if;
     end_for;
     
     // anyway -- we just need the norms:
     return(map(normsquares, sqrt));
     //return(S);
end_proc:
