/* --------------------------------------------------------------- 

   rootbound -- returns a bound on the absolute value of a
                polynomials roots.

Call:    polylib::rootbound(p);

Parameter:  p -- a univariate polynomial (DOM_POLY).  
                 Its coefficients must be integers or rationals 
                 or convertible to floats via MuPADs float command.

Return value: a float value.

Synopsis:    
              The return value is an upper bound on the absolute value
              for the roots of the polynomial p, i.e.:
              
              p(x) = 0 ==> |x| <= polylib::rootbound(p)
              
              Three different bounds are calculated, from which
              the smallest one is chosen.
              
              For details see W. Oevel: "Einfuehrung in die numerische
              Mathematik", Springer 1996, pp 69--70.
              
Remark:       The computation uses floating-point arithmetic.
              
Personal remark:  should be improved.

Examples:

  >> polylib::rootbound(poly(x^5+2*x^4+3*x^3+4*x^2+5*x+6,[x]));

                                    4.0
  >> polylib::rootbound(poly(x^5+x^3+x-1,[x]));

                                    2.0
  >> polylib::rootbound(poly((x-3)*(x-1)*(x+5),[x]));

                                8.246211251
  >> polylib::rootbound(poly(x^2+10*x+300,[x]));

                                   30.0

Author: Peter Horn 8-8-2001 hornp@upb.de
--------------------------------------------------------------- */

polylib::rootbound := proc(p : DOM_POLY)
option remember;
local b1, b2, b3, L, n, i, xdiv;
begin
	userinfo(9,"polylib::rootbound called with polynomial", p);
	
	// get the coefficient-vector of p 
	n := degree(p);
	L := array(0..n, [float(abs(coeff(p,i))) $ i=0..n]);

	// (a) first bound
	// max{|p0/pn|, 1+|p1/pn|,...,1+|pn-1/pn|}
	b1 := max(L[0]/L[n], 1+L[i]/L[n] $ i=1..n-1);
	userinfo(1,"polylib::rootbound: First bound: ", b1);

	// (b) second bound
	// max{|p0/p1|, 2*|p1/p2|,...,2*|pn-1/pn|}
	xdiv := proc(a,b) begin if iszero(b) then return(infinity) else
		return(a/b) end_if; end_proc;
	b2 := max(xdiv(L[0],L[1]), 2*xdiv(L[i],L[i+1]) $ i=1..n-1);
	userinfo(1,"polylib::rootbound: Second bound: ", b2);
	
	// (c) third bound
	// 2*max{|pn-1/pn|, |p0/pn|^(1/n),...,|pn-2/pn|^(1/2)}
	b3 := 2*max(L[n-1]/L[n], (L[i]/L[n])^(1/(n-i)) $ i=0..n-2);
	userinfo(1,"polylib::rootbound: Third bound: ", b3);
	
	// return the smallest one of the three.
	return(min(b1,b2,b3))
end_proc;
