/* --------------------------------------------------------------- 

   trace -- returns the i-th trace of a integer polynomial

Call:    polylib::trace(p,i);

Parameter:  p -- a monic univariate polynomial (DOM_POLY),
                 if p is not monic, it will be divided by its
                 lcoeff, if that is not possible, an error will be
                 returned.
                 
            i -- a natural number >= 1

Return value: an integer representing the i-th trace mods m.

Synopsis:    
              The i-th trace of a polynomial is the sum of
              the i-th powers of all its roots. 
              
              If a m is present, the symmetric remainder will be done
              in every step of the computation, i.e. it will be faster
              to call trace(p,i,m) than mods(trace(p,i),m).
                            
Remark:       The trace is determined without calculating the roots, 
              instead the Newton-Identities are used to calculate
              the trace efficiently.
              
              This function is mainly used in faclib::knapsack.
              
Personal remark: n/a

Examples: n/a

Author: Peter Horn 8-8-2001 hornp@upb.de
--------------------------------------------------------------- */

polylib::trace:=proc(f : DOM_POLY,
					i : DOM_INT)
local d, k, x, tmp, e, R, l, m, mmod;
option remember;
begin
	userinfo(1, i."th trace of",f,"requested.");
	if testargs() then
		if i<0 then 
			error ("can't compute ".i."th trace.");
		end_if;
	end_if;

	// Function e needed 'cause coeff will return error if number is < 0.
	e := n -> _if((n < 0), 0, coeff(f,n));
	mmod := (a,b) -> _if((b=0), a, a mod b); 

	x := op(op(f,2),1);
	R := op(f,3);
	
	if (op(R,0) = IntMod) then
		// We got a polynomial in Z/mZ so we can calculate efficiently
		// reducing our intermediate results modulo m.
		m := op(R,1);
	else
		m := 0;
	end_if; // poly over IntMod(m)
	
	d:=degree(f,x);
	if (l := lcoeff(f, [x])) <> 1 then
		if 0 <> traperror((f := mapcoeffs(f,_divide,l))) or (f = FAIL) then
			error ("Polynomial should be monic or divisible by lcoeaff.");
		end_if;
	elif i=0 then
		return(d);
	else
		// Newton identities
		tmp := - _plus(mmod(polylib::trace(f,k,m)* e(d-i+k), m) $ k=1..i-1)-i*e(d-i);
		return (mmod(tmp,m));
	end_if;
end_proc: // polylib::trace

