// $Date: 2001/11/25 21:44:55 $ $Author: bij $ $Revision: 1.1 $

/*
prog::maketest(f, <Options>)

  manipulates the procedure or function environment 'f'
    every following call of 'f' produces a 'prog::test'
    call that is written into a specified file

prog::maketest(f, Open = filename, <Proc = procedure, Text = message>)

  opens the file 'filename' (complete name must be given)
    and manipulates 'f'

  The optional argument 'procedure' is a procedure, that
  is called, when 'f' is entered and leaved.
  The arguments given to 'procedure' are
    1. the string "enter" or "leave" as first argument, when 'f'
       is entered resp. leaved
    2. all arguments given to 'f'
  'procedure' must return a string, that is printed to the
  opened file 'filename' instead of the default 'prog::test'
  call.

prog::maketest(f, Quit|Restore|Close)

  closes the opened file and rebuilds 'f' to the original
  procedure or function environment

*/

/* This table will contain the original code of all
   manipulated functions and is used for rebuilding */

prog::mktestTab := table():

// only first call, no recursive calls
prog::mktest_depth:= -1:

alias(AllOptions = {hold(Open), hold(Append), hold(Write), hold(File),
                    hold(Restore), hold(Quit), hold(Proc), hold(Text),
                    hold(Close),
                    hold(All)}):
// param options    hold(Open) = <filename>
//                  hold(File) = filename
//                  hold(Proc) = procedure
//                  hold(Text) = string

prog::maketest:=
proc(_f_)
  option hold, escape;
  local FILE, FILENAME, OPT, tabIndex,
        ff, p, mktest_internal, mktest_proc, mktest_func_env;
begin
  if args(0) = 0 then
    error("first argument must be a procedure or function environment")
  end_if;

  mktest_internal :=    /* mktest for kernel functions */
  proc(p,n)
    option escape, noDebug;
  begin
    subsop(proc()
             local  lastResult, _x_1_, _arguments_;
             option noDebug;
           begin
             hold(id)(NIL, "produces tests");
             sysassign(prog::mktest_depth, prog::mktest_depth + 1);
             _x_1_ := p;
             _arguments_ := args();
             if prog::mktest_depth < 1 then
               fprint(NoNL, FILE, "prog::test(".n."(".expr2text(_arguments_)."), ");
             end_if;
             hold(_x_1_)(_arguments_);
             lastResult := eval(%);
             if prog::mktest_depth < 1 then
               fprint(Unquoted, FILE, expr2text(lastResult)."):")
             end_if;
             sysassign(prog::mktest_depth, prog::mktest_depth - 1);
             lastResult;
           end_proc,
           6 = _f_,
           [4, 1, 1] = n) // name
  end_proc;
	       
  mktest_proc :=      /* mktest for procedures */
  proc(p, n)
    name mktest_proc;
    local mktest_exit;
    option escape;
  begin

    mktest_exit :=    /* print result on exit ; also used for return */
      subsop(proc(n)
               option escape;
               name mktest_exit;
             begin
               proc()
               begin
                 if prog::mktest_depth < 1 then
                   fprint(Unquoted, FILE,
                          expr2text(args()), "):")
                 end_if;
                 sysassign(prog::mktest_depth, prog::mktest_depth - 1);
                 args()
               end_proc
             end_proc);

    /* create new Proc with mktest-token, enter-info, old code, exit-info */
    subsop(p, 4 =
           hold(_stmtseq)(
                          hold(id)(n, "produces tests"),
                          NIL,
                          hold(_if)(NIL,
                                    hold(fprint)(NoNL, FILE,
                                                 "prog::test(",
                                                 prog::getname(_f_),
                                                 "(",
                                                  hold(expr2text(args())),
                                                 "), "),
                                    NIL),
                          hold(_exprseq)(subs(op(p,4),hold(return)=return@mktest_exit(n),
                                              Unsimplified)),
                          mktest_exit(n),
                          hold(%(%2))),
           [4,3,1] = hold(_less)(hold(slot)(prog, "mktest_depth"), 1),
           [4, 2] = hold(sysassign(prog::mktest_depth, prog::mktest_depth + 1)),
	   Unsimplified)
  end_proc;

  // mktest function environment 'ENV'
  mktest_func_env:=
    proc(ENV, ENVID)
      local Pr1;
      name prog::mktest::mktest_func_env;
    begin
      // manipulate only the "main" procedure
      userinfo(1, "manipulate function environment '".prog::getname(ENV)."'");
      Pr1:= op(ENV, 1);
      case domtype(Pr1)
	of DOM_PROC do
	  Pr1:= mktest_proc(op(ENV, 1), expr2text(ENVID));
	  break;
	  
	of DOM_EXEC do
	  Pr1:= mktest_internal(op(ENV, 1), expr2text(ENVID));
	  break;
      end_case;
        
      subsop(ENV, 1 = Pr1)
    end_proc;

  // -- BEGIN -------------------------

  // functions that cannot be manipulated
  if contains({hold(fprint), hold(hold), hold(DOM_SET::sort)}, _f_) then
    error("this function cannot be manipulated")
  elif contains({hold(_concat), hold(_procdef), 
                 hold(length), hold(Pref), hold(strmatch),
                 hold(anames), hold(type), hold(substring), hold(domtype),
                 hold(contains), hold(sysassign)}, _f_) then
    warning("this function should not be manipulated, unexpected sideeffects breaks out")
  elif Pref::typeCheck() <> hold(None) and contains({hold(testtype)}, _f_) then
    error("this function cannot be manipulated, when 'Pref::typeCheck' is not 'None'")
  end_if;

  ff:=context(_f_);
  prog::init(ff);

  // extract options
  OPT:= prog::getOptions(args(), AllOptions, 2); // skip filename
  if contains(OPT, hold(File)) then
    FILENAME:= OPT[hold(File)]
  elif contains(OPT, hold(Open)) and domtype(OPT[hold(Open)]) = DOM_STRING then
    FILENAME:= OPT[hold(Open)]
  else
    FILENAME:= prog::getname(_f_).".tst"
  end_if;

  if contains(OPT, hold(Open))
     or (not (contains(OPT, hold(Close))
              or contains(OPT, hold(Quit)) or contains(OPT, hold(Restore)))
         and not contains(prog::mktestTab, _f_)) then
    // open file
    if contains(OPT, hold(Append)) or contains(OPT, hold(Write)) then
      if contains(OPT, hold(Append)) then
        if (FILE:= fopen(FILENAME)) <> FAIL then
          fclose(FILE);
          if (FILE:= fopen(hold(Text), FILENAME, hold(Append))) = FAIL then
            error("cannot open '".FILENAME."' for append")
          else
            fprint(Unquoted, FILE, "") // newline
          end_if
        else
          error("cannot open '".FILENAME."' for append")
        end_if
      else // write
        if (FILE:= fopen(FILENAME)) <> FAIL then
          fclose(FILE);
          warning("file '".FILENAME."' was overwritten.")
        end_if;
        if (FILE:= fopen(hold(Text), FILENAME, hold(Write))) = FAIL then
          error("cannot open '".FILENAME."' for write")
        end_if
      end_if
    else
      // don't overwrite existing file, don't append automatically
      FILE:= fopen(FILENAME);
      if FILE = FAIL then // error
        if (FILE:= fopen(hold(Text), FILENAME, hold(Write))) = FAIL then
          error("cannot open '".FILENAME."' for write")
        end_if
      else // append
        fclose(FILE);
        error("file '".FILENAME."' exists. Use option 'Append' or 'Write'.")
      end_if
    end_if;
      
    // file is open, prepare test
    fprint(Unquoted, FILE, "// generated test file for '".prog::getname(_f_)."'");
    fprint(Unquoted, FILE, ""); // newline
    fprint(Unquoted, FILE, "prog::testinit(\"".prog::getname(_f_)."\"):"); // 
    fprint(Unquoted, FILE, "prog::testfunc(".prog::getname(_f_)."):"); // 
    fprint(Unquoted, FILE, ""); // newline

    // remember FILE for closing
    sysassign(prog::mktestTab[_f_, hold(File)], FILE);
      
    case domtype(ff)
      /*  ff is a internal function or procedure with slot's */
      of DOM_FUNC_ENV do
        userinfo(1, "manipulate function environment '".prog::getname(ff)."'");
        sysassign(prog::mktestTab[_f_], ff):
        stdlib::syseval(evalassign(_f_, mktest_func_env(ff, _f_), 1));
        break
      of DOM_PROC do
        userinfo(1, "manipulate procedure '".prog::getname(ff)."'");
        if type(_f_) = "slot"
           and domtype(context(op(_f_, 1))) = DOM_DOMAIN then
          sysassign(prog::mktestTab[_f_], ff);
          stdlib::syseval((evalassign(_f_,mktest_proc(ff, expr2text(_f_)), 1)))
        else
          sysassign(prog::mktestTab[_f_], ff):
          stdlib::syseval((evalassign(_f_,mktest_proc(ff, expr2text(_f_)), 1)))
        end_if;
        break
      otherwise
        error("first argument must be a procedure or function environment")
    end_case;

    fprint(Unquoted, 0, "Info: function '".prog::getname(_f_)
                        ."' is manipulated, file '".FILENAME."' is opened.")
  else // quit | close

    if contains(prog::mktestTab, _f_) then
      FILE:= prog::mktestTab[_f_, hold(File)];
      fprint(Unquoted, FILE, "");
      fprint(Unquoted, FILE, "prog::testexit():");
      fprint(Unquoted, FILE, "");
      fprint(Unquoted, FILE, "// end of testfile");
      fclose(FILE)
    elif contains(OPT, hold(Close)) then
      error("test file is not opened")
    else
      error("function was not manipulated")
    end_if;

    if type(_f_) = "slot" then
      if type(_f_) = "slot"
         and domtype(context(op(_f_, 1))) = DOM_DOMAIN then
        sysdelete(prog::mktestTab[op(_f_, 1)])
      end_if;
      tabIndex := hold(slot)(op(_f_,1),op(_f_,2));
      if not contains(prog::mktestTab, tabIndex) then
        error("function was not manipulated")
      end_if;
   
      stdlib::syseval(evalassign(_f_, prog::mktestTab[tabIndex], 2));
      sysdelete(prog::mktestTab[tabIndex]);
    else
      p:= eval(_f_);
      if not contains(prog::mktestTab, _f_) then
        error("function was not manipulated")
      end_if;
      stdlib::syseval(evalassign(_f_, prog::mktestTab[_f_], 2));
      sysdelete(prog::mktestTab[_f_]);
    end_if;
    
    fprint(Unquoted, 0, "Info: function '".prog::getname(_f_)
                        ."' is restored, file '".FILENAME."' is closed.")

  end_if
end_proc:
