// $Date: 2002/02/28 14:13:12 $ $Author: bij $ $Revision: 1.116 $

//------------------------------------------------------------
// W. Oevel, 15.11.01
// Special code for fast decisions
//       is(x < y), is(x <= y), is(x = y), is(x <> y)
// implemented (for x, y = numbers or constant numerical expressions)
// W. Oevel, 15.11.01
// Special code for fast decisions
//       is(x, Type::Real)
// implemented (for x = numbers or constant numerical expressions)
//------------------------------------------------------------

// Kern-is
is:=
  proc(ex, prop = property::Null)
    local x, y, z;
  begin

    //-----------------------------------------------------
    // First, overloading
    //-----------------------------------------------------
    // 'is' cannot be overloaded (?)
    //-----------------------------------------------------

    //-----------------------------------------------------
    // Treat the most important and simplest cases first.
    // For is(x < y) and is(x <= y) as well as is(x = y), 
    // is(x <> y), an immediate check via bool is possible
    // if both x and y are of a numerical type provided by
    // the kernel. Since x < y implies that both x and y 
    // must be real, any DOM_COMPLEX yields an immediate FALSE.
    // This can (and should) be done before any argument
    // testing (after overloading, in case kernel types
    // should overload 'is' in the future)
    //-----------------------------------------------------

    // convert is(x, Type::Positive) etc. to is(x > 0)
    // to treat this by the fast special cases
    if args(0) = 2 then
       case args(2)
       of Type::Positive do
          return(is(args(1) > 0));
       of Type::NonNegative do
          return(is(args(1) >= 0));
       of Type::Negative do
          return(is(args(1) < 0));
   //  of Type::NonPositive do      // presently, there is
   //     return(is(args(1) <= 0)); // no Type::NonPositive
       end_case;
    end_if;
     
    if args(0) = 1 then

      // For numbers, this test should be really fast.
      // Do this branch first.
      if has({"_less", "_leequal"}, type(ex)) then
         //----------------------
         // is(x < y), is(x <= y)
         //----------------------
        x:= op(ex, 1):
        y:= op(ex, 2):
        case domtype(x)
        of DOM_INT do
        of DOM_RAT do
        of DOM_FLOAT do
           case domtype(y)
           of DOM_INT do
           of DOM_RAT do
           of DOM_FLOAT do
              return(bool(ex));
           of DOM_COMPLEX do
              return(FALSE);
           end_case;
           break;
        of DOM_COMPLEX do
           return(FALSE)
        end_case;
      end_if;

      if has({"_equal", "_unequal"}, type(ex)) then
        //----------------------
        // is(x = y), is(x <> y)
        //----------------------
        x:= op(ex, 1):
        y:= op(ex, 2):
        if x = y then
          if type(ex) = "_equal" then
           return(TRUE)
          elif type(ex) = "_unequal" then
            return(FALSE)
          end_if
        end_if;
        case domtype(x)
        of DOM_INT do
        of DOM_RAT do
        of DOM_FLOAT do
        of DOM_COMPLEX do
           case domtype(y)
           of DOM_INT do
           of DOM_RAT do
           of DOM_FLOAT do
           of DOM_COMPLEX do
              return(bool(ex));
           end_case;
        end_case;
      end_if;

      //-----------------------------------------------
      // Resolve nested relations (a <= b) <= c.
      //-----------------------------------------------
      //-----------------------------------------------
      // avoid passing ex = (a<=b)<= c or ex = a<=(b<=c) 
      // to numeric::less. It would produce a useless 
      // UNKNOWN, anyway, and is *extremely* costly 
      // because if would invoke arithmetic (a<=b) - c etc.
      // Generally, handle this case by
      //          is(a<=b<=c) = is(a<=b) and is(b<=c)
      // (without double checking is(a<=c))
      // Use _lazy_and!
      //-----------------------------------------------

      if has({"_less", "_leequal", "_equal", "_unequal"}, type(ex)) then
         if has({"_equal", "_leequal", "_less", "_unequal"}, type(x)) then
           return(_lazy_and(is(x) and is(op(ex, 0)(op(x, 2), y))));
         end_if;
         if has({"_equal", "_leequal", "_less", "_unequal"}, type(y)) then
           return(_lazy_and(is(op(ex, 0)(x, op(y, 1))) and is(y)));
         end_if;
      end_if;
    

      /*-------------------------------------------------------------------------------
         the following is the code for 

          is(a<=b<=c) == is(a<=b) and is(b<=c) and is(a<=c)

         The case "_unequal" is missing.

         It is obsolete, since we decided to document
            is(a <= b) == is(a<=b) and is(b<=c)
 
      
      if has({"_less", "_leequal", "_equal"}, type(ex)) then
        if has({"_equal", "_leequal", "_less"}, type(x)) then
           case [type(x), type(ex)]
           of ["_equal", "_equal"]     do rel:= _equal;   break; // ex = (a =  b) =  c
           of ["_equal", "_leequal"]   do rel:= _leequal; break; // ex = (a =  b) <= c
           of ["_equal", "_less"]      do rel:= _less;    break; // ex = (a =  b) <  c
           of ["_leequal", "_equal"]   do rel:= _leequal; break; // ex = (a <= b) =  c
           of ["_leequal", "_leequal"] do rel:= _leequal; break; // ex = (a <= b) <= c
           of ["_leequal", "_less"]    do rel:= _less;    break; // ex = (a <= b) <  c
           of ["_less", "_equal"]      do rel:= _less;    break; // ex = (a <  b) =  c
           of ["_less", "_leequal"]    do rel:= _less;    break; // ex = (a <  b) <= c
           of ["_less", "_less"]       do rel:= _less;    break; // ex = (a <  b) <  c
           end_case;
           return(_lazy_and(                      // a < b < c
                      is(x),                      // a < b
                      is(op(ex, 0)(op(x, 2), y)), // b < c
                      is(rel(op(x, 1) , y))));    // a < c
        end_if;
        if has({"_equal", "_leequal", "_less"}, type(y)) then
           case [type(ex), type(y)]
           of ["_equal", "_equal"]     do rel:= _equal;   break; // ex =  a =  (b =  c)
           of ["_equal", "_leequal"]   do rel:= _leequal; break; // ex =  a =  (b <= c)
           of ["_equal", "_less"]      do rel:= _less;    break; // ex =  a =  (b <  c)
           of ["_leequal", "_equal"]   do rel:= _leequal; break; // ex =  a <= (b =  c)
           of ["_leequal", "_leequal"] do rel:= _leequal; break; // ex =  a <= (b <= c)
           of ["_leequal", "_less"]    do rel:= _less;    break; // ex =  a <= (b <  c)
           of ["_less", "_equal"]      do rel:= _less;    break; // ex =  a <  (b =  c)
           of ["_less", "_leequal"]    do rel:= _less;    break; // ex =  a <  (b <= c)
           of ["_less", "_less"]       do rel:= _less;    break; // ex =  a <  (b <  c)
           end_case;
           return(_lazy_and(                      // a < b < c
                      is(op(ex, 0)(x, op(y, 1))), // a < b
                      is(y),                      // b < c
                      is(rel(x , op(y, 2)))));    // a < c
        end_if;
      end_if;
      -------------------------------------------------------------------------------*/

      //------------------------------------------------
      // numeric::less provides a reasonably safe float
      // check: it returns TRUE/FALSE or UNKNOWN.
      // TRUE means a safe x < y, both x and y real
      // FALSE means 
      //          safe x > y, both x and y real
      //       or safe Im(x) <> 0
      //       or safe Im(y) <> 0
      // For 'UNKNOWN', we need to check further down
      // below via the property mechanism.
      //------------------------------------------------
      if has({"_less", "_leequal"}, type(ex)) then
        if indets(ex) minus Type::ConstantIdents = {} then
          case numeric::isless(x, y)
          of TRUE do return(TRUE);
          of FALSE do return(FALSE);
          end_case;
        end_if
      end_if;

      //------------------------------------------------
      // numeric::isnonzero provides a reasonably safe
      // float check: it returns TRUE or UNKNOWN.
      // TRUE means x - y <> 0, both x and y may be complex.
      // Heuristcally, UNKNOWN is taken as FALSE
      //------------------------------------------------
      if has({"_equal", "_unequal"}, type(ex)) then
         //----------------------
         // is(x = y), is(x <> y)
         //----------------------
        x:= op(ex, 1):
        y:= op(ex, 2):
        if indets(ex) minus Type::ConstantIdents = {} then
          if numeric::isnonzero(x - y) = TRUE then
            if type(ex) = "_equal" then
              return(FALSE)
            else
              return(TRUE)
            end_if;
          else // UNKNOWN or FALSE
            if {op(map(ex, type@float))}
               minus {DOM_FLOAT, DOM_COMPLEX} = {} then
              if(type(ex) = "_equal") then
                return(TRUE)
              else
                return(FALSE)
              end_if
            else
              if(type(ex) = "_equal") then
                return(FALSE)
              else
                return(TRUE)
              end_if
            end_if
          end_if
        end_if
      end_if;

     //----------------------------------------------
     // Note that, due to fast specialized code,
     // is(x, Type::Real) is pretty fast in those cases,
     // where x is numerical or symbolic without properties. 
     // If is(x, Type::Real) = FALSE/UNKNOWN,
     // we can return FALSE/UNKNOWN immediately without
     // invoking the deeper parts of the property mechanism.
     // Problem: when is(x, Type::Real) is expensive
     //          and TRUE, we do not use this result
     //          here. Is it inserted into the internal
     //          remember mechanism of properties or
     //          do we just loose runtime?
     // ToDo! Activate the following code when
     //       fast specialized code for is(x, Type::Real)
     //       is completely implemented.
     //----------------------------------------------
 /* to be activated:
      if has({"_less", "_leequal"}, type(ex)) then
        xr:= is(x, Type::Real):
        if xr <> TRUE then return(xr) end_if;
        y:= op(ex, 2):
        yr:= is(y, Type::Real):
        if yr <> TRUE then return(yr) end_if;
      end_if;
 */

    end_if;
    //----------------------------------------------------------
    // Fast specialized code for is(x rel y) with
    // rel = <, <=, = , <> is finished here.
    // Further, is(x rel1 y rel2) == is(x rel1 y) and is(y rel2)
    // is dealt with. From here, nested relations
    // x1 rel1 y rel2 z cannot turn up anymore.
    //----------------------------------------------------------


    //-----------------------------------------------------
    // Next special case: is(x, Type::Real) is one of the
    // most important calls. Tune it for speed!
    //-----------------------------------------------------

    if args(0) = 2 and args(2) = Type::Real then
       //-------------------------------------------------
       // ex = a number
       //-------------------------------------------------
       case domtype(ex) 
       of DOM_INT do
       of DOM_RAT do
       of DOM_FLOAT do
          return(TRUE);
       of DOM_COMPLEX do
          return(FALSE);
       end_case;

       //-------------------------------------------------
       // Test constant numerical expressions by a
       // float check.
       // numeric::isreal(x) yields TRUE/FLOAT/UNKNOWN.
       // TRUE is absolutely safe (happens only for exact
       //         numerical types of the kernel)
       // FALSE is absolutely safe for exact numerical types
       //       of the kernel. For constant algebraic expressions,
       //       it is heuristically ok with high probability
       // UNKNOWN does not provide any information (happens
       //         if Im(float(x)) = 0 or if x contains 
       //         symbolic objects.
       //-------------------------------------------------
       z:= numeric::isreal(ex);
       if z <> UNKNOWN then
          return(z);
       end_if;

       //-------------------------------------------------
       // Identifiers with and without properties.
       // Important: treat identifiers x and indexed
       // identifiers in exactly the same way
       //-------------------------------------------------
       if contains({DOM_IDENT, "_index"}, type(ex)) then
         if (domtype(ex) = DOM_IDENT and property::sysis(ex, Type::Real))
            or (contains(property::ITABLE, ex)
                and property::sysimplies(property::ITABLE[ex], Type::Real)) then
           if properties(hold(Global)) = FAIL or
              property::sysis(hold(Global), Type::Real) then
             return(TRUE)
           end_if
         end_if
       end_if;

    end_if;  // of is(ex, Type::Real)

    //--------------------------------------------------------
    // ab hier beginnt Bijs code. Keinerlei Modifikationen
    // von hier ab.
    //--------------------------------------------------------

//     if contains(["_less", "_leequal"], type(ex)) <> 0 then
//       // Relation - 'normal' for simplification
//       if not property::hasprop(ex)
//          and FALSE then
//       ex:= normal(ex);
//       if traperror((x:= map((ret:= indets(ex)), property::_getprop))) = 0 and
//          x = ret and ret <> {} then
//         // alle Bezeichner sind ohne Eigenschaften (Global wird bercksichtigt)
//         return(UNKNOWN)
//       end_if

    if type(ex) = DOM_IDENT then
      if contains(Type::ConstantIdents, ex) then
        //
      elif property::_getprop(ex) = ex then // Global wird bercksichtigt!
        //if property::_getprop(hold(Global)) = undefined then
          //fprint(Unquoted, 0, "Warning: kernel 'is' would have found this (2) [is]");
          return(UNKNOWN)
        //end_if
      elif property::sysis(ex, prop) then
        if property::_getprop(hold(Global)) = undefined then
          //fprint(Unquoted, 0, "Warning: kernel 'is' would have found this (3) [is]");
          return(TRUE)
        elif property::_getprop(hold(Global)) = property::_getprop(ex) then
          //fprint(Unquoted, 0, "Warning: kernel 'is' would have found this (5) [is]");
          return(TRUE)
        end_if
      elif properties(ex) <> FAIL and
           property::sysis(hold(Global), prop) then
        //fprint(Unquoted, 0, "Warning: kernel 'is' would have found this (4) [is]");
        return(TRUE)
      end_if
    elif contains([TRUE, FALSE, UNKNOWN, FAIL], ex) <> 0 then
      return(ex)
    end_if;

    //--------------------------------------------------------
    // Here, the generic part of the property mechanism starts
    // First, argument testing
    //-----------------------------------------------------
       
    if testargs() then
      if args(0) = 1 then
        if ex::dom::prop = TRUE then
          // globale Eigenschaft
        elif domtype(ex) = DOM_EXPR then
          //if misc::maprec(ex, {"NRange", "IV", "IVnat", "Prop",
          //                        "Rel", "Complex", "Residue"} = TRUE) = TRUE then
          if contains(["NRange", "IV", "IVnat", "Prop",
                       "Rel", "Complex", "Residue"], type(ex)) <> 0 then
            // globale Eigenschaft
          elif contains(["_less", "_leequal", "_equal", "_unequal",
                         "_and", "_or", "_not", "_in", "_xor", "_equiv",
                         "_implies"], type(ex)) <> 0 then
            // expression
          else
            error("wrong expression type of first argument")
          end_if
        elif domtype(ex) = DOM_BOOL then
          // boolean value
        else
          error("wrong type of first argument")
        end_if
      elif args(0) = 2 then
        if prop::dom::prop = TRUE then
          // globale Eigenschaft
        elif domtype(prop) = DOM_EXPR
             and contains(["NRange", "IV", "IVnat", "Prop",
                           "Rel", "Complex", "Residue"], type(ex)) <> 0 then
             //and misc::maprec(ex, {"NRange", "IV", "IVnat", "Prop",
             //                         "Rel", "Complex", "Residue"} = TRUE) = TRUE then
            // globale Eigenschaft
        else
          error("wrong type of second argument (property expected)")
        end_if
      else
        error("wrong number of arguments")
      end_if
    end_if;

    //fprint(Unquoted, 0, "Warning: kernel 'is' would not find this [is]");
    property::is(args())
  end_proc:

// gibt BOOL zuruck, ob prop eine Eigenschaft von ex ist
property::is:=
  proc(ex, prop = null())
    local ret, lev;
  begin
   // testtype
//     if domtype(prop) = Type and not(contains(prop::dom, "prop")) then
//       if nops(prop) > 4 then // provisorische Type's
//         return(slot(property, op(prop, 5))(ex))
//       else
//         error("wrong type of second argument (property expected)")
//         //userinfo(10, "testtype(".expr2text(args()).")");
//         //return(testtype(ex, prop) or UNKNOWN)
//       end_if
//     end_if;

    // enter recursion
    property::reset(1);
    ret:= args();
    if testtype(ex, Type::Property) then
      // globale Eigenschaft
      ret:= hold(Global), ret
    end_if;
    if property::RESET > 1 then // inline of recursion
      lev:= 0;
      ret:= property::_is(ret)
    elif property::NoErrors then
      //warning("Don't forget to call 'property::reset()' after any error!");
      lev:= 0;
      ret:= property::_is(ret)
    else // first call
      lev:= traperror((ret:= property::_is(ret)))
    end_if;
    if lev > 0 then
      sysassign(property::RESET, 0);
      property::reset();
      lasterror()
    end_if;
    // leave recursion
    property::reset(-1);
    if contains({TRUE, FALSE}, ret) then
      return(ret)
    else
      return(UNKNOWN)
    end_if
  end_proc:

// gibt BOOL zuruck, ob prop eine Eigenschaft von ex ist
property::_is:=
  proc(ex, prop = null())
    local ret, __f_, M, CONST, minprop, x, y, z;
    //option remember;
  begin
///////////////////////////////////////////////////////////////////////////////
//// Spezialcode von 'is' fr interne Aufrufe von 'property::_is' /////////////
///////////////////////////////////////////////////////////////////////////////
    if prop = null() then

      // For numbers, this test should be really fast.
      // Do this branch first.
      if has({"_less", "_leequal"}, type(ex)) then
         //----------------------
         // is(x < y), is(x <= y)
         //----------------------
        x:= op(ex, 1):
        y:= op(ex, 2):
        case domtype(x)
        of DOM_INT do
        of DOM_RAT do
        of DOM_FLOAT do
           case domtype(y)
           of DOM_INT do
           of DOM_RAT do
           of DOM_FLOAT do
              return(bool(ex));
           of DOM_COMPLEX do
              return(FALSE);
           end_case;
           break;
        of DOM_COMPLEX do
           return(FALSE)
        end_case;
      end_if;

      if has({"_equal", "_unequal"}, type(ex)) then
        //----------------------
        // is(x = y), is(x <> y)
        //----------------------
        x:= op(ex, 1):
        y:= op(ex, 2):
        if x = y then
          if type(ex) = "_equal" then
           return(TRUE)
          elif type(ex) = "_unequal" then
            return(FALSE)
          end_if
        end_if;
        case domtype(x)
        of DOM_INT do
        of DOM_RAT do
        of DOM_FLOAT do
        of DOM_COMPLEX do
           case domtype(y)
           of DOM_INT do
           of DOM_RAT do
           of DOM_FLOAT do
           of DOM_COMPLEX do
              return(bool(ex));
           end_case;
        end_case;
      end_if;

      //------------------------------------------------
      // numeric::less provides a reasonably safe float
      // check: it returns TRUE/FALSE or UNKNOWN.
      // TRUE means a safe x < y, both x and y real
      // FALSE means 
      //          safe x > y, both x and y real
      //       or safe Im(x) <> 0
      //       or safe Im(y) <> 0
      // For 'UNKNOWN', we need to check further down
      // below via the property mechanism.
      //------------------------------------------------
      if has({"_less", "_leequal"}, type(ex)) then
        if indets(ex) minus Type::ConstantIdents = {} then
          case numeric::isless(x, y)
          of TRUE do return(TRUE);
          of FALSE do return(FALSE);
          end_case;
        end_if
      end_if;

      //------------------------------------------------
      // numeric::isnonzero provides a reasonably safe
      // float check: it returns TRUE or UNKNOWN.
      // TRUE means x - y <> 0, both x and y may be complex.
      // Heuristcally, UNKNOWN is taken as FALSE
      //------------------------------------------------
      if has({"_equal", "_unequal"}, type(ex)) then
         //----------------------
         // is(x = y), is(x <> y)
         //----------------------
        x:= op(ex, 1):
        y:= op(ex, 2):
        if indets(ex) minus Type::ConstantIdents = {} then
          if numeric::isnonzero(x - y) = TRUE then
            if type(ex) = "_equal" then
              return(FALSE)
            else
              return(TRUE)
            end_if;
          else // UNKNOWN or FALSE
            if {op(map(ex, type@float))}
               minus {DOM_FLOAT, DOM_COMPLEX} = {} then
              if(type(ex) = "_equal") then
                return(TRUE)
              else
                return(FALSE)
              end_if
            else
              if(type(ex) = "_equal") then
                return(FALSE)
              else
                return(TRUE)
              end_if
            end_if
          end_if
        end_if
      end_if;

     //----------------------------------------------
     // Note that, due to fast specialized code,
     // is(x, Type::Real) is pretty fast in those cases,
     // where x is numerical or symbolic without properties. 
     // If is(x, Type::Real) = FALSE/UNKNOWN,
     // we can return FALSE/UNKNOWN immediately without
     // invoking the deeper parts of the property mechanism.
     // Problem: when is(x, Type::Real) is expensive
     //          and TRUE, we do not use this result
     //          here. Is it inserted into the internal
     //          remember mechanism of properties or
     //          do we just loose runtime?
     // ToDo! Activate the following code when
     //       fast specialized code for is(x, Type::Real)
     //       is completely implemented.
     //----------------------------------------------
 // to be activated:
      if has({"_less", "_leequal"}, type(ex)) then
        ret:= property::_is(x, Type::Real):
        if ret = FALSE then return(ret) end_if;
        y:= op(ex, 2):
        ret:= property::_is(y, Type::Real):
        if ret = FALSE then return(ret) end_if;
      end_if;
    end_if;

    //-----------------------------------------------------
    // Next special case: is(x, Type::Real) is one of the
    // most important calls. Tune it for speed!
    //-----------------------------------------------------

    if prop = Type::Real then
       //-------------------------------------------------
       // ex = a number
       //-------------------------------------------------
       case domtype(ex) 
       of DOM_INT do
       of DOM_RAT do
       of DOM_FLOAT do
          return(TRUE);
       of DOM_COMPLEX do
          return(FALSE);
       end_case;

       //-------------------------------------------------
       // Test constant numerical expressions by a
       // float check.
       // numeric::isreal(x) yields TRUE/FLOAT/UNKNOWN.
       // TRUE is absolutely safe (happens only for exact
       //         numerical types of the kernel)
       // FALSE is absolutely safe for exact numerical types
       //       of the kernel. For constant algebraic expressions,
       //       it is heuristically ok with high probability
       // UNKNOWN does not provide any information (happens
       //         if Im(float(x)) = 0 or if x contains 
       //         symbolic objects.
       //-------------------------------------------------
       z:= numeric::isreal(ex);
       if z <> UNKNOWN then
          return(z);
       end_if;

       //-------------------------------------------------
       // Identifiers with and without properties.
       // Important: treat identifiers x and indexed
       // identifiers in exactly the same way
       //-------------------------------------------------
       if contains({DOM_IDENT, "_index"}, type(ex)) then
         if (domtype(ex) = DOM_IDENT and property::sysis(ex, Type::Real))
            or (contains(property::ITABLE, ex)
                and property::sysimplies(property::ITABLE[ex], Type::Real)) then
           if properties(hold(Global)) = FAIL or
              property::sysis(hold(Global), Type::Real) then
             return(TRUE)
           end_if
         end_if
       end_if;

    end_if;  // of is(ex, Type::Real)
///////////////////////////////////////////////////////////////////////////////
//// Ende des Spezialcodes ////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////

    // Operand undefined
    if property::Fail(ex, op(ex), prop) then
      return(UNKNOWN)
    end_if;

    // zu grosse Eingaben
    if length(ex) > property::MAXLENGTH then
      userinfo(2, "Break: input expression is too complex (".expr2text(length(ex)).")! [property::is]");
      // fprint(Unquoted, 0, "Break: input expression is too complex: ", property::complexity(ex), " (",length(ex),  ")! [property::is]");
      return(UNKNOWN)
//     elif length(ex) > property::MAXLENGTH/2 then
//             userinfo(2, "Warning: input expression is complex (".expr2text(length(ex)).")! [property::is]");
    end_if;

    // Globale Eigenschaft
    if args(0) = 1 and testtype(ex, Type::Property) then
      prop:= ex;
      ex:= hold(Global)
    end_if;

    minprop:=
      if domtype(ex) = DOM_EXPR and domtype((minprop:= eval(op(ex, 0)))) = DOM_FUNC_ENV then
        slot(minprop, "minprop")
      else
        FAIL
      end_if;
    
    // Bezeichner ohne Eigenschaften
    // Ausdruck knnte zu "konstant" vereinfacht werden: x/(x + y) + y/(x + y) = 1
    
    if prop = null() then
      if contains({TRUE, FALSE, UNKNOWN}, ex) then
        return(ex)
      end_if
    end_if;

    // Bezeichner, Systemeigenschaft
    if domtype(ex) = DOM_IDENT then
      if property::sysis(ex, prop) then
        return(TRUE)
      end_if
    end_if;

    ret:= UNKNOWN;

    // prop keine Eigenschaft
    if prop <> null() and not testtype(prop, Type::Property) then
      if prop::dom::set2prop <> FAIL then
        prop:= prop::dom::set2prop(prop, hold(Exact));
        if prop = FAIL then
          return(UNKNOWN)
        end_if
      else
        userinfo(1, "wrong type of second argument, cannot recognize or convert to a property [is]");
        return(UNKNOWN)
      end_if
    end_if;

    __f_:= op(ex, 0);
    if domtype(ex) = DOM_EXPR then
//fprint(0, ">>>DOM_EXPR", __f_, slot(__f_, "is"));
      if contains(["_and", "_or", "_not"], type(ex)) <> 0 then
        ret:= property::_is(map(ex, property::_is, prop));
//         if ret = UNKNOWN then
//           // gegenseitiger Ausschluss von Relationen-Eigenschaften - kann teuer werden
//           if contains({TRUE, FALSE}, (M:= property::simpex(ex))) then
//             ret:= M
//           end_if
//         end_if
      elif type(ex) = "_in" then
        if testtype(op(ex, 2), Type::Property) then
          ret:= property::_is(op(ex))
        elif domtype(op(ex, 2)) = solvelib::BasicSet then
          // solvelib::BasicSet::contains wurde geloescht
          ret := property::_is(op(ex, 1), solvelib::set2prop(op(ex, 2)))
        else
          if contains((op(ex, 2))::dom, "simpexIn") and
            domtype(op(ex, 1)) <> piecewise then
            
            if contains({TRUE, FALSE}, (M:= slot((op(ex, 2))::dom, "simpexIn")
                                        (op(ex)))) then
              return(M)
            elif type(M) = "_in" then
              return(UNKNOWN)
            else
              return(property::_is(M))
            end_if
            
         /*
           // wird jetzt von ImageSet::simpexIn erledigt:
           // Methode ImageSet::expr2set wurde geloescht
          elif domtype(op(ex, 2)) = Dom::ImageSet then
            M:= Dom::ImageSet::expr2set(op(ex, 2), op(ex, 1));
            if type(M) = "_in" then
              return(UNKNOWN)
            else
              return(property::_is(M))
            end_if
          */
          elif (ret:= expand(ex)) <> ex then // simplification
            return(property::_is(ret))
          elif (ret:= solvelib::set2prop(op(ex, 2))) <> FAIL then
            ret:= property::_is(op(ex, 1), ret)
          else
            return(UNKNOWN)
          end_if
        end_if
      elif (__f_:= contains(["_less", "_leequal", "_equal", "_unequal"],
                            type(ex))) <> 0 then
        // numerisch  -> Kern-is
        if (M:= ({map(op(ex), domtype)} minus
            {DOM_INT, DOM_FLOAT, DOM_RAT})) = {} then
          return(bool(ex))
        //elif contains(M, DOM_COMPLEX) then
        //  if __f_ > 2 then
        //    return(bool(ex))
        //  else
//!!!!!!!!!!!!!!!!!!!!!!!!! COMPLEX !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
            //return(FALSE) // DOM_COMPLEX nicht reell!
        //  end_if
        end_if;
//	Constant Expression handled by numeric::isnonzero in line 40
//
//        __f_:= map(map(ex, float), float);
//        if ({map(op(__f_), domtype)} minus {DOM_INT, DOM_FLOAT, DOM_RAT}) = {} then
//fprint(0, "float bool", __f_);
//          return(bool(__f_))
//        end_if;

        // Relationen nur in einer Richtung 
        //if prop = null() then delete prop end_if;
//         if contains({">", "<"}, prop) then
//           if prop = "<" then
//             ret:= property::_is(op(ex, 1),
//                                     property::Rel(op(ex, 0),
//                                                   op(ex, 2)))
//           else // ">" 
//             ret:= property::_is(op(ex, 2),
//                                     property::Rel(-op(ex, 0),
//                                                   op(ex, 1)))
//           end_if
        if op(ex, 1) = op(ex, 2) then
          if contains([hold(_equal), hold(_leequal)], op(ex, 0)) <> 0 then
            ret:= TRUE
          else ret:= FALSE
          end_if
        else

          ret:= property::isrel(ex) // Relationen

        end_if;
      // DOM_FUNC_ENV 
      elif domtype(__f_) = DOM_FUNC_ENV and
           (__f_:= slot(__f_, "is")) <> FAIL then
        ret:= __f_(op(ex), prop)
      // DOM_EXPR
      else
        ret:= property::isfunc(ex, prop);
        if not(contains({TRUE, FALSE}, ret)) then
          // 
          __f_:= property::_getprop(ex); // zuerst ohne Konvertierung
          if minprop <> FAIL then
            if __f_ = ex then
              __f_:= minprop
            else
              __f_:= __f_ and minprop
            end_if
          end_if;
          ret:= property::implies(__f_, prop);
          if ret = UNKNOWN then
            ret:= property::implies(property::convert_to(__f_, prop), prop)
          end_if;
          // special case interval -> two relations
          if ret = UNKNOWN and type(prop) = "IV" then
            ret:= property::isrel(property::IV::expr(prop, ex))
          end_if
        end_if
      end_if
    else // not DOM_EXPR
      __f_:= property::Ex2Px(ex);
      if contains({"and", "or"}, type(__f_)) then
        ret:= property::lzeval(__f_, property::implies, prop)
      else
        ret:= property::implies(__f_, prop)
      end_if;
      if ret = UNKNOWN then
        ret:= property::implies(property::_getprop(__f_, prop), prop);
      end_if;
      // special case interval -> two relations
//       if ret = UNKNOWN and type(prop) = "IV" then
//         ret:= property::isrel(property::IV::expr(prop, ex))
//       end_if
    end_if;

    //sysassign(property::RECLEVEL, property::RECLEVEL - 1);
    ret
  end_proc:

// funktionale Eigenschaften 
property::isfunc:=
  proc(_func, prop)
    local ret, i;
  begin
    if op(_func) = hold(_in) then
      return(property::_is(op(_func, 1), op(_func, 2)))
    end_if;
    if nops(_func) <> 1 or domtype(op(_func)) <> DOM_IDENT then
      return(UNKNOWN)
    end_if;
    if (i:= contains([hold(Re), hold(Im)], op(_func, 0))) <> 0 then
      if property::implies(prop, Type::Real) = TRUE then
        return(TRUE)
      end_if;
      if type((ret:= getprop(op(_func, 1)))) = "Complex" then
        property::implies(ret, property::Complex(op(subsop([FAIL, FAIL], i = prop))))
      end_if
    else
      UNKNOWN
    end_if
  end_proc:

// gibt BOOL zuruck, ob aus _exp prop folgt 
// _exp - P-Ausdruck von Eigenschaften, prop - Eigenschaft 
property::implies:=
  proc(_exp, prop = null())
    local __ret, __rule;
  begin
    if prop = null() then
      return(UNKNOWN)
    end_if;
    
    // Grundeigenschaften
    if property::sysimplies(_exp, prop) then
      return(TRUE)
    end_if;
    
    if property::Fail(args(), op(_exp), op(prop)) then
      return(UNKNOWN)
    end_if;

    if _exp = prop then return(TRUE) end_if;

    case type(prop)
      of "_and" do
        return(_and(property::implies(_exp, k) $ k in op(prop)))
      of "_or" do
        return(_or(property::implies(_exp, k) $ k in op(prop)))
      of "_not" do
        return(not property::implies(_exp, op(prop)))
    end_case;

    __ret:= UNKNOWN;
    if domtype(_exp) = DOM_EXPR then
      if type(_exp) = "_not" then
        __ret:= not(property::implies(op(_exp), prop))
      elif contains({"_and", "_or"}, type(_exp)) then
        __ret:= property::lzeval(_exp, property::implies, prop)
      else
        if property::OWNRULES then
          __rule:= property::Prule(_exp);
          __ret:= property::_getrule((op(__rule, 2))::dom, __rule);
          if __ret <> FAIL then
            return(property::implies(__ret, prop))
          end_if;
          __ret:= property::_getfunc((op(__rule, 2))::dom, __rule);
          if __ret <> FAIL then
            __ret:= property::implies(__ret, prop)
          end_if
        end_if;

        if domtype((__rule:= float(_exp))) <> DOM_EXPR then
          __ret:= property::implies(__rule, prop)
        end_if
      end_if
    else
      // ungleiches Domain 
      //fprint(0, "isprop1", type(_exp), type(prop));
      if type(_exp) <> type(prop) then
        if not(contains({TRUE, FALSE},
                        (__ret:= prop::dom::is(_exp, prop)))) then
          //fprint(0, "isprop2", __ret);
          if not(contains({TRUE, FALSE},
                          (__ret:= _exp::dom::is(_exp, prop)))) then
            //fprint(0, "isprop3", __ret);
            if (__rule:= property::_convert(hold(Sloppy), _exp, prop)) <> FAIL then
              //fprint(0, "isprop4", __ret, prop::dom);
              if (__ret:= prop::dom::is(__rule, prop)) = UNKNOWN then
                // try extracting properties of identifiers inside of relations
//                 if indets(prop) = {}
//                    or (__rule:= property::_getprop(prop)) = prop
//                    or __rule::dom::prop <> TRUE
//                    or not(contains({TRUE, FALSE},
//                                    (__ret:= _rule::dom::is(_exp, __rule)))) then
//                   if indets(_exp) = {}
//                      or (__rule:= property::_getprop(_exp)) = _exp
//                      or __rule::dom::prop <> TRUE
//                      or not(contains({TRUE, FALSE},
//                                      (__ret:= _rule::dom::is(__rule, prop)))) then
                    //fprint(0, "isprop5", __ret);
                    if _exp::dom::prop = TRUE
                       and (__rule:= property::_convert(hold(Sloppy), prop, _exp)) <> FAIL then
                      __ret:= _exp::dom::is(_exp, __rule) and UNKNOWN; // sonst Informations-Verlust!!!
                      //fprint(0, "isprop6", __ret);
                    end_if
//                   end_if
//                 end_if
              end_if
            end_if
          end_if
        end_if
      else // gleiches Domain 
        __ret:= _exp::dom::is(_exp, prop)
      end_if
    end_if;
    if property::Fail(__ret) then
      __ret:= UNKNOWN
    end_if;
//     if __ret = UNKNOWN then
//       if property::constant(_exp) = TRUE and property::constant(prop) = TRUE then
//         __ret:= FALSE
//       end_if
//     end_if;
    __ret
  end_proc:

property::propend:=
  proc(ident, IN = {})
    local P;
  begin
    if indets((P:= property::_getprop(ident))) = {} then
      return(TRUE)
    elif indets(P) intersect IN <> {} then
      return(FALSE)
    else
      return(_and(map(op(indets(property::_getprop(ident))), property::propend, IN union {ident})))
    end_if
  end_proc:

property::_is:=
  prog::remember(property::_is, property::depends, hold(PreventRecursion), UNKNOWN):
property::implies:=
  prog::remember(property::implies, property::depends, hold(PreventRecursion), UNKNOWN):

