/*
 * series slot of lambertV - computes series(lambertV(f), x, n, dir)
 *
 * computes undirected expansions around x0 for -exp(-1) <= x0 < 0
 * returns FAIL for +-infinity and 0
 * returns a symbolic Taylor expansion for all other values of x0
 *
 * Author: Juergen Gerhard, 18.09.01
 */

lambertV::series := proc(f, x, n, dir)
  local s, t, k, mu, alpha;
begin
  // recursively expand the argument
  t := Series::series(f, x, n, dir);

  if domtype(t) = Series::Puiseux then
    k := Series::Puiseux::ldegree(t);
    if k = FAIL then // t = O(..)
      Series::error("order too small")
    end_if;
    if k > 0 then // expansion around 0
      return(FAIL);
    elif k < 0 then // expansion around infinity
      return(FAIL)
    else // k = 0, expansion around a finite point x0 <> 0
      t := lmonomial(t, Rem);
      if combine(t[1], exp) = -exp(-1) then
        // This is from the ISSAC'97 article of Corless, Jeffrey & Knuth,
        // section 3.
        mu[0] := -1;
        mu[1] := 1;
        alpha[0] := 2;
        alpha[1] := -1;
        for k from 2 to n - 1 do
          alpha[k] := _plus(mu[j]*mu[k + 1 - j] $ j = 2..(k - 1));
          mu[k] := (k - 1)/(k + 1)*(mu[k - 2]/2 + alpha[k - 2]/4)
                   - alpha[k]/2 - mu[k - 1]/(k + 1);
        end_for;
        s := Series::Puiseux::create(2, 0, n,
               [(-1)^k*mu[k]*2^(k/2)*exp(1)^(k/2) $ k = 0..(n-1)], x, 0, dir);
        return(s @ t[2]);
      end_if
    end_if
  end_if;

  Series::unknown(lambertV(f), x, n, dir)
end_proc:
