// Author:  T.C. Scott May 2001
//
//    hypergeom(a,b,z) : general hypergeometric function
//
//       hypergeom( [a ,a ,...a ], [b ,b ,...b ], z )  ==  F(a,d,z)
//                    1  2     p     1  2     q
//
//                        product( pochhammer(a[i],k), i=1..p ) z^k
// Defn : F(n,d,z) = sum( -----------------------------------------, k >= 0 )
//                        product( pochhammer(b[i],k), i=1..q ) k!
//
// where this sum converges, and its analytical continuation in the other
// cases.  In the divergent case, the function is regarded as the solution 
// of the generalized hypergeometric differential equation, which
// has the above series as asymptotic series at 0.
//
// Some elementary simplifications are handled automatically.
//
//
alias(Hypergeom=specfunc::Hypergeom):

alias(typenumcomplex=Hypergeom::typenumcomplex):
alias(sel=Hypergeom::sel):

hypergeom:=
proc(l1 : Type::ListOf(Type::Arithmetical),
     l2 : Type::ListOf(Type::Arithmetical), 
     z  : Type::Arithmetical)
local a0,b0,ak,bk,i,j,k,s,t;
begin
  if args(0) <> 3 then 
     error("wrong number of arguments")
  end_if:

  // Step 1: remove common terms
  if {op(l1)} intersect {op(l2)} <> {} then
    for i from nops(l1) downto 1 do
       j:=contains(l2, l1[i]):
       if j <> 0 then
         delete l1[i]:
         delete l2[j]:
       end_if:
    end_for:
  end_if:

  /*--------------------------------------------------
  // Tony, I do not understand the following.
  // I commented this out. Please, check.

  // Check that the function is defined.

  i:= select(l2, x -> _lazy_and(type(float(x))=DOM_FLOAT,
                                frac(x)=0,x<=0)):
  if nops(i) <> 0 then 
     i:= min(op(i)):
     j:= select(l1, x -> _lazy_and(type(float(x))=DOM_FLOAT,
                                   frac(x)=0, x<=0)):
     if nops(j) <> 0 and min(op(j)) < i then
        error("illegal parameters"):
     end_if:
  end_if:
  --------------------------------------------------*/

  if l1=[] and l2=[] then return(exp(z)): end_if:

  // Step 2: Sort l1 and l2 over i since F is symmetric in l1[i] and l2[i]
  if l1 <> [] then l1:= sort(l1): end_if:
  if l2 <> [] then l2:= sort(l2): end_if:

  // check possibilities of poles

  // sel(l) returns the largest nonpositive integer in
  // the list l. If there is none, sel(l) returns -infinity.
  a0:=sel(l1): 
  b0:=sel(l2):
  if b0 > a0 then
     error("illegal parameters"):
  end_if:

  if _and(op(map(l1,typenumcomplex)),
          op(map(l2,typenumcomplex)),
          typenumcomplex(z))
  and has([map((op(z),op(map(l1,op)),op(map(l2,op))),type)],DOM_FLOAT) then
      return(hypergeom::float(l1,l2,z)):
  end_if:

  // Simplification for special case
  if contains(l1, 0) <> 0 then return( 1 ) end_if:
  if iszero(z) then return(1) end_if:

  //-----------------------------------
  // Reduction to the polynomial case
  // Walter 15.2.02: changed the functionality.
  // Only reduce to the polynomial case, if a number
  // is produced. Otherwise, the polynomial may
  // be numerically unstable. The user should use
  // simplify to get the polynomial reduction.
  /* -----------------------------------------
  // Tonys original code:
  if a0 <> -infinity and a0 > b0 
     and 
    (testtype(z,Type::Rational) or 
     testtype(z,Type::Complex) or 
     testtype(z, Type::Indeterminate)) and
    _and(op(map(l1,testtype,Type::Rational)),
         op(map(l2,testtype,Type::Rational))) 
    then
    t:=1: s:=t:
    for k from 1 to RD_INF do
        ...
  ----------------------------------------- */
  // Walter's new code:
  if a0 <> -infinity and 
     a0 > b0 and
     (indets(l1) union
      indets(l2) union
      indets(z)
     ) minus Type::ConstantIdents = {} then
     // all indices and the argument z are numerical
     t:=1: s:=t:
     for k from 1 to round(-a0) do
       ak:=_mult(op(map(l1,((x,k)->x+k-1),k))):
       bk:=_mult(op(map(l2,((x,k)->x+k-1),k))):
       t:=t*ak*z/bk/k:
       s:=s+t:
       if iszero(t) then break: end_if:
     end_for:
     return(s): 
  end_if:

  //--------------------------
  // generic result
  return(procname(l1, l2, z)):

end_proc:

hypergeom := funcenv(hypergeom):
hypergeom:= slot(hypergeom,"type","hypergeom"):
hypergeom:= slot(hypergeom,"print","hypergeom"):
hypergeom:= slot(hypergeom,"info",
"hypergeom(as,bs,z) -- Hypergeometric function with lists of parameters as,bs and argument z"):

//hypergeom::Content:= (Out, data) -> Out::Capply(Out::Chypergeom, map(op(data),
//Out)):

hypergeom::float := loadproc(hypergeom::float, 
                             pathname("STDLIB","FLOAT"),"hypergeom"):

hypergeom::diff :=  loadproc(hypergeom::diff,
                             pathname("STDLIB","DIFF"),"hypergeom"):

hypergeom::series := loadproc(hypergeom::series,
                              pathname("SERIES"),"hypergeom"):

// -----------------------------------------------
// Tony never finished the complete simplify code.
// This is a very rudimentary version only doing
// the reduction to the polynomial case.
hypergeom::simplify:= proc(f)
local l1, l2, z, a0, b0, ak, bk, t, s, k;
begin
  [l1, l2, z]:= [op(f)];
  l1:= map(l1, simplify);
  l2:= map(l2, simplify);
  z:= simplify(z);
  a0:=sel(l1): 
  b0:=sel(l2):
  if a0 <> -infinity and a0 > b0 then
     t:=1: s:=t:
     for k from 1 to round(-a0) do
       ak:=_mult(op(map(l1,((x,k)->x+k-1),k))):
       bk:=_mult(op(map(l2,((x,k)->x+k-1),k))):
       t:=t*ak*z/bk/k:
       s:=s+t:
       if t=0 then break: end_if:
     end_for:
     return(s): 
  end_if:
  return(hold(hypergeom)(l1, l2, z));
end_proc:
// -----------------------------------------------

unalias(typenumcomplex):
unalias(Hypergeom):
