/*
this is an implementation of the shapiro wilk test for normality;
it follows closely the paper of Patrick Royston: ALGORITHM R94 APPL.
 STATIST. (1995), VOL.44, NO.4.
INPUT: A list of data l=[x.1, x.2, ..., x.n];
stats::swGOFT tests the hypothesis H0: l is a sample of independend 
N(m,v)- distributed members with m, v unknown.
RETURN: the shapiro wilk statistic w for the data l,
the attained significance level of w.
For definition of the shapiro-wilk statistic see the Mupad - documentation
for stats::swGOFT(l);
AUXILIARY ROUTINE: shapco = proc(n) computes the shapiro wilk 
coefficients of order n;
*/ 


stats::swGOFT:=
proc()
  local  l, data, i, s, n, k, A, w, pi6, st, pw, y, M, S, lnn, C3, C4, C5, C6,
  G, g, shapco, nodev;
begin
 if testargs() then
    //stats::testdata tests whether the data of the following types:
    //integers, rationals, floats, expressions, identifiers, complex numbers
     if stats::testdata("all_data", args(1..args(0))) = FALSE then
        error("some data are of illegal type")
     end_if:
  end_if:
  data:= stats::getdata(testargs(), "numeric_only", 1, args(1..args(0) )):
  if domtype(data) = DOM_STRING then
     error(data)
  end_if:

  // now, the data are a list 

  if data = [] then
     error("expecting a non-empty sample of data")
  end_if:
  l:=data:
  n:= nops(l):
  if n<3 then
    error("sample size too small (minimal samplesize is 3)"):
  end_if:
  if n > 5000 then
    error("maximal sample size 5000 exceeded"):
  end_if:
  l:= float(l):

  //----------------------------
  //Auxiliary procedure shapco:
  //----------------------------
  shapco:= proc(n)
  local nn2, A, summ2, ssumm2,  rsn, an25, a1, a2, C1 ,C2, i1,fac, i, n2;
  begin
    C1:= [0.0, 0.221157, -0.147981, -0.207119e1, 0.4434685e1, -0.2706056e1]:
    C2:= [0.0, 0.42981e-1, -0.293762, -0.1752461e1, 0.5682633e1, -0.35822633e1]:
    nn2:= n/2:
    n2:= nn2:
    if domtype(nn2) <> DOM_INT then
      nn2:= (n-1)/2:
      n2:=  (n+1)/2:
    end_if:
    A:= [0 $ i = 1..n2]: 
    if iszero(n-3) then
      A[1]:= 0.375^0.5:
    end_if:
    an25:= n+0.25:
    summ2:= 0.0:
    for i from 1 to n2 do 
      A[i]:= nodev((i-0.375)/an25):
      summ2:= summ2+A[i]^2:
    end_for:
    summ2:= 2*summ2:
    ssumm2:= summ2^0.5:
    rsn:= 1.0/n^0.5: 
    a1:= _plus(C1[i+1]*rsn^i $ i= 0..5) - A[1]/ssumm2: 
    //normalize coefficients:
    if n>5 then
      i1:= 3:
      a2:= -A[2]/ssumm2 + _plus(C2[i+1]*rsn^i$i= 0..5): 
      fac:= ((summ2-2*A[1]^2-2*A[2]^2)/(1.0-2*a1^2-2*a2^2))^(1/2):
      A[1]:= a1:
      A[2]:= a2: 
    else
      i1:= 2:
      fac:= ((summ2-2*A[1]^2)/(1.0-2*a1^2))^(1/2):
      A[1]:= a1:
    end_if:
    for i from i1 to nn2 do 
      A[i]:= -A[i]/fac: 
    end_for:
    A;
  end_proc:
  //------------------------------

  //---------------------------
  // auxiliary procedure nodev:
  //---------------------------
  nodev:= proc(p)
  local A, q, R, ppnd;
  begin
    A:= [3.3871327179, 5.0434271938e1, 1.5929113202e2, 5.9109374720e1,
        1.7895169469e1, 7.8757757664e1, 6.7187563600e1,
        1.423432777, 2.7568153900, 1.3067284816, 1.7023821103e-1,
        7.3700164250e-1, 1.2021132975e-1,
        6.6579051150, 3.0812263860, 4.2868294337e-1, 1.7337203997e-2,
        2.4197894225e-1, 1.2258202635e-2]:
    q:= p-0.5:
    if specfunc::abs(q)<= 0.425 then
      R:= 0.180625-q^2:
      ppnd:= q*(((A[4]*R+A[3])*R+A[2])*R+A[1])/
             (((A[7]*R+A[6])*R + A[5])*R + 1.0):
      return(ppnd);
    else
      if q<0 then
        R:= p:
      else
        R:= 1-p:
      end_if:
      if R<= 0 then 
        return(float(0));
      end_if:
      R:= (-ln(R))^0.5:
      if R<= 5.0 then
        R:= R-1.6:
        ppnd:= (((A[11]*R+A[10])*R+A[9])*R+A[8])/
               ((A[13]*R+A[12])*R+1.0):
      else
        R:= R-5.0:
        ppnd:= (((A[17]*R+A[16])*R+A[15])*R+A[14])/
               ((A[19]*R+A[18])*R+1.0):
      end_if:
      if q<0 then
        ppnd:= -ppnd:
      end_if:
      ppnd;
    end_if:
  end_proc:
  //------------------------------


  l:= sort(l):
  s:= 0:
  //shapco(n) computes list of shapiro wilk coefficients of order n:
  A:= shapco(n):
  k:= n/2:
  if domtype(k)<>DOM_INT then
    k:= (n-1)/2:
  end_if:

  //---------------------------------
  //calculate shapiro wilk statistic:
  //---------------------------------
  for i from 1 to k do
    s:= s + A[i]*(l[n+1-i] - l[i]):
  end_for:

  y:= stats::variance(l, Population):
  if iszero(y) then
     error("the variance of the sample is zero");
  end_if;
  w:= s^2/(n*y):

  //------------------------------------------
  // calculate the significance level pw for w:
  // It is uniquely deterimend by n and w:
  //------------------------------------------
  G:=  [-2.273,   0.459]:
  C3:= [ 0.544,  -0.39978,    0.25054e-1, -0.6714e-3]:
  C4:= [ 1.3822, -0.77857,    0.62767e-1, -0.20322e-2]:
  C5:= [-1.5861, -0.31082,   -0.83751e-1,  0.38915e-2]:
  C6:= [-0.4803, -0.82676e-1, 0.30302e-2]:
  pi6:= 1.909859:
  st:= 1.047198:
  if n = 3 then
    pw:= pi6*(arcsin(w^(1/2))-st):
    return([w, pw]);
  end_if:
  y:= ln(1-w):
  if n <= 11 then
    g:= _plus(G[i+1]*n^i $ i = 0..1):
    if y >= g then
      return([w, float(0)]):
    end_if:
    y:= -ln(g-y):
    M:=     _plus(C3[i+1]*n^i $ i= 0..3):
    S:= exp(_plus(C4[i+1]*n^i $ i= 0..3)):
  else
    lnn:= ln(float(n)):
    M:=     _plus(C5[i+1]*lnn^i $ i = 0..3):
    S:= exp(_plus(C6[i+1]*lnn^i $ i = 0..2)): 
  end_if:

  /*
  if you extend the shapiro wilk test to the data-censored case, use the
  following correcting factors of M, S :
  C7:= [0.164,  0.533]:
  C8:= [0.1736, 0.315]:
  C9:= [0.256, -0.635e-2]:
  bf:= float(1)+lnn*0.8378:
  z1:= 1.2816+bf*_plus(C7[i+1]*(0.556^lnn)^i$i= 0..1)^(-ln((n-k)/n)):
  z2:= 1.6449+bf*_plus(C8[i+1]*(0.622^lnn)^i$i= 0..1)^(-ln((n-k)/n)):
  z3:= 2.3263+bf*_plus(C9[i+1]*lnn^i$i= 0..1)^(-ln((n-k)/n)):
  z4:= (z1+z2+z3)/3:
  z5:= (1.2816(z1-z4)+1.6449*(z2-z4)+2.3263*(z3-z4))/0.56268:
  z6:= z4-z5*1.7509:
  M:= M+z6*S:
  S:= S*z5:
  */

  pw:= 1 - stats::normalCDF(M, S^2)(float(y)):
  return([w, pw])    
end_proc:
