// $Date: 2001/12/06 13:28:33 $ $Author: stefanw $ $Revision: 1.29 $ 

/*++
infinity -- representing the real positive infinity

An infinite number is represented as an element of the domain
stdlib::Infinity. An element of this domain has one operand:

- The sign, which is either -1 or 1.

- Given the sign s the elements represent limes(s*a, a = infinity).

The domain element 'infinity' (which represents the positive real
infinity) is defined here. Other domain elements are only created
via the operations of this domain and NOT via 'new'.
++*/

stdlib::Infinity:= newDomain("stdlib::Infinity"):
stdlib::Infinity::create_dom:=hold(infinity):

infinity:= new(stdlib::Infinity, 1):

stdlib::Infinity::info:= "Domain 'stdlib::Infinity': representing infinity":
stdlib::Infinity::interface:= {}:

stdlib::Infinity::expr:=
proc()
begin
    extop(args(1),1) * hold(infinity)
end_proc:

stdlib::Infinity::print:= stdlib::Infinity::expr:
stdlib::Infinity::isNeg:= x->bool(op(extop(x,1))<0):


stdlib::Infinity::TeX:=
proc()
begin
  case extop(args(1),1)
    of 1 do
      "\\infty";
      break;
    of -1 do
      "- \\infty";
      break;
    end_case
end_proc:

stdlib::Infinity::_negate:=
proc()
begin
  extsubsop(args(1), 1 = -extop(args(1),1))
end_proc:

stdlib::Infinity::_invert:=
proc()
begin
  0
end_proc:

stdlib::Infinity::_subtract:=
proc()
begin
  args(1)+(-args(2))
end_proc:

stdlib::Infinity::_divide:=
proc(x, y)
begin
  piecewise([y=0, undefined], [y<>0, x*(1/y)])  
end_proc:

stdlib::Infinity::_not:=
proc()
begin
  error("invalid operand")
end_proc:

stdlib::Infinity::_power:=
proc(x, pow)
begin
  case domtype(pow)
    of stdlib::Infinity do
      if // no! x = infinity or // not 'undefined' for infinity^infinity!
         x = -infinity or // could also be 'complexInfinity'
         x = complexInfinity or
         x = undefined then
        return(undefined)
      end_if;
      if pow = infinity then
        if is(x = 1) = TRUE then
          return(1)
        elif is(x>1) = TRUE then
          return(pow)
        elif (is(-1<x) and is(x<1)) = TRUE then
          return(0)
        end_if
      else
        return((x^extop(pow,1))^infinity)
      end_if;
      return(hold(_power)(args()));
    of DOM_FLOAT do
      if iszero(pow) then return(1.0) end_if;
    of DOM_INT do
      if pow = 0 then return(1) end_if;
    of DOM_RAT do
      if pow < 0 then return(0) end_if;
      (extop(x,1)^pow) * infinity;
      break;
    otherwise
      if is(pow<0) = TRUE then return(0) end_if;
      hold(_power)(args())
  end_case
end_proc:

stdlib::Infinity::_plus:=
proc()
  local a, b;
begin
  // select infinities and others 
  a:= select([args()], not (stdlib::Infinity)::thisDomain);
  // skip real part of constants
  if contains(a, complexInfinity) > 0 or
    contains(a, undefined) > 0 then
    return(undefined)
  end_if;

  // skip reals
  a:= select(a, z -> is(z, Type::Real) <> TRUE);
  
  a:= map(a, u-> if testtype(u, Type::Constant) then
                   I*Im(u)
                 else
                   u
                 end_if
                   );
  a:= _plus(op(a));
  if a = undefined or a = complexInfinity then
    return(undefined)
  end_if;
  
  b:= select({args()}, (stdlib::Infinity)::thisDomain);

  if nops(b) > 1 then
    return(undefined)
  elif nops(b)=0 then
    // should not happen!
    return(a)
  end_if;

  // now nops(b)=1

  if iszero(a) then
      op(b)
  elif type(a) = "_plus" then  // flat operands of a 
    hold(_plus)(op(a), op(b))
  else
    hold(_plus)(a, op(b))
  end_if
end_proc:

stdlib::Infinity::_mult:= 
proc() 
  local a, b, c, f, sgn:DOM_INT;
begin
  b:= [args()];
  if contains(b, undefined) <> 0 then return(undefined) end_if;
  if contains(b, complexInfinity) <> 0 then return(complexInfinity) end_if;

  // f is only a dummy arg:
  
  [a,c,f]:=split(b, (stdlib::Infinity)::thisDomain);

  // multiply the signs of all infinities
  sgn:=_mult(op(map(a, extop, 1)));

  // multiply by all arguments one after the other
  // if the multiplication cannot be done, coellect the factors in a

  a:=[];
  
  for b in c do
    case sign(b)
      of 0 do
        // 0*infinity is undefined 
        return(undefined)
      of 1 do
        break; // nothing to be done
      of -1 do
        sgn:= -sgn;
        break;
      otherwise
        if type(b) = "_mult" then  // flat operands of a 
          a:=a.[op(b)]
        else    
          a:=a.[b]
        end_if
    end_case;

  end_for;

  b:=new(stdlib::Infinity, sgn);  

  if a = [] then
    b
  else
    hold(_mult)(op(a), b)
  end_if
end_proc:

stdlib::Infinity::_less:= (x,y)-> bool(x<=y and x<>y):

stdlib::Infinity::_leequal:= (x,y) ->bool(x=-infinity or y=infinity):

stdlib::Infinity::min:=
proc()
  local b;
begin
  b:= { args() };
  if contains(map(b, domtype), DOM_COMPLEX) then
    error("complex numbers are not allowed as arguments")
  elif contains(b, -infinity) then
    -infinity
  else
    // remove all +infinity
    b:=select(b, _unequal, infinity);
    if b={} then
      infinity
    else
      min(op(b))
    end_if
  end_if
end_proc:


stdlib::Infinity::max:=
proc()
  local b;
begin
 b:= { args() };
  if contains(map(b, domtype), DOM_COMPLEX) then
    error("complex numbers are not allowed as arguments")
  elif contains(b, infinity) then
    infinity
  else
    // remove all +infinity
    b:=select(b, _unequal, -infinity);
    if b={} then
      -infinity
    else
      max(op(b))
    end_if
  end_if
end_proc:


stdlib::Infinity::intmult:=
proc()
begin
  if iszero(args(2)) then
    undefined
  elif
    args(2) > 0 then
    args(1)
  else
    extsubsop(args(1), 1 = -extop(args(1),1))
  end_if
end_proc:

stdlib::Infinity::thisDomain:=
proc()
begin
  bool(domtype(args(1)) = stdlib::Infinity)
end_proc:


stdlib::Infinity::simplify:= id:

// the only thing that an infinity "has" is the object infinity

stdlib::Infinity::has:=(x,y) -> bool(y=infinity):

stdlib::Infinity::sign:=
proc()
begin
  extop(args(1),1)
end_proc:

stdlib::Infinity::O:= proc() begin undefined end_proc:

stdlib::Infinity::Re := id:

stdlib::Infinity::Im := 0:

stdlib::Infinity::hull :=
proc(self)
begin
  case extop(self, 1)
  of 1 do
    hull(RD_INF);
    break;
  of -1 do
    hull(RD_NINF);
    break;
  end_case;
end_proc:


// end of file 
