// $Date: 2002/02/19 14:02:11 $ $Author: stefanw $ $Revision: 1.27 $

/* product(f(k), k) returns g(k) such that g(k+1)/g(k)=f(k)
Examples:
product(k^2,k=1..4);

                                    576
product(k^2,k);
                                         2
                                 gamma(k) 
product(a[k],k=1..4);
                            a[1] a[2] a[3] a[4]
product(n+k,k=1..4);
                      (n + 1) (n + 2) (n + 3) (n + 4)
product(n+k,k);
                               gamma(k + n)
product(n+k,k=1..m);
                             gamma(m + n + 1)
                             ----------------
                               gamma(n + 1)  
product(2*h(k+1)/h(k),k);

                                        k
                                  h(k) 2 
*/
product :=
proc(f, k)
  local S;
begin
  if args(0) = 0 then
    error("wrong number of arguments")
  end_if;
  
  if f::dom::product <> FAIL then
    return(f::dom::product(args()))
  elif testtype(f, Type::Arithmetical) = FALSE then
    error("Illegal argument")
  end_if;
  
  if domtype(k) = DOM_IDENT then
    stdlib::prod_indef(f,k)
  elif testtype(k, hold(_equal)(DOM_IDENT, "_range")) then
    stdlib::prod_def(f, op(k,1), op(op(k,2)))
  elif type(k) = "_in" then
    if type((S:= op(k, 2))) = DOM_SET then
      return(_mult(subs(f, op(k, 1) = op(S, i)) $i=1..nops(S)))
    else
      return(procname(args()))
    end_if
  else
    error("invalid argument")
  end_if
end_proc:

// indefinite product, k is an identifier 
stdlib::prod_indef :=
proc(f: Type::Arithmetical, k: DOM_IDENT)
  local a, b, c, dg, S,
  indef_rec: DOM_PROC,
  productOverSet: DOM_PROC;
  
begin
  // Should we do this: ?
  //save k;
  //assume(k, Type::Integer);


  // local method indef_rec
  // input: f, an arithmetical expression
  // returns [a, b] such that product(f, k) = a * b^k
  // this is necessary because MuPAD is unable to collect
  // terms with power k later on :-(

  indef_rec:=
  proc(f)
    
  begin
    if not has(f, k) then
      return([1, f])
    end_if;

    case type(f)  
      of "_power" do
        if not has(op(f, 2), k) then
          // product(f^a, k) = product(f, k)^a
          return(map(indef_rec(op(f, 1)), _power, op(f, 2)));
        else
          return([hold(product)(f, k), 1])
        end_if;
        // NOT REACHED
        assert(FALSE);
      of "_mult" do
        a := numer(f);
        b := denom(f);
        if not has((c:= a/subs(b, k = k+1)), k) then
          return([b, c])
        elif not has((c := subs(a, k = k+1)/b), k) then
          return(piecewise([a = 0, [0, 0]], [a <> 0, [1 / a, c]]))
        else
          map([op(f)], indef_rec);
          return([_mult(op(map(%, op, 1))), _mult(op(map(%, op, 2)))])
        // toDo: this may be a product of many gammas.
        // therefore, one should apply rules for simplifying
        // gamma(1+y*I) gamma(1-I*y) here, cf. Abramowitz-Stegun
        end_if;
      // NOT REACHED
        assert(FALSE);
    end_case;

    if testtype(f, Type::RatExpr(k)) then
      if (dg:= degree(f, [k])) = 1 then
        a:= coeff(f, [k], 1);
        b:= coeff(f, [k], 0);
      // Computing \prod_k a*k + b
      // we have to distinguish:
      // a=0. Then the result is b^k
      // Otherwise we write the producty as \prod_k a (k + b/a)
      // If the argument is not 0 or a negative integer, the
      // result g can be expressed in terms of gamma
      // For the moment, we do not pay respect to that restriction
      // 
        return(piecewise([a=0, [1, b]],
                         [a<>0, [gamma(k + b/a), a]]))
      elif contains({"_mult", "_power"},
                    type((a:= Factored::convert_to(factor(f), DOM_EXPR)))) then
        return(indef_rec(a))
      elif dg <> FAIL and dg <= 4 then
      // proceed by solving
        S:= solve(f, k, MaxDegree = 4, IgnoreProperties);

      // local method productOverSet:
      // computes product(f, n) where f(x) =0 iff x \in S  

        productOverSet:=
        proc(S)
          local mylcoeff: DOM_PROC;
        begin
        // local method mylcoeff
        // leading coefficient with case analysis

          mylcoeff:=
          proc(f)
            local coefflist;
          begin
            coefflist:= map(poly2list(poly(f, [k])), _index, 1);
            eval(hold(piecewise)([coefflist[i] <>0 and
                                  _and(coefflist[j]= 0 $j=1..i-1),
                                  coefflist[i]] $i=1..nops(coefflist),
                                 [_and(coefflist[j] = 0
                                       $j=1..nops(coefflist)), 0]))
          end_proc;
        
          case type(S)
            of piecewise do
              return(piecewise::extmap(S, productOverSet))
            of DOM_SET do
              [indef_rec(k-j, k) $j in S];
              return([_mult(op(map(%, op, 1))), _mult(op(map(%, op, 2))) *
                      mylcoeff(f)])
            of solvelib::BasicSet do
            // the polynomial has infinitely many roots and must therefore
            // be zero
              return([0, 0])
          end_case;
        // RootOf etc.
          FAIL
        end_proc;
      
        productOverSet(S);
        if not has(%, FAIL) then
          return(%)
        end_if
      end_if
    end_if;

  // give up:
    [hold(product)(f, k), 1];
  end_proc;


indef_rec(f);
%[1] * %[2] ^ k

end_proc:




stdlib::prod_def :=
proc(f: Type::Arithmetical, k: DOM_IDENT, a, b)
  local p, i, g, getProductFromIndefiniteProduct;
begin
  if type(b-a) = DOM_INT then // finite product
    p:= 1;
    if b-a >= 0 then
      for i from 0 to b-a do
        p:= p*eval(subs(f, k=a+i))
      end_for
    else
      for i from 1 to a-b-1 do
        p:= p/eval(subs(f, k=b+i))
      end_for
    end_if;
    return(p)
  else
    // if the borders are infinity or integers, use the
    // information that the product goes through integers only
    if map({a, b}, domtype) minus {stdlib::Infinity, DOM_INT} = {} then
      save k;
      assume(k, Type::Integer);
      f:= simplify(f)
    end_if;
    g:=stdlib::prod_indef(f, k);

    getProductFromIndefiniteProduct:=
    proc(g)
    begin
      if type(g) = piecewise then
        piecewise::extmap(g, getProductFromIndefiniteProduct)
      else  
        if hastype(g, "product") then
          hold(product)(f, k=a..b)
        else  
          if eval(Sum::myeval(f, k, a)) = 0 then
            return(0)
          end_if;
          return(eval(Sum::myeval(g, k, b+1)/Sum::myeval(g, k, a)))
        end_if
      end_if
    end_proc;

    return(getProductFromIndefiniteProduct(g));
    
  end_if;
  hold(product)(f, k=a..b)
end_proc:

product:=funcenv(product):
product:=slot(product, "type", "product"):
product:=slot(product, "print", "product"):
product::Content      := proc(Out, data)
                     begin
                       if type(op(data, 2)) = "_equal" then
                         Out::Capply(Out::Cproduct,
                                     Out::Cbvar(Out(op(data, [2,1]))),
                                     Out::Clowlimit(Out(op(data, [2,2,1]))),
                                     Out::Cuplimit( Out(op(data, [2,2,2]))),
                                     Out(op(data,1)));
                       else
                         Out::Capply(Out::Cproduct,
                                     Out::Cbvar(Out(op(data, 2))),
                                     Out(op(data,1)));
                       end_if;
                     end_proc:
