
// $Date: 2002/02/22 13:33:17 $ $Author: stefanw $ $Revision: 1.62 $ 

//  Original version: zimmerma, 14/12/95 

/*
        solve  --  solve equations and inequalities

        solve(eq,x)			returns a set of solutions
        solve(eq,x=t)			returns only roots in domain t
                              (obsolete, use option Domain=...)                        
		
        solve(eq,S)        returns a list of assignments
                           [x1=a1,..,xn=an] if S is a set {x1,..,xn}
                                            or a list [x1,..,xn]
        solve(eq)          equivalent to solve(eq, S) where S= set
                           of indeterminates of eq                   

        Options:                                      
        Multiple		   return a multiset
        PrincipalValue	return only principal values
        MaxDegree=n     solve by radicals up to degree n, use RootOf
                                              for higher degrees
        BackSubstitution=TRUE    enables back substitution
        BackSubstitution=FALSE <default>   disables back substitution
        Domain=d        returns solutions in domain d
        DontRewriteBySystem  do not reduce solving an equation to solving
                             a system 

possible types of return values:

        "RootOf" - polynomial equation could not be solved, or its
                   degree exceeds MaxDegree                             
        "solve"  - equation could not be solved
        DOM_SET  - finite number of solutions
                   or set of lists of equations (as a special
                   return type of the solver for systems)
        "_in" - var \in (some of the other sets mentioned)
        "_union"
        "_intersect"
        "_minus"
        "function" i.e. solvelib::Union(...)                                        
        solvelib::BasicSet - one of Z, Q, R, C
        Dom::Interval
        Dom::ImageSet - something of the form { f(x); x \in S}
        piecewise  - the solution depends on case analysis                                         
        Dom::Multiset
                                                      
*/


alias(OPTIONS={Multiple, PrincipalValue, MaxDegree, BackSubstitution,
               Domain, IgnoreProperties, IgnoreSpecialCases,
                 DontRewriteBySystem, MaxRecLevel}):


solve :=                   
proc(eq, var)
  name solve;
  local sol,options, eqs, vars,
  mygetprop: DOM_PROC,
  secondargisoption: DOM_BOOL,
  propset;
  
begin

  // local method
  
  mygetprop:=
  proc(x)
  begin
    if property::hasprop(x) then
      getprop(x)
    else  
      Type::Complex
    end_if
  end_proc;

  // m a i n   p r o g r a m
  
  if args(0)=0 then
    error("solve called without arguments")
  end_if;
  if args(0)>1 then
    eqs:=eq;
    vars:=var;  // to avoid warnings 
    userinfo(1, "solving ".expr2text(eqs)." for ".expr2text(vars));
  else
    eqs:=eq;
    userinfo(1, "Solving :".expr2text(eqs));
  end_if;
  // first try overloading 
  if eqs::dom::solve<>FAIL then
    return(eqs::dom::solve(args()))
  end_if;


  // handle special cases of equations and illegal first arguments 

  case type(eqs)
    of DOM_ARRAY do
      return(solve([op(eqs)], args(2..args(0))))
    of DOM_TABLE do
      return(solve([map(op(eqs), op, 2)], args(2..args(0))))
    of DOM_NIL do
    of DOM_FAIL do
    of DOM_POINT do
    of DOM_POLYGON do
    of DOM_EXEC do
    of DOM_FUNC_ENV do
    of DOM_PROC do
    of DOM_PROC_ENV do
    of DOM_STRING do  
    of DOM_DOMAIN do
    of DOM_VAR do  
      error("Illegal first argument ".expr2text(eqs))
  end_case;

  if args(0)>=2 and indets(args(2)) intersect OPTIONS <> {} then
    options:=solvelib::getoptions(args(2..args(0)));
    secondargisoption:= TRUE
  else
    options:=solvelib::getoptions(args(3..args(0)));
    secondargisoption:= FALSE
  end_if;    


  
  
  if args(0)=1 or secondargisoption then
    if domtype(eqs)=DOM_EXPR or domtype(eqs)=DOM_SET or
      domtype(eqs)=DOM_LIST or domtype(eqs)=DOM_IDENT then
      vars:= solvelib::indets(eqs);
      if nops(vars)=0 then
          error("Need variable to solve for")
      end_if
    elif domtype(eqs)=DOM_POLY then
      if op(eqs,3)=Expr then
        vars:=op(eqs,2)
      elif has(op(eqs,3),IntMod) then
        return(map(solve(poly(eqs, Dom::IntegerMod(op(eqs,[3,1])))),
                   expr));
      else
        vars:=op(eqs,2);
        if options["Domain"]=solvelib::defaultOptions["Domain"]
          and options["Domain"]<> op(eqs,3) then
          // solving over coefficient domain
          options["Domain"]:=op(eqs,3);
        end_if
      end_if
    else
      error("Need variable to solve for")
    end_if;
    if {op(vars)} intersect OPTIONS <> {} then
      warning("Name of option used as an identifier")
    end_if;
  end_if;


// maximal recursion depth reached ?
  if options["MaxRecLevel"] <= 0 then
    return(procname(eqs, vars))
  end_if;

  
  // handle the option PrincipalValue separately
  
  if options["PrincipalValue"]=TRUE then
    options["PrincipalValue"]:=FALSE;
    solve(eqs, vars, options);
    if %={} then
      return({})
    else
      solvelib::getElement(%)
    end_if;
    if %=FAIL then
      return(hold(solve)(eqs, vars, PrincipalValue))
    else
      return({%})
    end_if
  end_if;

  // domain solver


  if options["Domain"]<> solvelib::defaultOptions["Domain"] then
      return(solvelib::domsolve(eqs, vars, options))
  end_if;
  
  userinfo(10, "Solving for ".expr2text(vars));
  
  case type(vars)
    of "_equal" do
      // new syntax for domain solver !! 
      warning("Obsolete syntax for calling domain solver used");
      options["Domain"]:= rhs(vars);
      vars:=lhs(vars);
      return(solvelib::domsolve(eqs, vars, options))
    of "_index" do
    of "function" do
      userinfo(10, "Substituting solve variable by new ident");
      sol:=genident();
      return(solve(subs(eqs, vars=sol), sol, options));
    of DOM_IDENT do
      // solve for one variable

      // special case: the variable has properties
      if property::hasprop(vars) and options["IgnoreProperties"]=FALSE then
        // solve the equation, and output those solutions that
        // agree with the properties of the variable to solve for
        propset:= property::prop2set(mygetprop(vars));
        save vars;
        unassume(vars);
        // to avoid infinite recursion caused by global options:
        options["IgnoreProperties"]:= TRUE;
        return(solve(eqs, vars, options)
             intersect propset)
      end_if;
      
      case type(eqs)
        of DOM_SET do
        of DOM_LIST do
          if nops(eqs)=0 then
            error("Cannot solve empty set of equations")
          end_if;
          sol:=map({op(eqs)}, type);
          if {"_less","_leequal", "_unequal"} intersect sol <> {} then
            // inequalities          
            sol := solvelib::ineqsys({op(eqs)}, {vars});
            // should always produce a solution of the form "x in interval"
            assert(type(sol) = "_in");
            return(op(sol, 2))
          end_if;
         
          return(solvelib::solve_sys(eqs, {vars}, options))

        of "_in" do
          // interpret solve(a(x) in S,x) as
          // determining a^(-1)(S)
          return(solvelib::preImage(op(eqs,1), vars, op(eq,2)))
          
          // boolean cases
        of DOM_BOOL do
          if eqs=FALSE then
            return({})
          elif eqs=TRUE then
            return(C_)
          else
            error("Cannot solve boolean value UNKNOWN")
          end_if
        of "_or" do
          return(_union(map(op(eqs), solve, vars, options)))
        of "_and" do
          return(_intersect(map(op(eqs), solve, vars, options)))
        of "_not" do
          return(C_ minus solve(op(eqs,1), vars, options ))
        of "_less" do
        of "_leequal" do
            return(solvelib::ineqs(eqs, op(vars), options))
        of "_unequal" do
          return(solvelib::BaseSet minus
                 solvelib::solve_eq(op(eqs,1)-op(eqs,2), vars, options))
        of "_equal" do
          if not testtype(op(eqs, 1), Type::Arithmetical) or
            not testtype(op(eqs, 2), Type::Arithmetical) then
            error("Both sides of equation must be arithmetical expressions")
          end_if;
          eqs:=op(eqs,1)-op(eqs,2);
          break
        of DOM_POLY do
          if op(eqs,3)=hold(Expr) then
            break
          end_if;
          if op(op(eqs,3),0)= hold(IntMod) then
            options["Domain"]:=Dom::IntegerMod(op(op(eqs,3)));
            return(map(solvelib::domsolve
                   (expr(eqs), vars, options),
                       expr))
          end_if;
          // coefficient ring is a domain 
          options["Domain"]:=op(eqs,3);
          return(solvelib::domsolve(eqs, vars, options))
      end_case;
      return(solvelib::solve_eq(eqs, vars, options))
    of DOM_SET do
      vars:=sort([op(vars)]); // fall through
    of DOM_LIST do
      if nops(vars)=0 then
        error("Need variable to solve for")
      else
        case type(eqs)
          of DOM_LIST do
          of DOM_SET do
            // this handles also inequations (by returning them unsolved)
            return(solvelib::solve_sys(eqs, vars, options))
          of DOM_BOOL do  
          of "_and" do
          of "_or" do
            error("Boolean expressions not allowed in system solver")
          otherwise
            return(solvelib::solve_sys({eqs}, vars, options))
        end_case
      end_if
    otherwise
      error("Illegal variable(s) to solve for: ".
            expr2text(vars))
  end_case;

  // should not be reached
  
  assert(FALSE);
  procname(args())
  
end_proc:


/* solvelib::indets - extract the indeterminates of an expression
                     or set of expressions */

solvelib::indets:=
proc(eq)

begin
  if eq::dom::indets <> FAIL then
    return(eq::dom::indets(eq))
  end_if;
  case type(eq)
    of DOM_SET do
      return(_union(op(map(eq, solvelib::indets))))
    of DOM_POLY do
      return({op(op(eq,2))})
    of piecewise do
      _union(solvelib::indets(piecewise::expression(eq, i)) $i=1..extnops(eq))
    of DOM_IDENT do
      if contains(Type::ConstantIdents, eq) then
        return({})
      else  
        return({eq})
      end_if
    of "_index" do
      if type(op(eq,1))=DOM_IDENT then
        return({eq})
      else
        return({})
      end_if
    otherwise
      if nops(eq)=1 and op(eq,0)=FAIL then
        return({})
      else
        return(_union(op(map([op(eq)], solvelib::indets))))
      end_if;
  end_case
end_proc:

solve:= funcenv(solve):
solve:= slot(solve, "type", "solve"):
solve:= slot(solve, "print", "solve"):
solve:= slot(solve, "info",
    "solve -- solve equations and inequalities [try ?solve for options]"
):

solve:=slot(solve, "float", numeric::solve):

// end of file 



