/*
scilab::func( ["i1", "i2", ..., "iN"],
              ["r1", "r2", "r3", ..., "rM"],
              ["pure Scilab code", "more code", ... ],
              <ReturnTypes=[rt1, ..., rtM]>
            )
Here ["i1", "i2", ..., "iN"] is a list of Scilab variables
which are initialized by MuPAD, so the pure Scilab code can use this
predefined variables. ["r1", "r2", "r3", ..., "rM"] is a list of Scilab
variables where Scilab stores the results of its computation.
("pure Scilab code", "more code", ... ) is a sequence of strings which
contains the Scilab program. scilab::func() defines a MuPAD procedure which
will execute the Scilab program ["pure Scilab code", "more code", ... ] and
which will return the results as a list.
<ReturnTypes=[rt1, ..., rtM] specifies the return types of the
"result"-variables r1, ..., rM. Possible return types are Dom::Matrix,
Dom::SparseMatrix, DOM_ARRAY or Default. Where Default means do no conversion.
MuPAd tries to convert the results r1, ..., rM to the gives type specified
in ReturnTypes=[rt1, ..., rtM]. If the conversion is not possible the
function fails. If the function fails the result of the function is a list
with three elements, where the fist element is FAIL, the second element
is the string "Scilab Error" or "MuPAD error" specified if the error happens
on the Scilab or the MuPAD side and the third element is an additional
information. For "Scilab Error" it is the Scilab error code while executing
the Scilab operation

Example:
  f:=scilab::func( ["i1","i2"], ["r1", "r2", "r3"],
                   ["r1=i1^2","r2=i2^2", "r3=i1+i2"] ) ; 

Now f(3,5) gives the result [9,25,8].

*/

scilab::func:=
proc(inputlist:DOM_LIST, outputlist:DOM_LIST, program:DOM_LIST)
  option escape ;
  local i, testlist, scilab_prog, prog_array, tag, lprog ;
begin

  // The function tests if all elements in the gives list 'list' are of
  // type 'typi'
  testlist:=
  proc(list:DOM_LIST, typi)
    local i ;
  begin
    for i from 1 to nops(list) do
      if domtype(op(list,i)) <> typi then
        return( FALSE ) ;
      end_if;
    end_for ;
    return(TRUE);
  end_proc:
  
  // Test the arguments
  if args(0) <> 3  then
    error("wrong no of arguments");
  end_if;
  // Test the elements of every list
  if testlist(args(1), DOM_STRING) = FALSE then
    error("Wrong type in the first list") ;
  end_if;
  if testlist(args(2), DOM_STRING) = FALSE then
    error("Wrong type in the second list") ;
  end_if;
  if nops(args(2)) = 0 then
    error("Wrong number of names in the second list") ;
  end_if;
  if testlist(args(3), DOM_STRING) = FALSE then
    error("Wrong type in the third list");
  end_if;
  
  // Create the program which will be transfered to scilab and which will
  // execute the program, which is written in the third list
  scilab_prog:=[(op(args(1),i)." = _MUP_vars_(".
                 expr2text(i).") ; _MUP_vars_(".expr2text(i).") = 0 ;")
                 $i=1..nops(args(1))];
  // Initialize the result parameters
  scilab_prog:= scilab_prog.[(op(args(2),i)." = 0 ;") $i=1..nops(args(2))];
  scilab_prog:=scilab_prog.args(3) ;
  scilab_prog:=scilab_prog.["_MUP_isSendError_ = 0 ;"] ;
  // Getting (Sending) the results
  scilab_prog:=scilab_prog.
                ["udx_send(_MUP_channel_, _MUPANS_, _MUPANS_ ); "] ;
  scilab_prog:=scilab_prog.[(scilab::MUPpvm_sci2f77."(".op(args(2),i).
                            ") ; _MUP_sendE_(_MUP_channel_,".
                            op(args(2),i)." , _MUPANS_ ) ;")
                            $i=1..nops(args(2))];
  
  prog_array:=array(1..nops(scilab_prog),scilab_prog) ;

  lprog:=
  proc()
    local i, errtmp, err, nargs, quiet, arglist, arg_is_ok, tmplist,
          returnlist, tmplen, n, j, returntype, result, lcompute ;
  begin
    // Makes the computation via Scilab
    // It is a local procedure because in some error case it is reused
    lcompute:=
    proc()
    begin
      scilab::MUPstartComputing() ;
      // Send the number of arguments
      scilab::toscilab(nargs);
      // Send the mode
      scilab::toscilab(1);
      err:= FALSE ;
      // Send the arguments
      for i from 1 to nargs do
        errtmp:= scilab::toscilab(arglist[i]) ;
        if errtmp <> TRUE then
          err:= errtmp ;
        end_if:
      end_for ;
      
      // Send the programm
      scilab::toscilab(prog_array) ;
      // If there were special problems sending the data, try to handle this
      // 0: Scilab shouldn't compute the program
      // 1: Everything is OK Scilab should execute the program
      if  err <> FALSE then
        scilab::toscilab(0) ;
        if err = "CantConvert" then
          scilab::MUPendComputing() ;
          if quiet then
            return( [FAIL, "MuPAD Error",
                     "Cannot convert to double (out of range)"] ) ;
          else
            error("Cannot convert to double (out of range)") ;
          end_if;
        else
          scilab::MUPkillclient();
          error("Fatal error using hardware floats") ;
        end_if:
      else
        scilab::toscilab(1) ;
        // Till we have TAGS in the udx_send and udx_recv calls
        // we need another method to decide if there is an error
        if nops(outputlist) > 0 then
          tag := scilab::fromscilab() ;
          if tag = -4.0  then // Scilab program fails
            result := scilab::fromscilab() ;
            scilab::MUPendComputing() ;
            // Does we have a stacksize error
            if (result = 17.0) then
              // stacksize error: increase the stacksize and compute again
              result := scilab::MUPincStacksize(scilab::MUPincrementStacksize) ;
              if has(result, FAIL) then
                error("Cannot increase the Scilab stacksize") ;
              end_if;
              result := lcompute() ;
              return(result) ;    
            end_if;
            return([FAIL, "Scilab Error", result])
          else  // tag == -3.0
            result:= [scilab::fromscilab(returnlist[i])
                      $i=1..nops(outputlist)]  ;
            // Get the second message tag. Does the udx_send call fails?
            tag:= scilab::fromscilab() ;
            if tag = -4.0 then
              result:=scilab::fromscilab() ;
              scilab::MUPendComputing() ;
              return([FAIL, "Scilab Error", result])
            end_if;
            scilab::MUPendComputing() ;
            // Here everything is fine, eventually there exist a mistake
            // because a wrong returnType was used or it was not possible
            // to convert the Scilab data
            if ( has(result, FAIL) ) then
              for i from 1 to nops(outputlist) do
                if domtype(result[i]) = DOM_LIST then
                  if result[i][1] = FAIL then
                    warning(result[i][2]);
                    return( [FAIL, "MuPAD Error", FAIL] ) ;
                  end_if:
                end_if:
              end_for;
            else
              op(result);
            end_if:
          end_if;
        end_if;
      end_if:
    end_proc:
    
    // test if the options Quiet or ReturnType are set
    quiet:= FALSE ;
    returntype:=FALSE ;
    n:= args(0)-1 ;
    nargs:=args(0) ;
    returnlist:=[0 $k=1..nops(outputlist)] ;
     for i from n to args(0) do
      if i > 0 then
        if args(i) = Quiet then
          quiet:= TRUE ;
          nargs:= nargs-1;
        else
          // see if the option ReturnType is set
          if type(args(i)) = "_equal" then
            if op(args(i),1) = ReturnType then
              if domtype(op(args(i),2)) = DOM_LIST then
                if nops(op(args(i),2)) <> nops(outputlist) then
                  error("The list of the return types must contain as many \n".
                        "elements as the list of the return parameters.");
                else // create the list of the return types
                  tmplist:= op(args(i),2) ;
                  tmplen:= nops(op(args(i),2)) ; 
                  for j from 1 to tmplen do 
                    case tmplist[j] 
                      of Dom::Matrix() do
                        returntype:=TRUE ;
                        returnlist[j] := 2 ; break ;
                      of Dom::SparseMatrix() do
                        returntype:=TRUE ;
                        returnlist[j]:= 3 ; break ;
                      of DOM_ARRAY do
                        returntype:=TRUE ;
                        returnlist[j]:= 1 ; break ;
                      of Default do
                        returntype:=TRUE ;
                        break ;
                      otherwise
                        error("Wrong return type used. Possible values are:\n".
                         "Dom::Matrix(), Dom::SparseMatrix(), DOM_ARRAY\n".
                         "and Default") ;
                    end_case:
                  end_for:
                  nargs:= nargs-1;
                end_if;
              else
                error("Return types must be specified by the equation 'ReturnType = [...]' where the list [...] contains the types of the return values") ;
              end_if;
            else
              if (quiet = TRUE) or (returntype = TRUE) then
                error("Wrong option or wrong number or arguments used") ;
              end_if:
            end_if:
          else // of _equal
            if (quiet = TRUE) or (returntype = TRUE) then
              error("Wrong option or wrong number or arguments used") ;
            end_if;
          end_if;
        end_if; // else part of the ReturnType option test
      end_if;   // if i > 0
    end_for;

  if ( nops(inputlist) <> nargs ) then
    if quiet then
      return([FAIL, "MuPAD Error", "wrong no of arguments"]) ;
    else
      error("wrong no of arguments");
    end_if;
  end_if ;
  for i from 1 to nargs do
    arg_is_ok:= scilab::MUPtestarg( args(i) ) ;
    if arg_is_ok[1]  = FALSE then
      if quiet then
        return([FAIL, "MuPAD Error", "Argument No".expr2text(i).
                " from the list of input parameters ".
         "has a wrong type which cannot be converted from MuPAD to Scilab"]) ;
      else
        error("Argument No".expr2text(i).
              " from the list of input parameters\n".
         "has a wrong type which cannot be converted from MuPAD to Scilab");
      end_if;
    end_if;
    if arg_is_ok[2] = NIL then
      arglist[i]:= args(i) ;
    else
      arglist[i]:= arg_is_ok[2] ;
    end_if;
  end_for ;
    // Start Scilab if necessary
    if scilab::isup() = FALSE then
      if quiet then
        if scilab::start(Quiet) = FALSE then
          return(FAIL);
        end_if;
      else
         if scilab::start() = FALSE then
          return(FAIL);
        end_if;       
      end_if:
    end_if;

  result:= lcompute() ;

  end_proc:
end_proc:


