
// make_slot generically creates a function call for every scilab function,
// so that scilab::scilabfunction(), e.g. scilab::sqrt() execute
// scilabfunction() inside Scilab. 
scilab::make_slot:= proc(thisdomain, method:DOM_STRING) option escape;
local _method ;
begin
  // save the name of the method for later use
  _method := method ;
  // return value: a procedure generating some string 
  
  proc()
    local result, tag, i, arg_is_ok, mymethod, scilab_prog,
          prog_array, call, err, errtmp, nargs, n, quiet, arglist,
          returntype, lcompute ;
  begin
    // Makes the computation via Scilab
    // It is a local procedure because in some error case it is reused
    lcompute:=
    proc()
    begin
      scilab::MUPstartComputing() ;
      //scilab::number_of_data( nargs ) ;
      // Send the number of arguments
      scilab::toscilab(nargs) ;
      // Send the mode
      scilab::toscilab(0) ;
      // Send the Arguments
      err:= FALSE ;
      for i from 1 to nargs do
        errtmp:= scilab::toscilab(arglist[i]) ;
        if errtmp <> TRUE then
          err:= errtmp ;
        end_if:
      end_for ;
      // Send the program
      scilab::toscilab(prog_array) ;
      // If there were special problems sending the data, try to handle this
      // 0: Scilab shouldn't compute the program
      // 1: Everything is OK Scilab should execute the program
      if  err <> FALSE then
        scilab::toscilab(0) ;
        if err = "CantConvert" then
          scilab::MUPendComputing() ;
          if quiet then
            return( [FAIL, "MuPAD Error",
                     "Cannot convert to double (out of range)"] ) ;
          else
            error("Cannot convert to double (out of range)") ;
          end_if:
        else
          scilab::MUPkillclient();
          error("Fatal error using hardware floats") ;
        end_if:
      else
        scilab::toscilab(1) ;
        // Get the message tag
        // tag =  -2.0 means _MUPREQ_, a request from Scilab which is not
        //             used at the moment
        // tag =  -3.0 means _MUPANS_, a normal result no error
        // tag =  -4.0 means _MUPISERR_, an error occured while computing
        //             inside Scilab
        tag:= scilab::fromscilab() ;
        // Get the data
        result:= scilab::fromscilab(returntype);
        // Get the second message tag. Does the udx_send call fails?
        if tag = -3.0 then
          tag:= scilab::fromscilab() ;
        end_if;
        scilab::MUPendComputing() ;
        case tag
          of -3.0 do
            // Here it is posisble that one gets a wrong return type from
            // Scilab, which cannot be handled by MuPAD even if the Scilab
            // computation was successful. So we have to check if this is
            // the case
            if (domtype(result) = DOM_LIST) and (result[1] = FAIL) then
              warning(result[2]);
            return( [FAIL, "MuPAD Error", FAIL] ) ;
            else
              return( result ) ;
            end_if:
          of -4.0 do
            if (result = 17.0) and (_method <> "MUPincStacksize") then
              // stacksize error: increase the stacksize and compute again
              result := scilab::MUPincStacksize(scilab::MUPincrementStacksize) ;
              if has(result, FAIL) then
                error("Cannot increase the Scilab stacksize") ;
              end_if;
              result := lcompute();
              return( result ) ;
            else
              return( [FAIL, "Scilab Error", result] ) ;
            end_if;
          otherwise
            scilab::MUPkillclient();
            error("Fatal error using hardware floats");
        end_case ;
      end_if:
    end_proc:

    // test if the options Quiet or ReturnType are set
    quiet:= FALSE ;
    returntype:= 0 ;
    n:= args(0)-1 ;
    nargs:=args(0) ;
    for i from n to args(0) do
      if i > 0 then
        if args(i) = Quiet then
          quiet:= TRUE ;
          nargs:= nargs-1;
        else
          // see if the option ReturnType is set
          if type(args(i)) = "_equal" then
            if op(args(i),1) = ReturnType then
              case op(args(i),2)
                of Dom::Matrix() do
                  returntype:= 2 ; break ;
                of Dom::SparseMatrix() do
                  returntype:= 3 ; break ;
                of DOM_ARRAY do
                  returntype:= 1 ; break ;
                of Default do
                  // Means no conversion should happen
                  returntype:= 4 ; break ;
                otherwise
                  error("Wrong return type used. Possible values are:\n".
                        "Dom::Matrix(), Dom::SparseMatrix(), DOM_ARRAY or Default") ;
              end_case:
              nargs:= nargs-1;
            else
              if (quiet = TRUE) or (returntype <> 0) then
                error("Wrong option used") ;
              end_if:
            end_if:
          else // of _equal
            if (quiet = TRUE) or (returntype <> 0) then
              error("Wrong option used") ;
            end_if;
          end_if;
        end_if; // else part of the ReturnType option test
      end_if;   // if i > 0
    end_for;
    if returntype = 4 then
      returntype := 0 ;
    end_if:
    case method
      of "_plus"  do mymethod:="+"; break ;
      of "_minus" do mymethod:="-"; break ;
      of "_mult"  do mymethod:="*"; break ;
      of "_power" do mymethod:="^"; break ;
      of "_divide" do mymethod:="/"; break ;
      of "_and" do 
          if nargs = 1 then 
            mymethod:="and";
          else
            mymethod:="&";
          end_if; 
          break ;
      of "_or" do 
          if nargs = 1 then 
            mymethod:="or";
          else
            mymethod:="|";
          end_if; 
          break ;
      of "_not" do mymethod:="~"; break ;
      of "_transpose" do mymethod:=".'"; break ;
      otherwise mymethod:=method ;
    end_case;
    
    // Test if the arguments can be handled by Scilab and MuPAD
    for i from 1 to nargs do
      arg_is_ok:=scilab::MUPtestarg(args(i));
      if arg_is_ok[1] =  FALSE then
        if quiet then
          return([FAIL, "MuPAD Error", "Argument No".expr2text(i).
                " has a wrong type which cannot be".
                " converted from MuPAD to Scilab" ]);
        else
          error("Argument No".expr2text(i).
                " has a wrong type which cannot be\n".
                "converted from MuPAD to Scilab");
        end_if;
      end_if;
      if arg_is_ok[2] = NIL then
        arglist[i]:= args(i) ;
      else
        arglist[i]:= arg_is_ok[2] ;
      end_if;
    end_for;
    // Start Scilab if necessary
    if scilab::isup() = FALSE then
      if quiet then
        if scilab::start(Quiet) = FALSE then
          return(FAIL);
        end_if;
      else
         if scilab::start() = FALSE then
          return(FAIL);
        end_if;       
      end_if:
    end_if;

    if mymethod = "+" or mymethod = "-" or mymethod = "*" or \
       mymethod = "^" or mymethod = "/" or mymethod = "&" or \
       mymethod = "|" then
      // Create the Scilab program useing UDX for the communication part
      // The program does not contain the part which receives the
      // parameters. This is done inside the Scilab startup script
      scilab_prog:=["_MUP_result_ = "  .
      _concat(("_MUP_vars_(",expr2text(i),")".
               (if i<nargs then " ".mymethod." " else null() end_if)) 
                $i=1..nargs)." ;",
      scilab::MUPpvm_sci2f77."(_MUP_result_) ;"];
      //"udx_send(_mupad_channel_, _MUP_result_, _MUPANS_ ) ;",
      //"udx_send(_mupad_channel_, _MUPANS_, _MUPANS_ ) ;" ;
      // "clean("._concat(("_MUP_arg".expr2text(i)."_, ")
      // $i=1..nargs)."_MUP_result_ ) ;";  
    elif mymethod = ".'" or mymethod = "~" then
      if nargs <> 1 then
        if quiet then
          return([FAIL, "MuPAD Error", "Wrong number of arguments"]) ;
        else
          error("Wrong number of arguments");
        end_if:
      end_if ;
      case mymethod
        of ".'" do call:="_MUP_result_ = "."_MUP_vars_(1)".mymethod." ;" ;
          break ;
        of "~" do call:="_MUP_result_ = ".mymethod."_MUP_vars_(1) ; " ;
          break ;
      end_case;
      scilab_prog:=[call,
      scilab::MUPpvm_sci2f77."(_MUP_result_) ;"];
      //"udx_send(_mupad_channel_, _MUP_result_, _MUPANS_ ) ;",
      //"udx_send(_mupad_channel_, _MUPANS_, _MUPANS_ ) ;" ;
    else
      scilab_prog:=["_MUP_result_ = ".mymethod."("  .
      _concat(("_MUP_vars_(",expr2text(i),")".
               (if i<nargs then ", " else null() end_if)) $i=1..nargs).
              ");",
      scilab::MUPpvm_sci2f77."(_MUP_result_) ;"];
      //"udx_send(_mupad_channel_, _MUP_result_, _MUPANS_ ) ;",
      //"udx_send(_mupad_channel_, _MUPANS_, _MUPANS_ ) ;" ;
      //"clean("._concat(("_MUP_arg".expr2text(i)."_, ")
      // $i=1..nargs)."_MUP_result_ ) ;";
    end_if;
    prog_array:=array(1..nops(scilab_prog),scilab_prog) ;
    result:= lcompute() ;
  end_proc:

  // end of return value 

end_proc:

