
// Internal slots of the scilab domain
// Every internal slot will start with the prefix MUP

// At the moment the following external Slots exist, which should be directly
// available to the user:
//   scilab::start()
//   scilab::avail()
//   scilab::isup()
//   scilab::stop()

// Procedure to start scilab
scilab::start:=
proc()
  local res, quiet ;
begin
  // test the arguments
  case args(0)
    of 0 do  quiet:=FALSE ;
      break ;
    of 1 do if args(1) = Quiet then
              quiet:=TRUE;
            else
              error("First parameter must be the option Quiet") ;
            end_if;
      break ;
    otherwise error("wrong number of arguments") ;
  end_case;

  // Handle the case if a previous scilab computation was interrupted
  if ( scilab::isup() and not scilab::MUPisComputing() ) then
    return(TRUE) ;
  end_if;

  if scilab::MUPisComputing() then
    scilab::MUPkillclient() ;
  end_if:

  if quiet then
    res:=scilab::MUPstart(Quiet) ;
  else
    res:=scilab::MUPstart() ;
  end_if:
  
  if res = TRUE then
    // The following is also responsible for the synchronization
    // of the Shared Memory interface 
    if traperror((scilab::fromscilab()), 30) <> 0 then
      scilab::MUPkillclient() ;
      if quiet = FALSE then
        error("Cannot start the Scilab client. It seems to be that the\n".
            "    MUP_SCI environment variable is not set correctly.\n".
            "    Remember that MuPAD needs Scilab2.6 to work as expected.\n");
      else
        return(FALSE);
        end_if;
    end_if;
  else
    return(FALSE) ;
  end_if;
  sysassign(scilab::MUPScilabTmpDir, scilab::MUPgetScilabTmpDir());
  return(TRUE) ;
end_proc:

// Procedure to restart scilab
// This procedure is used if Scilab was aborted
scilab::MUPrestart:=
proc()
  local res ;
begin
  // test the arguments
  res:=scilab::start(Quiet) ;
  if res = FALSE then
        error("Cannot restart the Scilab client after an aborted computation.".
              " It seems to be that the MUP_SCI environment variable is not".
              " set correctly. Remember that MuPAD needs Scilab2.6 to work".
              " as expected.\n");
  end_if:
  sysassign(scilab::MUPScilabTmpDir, scilab::MUPgetScilabTmpDir());
end_proc:

// Procedure to restart Scilab test if Scilab can be started correctly
// This procedure also starts Scilab
scilab::avail:=
proc()
begin
  if scilab::isup() then
    return(TRUE) ;
  else
    return(scilab::start(Quiet)) ;
  end_if;
end_proc:

// Procedure which send the right program to Scilab to stop it
scilab::MUPstopit:=
proc()
  local scilab_prog, prog_array ;
begin
  scilab_prog:=["udx_exit(_MUP_channel_) ;",
                "quit ;"] ;
  prog_array:=array(1..nops(scilab_prog),scilab_prog) ;
  // Sending the
  // - Number of arguments
  // - Mode
  scilab::toscilab(0);
  scilab::toscilab(0);
  // Send the program to Scilab
  scilab::toscilab(prog_array) ;
  // Send the action mode
  scilab::toscilab(1);
  return( TRUE ) ;
end_proc:

scilab::MUPSetCallOnExit:=
proc()
  local elem ;
begin
  elem:= Pref::callOnExit() ;
  case domtype(elem)
    of DOM_NIL do
      Pref::callOnExit(scilab::stop):
      break :
    of DOM_LIST do
      elem:= append(elem,scilab::stop):
      Pref::callOnExit(elem):
      break:
    otherwise
      Pref::callOnExit([elem, scilab::stop])
  end_case ;
  return(TRUE) ;
end_proc:

// Procedure which return the coeffizient of a polynomial as a list
// There is no type check, because it is assumed that the typechecking is
// done elsewhere (see scilab::MUPtestarg())
scilab::MUPcoeff2list:=
proc(poly)
begin
  (domtype(poly))::vectorize(poly) ;
end_proc:

// Procedure which creates a MuPAD polynomial from a list and an identifier
scilab::MUPcreateRealPoly:=
proc(list:DOM_LIST, ident:DOM_IDENT )
begin
  Dom::UnivariatePolynomial(ident, Dom::ExpressionField())(
                                                 poly( list,[ident] )) ;
end_proc:

// Procedure which creates a MuPAD polynomial from a list and an identifier
scilab::MUPcreateComplexPoly:=
proc(list:DOM_LIST, ident:DOM_IDENT )
begin
  Dom::UnivariatePolynomial(ident, Dom::ExpressionField())(
                                                 poly( list,[ident] )) ;
end_proc:

// Procedure which creates a matrix with real polynomials as entries
// from a list and an identifier
scilab::MUPcreateRePoMatrix:=
proc(rows:DOM_INT, columns:DOM_INT, list:DOM_LIST, ident:DOM_IDENT )
begin
  Dom::Matrix(Dom::UnivariatePolynomial(ident, Dom::ExpressionField()))(rows,columns,list)
end_proc:

// Procedure which creates a matrix with complex polynomials as entries
// from a list and an identifier
scilab::MUPcreateCoPoMatrix:=
proc(rows:DOM_INT, columns:DOM_INT, list:DOM_LIST, ident:DOM_IDENT )
begin
  Dom::Matrix(Dom::UnivariatePolynomial(ident, Dom::ExpressionField()))(rows,columns,list)
end_proc:

// Procedure which creates a sparse matrix from a list containing polynomials
// representing the sparse matrix
scilab::MUPcreateSparseMatrix:=
proc(rows:DOM_INT, columns:DOM_INT, polylist:DOM_LIST )
begin
  new(Dom::SparseMatrix(), rows, columns, polylist)
end_proc:

// Procedure which tests if the data can be send to Scilab
// The procedure also converts the argument to a floating point
// number if this is necessary and possible
// Return values:
//   [FALSE, NIL] : The data can not be converted to a Scilab object
//   [TRUE, NIL]  : The data can be converted to a Scilab object
//   [TRUE, value]: The data can be converted to a Scilab object and
//                  value is the object which should be used for this
//                  conversion instead of the given data
//
scilab::MUPtestarg:=
proc( arg )
  local typi, dim, elemtype, i, j, domext, typus, value, tmparg, result, data ;
begin
  typi:=domtype(arg) ;
  tmparg:= FAIL;
  if typi = DOM_EXPR then
    tmparg:= float(arg);
    typi:= domtype(tmparg) ;
  end_if;
  case typi
    of DOM_IDENT do
      if contains(Type::ConstantIdents, arg) then
        tmparg:=float(arg) ;
      else
        return([FALSE,NIL]);
      end_if;
    of DOM_FLOAT do ;
    of DOM_RAT do ;
    of DOM_INT do ;
    of DOM_COMPLEX do;
    of DOM_BOOL do;
    of DOM_STRING do
      if tmparg = FAIL then
        return( [TRUE, NIL] );
      else
        return([TRUE, tmparg]) ;
      end_if;
    of DOM_LIST do
    of scilab::Tlist do
      result:= [0 $ j=1..nops(arg)] ;
      for i from 1 to nops(arg) do
        value := scilab::MUPtestarg(op(arg,i));
        if value[1] = FALSE then
          return( [FALSE, NIL] ) ;
        else
          // Create the return list
          if value[2] = NIL then
            result[i]:= op(arg,i) ;
          else
            result[i]:= value[2] ;
          end_if:
        end_if;
      end_for;
      return([TRUE, result]);
    of DOM_ARRAY do
      dim := op(op(arg,0),1);
      if dim > 2 then
        return( [FALSE, NIL] );
      end_if;
      elemtype := domtype(op(arg,1));
      case elemtype
        of DOM_STRING do ;
        of DOM_BOOL   do ;
          for i from 2 to nops(arg) do
            if domtype(op(arg,i)) <> elemtype then
              return( [FALSE, NIL] );
            end_if ;
          end_for;
          return([TRUE, NIL]);
        otherwise
          // Here it is possible that we have an array representing a matrix
          // Create the matrix and do the typicall test
          if ( dim = 1 ) then
            arg := array(1..1,1..nops(arg),[[op(arg)]]);
            data:= (Dom::Matrix())::create(arg);
          else
            data:= (Dom::Matrix())::create(arg);
          end_if;
          result := scilab::MUPTestMatrixEntries(data) ;
          if result[1] <> FAIL then
            if result[2] = NIL then
              return([TRUE,data]) ;
            else
              return([TRUE,result[2]]) ;
            end_if:
          else
            return([FALSE,NIL]) ;
          end_if:
      end_case;
    otherwise
      // Now the case of domains is handled
      domext:= (typi)::constructor ;
      if domext = FAIL then
        return( [FALSE, NIL] ) ;
      end_if;
      typus:= scilab::MUPgetdomtype(arg) ;
      if typus[1] = (-1) then
        return( [FALSE, NIL] );
      end_if;
      return([TRUE,typus[2]]) ;
  end_case;
end_proc:

// Procedure which returns the appropiate Scilab data type matching the
// data type of the MuPAD object 'data'. Here 'data' must be of type
// DOM_EXT where the appendant domain contains a domain constructor
//
// Return values:
//   [special_typ, newvalue] were 'special type' gives information about the
//   Scilab datatype of the 'data' and 'newvalue' can be NIL if
//   special_typ == -1 or a mupad object, which should be used for transfer
//   to Scilab instead of data. It's still possible that 'newvalue' and
//   the given 'data' are equal.
//
//   Values and there meanings for special_typ:
//     -1 : The data typ is not known (convertable) to Scilab (a Scilab data
//          object.
//      1 : The appropriate Scilab data typ is real matrix
//      2 : The appropriate Scilab data typ is complex matrix
//      3 : The appropriate Scilab data typ is real polynomial 
//      4 : The appropriate Scilab data typ is complex polynomial 
//      5 : The appropriate Scilab data typ is real polynomial matrix
//      6 : The appropriate Scilab data typ is complex polynomial matrix
//      7 : The appropriate Scilab data typ is a Tlist
//      8 : The appropriate Scilab data typ is a sparse boolean matrix
//      9 : The appropriate Scilab data typ is a real sparse matrix
//     10 : The appropriate Scilab data typ is a complex sparse matrix
//
scilab::MUPgetdomtype:=
proc( data )      
  local typi, coeffring, result, ident ;
begin
  if args(0) <> 1 then
    return( [-1, NIL] ) ;
  end_if;
  typi:= domtype(data) ;
  case typi
    of scilab::Tlist do //scilab::Tlist do
      return([7,NIL]) ; // Is a Tlist
    of Dom::SparseMatrixF2 do
      return([8,NIL]) ;  // Is a SparseBooleanMatrix in Scilab
  end_case;
  typi:= (domtype(data))::constructor ;
  case typi
    of Dom::SparseMatrix do
      result := scilab::MUPTestSparseMatrixEntries(data) ;
      if result[1] <> FAIL then
        if result[1] = "Real" then
          return([9,result[2]]) ;
        else
          return([10,result[2]]) ;
        end_if:
      else
        return([-1,NIL]) ;
      end_if:
    of Dom::Matrix do ;
    of Dom::SquareMatrix do ;
      coeffring:= (domtype(data))::coeffRing ;
      if coeffring::constructor = Dom::UnivariatePolynomial then
        ident:=coeffring::indets();
        if ( nops(ident) <> 1 ) then
          return([-1, NIL]);
        end_if;
        if domtype(op(ident,1)) <> DOM_IDENT then
          return([-1, NIL]);
        end_if;
        result:= scilab::MUPTestPolyMatrixData(data) ;
        if result[1] <> FAIL then
          if result[1] = "Real" then
          return([5,result[2]]) ; // Is a RealPolynomialMatrix
          else
            return([6,result[2]]) ; // Is a ComplexPolynomialMatrix
          end_if:
        else
          return([-1,NIL]) ;
        end_if:
      else  // normal matrices
        result := scilab::MUPTestMatrixEntries(data) ;
        if result[1] <> FAIL then
          if result[1] = "Real" then
            return([1,result[2]]) ;
          else
            return([2,result[2]]) ;
          end_if:
        else
          return([-1,NIL]) ;
        end_if:
      end_if;
    of Dom::UnivariatePolynomial do
      result := scilab::MUPTestPolyData(data) ;
      if result[1] <> FAIL then
        if result[1] = "Real" then
          return([3,result[2]]) ; // Is a real Polynomial
        else
          return([4,result[2]]) ; // Is a complex Polynomial
        end_if:
      else
        return([-1,NIL]) ;
      end_if:
    otherwise
      return([-1,NIL]);
  end_case;
end_proc:

// Test if the entries of a given sparse matrix can be converted to a
// floating point number
// Return value:
//   [FAIL, NIL]     : The entries cannot be converted
//   ["Real", NIL]   : 'data' is real sparse matrix
//   ["Complex", NIL]: 'data' is a complex sparse matrix
//   ["Real", val]   : 'data' is real sparse matrix  but 'val' is the MuPAD
//                     object which should be sended to Scilab
//   ["Complex", val]: 'data' is complex sparse matrix  but 'val' is the MuPAD
//                     object which should be sended to Scilab
//
scilab::MUPTestSparseMatrixEntries:=
proc(data)
  local real_valid_types, complex_valid_types, columns, has_complex_values,
        i, result,  symbolic_entries,  mat ;
begin
  // test if the entries of the matrix have the right type
  // eventually we need a C-routine to do this if it is too slow
  // on the library side
  real_valid_types := {DOM_FLOAT, DOM_INT, DOM_RAT} ;
  complex_valid_types := {DOM_COMPLEX};
  columns := extop(data,2) ;
  has_complex_values := FALSE ;
  symbolic_entries := FALSE ;
  // First strategy: Test if all entries are of type DOM_FLOAT, DOM_INT,
  // DOM_RAT or DOM_COMPLEX
  for i from 1 to columns do
    // See if one of the matrix elements of the j-th columns is
    // of a wrong type
    result := map({coeff(extop(data,3)[i])},domtype) minus
              real_valid_types ;
    if result = complex_valid_types then
      has_complex_values := TRUE ;
    elif result <> {} then
      symbolic_entries := TRUE ;
      break ;
    end_if:
  end_for:
  if symbolic_entries = FALSE then
    if has_complex_values = TRUE then
      return(["Complex", NIL]) ;
    else
      return(["Real", NIL]) ;
    end_if:
  else
    // Second strategy: Test if all entries can be converted to a floating
    // number. E.g. for sin(2) this is the case.
    has_complex_values := FALSE ;
    mat:=float(data);
    for i from 1 to columns do
    // See if one of the matrix elements of the j-th columns is
    // of a wrong type
      result := map({coeff(extop(mat,3)[i])},domtype) minus
                real_valid_types ;
      if result = complex_valid_types then
        has_complex_values := TRUE ;
      elif result <> {} then
        return([FAIL,NIL]);
      end_if:
    end_for:
    if has_complex_values = TRUE then
      return(["Complex",mat]) ;
    else
      return(["Real",mat]) ;
    end_if:    
  end_if:
  return([FAIL,NIL]) ;
end_proc:

// Test if the entries of a given matrix can be converted to a
// floating point number
// Return value:
//   [FAIL, NIL]     : The entries cannot be converted
//   ["Real", NIL]   : 'data' is real  matrix
//   ["Complex", NIL]: 'data' is a complex matrix
//   ["Real", val]   : 'data' is real  matrix  but 'val' is the MuPAD
//                     object which should be sended to Scilab
//   ["Complex", val]: 'data' is complex matrix  but 'val' is the MuPAD
//                     object which should be sended to Scilab
//
scilab::MUPTestMatrixEntries:=
proc(data)
  local real_valid_types, complex_valid_types, has_complex_values,
        result,  symbolic_entries,  mat, coeffring  ;
begin
  // Make a fast test, if there are type information available.
  coeffring:= (domtype(data))::coeffRing ;
  case coeffring
    of Dom::Float do
      return(["Real",NIL]);    // Is a RealMatrix
    of Dom::Numerical do
      return(["Complex",NIL]); // 2 Is a ComplexMatrix
  end_case:
  // test if the entries of the matrix have the right type
  // eventually we need a C-routine to do this if it is too slow
  // on the library side
  real_valid_types := {DOM_FLOAT, DOM_INT, DOM_RAT} ;
  complex_valid_types := {DOM_COMPLEX};
  has_complex_values := FALSE ;
  symbolic_entries := FALSE ;
  // First strategy: Test if all entries are of type DOM_FLOAT, DOM_INT,
  // DOM_RAT or DOM_COMPLEX
  result := map({op(extop(data,3))},domtype) minus
            real_valid_types ;
  if result = complex_valid_types then
    has_complex_values := TRUE ;
  elif result <> {} then
    symbolic_entries := TRUE ;
  end_if:
  if symbolic_entries = FALSE then
    if has_complex_values = TRUE then
      return(["Complex", NIL]) ;
    else
      return(["Real", NIL]) ;
    end_if:
  else
    // Second strategy: Test if all entries can be converted to a floating
    // number. E.g. for sin(2) this is the case.
    has_complex_values := FALSE ;
    mat:=float(data);
    result := map({op(extop(mat,3))},domtype) minus
              real_valid_types ;
    if result = complex_valid_types then
      has_complex_values := TRUE ;
    elif result <> {} then
      return([FAIL,NIL]);
    end_if:
    if has_complex_values = TRUE then
      return(["Complex",mat]) ;
    else
      return(["Real",mat]) ;
    end_if:    
  end_if:
  return([FAIL,NIL]) ;
end_proc:

// Test if the entries of a given polynomial can be converted to a
// floating point number
// Return value:
//   [FAIL, NIL]     : The entries cannot be converted
//   ["Real", NIL]   : 'data' is real polynomial 
//   ["Complex", NIL]: 'data' is a complex polynomial 
//   ["Real", val]   : 'data' is real  polynomial  but 'val' is the MuPAD
//                     object which should be sended to Scilab
//   ["Complex", val]: 'data' is complex polynomial  but 'val' is the MuPAD
//                     object which should be sended to Scilab
//
scilab::MUPTestPolyData:=
proc(p)
  local real_valid_types, complex_valid_types, has_complex_values,
        result,  symbolic_entries,  mat  ;
begin
  if domtype(op(indets(p))) <> DOM_IDENT then
    return([FAIL, NIL]);
  end_if;
  
  // test if the coefficients of the polynomial have the right type
  real_valid_types := {DOM_FLOAT, DOM_INT, DOM_RAT} ;
  complex_valid_types := {DOM_COMPLEX};
  has_complex_values := FALSE ;
  symbolic_entries := FALSE ;
  // First strategy: Test if all coefficients are of type DOM_FLOAT, DOM_INT,
  // DOM_RAT or DOM_COMPLEX
  result := map({coeff(p)},domtype) minus
            real_valid_types ;
  if result = complex_valid_types then
    has_complex_values := TRUE ;
  elif result <> {} then
    symbolic_entries := TRUE ;
  end_if:
  if symbolic_entries = FALSE then
    if has_complex_values = TRUE then
      return(["Complex", NIL]) ;
    else
      return(["Real", NIL]) ;
    end_if:
  else
    // Second strategy: Test if all entries can be converted to a floating
    // number. E.g. for sin(2) this is the case.
    has_complex_values := FALSE ;
    mat:=mapcoeffs(p,float);
    result := map({coeff(mat)},domtype) minus
              real_valid_types ;
    if result = complex_valid_types then
      has_complex_values := TRUE ;
    elif result <> {} then
      return([FAIL,NIL]);
    end_if:
    if has_complex_values = TRUE then
      return(["Complex",mat]) ;
    else
      return(["Real",mat]) ;
    end_if:
  end_if:
  return([FAIL,NIL]) ;
end_proc:

// Test if the entries of a given polynomial matrix can be converted to a
// floating point number
// Return value:
//   [FAIL, NIL]     : The entries cannot be converted
//   ["Real", NIL]   : 'data' is real polynomial matrix
//   ["Complex", NIL]: 'data' is a complex polynomial matrix
//   ["Real", val]   : 'data' is real  polynomial matrix but 'val' is the
//                     MuPAD object which should be sended to Scilab
//   ["Complex", val]: 'data' is complex polynomial matrix  but 'val' is the 
//                     MuPAD object which should be sended to Scilab
//
scilab::MUPTestPolyMatrixData:=
proc(data)
  local real_valid_types, complex_valid_types, has_complex_values,
        result,  symbolic_entries,  mat ;
begin
  // test if the entries of the matrix have the right type
  // eventually we need a C-routine to do this if it is too slow
  // on the library side
  real_valid_types := {DOM_FLOAT, DOM_INT, DOM_RAT} ;
  complex_valid_types := {DOM_COMPLEX};
  has_complex_values := FALSE ;
  symbolic_entries := FALSE ;
  // First strategy: Test if all polynomials have coefficients of type
  // DOM_FLOAT, DOM_INT, DOM_RAT or DOM_COMPLEX
  result:= map(map({op(extop(data,3))},coeff),domtype) minus
           real_valid_types ;
  if result = complex_valid_types then
    has_complex_values := TRUE ;
  elif result <> {} then
    symbolic_entries := TRUE ;
  end_if:  
  if symbolic_entries = FALSE then
    if has_complex_values = TRUE then
      return(["Complex", NIL]) ;
    else
      return(["Real", NIL]) ;
    end_if:
  else
    has_complex_values := FALSE ;
    mat:=map(data,mapcoeffs,float);
    // Second strategy: Test if all entries can be converted to a floating
    // number. E.g. for sin(2) this is the case.
    result:= map(map({op(extop(mat,3))},coeff),domtype) minus
           real_valid_types ;
    if result = complex_valid_types then
      has_complex_values := TRUE ;
    elif result <> {} then
      return([FAIL,NIL]);
    end_if:
    if has_complex_values = TRUE then
      return(["Complex",mat]) ;
    else
      return(["Real",mat]) ;
    end_if:
  end_if:
  return([FAIL,NIL]) ;
end_proc:

// Defining a fuunction to increase the Scilab stacksize
scilab::MUPincStacksize:=
scilab::func( ["x"],["r"],
              ["incr = stacksize() ; stacksize(incr(1)+x);"] ):


// Gets the Scilab tmp directory TMPDIR
scilab::MUPgetScilabTmpDir:= scilab::func( [],["r"], ["r=TMPDIR;"] ):
