/* $Date: 2002/08/22 08:41:31 $  $Author: andi $  $Revision: 1.29 $ */

/**************************************************************************/
/* FILE   : MUP_env.h                                                     */
/* AUTHOR : Andreas Sorgatz                                               */
/* CREATED: 31.08.1999                                                    */
/* CHANGED: 15.08.2002                                                    */
/* DESCRIP: This file declares the MuPAD Kernel Environment class.        */
/**************************************************************************/

// doxygen Kommentar
// The file section is only here, so that the file is shown in the file 
// list of the created documentation
/** \file MUP_env.h
    \brief MuPAD Kernel environment

*/

#ifndef __MUP_env__
#define __MUP_env__


#include "MUP_constants.h"


#if (!defined MSDOS)
class MCLT_classic_switch;
#endif


extern bool MDB_gcov_initOK(char *);            // needed for test coverage
extern void MDB_gcov_exit();                    // needed for test coverage

extern int  MUT_set_alarm(long s,void (*h)());  // needed for the watchdog
extern void MUT_reset_alarm();                  // needed for the watchdog
extern void MUP_catch_system_error();           // needed for global use
void        MUP_env_catchWatchdog();            // Global watchdog handler




/** \brief MuPAD Kernel Status
    
    stores the status of the evaluator
*/
class MUPT_status {
public:
   /** possible values for MuPAD Kernel Status represents the status 
       of the evaluator */
  enum KernelStatus {
    Continue, Return, Quit, Next, Break
  };
  KernelStatus status_;  // continue, return, quit, next or break
  short count_;   // next or break level counter
};

class MCLT_bytestream;                          // external declaration

/** \brief MuPAD Kernel Environment
    
    The MuPAD kernel environment specifies and controlles the behaviour of 
    the kernel.  This includes features like suppressing a banner message,
    specifying the library path, etc. The environment is first set to some
    default values.  It can be changed via command line options as well as
    by directly accessing method and variables of the MUPT_env class.
*/
class MUPT_env {  

public:
  /// Constructor
  MUPT_env();

  /// Destructor
  ~MUPT_env();

  /// Sets default values for all options of the MuPAD kernel environment
  void setDefaults ( int argc, char* argv[] );

  /// Checks is MuPAD is registered correctly
  bool isRegistered ( );

  /// Frees allocated S_Pointers
  void freeEnvironment();

  /// Parses command line options to initialise the kernel environment
  int  setCommandLineOptions ( int argc, char* argv[] );

  /// Displays information about the usage of command line options
  void printUsage ();

  /// Prints out the current values of the MuPAD kernel environment.
#if (defined MUP_ENV_DEBUG) || (defined DEBUG_ON)
  void debugDisplay ( );
#endif

  /// Sets/Gets the type of communication switch that is used
  inline void setTermSwitch()  {        terminalMode =  'T' ; };
  inline int   isTermSwitch()  { return(terminalMode == 'T'); };
  inline void setMotifSwitch() {        terminalMode =  'M' ; };
  inline int   isMotifSwitch() { return(terminalMode == 'M'); };
  inline void setEmacsSwitch() {        terminalMode =  'E' ; };
  inline int   isEmacsSwitch() { return(terminalMode == 'E'); };

#if (!defined MSDOS)
  /// This variables stores the handle of the communication switch.
  /// The classic (old) communication is only used on Macintosh and
  /// UNIX systems.
  MCLT_classic_switch *theSwitch;
#endif

  // switches global test coverage on/off
  inline void setGcovOn ()    { MDB_gcov_initOK(gcovOutputFile); }
  inline void setGcovOff (){ 
    if (gcovOutputFile != NULL)   
      MDB_gcov_exit(); 
  }

  /// Sets/Gets the debug modes. Note that on Macintosh and Windows
  /// systems a graphical debugger front end is used by default and
  /// thus debugging takes place always in connect mode.
#if (defined MACINTOSH) || (defined MSDOS)
  inline void setDebugOn ()            { debugMode |=  1 ; 
                                         setDebugConnect(); };
#else
  inline void setDebugOn()             { debugMode |=  1 ; };
#endif
  inline void setDebugMode (long dm)   { debugMode  = dm ; };
  inline void setDebugOff()            { debugMode  =  0 ; };
  inline void setDebugVerbose()        { debugMode |=  4 ; };
  inline void setDebugConnect()        { debugMode |=  8 ; };
  inline void setDebugInteractive()    { debugMode |= 17 ; };
  inline void setDebugNotInteractive() { debugMode &= ~16; };
  inline void setDebugQuitRequest()    { 
    dbgQuitRequested_ = true;
    setUserInterrupt();
  };
  inline void resetDebugQuitRequest()    { 
    dbgQuitRequested_ = false;
  };
  inline bool getDebugQuitRequest()    { return(dbgQuitRequested_);};
  inline int  getDebugMode()           { return(debugMode      ); };
  inline int   isDebugOn()             { return(debugMode &   1); };
  inline int   isDebugVerbose()        { return(debugMode &   4); };
  inline int   isDebugConnect()        { return(debugMode &   8); };
  inline int   isDebugInteractive()    { return(debugMode &  16); };

  /// PARI Stacksize in KB
  long pariStackSize;

  /// PARI prime number limit
  long primeNumberLimit;

  /// These streams define the default output and error devices
  MCLT_bytestream *stdOutput;
  MCLT_bytestream *errOutput;

  /// These time stamps are set when the kernel is started up.
  /// Varable sessionBeginMSec is used for real time meassuring.
  long sessionBegin;
  long sessionBeginMSec;    // process start real time in msec's 
  
  /// This flag specifies if a banner is printed or not.
  bool printBanner;

  /// Sets/Gets the watchdog modes and installs a new watchdog
  inline void setWatchdogDoTerminate()  { watchdogAction  = 'T' ; };
  inline void setWatchdogThrowError()   { watchdogAction  = 'E' ; };
  inline void setWatchdogMode(char wdm) { watchdogAction  = wdm ; };
  inline char getWatchdogMode()         { return(watchdogAction); };
  void        resetWatchdog()           { MUT_reset_alarm()     ; };
  inline int  installWatchdog(int sec)  
          { return( MUT_set_alarm(sec, &MUP_env_catchWatchdog) ); };

  /// Controls user interrupt in critical kernel sections
  bool setUserInterrupt       ( bool interrupt=true );
  bool setBlockUserInterrupts ( bool block=true );
  bool isUserInterrupt        ( );

  /// Tests if the kernel is in a critical section of initialization
  bool isInitializing( );

  /// Variables to control critical sections of kernel initialization.
  bool inInit;
  bool inPreInit;

  /// Is set to 'true' when the memory management has been set up
  bool isMemUp;

  /// Default new Handler of the kernel
#ifdef WIN32
  new_handler origNewHandler;
#else
  std::new_handler origNewHandler;
#endif

  /// This flag specifies whether the userinit file is read or not. It
  /// is needed when the kernel is launched or reset.  The default value
  /// should be true.
  bool readUserInit;

  /// This flag specifies whether the init file of user packages are read or
  /// not. It is needed when the kernel is launched or reset.  The default 
  /// value is true.
  bool readUserPack;

  /// Process-id of the MuPAD kernel
  int parentPid;

  /// Pathname of the MuPAD kernel binary file
  char *progName;

  /// Pathname of the MuPAD installation directory
  char rootPath [MUPC_MAX_PATH];

  /// Pathname of the MuPAD binary directory
  char binPath [MUPC_MAX_PATH];

  /// Pathname of the MuPAD dynamic modules directory
  char modPath [MUPC_MAX_PATH];
  char modPathAbs [MUPC_MAX_PATH];

  /// Pathname of the MuPAD library directory
  inline char *libPath() { return libPath_; }  

  /// sets the pathname of the MuPAD library directory
  inline void setLibPath(char *tmp) { libPath_ = tmp; }  

  /// Pathname of the MuPAD standard library directory
  inline char *libMainPath() { return libMainPath_; }  

   /** 
       sets the pathname of the MuPAD standard library directory
       using the existing lib path,. which contains the standard
       library as the last component
   */
  void setLibMainPath() ;

  /// Pathname of the MuPAD package directory
  inline const char *packPath() { return packPath_; }

  /// Pathname of the MuPAD user init file and preferences/resources
  /// On Macintosh systems it is not used.  On Windows systems it is
  /// used only if it is defined via a command line option.
  char userPath [MUPC_MAX_PATH];

  /// Pathname of the MuPAD help pages and help index file
  char helpPath      [MUPC_MAX_PATH];
  char helpIndexPath [MUPC_MAX_PATH];

  /// Pathname which can be used to temporary files
  char tempPath      [MUPC_MAX_PATH];
  
  /// Output file when batch mode is active. This variable either 
  /// contains a file name,  the value - for the default output
  /// device or NULL  when the output file could not be opened or
  /// batch mode is inactive.
  char *batchOutputFile;

  /// file used for test coverage outputs
  char *gcovOutputFile;

  /// Stores a user option which may be specified via the command line
  char *userOption;

  /// functions to get/set the current read file name
  const char* get_READ_FILENAME();
  void set_READ_FILENAME(const char *str);
  

  /// Methods to access Pref options stored in MUPT_env.
  inline void setSecureKernel(bool value) { SecureKernel = value; };
  inline bool getSecureKernel()           { return(SecureKernel); };
  inline bool getPrintEcho()              { return(PrintEcho   ); };
  inline bool getPrintPrompt()            { return(PrintPrompt ); };
  inline bool getPrintWarn15()            { return(PrintWarn15 ); };

  /// Methods to access the SecureReadPath, SecureWritePath, SecureExecPath
  inline char *getSecureReadPath()             { return(SecureReadPath ); };
  inline char *getSecureWritePath()            { return(SecureWritePath); };
  inline char *getSecureExecPath()             { return(SecureExecPath);  };
  inline void  setSecureReadPath(char *value)  { 
    if (SecureReadPath!=NULL) free(SecureReadPath);
    SecureReadPath  = value; 
  };
  inline void  setSecureWritePath(char *value) { 
    if (SecureWritePath!=NULL) free(SecureWritePath);
    SecureWritePath = value; 
  };
  inline void  setSecureExecPath(char *value) { 
    if (SecureExecPath!=NULL) free(SecureExecPath);
    SecureExecPath = value; 
  };

  /// Methods to access the current state or the evaluator
  inline bool getStatusContinue() 
  { return(status.status_ == MUPT_status::Continue); };
  inline void setStatusContinue() 
    { status.status_ = MUPT_status::Continue; };

  inline bool getStatusQuit() 
    { return(status.status_ == MUPT_status::Quit); };
  inline void setStatusQuit() 
    { status.status_ = MUPT_status::Quit; };

  inline bool getStatusReturn() 
    { return(status.status_ == MUPT_status::Return); };
  inline void setStatusReturn() 
    { status.status_ = MUPT_status::Return; };

  inline short getStatusBreak() 
    { return((status.status_ == MUPT_status::Break) ? status.count_ : 0); };
  inline short decreaseStatusBreak() { 
    if (status.status_ == MUPT_status::Break) {
      status.count_ -= 1;
      if (status.count_ == 0) {
	status.status_ = MUPT_status::Continue;
	return true;
      }
    }
    return false;
  };
  inline void setStatusBreak( short level ) {
    assert(level > 0);
    status.status_ = MUPT_status::Break;  status.count_ = level; 
  };
  
  inline short getStatusNext() 
    { return((status.status_ == MUPT_status::Next) ? status.count_ : 0); };
  inline short decreaseStatusNext() { 
    if (status.status_ == MUPT_status::Next) {
      status.count_ -= 1;
      if (status.count_ == 0) {
	status.status_ = MUPT_status::Continue;
	return true;
      }
    }
    return false;
  };
  inline void setStatusNext( short level ) {
    assert(level > 0);
    status.status_ = MUPT_status::Next;  status.count_ = level; 
  };
  

  // needed for saving and restoring the status
  inline MUPT_status getStatus() 
    { return(status); };
  inline void setStatus( MUPT_status stat ) 
    { status = stat; };

  // functions for getting/setting internal representations
  // of environment variables
  inline S_Pointer getLIBPATH()   { return(LIBPATH); };
  inline void setLIBPATH(S_Pointer path) {
    if (LIBPATH != MMMNULL) MMMfree(LIBPATH);
    LIBPATH = path; 
  };
  inline S_Pointer getREADPATH()   { return(READPATH); };
  inline void setREADPATH(S_Pointer path) {
    if (READPATH != MMMNULL) MMMfree(READPATH);
    READPATH = path; 
  };
  inline S_Pointer getWRITEPATH()   { return(WRITEPATH); };
  inline void setWRITEPATH(S_Pointer path) {
    if (WRITEPATH != MMMNULL) MMMfree(WRITEPATH);
    WRITEPATH = path; 
  };
  inline S_Pointer getPACKAGEPATH()   { return(PACKAGEPATH); };
  inline void setPACKAGEPATH(S_Pointer path) {
    if (PACKAGEPATH != MMMNULL) MMMfree(PACKAGEPATH);
    PACKAGEPATH = path; 
  };

private:
  /// This flag specifies the terminal mode that is used to launch the
  /// kernel. Currently supported modes are (I)nteractive, (M)otif and
  /// (X)view.
  char terminalMode;

  /// This flag specifies the debug modes that are used by the kernel.
  /// Currently the following bits are supported: 1=active, 4=verbose, 
  /// 8=connected, 16=interactive. Bit 4,8,16 requires also to set bit
  /// 1. If no bit is set, debugging in inactive.
  long debugMode;


  /// This flag controls the action of isUserInterrupt.  If true 
  /// and the debugger is active then we want to quit the debugger.
  /// If the debugger is active and this flag is set to false
  /// we switch into step mode.
  bool dbgQuitRequested_;


  /// Stores the action mode of the watchdog timer. Supported modes
  /// are: (E)xit program and (C)atch error code.
  volatile char watchdogAction;

  /// Variables to control user interrupt in critical kernel sections.
  bool UserInt;
  bool BlockUserInt;

  /// Variables that store command line flags which influence options
  /// that can be set with the Pref command. This is needed for the
  /// MuPAD reset command.
  bool SecureKernel;
  bool PrintEcho;
  bool PrintPrompt;
  bool PrintWarn15;

  /// Pathname list of secure read paths
  char *SecureReadPath;

  /// Pathname list of secure write paths
  char *SecureWritePath;

  /// Pathname list of secure execute paths
  char *SecureExecPath;

  ///  the status of the evaluator
  MUPT_status  status;

  /// the values of the corresponding environment variables
  S_Pointer LIBPATH; 
  S_Pointer READPATH; 
  S_Pointer WRITEPATH; 
  S_Pointer PACKAGEPATH; 

  // filename of the file currently read
  const char *READ_FILENAME;

  // package path
  char *packPath_;

  // lib path
  char *libPath_;
 
  // lib path of the standard library
  char libMainPath_[MUPC_MAX_PATH];
  
};

//@Include: MUP_env.cpp

#endif /* __MUP_env__ */

